/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "shared/core/RCEngine.h"
#include "core/Transaction.h"
#include "shared/core/RCTableImpl.h"

BHEngineReturnValues RCEngine::RunLoader(THD* thd, sql_exchange* ex, TABLE_LIST* table_list, BHError& bherror)
{
	return ExternalLoad(thd, ex, table_list, bherror);
}

BHEngineReturnValues RCEngine::ExternalLoad(THD* thd, sql_exchange* ex,	TABLE_LIST* table_list, BHError& bherror)
{
	BHEngineReturnValues ret = LD_Successed;

	char name[FN_REFLEN];
	TABLE *table;
	int error=0;
	String *field_term = ex->field_term;
	int transactional_table = 0;
	boost::shared_array<char> BHLoader_path(new char[4096]);
	ConnectionInfo* tp = NULL;

	COPY_INFO info;

	try
	{
		// GA
			size_t len = 0;
			dirname_part(BHLoader_path.get(), my_progname, &len);
			BHASSERT_WITH_NO_PERFORMANCE_IMPACT(BHLoader_path[len] == 0);

		// 	dirname_part(BHLoader_path.get(), my_progname);

		strcat(BHLoader_path.get(), BHLoaderAppName);
		uint counter = 0;
		open_tables(thd, &table_list, &counter, MYSQL_LOCK_IGNORE_FLUSH);
		if(counter != 1 || !RCEngine::IsBHTable(table_list->table))
			return LD_Continue;

		if (field_term->length() > 1)
		{
			my_message(ER_WRONG_FIELD_TERMINATORS,ER(ER_WRONG_FIELD_TERMINATORS), MYF(0));
			bherror = BHError(BHERROR_SYNTAX_ERROR, "Wrong field terminator.");
			return LD_Failed;
		}

		if (open_and_lock_tables(thd, table_list)) {
			bherror = BHError(BHERROR_UNKNOWN, "Could not open or lock required tables.");
			return LD_Failed;
		}
		auto_ptr<Loader> bhl = LoaderFactory::CreateLoader(LOADER_BH);

		table = table_list->table;
		transactional_table = table->file->has_transactions();

		IOParameters iop;
		BHError bherr;
		if((bherror = RCEngine::GetIOParameters(iop, *thd, *ex, table)) != BHERROR_SUCCESS)
			throw LD_Failed;

		string table_path = GetTablePath(thd, table);
		// already checked that this is a BH table
		if (rsi_manager)
			rsi_manager->UpdateDef();

		tp = GetThreadParams(*thd);


		table->copy_blobs=0;
		thd->cuted_fields=0L;

		/*
		We must invalidate the table in query cache before binlog writing and
		ha_autocommit_...
		*/
		query_cache_invalidate3(thd, table_list, 0);

		bhl->Init(BHLoader_path.get(), &iop);
		bhl->MainHeapSize() = this->m_loader_main_heapsize;
		bhl->CompressedHeapSize() = this->m_loader_compressed_heapsize;

		uint transaction_id = tp->GetTransactionID();
		bhl->TransactionID() = transaction_id;
		//rcdev << lock << "bhl.Proceed()" << unlock;
		
		uint lask_packrow = JustATable::PackIndex(GetTable(table_path, table->s)->NoObj());
		
#ifdef __WIN__
        if(bhl->Proceed(tp->pi.hProcess) && !tp->killed())
#else
        if(bhl->Proceed(tp->pi) && !tp->killed())
#endif
		{
			if((_int64)bhl->NoRecords() > 0) {
				tp->GetTransaction()->ReleasePackRow(GetTable(table_path, table->s)->GetID(), lask_packrow);
				RefreshTable(table_path.c_str(), bhl->TransactionID());  // thread id used temp to store transaction id ...
			}
		} else {
			if (tp->killed()) {
				bherror = BHError(BHERR0R_KILLED, error_messages[BHERR0R_KILLED]);
				thd->send_kill_message();
			}
			else if(*bhl->ErrorMessage() != 0)
				bherror = BHError(bhl->ReturnCode(), bhl->ErrorMessage());
			else
				bherror = BHError(bhl->ReturnCode());
			error = 1; //rollback required
			throw LD_Failed;
		}

		info.records = ha_rows((_int64)bhl->NoRecords());
		info.copied  = ha_rows((_int64)bhl->NoCopied());
		info.deleted = ha_rows((_int64)bhl->NoDeleted());

		sprintf(name, ER(ER_LOAD_INFO), (ulong) info.records, (ulong) info.deleted,
				(ulong) (info.records - info.copied), (ulong) thd->cuted_fields);
		my_ok(thd, info.copied+info.deleted, 0L, name);
		//send_ok(thd,info.copied+info.deleted,0L,name);
	} catch(BHEngineReturnValues bherv) {
		ret = bherv;
	}

	if (thd->transaction.stmt.modified_non_trans_table)
		thd->transaction.all.modified_non_trans_table= TRUE;

	if (transactional_table)
		error = ha_autocommit_or_rollback(thd, error);




	if (thd->lock)
	{
		mysql_unlock_tables(thd, thd->lock);
		thd->lock=0;
	}
	thd->abort_on_warning = 0;
	//rcdev << lock << "RCEngine::LoadData(...) ended." << unlock;
	return ret;
}
