/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef COLUMN_H_
#define COLUMN_H_

#include "RCAttrTypeInfo.h"
#include "types/RCDataTypes.h"
#include "common/bhassert.h"
#include "common/CommonDefinitions.h"
#include "ColumnType.h"
#include "Descriptor.h"

class ConnectionInfo;

enum PackOntologicalStatus	{ NULLS_ONLY, UNIFORM, UNIFORM_AND_NULLS, SEQUENTIAL, NORMAL };

/*! \brief Base class for columns.
 *
 * Defines the common interface for RCAttr, Attr and VirtualColumn
 */
class Column {
public:

	Column(ColumnType ct = ColumnType(), bool is_unique = false, bool is_unique_updated = false) :
		ct(ct), is_unique(is_unique), is_unique_updated(is_unique_updated) {}

	Column(const Column& c) { if(this != &c) *this = c; }


	/*! \brief Get the full type of the column
	 *
	 */
	inline const ColumnType& Type() const { return ct; }

	/*! \brief Get simple type of the column
	 *
	 */
	inline const AttributeType TypeName() const { return ct.GetTypeName(); }

	/*! \brief Are all the values unique?
	 *
	 */
	virtual bool IsUnique() const { return is_unique; }

	/*! \brief Sets the uniqueness status
	 *
	 * The status is changed in memory only
	 * \param unique true if column has been verified to contain unique values
	 */
	virtual void SetUnique(bool unique) { is_unique = unique; }

	/*! \brief Is Unique status updated (valid)?
	 *
	 */
	virtual bool IsUniqueUpdated() const { return is_unique_updated; }

	/*! \sa  IsUniqueUpdated
	 *
	 */
	virtual void SetUniqueUpdated(bool updated) { is_unique_updated = updated; }

	//! calculate the number of 1's in histograms and other KN stats
	virtual void RoughStats(double &hist_density,int &trivial_packs,double &span) { BHASSERT(false, "not overridden"); }

	void SetCollation(DTCollation collation) { ct.SetCollation(collation); }
	DTCollation GetCollation() { return ct.GetCollation(); }
protected:
	ColumnType ct;

private:
	bool is_unique;
	bool is_unique_updated;
};
#endif
