/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef _COMPILEDQUERY2_H_
#define _COMPILEDQUERY2_H_

#include <vector>
#include "common/CommonDefinitions.h"
#include "JustATable.h"
#include "TempTable.h"
#include "CQTerm.h"

//////////////////////////////////////////////////////////////////////////////////
//
//  CompiledQuery - for storing execution plan of a query (sequence of primitive operations) and output data definition
//	version 2.0, will eventually replace the current CompiledQuery.
//
/////////////////////////////////////////////////////////////////////////////////
class MysqlExpression;
class Item_bhfield;
class CompiledQuery
{
public:
	////////////////////// Definition of one step
	enum StepType { TABLE_ALIAS, TMP_TABLE, CREATE_FILTER, AND_F, OR_F, NOT_F, COPY_F, AND_E, /*OR_E, */
		T_MODE, JOIN_T, DEFINE_PARAM, FILTER_T, ADD_COLUMN,
		ADD_ORDER, UNION, INTERSECT, MINUS, RESULT, STEP_ERROR, CREATE_VC };
	// ADD_GROUP deprecated

	class CQStep
	{
	public:
		StepType	type;
		TabID		t1;
		TabID t2;
		TabID t3;
		AttrID		a1, a2;
		FilterID	f1, f2, f3;
		ParamID		p1;
		CQTerm		e1, e2, e3;
		Operator	op;						// predicate: O_EQ, O_LESS etc.
		TMParameter tmpar;					// Table Mode Parameter
		JoinType	jt;
		ColOperation cop;
		char*		alias;
		std::vector<MysqlExpression*> mysql_expr;
		std::vector<int> virt_cols;
		_int64			n1, n2;						// additional parameter (e.g. descending order, TOP n, LIMIT n1..n2)

		CQStep(): type(TABLE_ALIAS),t1(NULL_VALUE_32),t2(NULL_VALUE_32),t3(NULL_VALUE_32),a1(NULL_VALUE_32),a2(NULL_VALUE_32),f1(NULL_VALUE_32),f2(NULL_VALUE_32),f3(NULL_VALUE_32),p1(NULL_VALUE_32),e1(),e2(),e3(),op(O_EQ),tmpar(TM_DISTINCT),jt(JO_INNER),cop(LISTING),alias(NULL), mysql_expr(NULL), n1(NULL_VALUE_64), n2(NULL_VALUE_64) {};
		CQStep(const CQStep & );

		~CQStep() { delete [] alias; }

		CQStep & operator=(const CQStep &);
		/*! \brief Swap contents with another instance of CQStep.
		 * \param s - another instance of CQStep.
		 */
		void swap(CQStep& s);
		void Print(char ** str = 0, int *size_left = 0);						// display the step
	private:
		int N(int x) const { if(x == NULL_VALUE_32) return 0; else return x; };	// if x is NullValue returns 0 otherwise x
	};
	////////////////////// Initialization

	CompiledQuery();
	CompiledQuery(CompiledQuery const &);
	CompiledQuery& operator=(CompiledQuery const &);
	~CompiledQuery();

	////////////////////// Add a new step to the execution plan

	/// <summary> Creates a proxy view <param ord="1"/> to a physical table <param ord="2"/> </summary>
	void TableAlias(TabID &t_out, TabID n, const char *tab_name = NULL);
	/// <summary> Creates a filtered view <param ord="1"/> out of the given view <param ord="2"/> </summary>
	void TmpTable(TabID &t_out, TabID t1); // f1 can be ommited
	// filter can be created for source tables (on_result = false) or
	//						 for output columns (on_result = true)
	void CreateFilter(FilterID &f_out, TabID t1, CQTerm e1, Operator pr, CQTerm e2, CQTerm e3=CQTerm(), FilterType type = WHERE_FILTER);
	void And(FilterID &f_out, FilterID f1, FilterID f2, TabID t);
	void Or(FilterID &f_out, FilterID f1, FilterID f2);
	void Not(FilterID &f_out, FilterID f1);
	void Copy(FilterID &f_out, FilterID f1);
	void And(FilterID &f_out, FilterID f1, TabID t, CQTerm e1, Operator pr, CQTerm e2=CQTerm(), CQTerm e3=CQTerm());
	void SwitchOperator(Operator &op);
	// compatibility interfaces, must share type
	typedef void impose_t(TabID const &, FilterID &p_t, FilterID const &p_s, Operator const &p_o, CQTerm const p_v[03], bool);
public:
	//impose_t And, Or;
	//void Or(FilterID &f_out, FilterID f1, CQTerm e1, Operator pr, CQTerm e2=CQTerm(), CQTerm e3=CQTerm());
	void Mode(TabID t1, TMParameter mode, _int64 n1 = 0, _int64 n2 = 0);
	void Join(TabID t1, TabID t2);
	void Define_Param(ParamID &p_out, CQTerm e1);
	void Filter(TabID t1, FilterID f1);
	void AddColumn(AttrID& a_out, TabID t1, CQTerm e1, ColOperation op, char const alias[] = 0, bool distinct = false);

	/*! \brief Create compilation step CREATE_VC for mysql expression
	 * \param a_out - id of created virtual column
	 * \param t1 - id of TempTable for which virtual column is created
	 * \param expr - the expression to be contained in the created column
	 * \param src_table - src_table == t1 if expression works on aggregation results, == NULL_VALUE_32 otherwise
	 * \return void
	 */
	void CreateVirtualColumn(AttrID& a_out, TabID t1, MysqlExpression* expr, TabID src_table = TabID(NULL_VALUE_32));

	/*! \brief Create compilation step CREATE_VC for subquery
	 * \param a_out - id of created virtual column
	 * \param t1 - id of TempTable for which virtual column is created
	 * \param subquery - id of TempTable representing subquery
	 * \param on_result - indicator of which MultiIndex should be passed to Virtual column: one representing source data of t1 (\e false) or
	 * one representing output columns of t1 (\e true), e.g. in case of subquery in HAVING
	 * \return void
	 */
	void CreateVirtualColumn(AttrID& a_out, TabID t1, TabID subquery, bool on_result);
	void CreateVirtualColumn(AttrID& a_out, TabID t1, std::vector<int>& vcs, AttrID vc_id);
	void CreateVirtualColumn(int& a_out, TabID t1, TabID table_alias, AttrID col_number);
	void Add_Order(TabID t1, AttrID a1, int d=0);			// d=1 for descending
	void Add_Order(TabID p_t, ptrdiff_t p_c) {
		int a_c(static_cast<int>(p_c)); if(p_c == a_c) return this->Add_Order(p_t, AttrID(-abs(a_c)), a_c < 0);
	}
	void Union(TabID &t_out, TabID t2, TabID t3, int all=0);
	void Intersect(TabID &t_out, TabID t2, TabID t3, int all=0);
	void Result(TabID t1);

	////////////////////// Informations

	int NoTabs()		{ return no_tabs; }
	int NoAttrs(TabID tab)	{ return no_attrs[-tab.n - 1] ; }
	int NoFilters()		{ return no_filters; }
	int NoParams()		{ return no_params; }
	int NoVirtualColumns(TabID tt) { /*assert(no_virt_cols.find(tt) != no_virt_cols.end());*/ return no_virt_cols[tt]; }
	void Print(char ** str = 0, int *size_left = 0);			// display the CompiledQuery

	int NoSteps()			{ return (int)steps.size(); }
	CQStep& Step(int i)		{ return steps[i]; }

	bool CountColumnOnly(int table);
	bool IsSimpleFilter(int filter);

	/*! \brief Checks if given TempTable has WHERE clause without OR logical connector 
	// \param tab_id - id of TempTable
	// \return bool - true if \e tab_id has WHERE clause without OR, false otherwise
	*/
	bool ExistsWhereFilterWithoutOr(int tab_id);
	
	/*! \brief Checks if given TempTable has WHERE clause
	// \param tab_id - id of TempTable
	// \return bool - true if \e tab_id has WHERE clause, false otherwise
	*/
	bool ExistsWhereFilter(int tab_id);

	/*! \brief Checks if one filter is created as a copy of another filter (step COPY_F)
	 * \param first - source filter
	 * \param second - destination filter
	 * \return true if \e first is copy of \e second, false otherwise
	 */
	bool IsCopyOf(FilterID first, FilterID second);

	/*! \brief verify is given table alias is used in the query defined by a temp table
	 * \param tab_id alias of a searched table
	 * \param tmp_table alias of the temp table to be searched through
	 * returns true if tab_id == tmp_table or tab_id is one of the source tables used in tmp_table
	 */
	bool ExistsInTempTable(TabID tab_id, TabID tmp_table);

	/*! \brief Returns set of all dimensions used by output columns (ADD_COLUMN step) for a given TempTable.
	 * \param table_id - id of given TempTable
	 * \param ta - vector of pointers to JustATables necessary to get access to virtual column inside relevant TempTable
	 * \return set<int>
	 */
	set<int> GetUsedDims(int table_id, vector<JustATablePtr>& ta);

	/*! \brief Finds the first TempTable in compilation steps that contains given table as source table
	 * \param tab_id - id of search table
	 * \param tmp_table - id of TempTable to start search from
	 * \param is_group_by - flag set to true if searched table (one that is returned) represents GROUP BY query
	 * \return returns id of TempTable containing tab_id as source table
	 */
	TabID FindSourceOfParameter(TabID tab_id, TabID tmp_table, bool& is_group_by);

	/*! \brief Finds if TempTable represents query with GroupBy clause
	 * \param tab_id - TempTable to be checked
	 * \return true if it is GroupBy query, false otherwise
	 */
	bool IsGroupByQuery(TabID tab_id);
	std::vector<int> GetOnFilterIds(int tab_id);

	/*! \brief Modifies relevant compilation steps CREATE_FILTER and FILTER_AND by specifying dimension of
	 * LOJ descriptor from this step. Dimension of LOJ descriptor (outer_dim field) is the joined
	 * table having this descriptor in ON clause.
	 * \param tab_id - TempTable representing join
	 * \param joined_tab_id - id of joined table
	 * \param joined_tab_dim - dimension of relevant descriptor
	 * \return void
	 */
	void SetLOJOuterDim(TabID tab_id, TabID joined_tab_id, int joined_tab_dim);
	void FixLOJConditionsTablesOrder(TabID tab_id, TabID joined_tab_id);
	bool IsFilterAppliedToTable(FilterID f, TabID tab_id);
	_int64 FindLimit(int table);				// return LIMIT value for table, or -1
	bool NoAggregationOrOrdering(int table);
	void TakeNoVirtColumns(CompiledQuery* q) {no_virt_cols = q->no_virt_cols;}


private: // NOTE: new objects' IDs (e.g. while declaring new aliases, filters etc.) are reserved and obtained by these functions.
	TabID NextTabID()		{ no_tabs++; no_attrs.push_back(0); return TabID(-no_tabs); }
	AttrID NextVCID(TabID tt) { return AttrID(no_virt_cols[tt]++); }
	AttrID NextAttrID(TabID tab) { no_attrs[-tab.n - 1]++;
								   return AttrID( -no_attrs[-tab.n - 1] ); }
	ParamID NextParamID()	{ return ParamID(no_params++); }	// Actual new objects are NOT created by CompiledQuery, but by a query executor.
	FilterID NextFilterID()	{ return FilterID(no_filters++); }	// IDs of tables and attributes start with -1 and decrease;
	
	/*! \brief Checks if given TabID represents TempTable or just some table alias
	 * \param t - TabID of table to be checked
	 * \return true if t is an alias of TempTable, false otherwise
	 */
	bool IsTempTable(TabID t);

	// IDs of params and filters start with 0 and increase.
	int no_tabs, no_params, no_filters;		// counters: which IDs are not in use?
	std::map<TabID,int> no_virt_cols;

	std::vector<int> no_attrs;	// repository of attr_ids for each table
	std::vector<CompiledQuery::CQStep> steps;			// repository of steps to be executed
	// complex terms can define particular steps
};

inline void SwitchOperator(Operator &op)
{
	if(op==O_LESS)
		op=O_MORE;
	else if(op==O_MORE)
		op=O_LESS;
	else if(op==O_LESS_EQ)
		op=O_MORE_EQ;
	else if(op==O_MORE_EQ)
		op=O_LESS_EQ;
}

void swap(CompiledQuery::CQStep& a, CompiledQuery::CQStep& b);

#endif
