/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "DataCache.h"
#include "core/RCAttr.h"
//#include "core/AttrDataCache.h"


SynchronizedDataCache SynchronizedDataCache::global_data_cache;
#if 0


boost::shared_ptr<AttrDataCache> DataCache::GetAttrDataCache(RCAttr& attr)
{
	return boost::shared_ptr<AttrDataCache>(new AttrDataCache(attr, attr.PackType(), attr.GetNoPackPtr()));
}
#endif

AttrPack* DataCache::GetAttrPackRO(int table, int column, int dp) {
	PackCoordinate coord;
	coord.table = table;
	coord.column = column;
	coord.dp = dp;
	PackContainer::iterator it = m_datapacks.find(coord);

	if (it == m_datapacks.end())
		return NULL;
	else
		return (*it).second;
}

bool DataCache::PutAttrPackRO(int table, int column, int dp, AttrPack *p)
{
	PackCoordinate coord;
	coord.table = table;
	coord.column = column;
	coord.dp = dp;
	std::pair<PackContainer::iterator,bool> result = m_datapacks.insert( std::make_pair(coord, p) );
	if(!result.second && m_datapacks.find(coord)->second != p) {
		AttrPack* z = m_datapacks.find(coord)->second;
		ReleasePack(z);
		result = m_datapacks.insert( std::make_pair(coord, p) );
	}
	return result.second;
}

void DataCache::ReleasePack(int table, int column, int dp)
{
	PackCoordinate coord;
	coord.table = table;
	coord.column = column;
	coord.dp = dp;
	PackContainer::iterator it = m_datapacks.find( coord );
	if( it != m_datapacks.end() ) {
		AttrPack* p = (*it).second;
		m_datapacks.erase(it);
		delete p;
	}
}

void DataCache::ReleasePack(AttrPack* p)
{
	ReleasePack(p->table, p->column, p->dp);
}

void DataCache::ReleasePacks(int table)
{
	PackContainer::iterator it = m_datapacks.begin();
	while(it != m_datapacks.end()) {
		if(it->first.table == table) {
			PackContainer::iterator tmp = it++;
			AttrPack* p = tmp->second;
			m_datapacks.erase(tmp);
			delete p;
		} else
			it++;
	}
}

void DataCache::ReleaseColumnPacks(int table, int column)
{
	PackContainer::iterator it = m_datapacks.begin();
	while(it != m_datapacks.end()) {
		if(it->first.table == table && it->first.column == column) {
			PackContainer::iterator tmp = it++;
			AttrPack* p = tmp->second;
			m_datapacks.erase(tmp);
			delete p;
		} else
			it++;
	}
}

void DataCache::ReleasePackRow(int table, int pack_row)
{
	PackContainer::iterator it = m_datapacks.begin();
	while(it != m_datapacks.end()) {
		if(it->first.table == table && it->first.dp == pack_row) {
			PackContainer::iterator tmp = it++;
			AttrPack* p = tmp->second;
			m_datapacks.erase(tmp);
			delete p;
		} else
			it++;
	}
}

void DataCache::MoveCache(DataCache& local)
{
	PackContainer::iterator it = local.m_datapacks.begin();
	while(it != local.m_datapacks.end()) {
		PackContainer::iterator tmp = it;
		it++;
		if(!tmp->second->IsEmpty()) //TODO<michal> why this causes problems?
			PutAttrPackRO(tmp->first.table, tmp->first.column, tmp->first.dp, tmp->second);
		else
			delete tmp->second;
		local.m_datapacks.erase(tmp);
	}
	local.m_datapacks.clear();
}



AttrPack * 
SynchronizedDataCache::GetAttrPackRO(int table, int column, int dp)
{
	MEMORY_LOCK_GUARD_USE_ANYWHERE
	//IBGuard guard(m_cache_mutex);
	return DataCache::GetAttrPackRO(table,column,dp);
}

// what when the dp is already in the container?
//Currently:	if this is the same one then do nothing
//				if it is new one then replace the old one
bool 
SynchronizedDataCache::PutAttrPackRO(int table, int column, int dp, AttrPack *p)
{
	MEMORY_LOCK_GUARD_USE_ANYWHERE
	//IBGuard guard(m_cache_mutex);
	return DataCache::PutAttrPackRO(table,column,dp,p);
}


// what when dp isn't in the map?
// TBD: does caller unlock pack from use?
void 
SynchronizedDataCache::ReleasePack(int table, int column, int dp)
{
	MEMORY_LOCK_GUARD_USE_ANYWHERE
	//IBGuard guard(m_cache_mutex);
	DataCache::ReleasePack(table,column,dp);
}

void 
SynchronizedDataCache::ReleasePack(AttrPack* p)
{
	MEMORY_LOCK_GUARD_USE_ANYWHERE
	//IBGuard guard(m_cache_mutex);
	DataCache::ReleasePack(p);
}

void 
SynchronizedDataCache::ReleasePacks(int table)
{
	MEMORY_LOCK_GUARD_USE_ANYWHERE
	//IBGuard guard(m_cache_mutex);
	DataCache::ReleasePacks(table);
}

void 
SynchronizedDataCache::ReleaseColumnPacks(int table, int column)
{
	MEMORY_LOCK_GUARD_USE_ANYWHERE
	//IBGuard guard(m_cache_mutex);
	DataCache::ReleaseColumnPacks(table, column);
}

void
SynchronizedDataCache::ReleasePackRow(int table, int pack_row)
{
	MEMORY_LOCK_GUARD_USE_ANYWHERE
	//IBGuard guard(m_cache_mutex);
	DataCache::ReleasePackRow(table,pack_row);
}

void 
SynchronizedDataCache::MoveCache(DataCache& local)
{
	MEMORY_LOCK_GUARD_USE_ANYWHERE
	//IBGuard guard(m_cache_mutex);
	return DataCache::MoveCache(local);
}
