/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "Filter.h"
#include "core/PackOrderer.h"
#include "common/bhassert.h"
#include "tools.h"
#include "bintools.h"

using namespace bh;


///////////////////////////////////////////////////////////////////////////////////////

const int Filter::Block::posOf1[] = {
		8, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
		4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
		5, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
		4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
		6, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
		4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
		5, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
		4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
		7, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
		4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
		5, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
		4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
		6, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
		4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
		5, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
		4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
};

Filter::Block::Block(Filter* owner, int _no_obj, bool all_full)
{
#ifdef FUNCTIONS_EXECUTION_TIMES
	FETOperator fet("Filter::Block::");
#endif
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(_no_obj>0 && _no_obj<=65536);

	CommonInit(owner);
	no_obj = ushort( _no_obj - 1 );
	block_size = (ushort)((NoObj() + 31)/32);
	block_table = 0;
	Clean(all_full);
}

Filter::Block::Block(Block const& block, Filter* owner)
{
#ifdef FUNCTIONS_EXECUTION_TIMES
	FETOperator fet("Filter::Block::");
#endif
	CommonInit(owner);
	block_size = block.block_size;
	no_set_bits = block.no_set_bits;
	no_obj = block.no_obj;
	block_table = 0;
	if(block_size != 0)
	{
		block_table = (uint*)bit_block_pool->malloc();
		memcpy(block_table, block.block_table, block_size*4);
	}
}

Filter::Block::Block(Filter* owner, IBFile* f)
{
    char buf[COMPRESSED_BIT_BLOCK_SIZE];
    BitstreamCompressor bs;

    BHeader bheader;
    CommonInit(owner);

	f->ReadExact(&bheader, sizeof(bheader));

	no_obj = bheader.nob - 1;
	block_size = (ushort)((NoObj() + 31)/32);
	no_set_bits = bheader.no1;
	block_table =(uint*) bit_block_pool->malloc();

    f->ReadExact(buf,bheader.len);

    int res = bs.Decompress((char*)block_table, bheader.len, (char*)buf, bheader.nob, bheader.no1);

	if(process_type != ProcessType::DIM) {
		int test_bits = 0;
		if(!IsNumberOfOnesInHeaderProper(test_bits)) {
			rccontrol.lock() << "WARNING: inconsistent data on disk detected." << unlock;
			no_set_bits = test_bits;
		}
	}

	BHASSERT(res != CPRS_ERR_BUF, "Decompression error in Filter::Block::Block(..)");
}

bool Filter::Block::IsNumberOfOnesInHeaderProper(int& tst)
{	
	int test_bits = 0;		// verify filter pack consistency
	for(int i = 0; i < no_obj + 1; i++)
		if(Get(i)) test_bits++;

	tst = test_bits;
	return (test_bits == no_set_bits);
}

void Filter::Block::CommonInit(Filter * owner)
{
	compress_buffer_size = 0;
	no_set_bits = 0;
	compress_up_to_date = 0;
	this->owner = owner;
	bit_block_pool= owner->GetBitBlockPool();
	b = bitsLeft =0;
	lastn = -2;
	bln = 0;
}

Filter::Block::~Block()
{
#ifdef FUNCTIONS_EXECUTION_TIMES
	FETOperator fet("Filter::Block::");
#endif
	if (block_table)
		bit_block_pool->free(block_table);
}

bool Filter::Block::Set(ushort n)
{
	if(no_set_bits==0)	// just created as empty
	{
		block_table[n/32] |= 1 << n%32;
		no_set_bits = 1;
	}
	else
	{
		if(!(block_table[n/32] & (1 << n%32))) //if this operation change value
		{
			no_set_bits++;
			block_table[n/32] |= 1 << n%32;
		}
	}
	compress_up_to_date = 0;
	return no_set_bits==0 || (no_set_bits-1 == no_obj);	// note that no_set_bits==0 here only when the pack is full
}

bool Filter::Block::Set(ushort  n1, ushort n2)
{
	bool last_result = false;
	for(int i = n1; i <= n2; i++)
		last_result = Set((ushort)i);
	return last_result;
}

bool Filter::Block::Reset(ushort n)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(n<=no_obj);
	if(no_set_bits-1==no_obj)	// just created as full
	{
		block_table[n/32] &= ~(1 << n%32);
		no_set_bits = no_obj;				// note that (no_obj+1) is the actual number of objects
	}
	else
	{
		if((block_table[n/32] & (1 << n%32))) //if this operation change value
		{
			no_set_bits--;
			block_table[n/32] &= ~(1 << n%32);
		}
	}
	compress_up_to_date = 0;
	return (no_set_bits==0);
}

bool Filter::Block::Reset(ushort n1, ushort n2)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(n1 <= n2 && n2 <=no_obj);
	int bl1 = n1/32;
	int bl2 = n2/32;
	if(no_set_bits-1==no_obj)	// just created as full
	{
		if(bl1==bl2)
		{
			for(int i=n1%32;i<=n2%32;i++)
 				block_table[bl1] &= ~(1 << i%32);
		}
		else
		{
			for(int i=n1%32;i<32;i++)
				block_table[bl1] &= ~(1 << i%32);
			for(int i=bl1+1;i<bl2;i++)
				block_table[i] = 0;
			for(int i=0;i<=n2%32;i++)
				block_table[bl2] &= ~(1 << i%32);
		}
		no_set_bits = int(no_obj)+1-(int(n2)-int(n1)+1);		// note that (no_obj+1) is the actual number of objects
	}
	else
	{
		if(bl1==bl2)
		{
			for(int i = n1; i <= n2; i++)
				Reset((ushort)i);
		}
		else
		{
			for(int i=n1%32;i<32;i++)
				if((block_table[bl1] & (1 << i%32))) //if this operation change value
				{
					no_set_bits--;
					if(no_set_bits == 0)
						break;
					block_table[bl1] &= ~(1 << i%32);
				}
				if(no_set_bits>0)
					for(int i=bl1+1;i<bl2;i++)
					{
						no_set_bits -= CalculateBinSum(block_table[i]);
						block_table[i] = 0;
						if(no_set_bits == 0)
							break;
					}
					if(no_set_bits>0)
						for(int i=0;i<=n2%32;i++)
							if((block_table[bl2] & (1 << i%32))) //if this operation change value
							{
								no_set_bits--;
								if(no_set_bits == 0)
									break;
								block_table[bl2] &= ~(1 << i%32);
							}
		}
	}
	compress_up_to_date = 0;
	return (no_set_bits==0);
}

bool Filter::Block::IsEmptyBetween(ushort n1, ushort n2)
{
	int bl1 = n1/32;
	int bl2 = n2/32;
	if(bl1 == bl2)
	{
		if(block_table[bl1] == 0)
			return true;
		for(int i = n1%32; i <= n2%32; i++)
			if((block_table[bl1] & (1 << i)))
				return false;
	}
	else
	{
		int i_start = (n1%32 == 0? bl1 : bl1 + 1);
		int i_stop  = (n2%32 == 31? bl2 : bl2 - 1);
		for(int i = i_start; i <= i_stop; i++)
			if(block_table[i] != 0)
				return false;
		if(bl1 != i_start) {
			for(int i = n1%32; i <= 31; i++)
				if((block_table[bl1] & (1 << i)))
					return false;
		}
		if(bl2 != i_stop) {
			for(int i = 0; i <= n2%32; i++)
				if((block_table[bl2] & (1 << i)))
					return false;
		}
	}
	return true;
}

bool Filter::Block::IsFullBetween(ushort n1, ushort n2)
{
	int bl1 = n1/32;
	int bl2 = n2/32;
	if(bl1 == bl2)
	{
		if(block_table[bl1] == 0xFFFFFFFF)
			return true;
		for(int i = n1%32; i <= n2%32; i++)
			if((block_table[bl1] & (1 << i)) == 0)
				return false;
	}
	else
	{
		int i_start = (n1%32 == 0? bl1 : bl1 + 1);
		int i_stop  = (n2%32 == 31? bl2 : bl2 - 1);
		for(int i = i_start; i <= i_stop; i++)
			if(block_table[i] != 0xFFFFFFFF)
				return false;
		if(bl1 != i_start) {
			for(int i = n1%32; i <= 31; i++)
				if((block_table[bl1] & (1 << i)) == 0)
					return false;
		}
		if(bl2 != i_stop) {
			for(int i = 0; i <= n2%32; i++)
				if((block_table[bl2] & (1 << i)) == 0)
					return false;
		}
	}
	return true;
}

uint Filter::Block::NoOnesBetween(ushort n1, ushort n2)
{
	uint result = 0;
	int bl1 = n1/32;
	int bl2 = n2/32;
	if(bl1 == bl2)
	{
		if(block_table[bl1] == 0)
			return 0;
		for(int i = n1%32; i <= n2%32; i++)
			result += (block_table[bl1] & (1 << i)) ? 1 : 0;
	}
	else
	{
		int i_start = (n1%32 == 0? bl1 : bl1 + 1);
		int i_stop  = (n2%32 == 31? bl2 : bl2 - 1);
		for(int i = i_start; i <= i_stop; i++)
			result += CalculateBinSum(block_table[i]);
		if(bl1 != i_start) {
			for(int i = n1%32; i <= 31; i++)
				result += (block_table[bl1] & (1 << i)) ? 1 : 0;
		}
		if(bl2 != i_stop) {
			for(int i = 0; i <= n2%32; i++)
				result += (block_table[bl2] & (1 << i)) ? 1 : 0;
		}
	}
	return result;
}

bool Filter::Block::IsEqual(Block &b2)
{
	int mn = no_obj / 32;
	for(int n = 0; n < mn; n++) {
		if(block_table[n] != b2.block_table[n])
			return false;
	}
	unsigned int mask = 0xffffffff >> ( 32 - ((no_obj % 32) + 1) );
	if( (block_table[mn] & mask) != (b2.block_table[mn] & mask))
		return false;
	return true;
}

bool Filter::Block::And(Block &b2)
{
	int mn = b2.NoObj() < NoObj() ? b2.NoObj() : NoObj();
	for(int n = 0; n < mn; n++)
	{
		if(Get((ushort)n) && !b2.Get((ushort)n))
			Reset((ushort)n);
	}
	return (no_set_bits==0);
}

bool Filter::Block::Or(Block &b2)
{
	int mn = b2.NoObj() < NoObj() ? b2.NoObj() : NoObj();
	for(int n = 0; n < mn; n++)
	{
		if(!Get((ushort)n) && b2.Get((ushort)n))
			Set((ushort)n);
	}
	return no_set_bits==0 || (no_set_bits-1==no_obj);			// 0 here means only a full pack, other possibilities are excluded by an upper level
}

void Filter::Block::Not()
{
	ushort new_set_bits;
	if(no_obj != 65535)
		new_set_bits = no_obj - no_set_bits + 1;
	else
		new_set_bits = ~no_set_bits + 1;
	if(NoObj()%32)
		for(uint n = NoObj()/32 * 32; n < NoObj(); n++)
			if(Get((ushort)n))
				Reset((ushort)n);
			else
				Set((ushort)n);
	for(uint b = 0; b < NoObj()/32; b++)
		block_table[b] = ~(block_table[b]);

	no_set_bits = new_set_bits;
	compress_up_to_date = 0;
}

bool Filter::Block::Reset(Block &b2)
{
	int mn = b2.NoObj() < NoObj() ? b2.NoObj() : NoObj();
	for(int n = 0; n < mn; n++)
	{
		if(Get((ushort)n) && b2.Get((ushort)n))
			Reset((ushort)n);
	}
	return (no_set_bits==0);
}

bool Filter::Block::GetNext1(int &n) // the next iterator position
{
	if(n>no_obj) {
		lastn = -2;
		bln = 0;
		return false;
	}
	if(lastn+1 == n) {
		if(bitsLeft>0)
			bitsLeft--;
		b >>= 1;
	} else {
		bln = n >> 5;
		uint bitno = n & 0x1f;
		b = block_table[bln];
		b >>= bitno;
		bitsLeft = 32 - bitno;
	}
	while(1) {
		// find bit == 1
		// first find a portion with 1
		if(b == 0) {
			n+= bitsLeft;
			if(++bln >= block_size) {
				lastn = -2;
				return false;
			}
			while(block_table[bln]==0) {
				n+=32;
				if(n>no_obj) {
					lastn = -2;
					return false;
				}
				++bln;
			}
			b = block_table[bln];
			bitsLeft=32;
		}

		while(1) {
			//process 32bit portion
			int skip = posOf1[b & 0xff];
			if(skip > bitsLeft) {
				//end of 32bit portion
				n+= bitsLeft;
				if(n>no_obj) {
					lastn = -2;
					return false;
				}
				break;
			} else {
				n+= skip;
				if(n>no_obj) {
					lastn = -2;
					return false;
				}
				bitsLeft -= skip;
				b >>= skip;
				if(skip < 8) {
					// found 1
					lastn = n;
					return true;
				}
			}
		}
	}
}

bool Filter::Block::GetNext0(int &n)		// the next iterator position
{
	if(n>no_obj)
		return false;
	int bln = n/32;
	uint mask = (uint(1) << n%32);
	do
	{
		if(((~(block_table[bln])) & mask) != 0)		// zero found
		{
			return true;
		}
		n++;
		if(n>no_obj)
			return false;
		mask <<= 1;
		if(mask==0)			// end of 32-bit portion
		{
			mask=1;
			bln++;
			while(block_table[bln]==0xFFFFFFFF)
			{
				n+=32;
				if(n>no_obj)
					return false;
				bln++;
			}
		}
	}
	while(1);
	return false;
}

int Filter::Block::Find1(int k)				// note: k is now increased by 1, i.e. k=1 means the first 1
{
	int ret = -1;
	if(k - 1 > no_obj || k > no_set_bits)
		return ret;
	int count1 = 0;
	int b = 0;
	uint umax = ~(uint)0;
	while(count1 < k && b <= block_size)
	{
		if(block_table[b] == umax && b != block_size)
		{
			if(count1 + 32 >= k)
			{
				ret = b * 32 + (k - count1) - 1;
				break;
			}
			count1 += 32;
		}
		else
		{
			int n = 0;
			while(count1 < k && n < (int)(b == block_size ? (NoObj() - b * 32) : 32))
				if(block_table[b] & 1 << n++)
					count1++;
			if(count1 == k)
				ret = b * 32 + n - 1;
		}
		b++;
	}
	return ret;
}

void Filter::Block::Clean(bool value)
{
	if(!block_table)
	{
		block_table =(uint*) bit_block_pool->malloc(); //(uint*)alloc(block_size*sizeof(uint), BLOCK_UNCOMPRESSED, owner);
	}
	if(value)
	{
		memset(block_table, 255, Filter::bitBlockSize);
		no_set_bits = no_obj + 1;					// Note: when the block is full, the overflow occurs and no_set_bits=0.
	}
	else
	{
		memset(block_table, 0, Filter::bitBlockSize);
		no_set_bits = 0;
	}
	dirty = 0;
}

//set bits to value, do not pay attention to previous values
void Filter::Block::InitBits(ushort  n1, ushort n2, bool value)
{
	if (value)
		for(int i = n1; i <= n2; i++) {
				block_table[i/32] |= 1 << i%32;
				no_set_bits ++;
		}
	else
		for(int i = n1; i <= n2; i++)
				block_table[i/32] &= ~(1 << i%32);

	compress_up_to_date = 0;
}

void Filter::Block::GrowBlock(int grow_size, bool value)
{
	BHASSERT(NoObj() + grow_size <= 0x10000, "Filter::Block grow beyond pack size");
	InitBits(no_obj+1, no_obj+ grow_size, value);
	no_obj+= grow_size;
	block_size = (ushort)((NoObj() + 31)/32);
}

void Filter::Block::ShrinkBlock(int new_size)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(new_size <= no_obj+1);
	if (new_size == no_obj+1)
		return;
	Reset(new_size, std::min<ushort>(0xFFFF, no_obj));
	no_obj = new_size - 1;
	block_size = (ushort)((NoObj() + 31)/32);
}

void Filter::Block::Store(IBFile &f)
{
	char buf[COMPRESSED_BIT_BLOCK_SIZE];
	BitstreamCompressor bs;
	BHeader bheader;

	bheader.nob = NoObj();
	bheader.no1 = no_set_bits;
	bheader.len = COMPRESSED_BIT_BLOCK_SIZE;

	int res = bs.Compress((char*)buf, bheader.len, (char*)block_table, bheader.nob, bheader.no1);
	BHASSERT(res != CPRS_ERR_BUF, "Compression error in Filter::Block::Store()");
#ifndef NDEBUG
	int test = 0;		// verify filter pack consistency
	for(int i = 0; i < bheader.nob; i++)
		if(Get(i)) test++;
	BHASSERT(test == bheader.no1, "Wrong filter contents");
#endif

	f.WriteExact(&bheader, sizeof(bheader));
	f.WriteExact(buf,bheader.len);
}

///////////////////////////////////////////////////////////////////////////////////////

FilterOnesIteratorOrdered::FilterOnesIteratorOrdered() : FilterOnesIterator(), po(NULL) {};

FilterOnesIteratorOrdered::FilterOnesIteratorOrdered(Filter *ff, PackOrderer* po)
{
	Init(ff,po);
}

void FilterOnesIteratorOrdered::Init(Filter *ff, PackOrderer* po)
{
	FilterOnesIterator::Init(ff);
	this->po = po;
	FilterOnesIteratorOrdered::Rewind();
}

void FilterOnesIteratorOrdered::Rewind()
{
    Reset();
    po->Rewind();
    FilterOnesIteratorOrdered::operator++();
}

FilterOnesIteratorOrdered& FilterOnesIteratorOrdered::operator++()
{

	//the code is copied from FilterOnesIterator::operator++(),
	// only iterator_b++ and end-of-iteration checks are changed.
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(valid);		// cannot iterate invalid iterator
	if(iterator_n == 65535 || (iterator_b == f->no_blocks - 1 && iterator_n == f->no_of_bits_in_last_block - 1)) {
		iterator_n = 0;		// switching to the next block
		iterator_b = (++(*po)).Current();
		if(!po->IsValid()) {
			valid = false;
			return *this;
		}
	} else
		iterator_n++; // now we are on the first suspected position - go forward if it is not 1

	bool found = false;
	while(!found) {
		if(f->block_status[iterator_b] == f->FB_FULL) {
//			if( po->IsLast() && (iterator_n == 65536 ||
//				(iterator_b == f->no_blocks - 1 && iterator_n == f->no_of_bits_in_last_block))) {// end of filter
//				valid = false;
//				return *this;
//			}
			found = true;
		} else if(f->block_status[iterator_b] == f->FB_EMPTY) {
			iterator_b = (++(*po)).Current();
			iterator_n = 0;
			if(!po->IsValid()) {
				valid = false;
				return *this;
			}
		} else {
			///////////////////// inter-block iteration //////////////////////////
			Filter::Block *cur_block = f->blocks[iterator_b];
			if(iterator_b != prev_block) {
				b = bitsLeft = 0;
				lastn = -2;
				bln = 0;
				prev_block = iterator_b;
				ones_left_in_block = cur_block->no_set_bits;
			} else {		// we already were in this block and the previous bit is after us
				ones_left_in_block--;
			}

			if(iterator_n <= cur_block->no_obj) {
				// else leave "found == false", which will iterate to the next block
				if(lastn + 1 == iterator_n) {
					if(bitsLeft > 0)
						bitsLeft--;
					b >>= 1;
				} else {
					bln = iterator_n >> 5;
					uint bitno = iterator_n & 0x1f;
					b = cur_block->block_table[bln];
					b >>= bitno;
					bitsLeft = 32 - bitno;
				}
				// find bit == 1
				while(!found && iterator_n <= cur_block->no_obj) {
					// first find a portion with 1
					if(b == 0) {
						iterator_n += bitsLeft;
						if(++bln >= cur_block->block_size)
							break;				// leave found == false
						while(cur_block->block_table[bln] == 0) {
							iterator_n += 32;
							if(iterator_n > cur_block->no_obj)
								break;
							++bln;
						}
						if(iterator_n > cur_block->no_obj)
							break;
						b = cur_block->block_table[bln];
						bitsLeft=32;
					}

					while(iterator_n <= cur_block->no_obj) {
						// non-zero value found - process 32bit portion
						int skip = cur_block->posOf1[b & 0xff];
						if(skip > bitsLeft) {
							//end of 32bit portion
							iterator_n += bitsLeft;
							break;
						} else {
							iterator_n += skip;
							if(iterator_n > cur_block->no_obj)
								break;
							bitsLeft -= skip;
							b >>= skip;
							if(skip < 8) {
								// found 1
								lastn = iterator_n;
								found = true;
								break;
							}
						}
					}
				}
			}
			//////////////////////////////////////////////////////////////////////
			if(!found) {
				iterator_b = (++(*po)).Current();
				iterator_n = 0;
				if(!po->IsValid()) {
					valid = false;
					return *this;
				}
			}
		}
	}
	cur_position = ((_int64(iterator_b)) << 16) + iterator_n;
	return *this;
}


FilterOnesIteratorOrdered* FilterOnesIteratorOrdered::Copy()
{
	FilterOnesIteratorOrdered* f = new FilterOnesIteratorOrdered();
	*f = *this;
	f->po = new PackOrderer();
	*(f->po) = *po;
	return f;
}

void FilterOnesIteratorOrdered::NextPack()
{
	iterator_b = (++(*po)).Current();

	while( po->IsValid() && f->block_status[iterator_b] == f->FB_EMPTY)
		iterator_b = (++(*po)).Current();

	if(po->IsValid()) {
		iterator_n = 0;
		if(f->block_status[iterator_b] == f->FB_FULL) {
			return;
		} else {
			ones_left_in_block = f->blocks[iterator_b]->no_set_bits;
			if( !( f->blocks[iterator_b]->block_table[0] & 1))
				FilterOnesIteratorOrdered::operator++();
		}
	} else
		valid = false;
}


///////////////////////////////////////////////////////////////////////////////////////
FilterOnesIterator::FilterOnesIterator()
{
	valid = false;
	f = NULL;
	cur_position = -1;
	b = lastn = bln = bitsLeft = 0;
	prev_block = -1;
	iterator_n = -1;
	iterator_b = -1;
	inside_one_pack = true;
}

FilterOnesIterator::FilterOnesIterator(Filter *ff)
{
	Init(ff);
}

void FilterOnesIterator::Init(Filter *ff)		//!< reinitialize iterator (possibly with the new filter)
{
	f = ff;
	inside_one_pack = true;
	bool nonempty_pack_found = false;
	iterator_b = 0;
	while(iterator_b < f->no_blocks) { 			// check whether there is more than one pack involved
		if(f->block_status[iterator_b] != f->FB_EMPTY) {
			if(nonempty_pack_found) {
				inside_one_pack = false;
				break;
			}
			nonempty_pack_found = true;
		}
		iterator_b++;
	}
	FilterOnesIterator::Rewind();									//!< initialize on the beginning
}

void FilterOnesIterator::Reset()
{
    cur_position = -1;
    b = lastn = bln = bitsLeft = 0;
    prev_block = -1;
    valid = true;
    iterator_n = 65535;
    iterator_b = -1;
}

void FilterOnesIterator::Rewind()				//!< iterate from the beginning (the first 1 in Filter)
{
    Reset();
    FilterOnesIterator::operator++();
}

void FilterOnesIterator::Rewind(int pack)	// iterate from the beginning of the given pack (the first 1 in Filter)
											// note: if the pack is empty, set IsValid() to false
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(pack >= 0 && pack < f->no_blocks);
	if(f->block_status[iterator_b] == f->FB_EMPTY) {
		valid = false;
		return;
	}
	valid = true;
	iterator_n = 65535;		// simulate the end of the previous pack
	iterator_b = pack - 1;
    FilterOnesIterator::operator++();
}

void FilterOnesIterator::NextPack()
{
	iterator_b++;
	while(iterator_b < f->no_blocks && f->block_status[iterator_b] == f->FB_EMPTY)
		iterator_b++;
	if(iterator_b < f->no_blocks)
		FilterOnesIterator::Rewind(iterator_b);
	else
		valid = false;
}

_int64 FilterOnesIterator::GetPackSizeLeft()	// how many 1-s in the current block left (including the current one)
{
	if(!valid)
		return 0;
	if(f->block_status[iterator_b] == f->FB_FULL) {
		if( iterator_b == f->no_blocks - 1) {
			return f->no_of_bits_in_last_block - iterator_n;
		}
		return 65536 - iterator_n;
	}
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(f->block_status[iterator_b] != f->FB_EMPTY);
	return ones_left_in_block;
}

FilterOnesIterator& FilterOnesIterator::operator++()
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(valid);		// canot iterate invalid iterator
	if(iterator_n == 65535) {
		iterator_n = 0;
		iterator_b++;
		if(iterator_b == f->no_blocks) {
			valid = false;
			return *this;
		}
	} else
		iterator_n++; // now we are on the first suspected position - go forward if it is not 1

	bool found = false;
	while(!found) {
		if(f->block_status[iterator_b] == f->FB_FULL) {
			if( iterator_b == f->no_blocks - 1 &&
				iterator_n == f->no_of_bits_in_last_block) {// end of filter
				valid = false;
				return *this;
			}
			found = true;
		} else if(f->block_status[iterator_b] == f->FB_EMPTY) {
			iterator_b++;
			iterator_n = 0;
			if(iterator_b == f->no_blocks) {
				valid = false;
				return *this;
			}
		} else {
			///////////////////// inter-block iteration //////////////////////////
			Filter::Block *cur_block = f->blocks[iterator_b];
			if(iterator_b != prev_block) {
				b = bitsLeft = 0;
				lastn = -2;
				bln = 0;
				prev_block = iterator_b;
				ones_left_in_block = cur_block->no_set_bits;
			} else {		// we already were in this block and the previous bit is after us
				ones_left_in_block--;
			}

			if(iterator_n <= cur_block->no_obj) {
				// else leave "found == false", which will iterate to the next block
				if(lastn + 1 == iterator_n) {
					if(bitsLeft > 0)
						bitsLeft--;
					b >>= 1;
				} else {
					bln = iterator_n >> 5;
					uint bitno = iterator_n & 0x1f;
					b = cur_block->block_table[bln];
					b >>= bitno;
					bitsLeft = 32 - bitno;
				}
				// find bit == 1
				while(!found && iterator_n <= cur_block->no_obj) {
					// first find a portion with 1
					if(b == 0) {
						iterator_n += bitsLeft;
						if(++bln >= cur_block->block_size)
							break;				// leave found == false
						while(cur_block->block_table[bln] == 0) {
							iterator_n += 32;
							if(iterator_n > cur_block->no_obj)
								break;
							++bln;
						}
						if(iterator_n > cur_block->no_obj)
							break;
						b = cur_block->block_table[bln];
						bitsLeft=32;
					}

					while(iterator_n <= cur_block->no_obj) {
						// non-zero value found - process 32bit portion
						int skip = cur_block->posOf1[b & 0xff];
						if(skip > bitsLeft) {
							//end of 32bit portion
							iterator_n += bitsLeft;
							break;
						} else {
							iterator_n += skip;
							if(iterator_n > cur_block->no_obj)
								break;
							bitsLeft -= skip;
							b >>= skip;
							if(skip < 8) {
								// found 1
								lastn = iterator_n;
								found = true;
								break;
							}
						}
					}
				}
			}
			//////////////////////////////////////////////////////////////////////
			if(!found) {
				iterator_b++;
				iterator_n = 0;
				if(iterator_b == f->no_blocks) {
					valid = false;
					return *this;
				}
			}
		}
	}
	cur_position = ((_int64(iterator_b)) << 16) + iterator_n;
	return *this;
}

FilterOnesIterator* FilterOnesIterator::Copy()
{
	FilterOnesIterator* f = new FilterOnesIterator();
	*f = *this;
	return f;
}
