/* Copyright (C)  2005-2009 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "Joiner.h"
#include "Descriptor.h"
#include "vc/VirtualColumn.h"
#include "MIUpdatingIterator.h"

int JoinerGeneral::ExecuteJoinConditions(vector<Descriptor> &desc)
{
	int no_desc = desc.size();
	//if(desc[0].val1.vc->IsConst() && desc[0].val2.vc == NULL) {
	//	// Special case: if there is a chance for one-dimensional filtering, execute one condition only.
	//	no_desc = 1;
	//}
	DimensionVector all_dims(mind);		// "false" for all dimensions
	vector<bool> pack_desc_locked;
	bool false_desc_found = false;
	for(int i = 0; i < no_desc; i++) {
		if(desc[i].IsFalse())
			false_desc_found = true;
		pack_desc_locked.push_back(false);
		desc[i].DimensionUsed(all_dims);
	}
	mind->MarkMaterialClasses(all_dims);		// e.g. (t1 lj t2) join (t3 lj t4) - find and add results of previous joins
	if(false_desc_found) {
		for(int i = 0; i < mind->NoDimensions(); i++)
			if(all_dims[i])
				mind->Empty(i);
		return no_desc;			// all done
	}
	MIUpdatingIterator mit(mind, all_dims);
	_int64 tuples_in_output = 0;
	bool loc_result;
	bool stop_execution = false;	// early stop for LIMIT
	rccontrol.lock(m_conn.GetThreadID()) << "Starting joiner loop (" << mit.NoTuples() << " rows)." << unlock;
	for(int i = 0; i < no_desc; i++) {
		if(RequiresUTFConversions(desc[i].GetCollation()))
			desc[i].InitUTFBuffers();
	}
	// The main loop for checking conditions
	while(mit.IsValid()) {
		if(mit.PackrowStarted()) {
			if(stop_execution) {
				mit.ResetCurrentPack();
				mit.NextPackrow();
				continue;
			}
			for(int i = 0; i < no_desc; i++)
				pack_desc_locked[i] = false;		// delay locking
		}
		loc_result = true;
		for(int i = 0; i < no_desc; i++) {
			if(!pack_desc_locked[i])				// delayed locking - maybe will not be locked at all?
				desc[i].LockSourcePacks(mit);
			if(RequiresUTFConversions(desc[i].GetCollation())) {
				if(desc[i].CheckCondition_UTF(mit) == false) {
					loc_result = false;
					break;
				}
			} else {
				if(desc[i].CheckCondition(mit) == false) {
					loc_result = false;
					break;
				}
			}
		}
		if(loc_result == false) {
			if(!count_only_now)
				mit.ResetCurrent();
		} else
			tuples_in_output++;
		++mit;
		if(m_conn.killed())
			throw KilledRCException();
		if(output_limit > -1 && tuples_in_output >= output_limit)
			stop_execution = true;
	}

	// Postprocessing and cleanup
	if(count_only_now)
		mind->MakeCountOnly(tuples_in_output * mit.Factor());
	else
		mit.Commit();
	for(int i = 0; i < no_desc; i++) {
		desc[i].UnlockSourcePacks();
		desc[i].done = true;
	}
	why_failed = NOT_FAILED;
	return no_desc;
}

