/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef JOINERHASH_H_
#define JOINERHASH_H_

#include "Joiner.h"
#include "JoinerHashTable.h"

class JoinerHash : public TwoDimensionalJoiner
{
	/*
	 * There are two sides of join:
	 * - "traversed", which is put partially (chunk by chunk) into the hash table,
	 * - "matched", which is scanned completely for every chunk gathered in the hash table.
	 *
	 * Algorithm:
	 * 1. determine all traversed and matched dimensions,
	 * 2. create hash table,
	 * 3. traverse the main "traversed" dimension and put key values into the hash table,
	 * 4. put there also information about row numbers of all traversed dimensions
	 *    (i.e. the main one and all already joined with it),
	 * 5. scan the "matched" dimension, find the key values in the hash table,
	 * 6. submit all the joined tuples as the result of join
	 *    (take all needed tuple numbers from the hash table and "matched" part of multiindex),
	 * 7. if the "traversed" side was not fully scanned, clear hash table and go to 4 with the next chunk.
	 *
	 */
public:
	JoinerHash( MultiIndex *_mind, RoughMultiIndex *_rmind,	bool _count_only_now, _int64 _output_limit);
	~JoinerHash();

	int ExecuteJoinConditions(std::vector<Descriptor> &_desc);

private:
	_int64 TraverseDim(MIIterator &mit);
	_int64 MatchDim(MINewContents &new_mind);

	void ExecuteJoin();

	//////////////////////////////////////////////////////
	JoinerHashTable jhash;

	/////////////// dimensions description ///////////////
	DimensionVector traversed_dims;	// the mask of dimension numbers of traversed dimensions to be put into the join result
	DimensionVector matched_dims;		// the mask of dimension numbers of matched dimensions to be put into the join result
	int no_of_traversed_dims;
	vector<int> traversed_hash_column;	// the number of hash column containing tuple number for this dimension

	VirtualColumn **vc1;
	VirtualColumn **vc2;
	int cond_hashed;

	bool other_cond_exist;				// if true, then check of other conditions is needed on matching
	std::vector<Descriptor> other_cond;

	// Statistics
	_int64 packrows_omitted;			// roughly omitted by by matching
	_int64 packrows_matched;

	/////////////////////////////////////////////////////

	void SubmitJoinedTuple(_int64 hash_row, MIIterator &mit, MINewContents &new_mind);
};




#endif /*JOINERHASH_H_*/
