/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef _JUST_A_TABLE_H_
#define _JUST_A_TABLE_H_

#include <memory>
#include <vector>
#include <boost/shared_ptr.hpp>
#include <boost/weak_ptr.hpp>
#include <boost/enable_shared_from_this.hpp>

#ifndef __BH_COMMUNITY__
#	include "enterprise/local_fwd.h"
#else /* not __BH_COMMUNITY__ */
#	include "community/local_fwd.h"
#endif /* __BH_COMMUNITY__ */
#include "Filter.h"
#include "MysqlExpression.h"

enum TType { RC_TABLE, TEMP_TABLE };

class ConnectionInfo;
class RCBString;
class AttributeTypeInfo;
class PhysicalColumn;

class TempTable;
class Descriptor;


typedef AttributeTypeInfo ATI;

class JustATable : public boost::enable_shared_from_this<JustATable>
{
public:

	int NoPacks() { return bh::common::NoObj2NoPacks(NoObj()); }
	static unsigned PackIndex(_uint64 obj)
	{ return (obj==NULL_VALUE_64 ? 0xFFFFFFFF : (unsigned)(obj >> 16) ); }	// null pack number (interpreted properly)
	virtual void LockPackForUse(unsigned attr, unsigned pack_no, ConnectionInfo& conn) = 0;
	virtual void UnlockPackFromUse(unsigned attr, unsigned pack_no) = 0;
	virtual _uint64 NoObj() = 0;
	virtual uint NoAttrs() = 0;
	virtual uint NoDisplaybleAttrs() = 0;
	virtual TType TableType() const = 0;
//	virtual void SortStatistics(Filter *f, int n_a, class SortEncoder *stat, bool multiple, bool approx_only, ConnectionInfo *conn) = 0;
//	virtual _uint64  SortEncode(_uint64 obj, int n_a, SortEncoder *stat) = 0;

	virtual int GetTable32(_uint64 obj, int attr) = 0;
	virtual _int64 GetTable64(_uint64 obj, int attr ) = 0;
	virtual RCBString GetTable_S(_uint64 obj, int attr, bool bin_as_hex = true) = 0;
	virtual void GetTable_B(_uint64 obj, int attr, int &size, char *val_buf) = 0;
	virtual int GetLength(_uint64 obj, int attr, ConnectionInfo *conn) = 0;

	virtual bool IsNull(_uint64 obj, int attr) = 0;
	/*
	virtual _uint64 ApproxDistinctVals(int attr, Filter *f=NULL) = 0;	// VC obsolete: provide the best upper approximation of number of diff. values (incl. null)
	virtual _uint64 ApproxAnswerSize(int attr, Descriptor &d) = 0;	// provide the most probable approximation of number of objects matching the condition
	virtual _int64 AttrSum(int n_a, Filter *f, _int64 &counter, _int64 *quant = 0, int q_pack_start = 0, int q_pack_stop = 0, _int64 quant_factor = 1, bool distinct = false,ConnectionInfo *conn=NULL) = 0;
	virtual _int64 AttrMin(int n_a, Filter *f, ConnectionInfo *conn) = 0;
	virtual _int64 AttrMax(int n_a, Filter *f, ConnectionInfo *conn) = 0;
	virtual double AttrAvg(int n_a, Filter *f, _int64* quant = 0, bool distinct = false,ConnectionInfo *conn=NULL) = 0;
	typedef double AttrAvg_vt(int, Filter *, _int64 *, bool);
	virtual _int64 AttrCountNotNull(int n_a, Filter *f, _int64 *quant = 0, int q_pack_start = 0, int q_pack_stop = 0, ConnectionInfo *conn=NULL) = 0;
	*/
	virtual RSValue	AttrIsDistinct(int n_a) = 0;					// three-valued answer

//	virtual RCBString AttrMinS(int n_a, Filter *f, ConnectionInfo *conn) = 0;
//	virtual RCBString AttrMaxS(int n_a, Filter *f, ConnectionInfo *conn) = 0;
//	virtual _int64 RoughMin(int n_a, Filter *f = NULL) = 0;		// for numerical: best rough approximation of min for a given filter (or global min if filter is NULL)
//	virtual _int64 RoughMax(int n_a, Filter *f = NULL) = 0;		// for numerical: best rough approximation of max for a given filter (or global max if filter is NULL)
	virtual ushort MaxStringSize(int n_a, Filter *f = NULL) = 0;

	virtual AttributeType AttrType(int n_a) = 0;
	virtual std::vector<ATI> GetATIs() = 0;
	virtual int GetAttrScale(int n_a) = 0;
	virtual int GetAttrSize(int n_a) = 0;
	virtual int GetFieldSize(int n_a) = 0;
	virtual NullMode GetAttrNullMode(int n_a) = 0;

	//virtual bool CheckCondition(_uint64 obj, Descriptor &d,ConnectionInfo *conn=NULL) = 0; // checks if d is satisfied by obj
	//virtual void CheckPackCondition(int pack, Filter &f, Descriptor& d,ConnectionInfo *conn) = 0;
	int SortPack(int pack, Filter *f, _int64 *index, int attr, ConnectionInfo *conn = NULL);
	bool CheckJoinCondition(_int64 obj1,int attr1,JustATable &sec,_int64 obj2,int atttr2,Operator op, ConnectionInfo *conn=NULL, _int64 factor1=1,_int64 factor2=1,_int64 vmax=PLUS_INF_64);	// WARNING: packs (if any) must be locked externally!
	//virtual void FilterAttrsAnd(Filter &, Descriptor&,ConnectionInfo *conn=NULL) = 0;
	virtual const PhysicalColumn* GetColumn(int col_no) = 0;

	virtual const ColumnType& GetColumnType(int n_a)	= 0;

	//! Returns column value in the form required by complex expressions
	ValueOrNull GetComplexValue(const _uint64 obj, const int attr);

	virtual ~JustATable(){};
private:
	void QuickSortPack(_int64 *index,int attr,int s1,int s2,ConnectionInfo *conn=NULL);
	int Compare(int attr,_int64 obj1,_int64 obj2,ConnectionInfo *conn);

public:
	class RecordIteratorInterface;
	template<typename lock_policy_t>
	class RecordIteratorImplementation;

	struct TempTableLockPolicy
	{
		typedef TempTable table_t;
		typedef RecordIteratorImplementation<TempTableLockPolicy> iterator_t;
		static inline void Destroy( iterator_t& ) {}
		static inline void LockPacks( iterator_t&, _uint64 const& ) {}
		static inline void UnlockPacks( iterator_t&, _uint64 const& ) {}
		static void Next( iterator_t& it );
		static iterator_t& Assign( iterator_t&, iterator_t const& );
		typedef struct DataTag
		{
			DataTag( table_t* table = NULL ) : table( table ) {}
			DataTag( DataTag const& d ) : table( d.table ) {}
			table_t* table;
			/*! \brief Get number of columns.
			 */
			uint Size();
		} data_t;
		typedef struct InitializerTag
		{
			table_t* table;
			ConnectionInfo* conn;
			InitializerTag( table_t* table, ConnectionInfo* conn )
				: table( table ), conn( conn ) {}
			InitializerTag( InitializerTag const& i )
				: table( i.table ), conn( i.conn ) {}
		} initializer_t;
	};

	struct RcTableLockPolicy
	{
		typedef FilterPtr filter_ptr_t;
		typedef RCTableImpl table_t;
		typedef boost::weak_ptr<JustATable> table_ptr_t;
		typedef RecordIteratorImplementation<RcTableLockPolicy> iterator_t;
		typedef std::vector<int> attr_uses_t;
		static void Destroy( iterator_t& );
		static void LockPacks( iterator_t&, _uint64 const& );
		static void UnlockPacks( iterator_t&, _uint64 const& );
		static void Next( iterator_t& );
		static iterator_t& Assign( iterator_t&, iterator_t const& );
		typedef struct DataTag
		{
			DataTag( table_ptr_t table = table_ptr_t() ) : table( table ), locked( false ) {}
			DataTag( DataTag const& d ) : table( d.table ), locked( d.locked ), attr_uses( d.attr_uses ), filter( d.filter ), it( d.it ) {}
			table_ptr_t table;
			bool locked;
			attr_uses_t attr_uses;
			FilterPtr filter;
			Filter::iterator it;
			/*! \brief Get number of columns.
			 */
			uint Size();
		} data_t;
		typedef struct InitializerTag
		{
			table_ptr_t table;
			ConnectionInfo* conn;
			FilterPtr filter;
			attr_uses_t attr_uses;
			InitializerTag( table_ptr_t table, ConnectionInfo* conn, FilterPtr filter, attr_uses_t const& attr_uses )
				: table( table ), conn( conn ), filter( filter ), attr_uses( attr_uses ) {}
			InitializerTag( InitializerTag const& i )
				: table( i.table ), conn( i.conn ), filter( i.filter ), attr_uses( i.attr_uses ) {}
		} initializer_t;
	};

	/*! \brief A record interface
	 */
	class RecordInterface
	{
	public:
		typedef std::auto_ptr<RecordInterface> record_ptr_t;
		virtual ~RecordInterface() {}
		RCValueObject get_item( uint item ) { return ( do_get_item( item ) ); }
		_uint64 GetRowId() { return ( do_get_row_id() ); }
		uint size() const { return ( do_size() ); }
		record_ptr_t clone() const { return do_clone(); }
	protected:
		virtual RCValueObject do_get_item( uint ) = 0;
		virtual _uint64 do_get_row_id() = 0;
		virtual uint do_size() const = 0;
		virtual record_ptr_t do_clone() const = 0;
	};

	class Record
	{
		Record( RecordInterface::record_ptr_t record ) : _record( record ) {}
	public:
		Record( Record const& r ) : _record( r._record.get() ? r._record->clone() : RecordInterface::record_ptr_t() ) {}
		Record& operator = ( Record const& r )
		{
			if ( &r != this ) {
				if( r._record.get() )
					_record = r._record->clone();
				else
					_record.reset();
			}
			return ( *this );
		}
		RCValueObject operator[]( uint item ) const { return ( _record->get_item( item ) ); }
		_uint64 GetRowId() { return ( _record->GetRowId() ); }
		uint size() const { return ( _record->size() ); }
	protected:
		RecordInterface::record_ptr_t _record;
		friend class RecordIteratorImplementation<RcTableLockPolicy>;
		friend class RecordIteratorImplementation<TempTableLockPolicy>;
	};

	template<typename table_iterator_t>
	class RecordImplementation : public RecordInterface
	{
	public:
		RecordImplementation( table_iterator_t& it, _uint64 atRow ) : _it( it ), _atRow( atRow ) {}
		virtual ~RecordImplementation() {}
	protected:
		virtual RCValueObject do_get_item( uint );
		virtual _uint64 do_get_row_id()
		{
			return ( _atRow );
		}

		virtual uint do_size() const
		{
			return ( _it._data.Size() );
		}

		virtual record_ptr_t do_clone() const
		{
			return record_ptr_t( new RecordImplementation<table_iterator_t>( _it, _atRow ) );
		}

	private:
		RecordImplementation( RecordImplementation const& );
		table_iterator_t& _it;
		_uint64 _atRow;
	};

	/*! \brief Interface for RecordIterator classes.
	 */
	class RecordIteratorInterface
	{
	public:
		typedef std::auto_ptr<RecordIteratorInterface> iterator_ptr_t;
		virtual ~RecordIteratorInterface( void ) {}
		bool operator == ( RecordIteratorInterface const& i )
		{
			return ( do_equals( i ) );
		}
		bool operator != ( RecordIteratorInterface const& i ) { return ( ! operator == ( i ) ); }
		void next( void ) {	do_next(); }
		Record operator*() { return ( do_get_record() ); }
		iterator_ptr_t clone() const { return do_clone(); }
	protected:
		virtual iterator_ptr_t do_clone() const = 0;
		virtual bool do_equals( RecordIteratorInterface const& ) const = 0;
		virtual void do_next( void ) = 0;
		virtual Record do_get_record() = 0;
	};

	/*! \brief RecordIterator class for _table records.
	 */
	template<typename lock_policy_t>
	class RecordIteratorImplementation : public RecordIteratorInterface
	{
	public:
		friend class JustATable;
		friend class TempTable;
		USE_THIS_MACRO_TO_DO_RC_TABLE_A_FIERND
		typename lock_policy_t::data_t _data;
		_uint64 _currentRNo;
		ConnectionInfo* _conn;

		RecordIteratorImplementation( typename lock_policy_t::initializer_t const&, _uint64 = 0 );
	public:
		RecordIteratorImplementation( void );
		RecordIteratorImplementation( const RecordIteratorImplementation& riter );
		virtual ~RecordIteratorImplementation();

		RecordIteratorImplementation& operator = ( const RecordIteratorImplementation& ri );
	protected:
		virtual bool do_equals( RecordIteratorInterface const& i ) const;
		virtual void do_next( void ) { lock_policy_t::Next( *this ); }
		virtual iterator_ptr_t do_clone() const
		{
			return iterator_ptr_t( new RecordIteratorImplementation<lock_policy_t>( *this ) );
		}
		virtual Record do_get_record()
		{
			return ( Record( RecordInterface::record_ptr_t(
				new RecordImplementation<RecordIteratorImplementation<lock_policy_t> >( *this, _currentRNo ) ) ) );
		}

		_uint64 currentRowNumber() { return _currentRNo; }
		void LockPack( _uint64 obj_no );
		void UnlockPack( _uint64 obj_no );

	private:
		friend struct TempTableLockPolicy;
		friend struct RcTableLockPolicy;
		template<typename table_iterator_t>
		friend class RecordImplementation;
  };

	/*! \brief Wrapper for different iterator implementations.
	 */
	class RecordIterator
	{
		RecordIteratorInterface::iterator_ptr_t _it;
	public:
		RecordIterator( void ) : _it() {}
		RecordIterator(RecordIterator const& ri) : _it() {
			if(ri._it.get())
				_it = ri._it->clone();
		}
		RecordIterator& operator = ( RecordIterator const& it ) {
			if(&it != this) {
				if(it._it.get())
					_it = it._it->clone();
				else
					_it.reset();
			}
			return ( *this );
		}

		bool operator == ( RecordIterator const& i ) const { return ( *_it == *i._it ); }
		bool operator != ( RecordIterator const& i ) const { return ( ! operator == ( i ) ); }
		RecordIterator& operator++( void )
		{
			_it->next();
			return ( *this );
		}

		Record operator*() { return ( **_it ); }
	private:
		friend class TempTable;
		USE_THIS_MACRO_TO_DO_RC_TABLE_A_FIERND
		RecordIterator( RecordIteratorInterface::iterator_ptr_t it ) : _it( it ) {}
	};

	RecordIterator find( ConnectionInfo *conn, _uint64 pos, Filter const* f, RcTableLockPolicy::attr_uses_t const& attrUses )
		{ return ( do_find( conn, pos, f, attrUses ) ); }
	RecordIterator find( ConnectionInfo *conn, _uint64 pos, RcTableLockPolicy::attr_uses_t const& attrUses )
		{ return ( do_find( conn, pos, attrUses ) ); }
	RecordIterator begin( ConnectionInfo *conn = NULL, RcTableLockPolicy::attr_uses_t const& attrUses = RcTableLockPolicy::attr_uses_t())
		{ return ( do_begin( conn, attrUses ) ); }
	RecordIterator begin( ConnectionInfo *conn, const Filter& filter, RcTableLockPolicy::attr_uses_t const& attrUses = RcTableLockPolicy::attr_uses_t() )
		{ return ( do_begin( conn, filter, attrUses ) ); }
	RecordIterator end(ConnectionInfo *conn = NULL) { return ( do_end( conn ) ); }
protected:
	virtual RecordIterator do_find( ConnectionInfo *conn, _uint64 pos, RcTableLockPolicy::attr_uses_t const& ) = 0;
	virtual RecordIterator do_find( ConnectionInfo *conn, _uint64 pos, Filter const*, RcTableLockPolicy::attr_uses_t const& ) = 0;
	virtual RecordIterator do_begin( ConnectionInfo *conn, RcTableLockPolicy::attr_uses_t const& ) = 0;
	virtual RecordIterator do_begin( ConnectionInfo *conn, Filter const& filter, RcTableLockPolicy::attr_uses_t const& ) = 0;
	virtual RecordIterator do_end( ConnectionInfo *conn ) = 0;
 };


template<typename lock_policy_t>
JustATable::RecordIteratorImplementation<lock_policy_t>::RecordIteratorImplementation( void )
	: _data( typename lock_policy_t::data_t( NULL ) ), _currentRNo( 0 ), _conn( NULL )
{
}

template<typename lock_policy_t>
JustATable::RecordIteratorImplementation<lock_policy_t>::~RecordIteratorImplementation()
{
	lock_policy_t::Destroy( *this );
}

template<typename lock_policy_t>
JustATable::RecordIteratorImplementation<lock_policy_t>::RecordIteratorImplementation( RecordIteratorImplementation const& it )
	: _data( typename lock_policy_t::data_t() ), _currentRNo( 0 ), _conn( NULL )
{
	operator = ( it );
}

template<typename lock_policy_t>
JustATable::RecordIteratorImplementation<lock_policy_t>& JustATable::RecordIteratorImplementation<lock_policy_t>::operator = ( RecordIteratorImplementation const& it )
{
	return ( lock_policy_t::Assign( *this, it ) );
}

template<typename lock_policy_t>
bool JustATable::RecordIteratorImplementation<lock_policy_t>::do_equals( JustATable::RecordIteratorInterface const& it ) const
{
	JustATable::RecordIteratorImplementation<lock_policy_t> const& i = static_cast<JustATable::RecordIteratorImplementation<lock_policy_t> const& >( it );
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT( ( ! ( _data.table < i._data.table ) ) && ( ! ( i._data.table < _data.table ) ) );
	return ( _currentRNo == i._currentRNo );
}


#endif

