/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "MIOrderedIterator.h"
#include "core/PackOrderer.h"


MIOrderedIterator::MIOrderedIterator(MultiIndex *mind, std::vector<PackOrderer>& po) : MIIterator(mind,false), po(po)
{
	DimensionVector dimensions(mind);
	for(int i = 0; i < no_dims; i++)
		dimensions[i] = true;
	Init(dimensions);
}

MIOrderedIterator::MIOrderedIterator(MultiIndex *mind, int one_dimension, std::vector<PackOrderer>& po)  : MIIterator(mind, one_dimension, false), po(po)
{
	DimensionVector dimensions(mind);
	if(one_dimension == -1) {
		for(int i = 0; i < no_dims; i++)
			dimensions[i] = true;
	} else
		dimensions[one_dimension] = true;
	Init(dimensions);
}

MIOrderedIterator::MIOrderedIterator(MultiIndex *mind, DimensionVector &dimensions, std::vector<PackOrderer>& po)  : MIIterator(mind, dimensions, false), po(po)
{
	Init(dimensions);
}


void MIOrderedIterator::Init(DimensionVector &dimensions)
{
//	for(int i = 0; i < no_dims; i++) {
//		if(dimensions[i] && !po[i].Initialized())
//			BHERROR("Ordered MIIterator :: PackOrderer mismatch");
//	}

	Rewind();
	if(!IsValid())
		no_obj = 0;
}

void MIOrderedIterator::RewindFilterDim(int i)
{
	if(!fi[i]) {
		if(po[i].Initialized())
  //  		fi[i] = new FilterOnesIterator(mind->GetFilter(i));
			fi[i] = new FilterOnesIteratorOrdered(mind->GetFilter(i), &po[i]);
		else
			fi[i] = new FilterOnesIterator(mind->GetFilter(i));
	} else {
		if(dynamic_cast<FilterOnesIteratorOrdered*>(fi[i]))
			static_cast<FilterOnesIteratorOrdered*>(fi[i])->Init(mind->GetFilter(i), &po[i]);	// = Rewind()
		else
			fi[i]->Init(mind->GetFilter(i));
	}
	if(fi[i]->IsValid()) {				// 0 tuples?
		valid = true;
		cur_pos[i] = **(fi[i]);
	} else
		valid = false;
}

void MIOrderedIterator::swap(MIOrderedIterator&mio)
{
	MIIterator::swap(mio);
	using std::swap;
	swap(po, mio.po);
}



