/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef MIORDEREDITERATOR_H_INCLUDED
#define MIORDEREDITERATOR_H_INCLUDED

#include "MIIterator.h"
class PackOrderer;

class MIOrderedIterator : public MIIterator
{
public:
	/*! create an iterator for a MultiIndex and set to the \e mind begin
	 *
	 * \param mind the MultiIndex for which the iterator is created
	 * Example:
	 * 		DimensionVector dimensions(mind);
	 *		for(...)
	 *			dimensions[...we are interested in...] = true;
	 *		MIIterator mit(mind, dimensions);
	 *
	 */
	MIOrderedIterator(MultiIndex* mind, std::vector<PackOrderer>& po);							// note: mind=NULL will create a local multiindex - one dim., one row
	MIOrderedIterator(MultiIndex* mind, int one_dimension, std::vector<PackOrderer>& po);		// -1 is an equivalent of all dimensions
	MIOrderedIterator(MultiIndex* mind, DimensionVector &dimensions, std::vector<PackOrderer>& po);

	void swap(MIOrderedIterator&);


private:
	void Init(DimensionVector &dimensions);
	void RewindFilterDim(int i);

	std::vector<PackOrderer>& po;
};

/*! \brief Used for storing position of another iterator and for combining positions from several iterators.
 *
 */

#endif /*MIORDEREDITERATOR_H_INCLUDED*/
