/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef OUTERJOINER_H_
#define OUTERJOINER_H_

#include "MultiIndex.h"
#include "Descriptor.h"
#include "Joiner.h"


/* Outer join algorithm:
 *
 * - Constructor: take a snapshot of initial multiindex state.
 *   Depending on multiindex contents and joining conditions,
 *   there ale the following modes:
 *
 *   -- find reference_dim and create a reference_filter for it. The dimension may be:
 *      --- the only one, if there is no more dimensions on the left side,
 *      --- a dimension used in conditions, if there is only one left dimension used,
 *          (reason: even if there are repeated rows for such dimension,
 *           the whole abstract class is either inside, or outside join result; assumption: conditions are deterministic)
 *      --- any dimension discovered to be a key (i.e. no repetitions), except a case of nontrivial virtual dimensions
 *
 *   -- if no reference_dim exists, fail and switch to MySQL (for now).
 *      In future we will have to copy the multiindex somehow.
 *
 * - AddOuterTuples: add all tuples found to be not present after join (left side).
 *
 *   -- in case a reference_dim exists, go through the post-join multiindex and check whether
 *      reference_filter contains these rows. If yes, reset these rows in reference_filter.
 *      Then go through the old multiindex and add all tuples remaining in reference_filter, with nulls.
 *
 *   -- in case of no reference_dim, we will have to find somehow every new tuple in the old multiindex.
 *      Probably hashing etc. will be needed.
 */

class OuterJoiner
{
public:
	OuterJoiner(MultiIndex& mind, std::vector<Descriptor>& desc, bool _count_only);
	~OuterJoiner();

	void AddOuterTuples(MultiIndex &new_mind);

private:
	/*
	_int64 AddOuterTuples(int dim);		// return: the number of added tuples
	void Init();
	FilterAutoPtr InitNonFilterDimension(int dim);
	void InitDimension(int orig_dim, int dim);

	struct MaterializedRows {
		_int64 materialized_before_join;	// number of records (materialized) before outer join
		int joining_ndx[2];
		std::vector<int> dims;
		std::vector<IndexTable*> i_tables;
		MaterializedRows(int join_dim1, int join_dim2, MultiIndex& mind);
		~MaterializedRows();
	};
	std::auto_ptr<MaterializedRows> materialized_rows;

	struct DimForJoinInfo {
		int joining_ndx;
		int dim_no;
		bool dim_is_filter;
		FilterAutoPtr 	filter;
	};

	struct DimForJoinInfo dims_info[2];

	JoinType 	join_type;

	// new:
	 */
	int				outer_dim;				// right side of outer join
	DimensionVector	left_dims;				// left side of outer join
	int				no_of_left_dims;		// a number of left dims (not only these used in conditions)
	int 			reference_dim;			// a number of (left side) dimension used as a reference/key, or -1
	FilterAutoPtr 	reference_filter;		// a filter of original size of reference_dim; 1 means that the row was present before join
	bool			add_old_nulls;			// true if we should add all tuples with null values on old reference_dim
	bool			count_only;				// true if the materialization of results can be done virtually
	MultiIndex		old_mind;				// a copy of the previous mind
};


#endif /*OUTERJOINER_H_*/
