/* Copyright (C)  2005-2008 Infobright Inc.

	 This program is free software; you can redistribute it and/or modify
	 it under the terms of the GNU General Public License version 2.0 as
	 published by the Free  Software Foundation.

	 This program is distributed in the hope that  it will be useful, but
	 WITHOUT ANY WARRANTY; without even  the implied warranty of
	 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	 General Public License version 2.0 for more details.

	 You should have received a  copy of the GNU General Public License
	 version 2.0  along with this  program; if not, write to the Free
	 Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
	 02111-1307 USA  */

#ifndef CORE_QUERY_H_INCLUDED
#define CORE_QUERY_H_INCLUDED

#include "system/ConnectionInfo.h"
#include "MysqlExpression.h"
#include "Joiner.h"
#include "Item_bhfield.h"

#define RETURN_QUERY_TO_MYSQL_ROUTE 0
#define RCBASE_QUERY_ROUTE 1

class RoughMultiIndex;
class ParameterizedFilter;
class CompiledQuery;
class MysqlExpression;

class Query
{
public:
	struct Context {
		typedef enum { SELECT, WHERE, HAVING, GROUPBY } context_t;
	};

	struct WrapStatus {
		typedef enum { SUCCESS, FAILURE, OPTIMIZE } wrap_status_t;
	};

public:
	Query(ConnectionInfo *conn_info);
	~Query();

	void Add(RCTable* tab);

	/*! \brief Remove the given RCTable from the list of tables used in a query
	 */
	void RemoveFromManagedList(const RCTablePtr tab);

	int NoTabs() const { return no_tab; };
	RCTable* Table(int table_num) const;
	void Clear() { no_tab = 0; };
	void LockAll();
	void UnlockAll();
	void LockPackInfoForUse();
	void UnlockPackInfoFromUse();
	bool TryToMerge(Descriptor & d1, Descriptor & d2);

	struct Distinct	{
		std::map<bool,int> el;
	};
	struct Operation {
		std::map<ColOperation,struct Distinct> el;
	};
	struct FieldAlias {
		std::map<std::string,struct Operation> el;
	};
	struct TableAlias {
		std::map<std::string,struct FieldAlias> el;
	};
	typedef struct TableAlias map_of_fields;
	int Compile(CompiledQuery *compiled_query, SELECT_LEX *selects_list, SELECT_LEX *last_distinct, std::map<std::string,int> *outer_table_alias2index, std::map<std::string,TABLE*> *outer_table_alias2ptr, TabID *res_tab = NULL, bool ignore_limit = false, Item *left_expr_for_subselect = NULL, Operator *oper_for_subselect = NULL, bool ignore_minmax = false);
	TempTable *Execute(CompiledQuery & qu);
	TempTable *Preexecute(CompiledQuery & qu, bool display_now = true);
	int BuildFilter(COND *conds, CompiledQuery *cq, std::map<std::string,int> *table_alias2index, std::map<std::string,TABLE*> *table_alias2ptr, const TabID & tmp_table, FilterType filter_type, bool is_zero_result = false, JoinType join_type = JO_INNER);
	void UpdateMultiIndex(MultiIndex & mind, std::vector<Descriptor> & desc, TempTable *table, bool count_only, _int64 limit);
	void UpdateParamFilter(ParameterizedFilter *filter, bool count_only, _int64 limit, bool display_now);

	typedef std::vector<MysqlExpression*> expressions_t;
private:
	class ColumnData {
	private:
		AttrID atr;
		TabID tab;
		CQTerm term;
		ColOperation oper;
		bool dist;
	public:
		ColumnData(AttrID _atr, TabID _tab, CQTerm _term, ColOperation _oper, bool _dist)
			: atr(_atr), tab(_tab), term(_term), oper(_oper), dist(_dist)
			{ }
		AttrID atribute() { return atr; }
		TabID table() { return tab; }
		CQTerm cqterm() { return term; }
		ColOperation operation() { return oper; }
		bool distinct() { return dist; }
	};

	CompiledQuery *cq;
	std::vector<std::pair<TabID,bool> > subqueries_in_where;
	std::map<std::string,int> *table_alias2index;
	std::map<std::string,TABLE*> *table_alias2ptr;
	typedef std::pair<int,std::string> TabIDColAlias;
	std::map<TabIDColAlias,int> field_alias2num;
	std::map<std::string,unsigned > path2num;
	std::map<std::pair<int,int>,int> coltab_to_vc;
	typedef std::pair<int, MysqlExpression*> virtual_column_meta_t;
	typedef std::multimap<TabID,virtual_column_meta_t> tab_id2expression_t; // all expression based virtual columns for a given table
	tab_id2expression_t tab_id2expression;
	expressions_t gc_expressions;
	std::multimap<std::pair<int,int>, std::pair<int,int> > phys2virt;
	std::multimap<TabID, std::pair<int, TabID> > tab_id2subselect;

	std::vector<JustATablePtr> ta;					// table aliases - sometimes point to TempTables (maybe to the same one), sometimes to RCTables
	//int no_taba;
	RCTable **t;
	int no_tab;
	int capacity;

	bool FieldUnmysterify(Item * item, TabID &tab, AttrID &col);
	int FieldUnmysterify(Item * item, const char *& database_name, const char *& table_name, const char *& table_alias, const char*& table_path, const char *& field_name, const char *& field_alias);
	int FieldUnmysterify(Item * item, TabID &tab, AttrID &col, const char *& field_alias, const char *& table_alias);
	int SomeSpecialColumnsRemoval(const char * table_alias, ColOperation oper, CQTerm & const_term, AttrID &col, bool group_by_clause);
	int PrefixCheck(COND * conds);
	int GetTableIDFromFieldItem(COND * conds);
	int SimpleCaseOfLeftJoin(COND * conds, int to_be_joined_table, int &already_joined_table);

	int TabNum(const char *table_name) const;

	/*! \brief Checks if exists virtual column defined by physical column:
	 * \param tmp_table - id of TempTable for which VC is searched for
	 * \param tab - table id
	 * \param col - column id
	 * \return table id and number of virtual column if it exists or NULL_VALUE_32 otherwise
	 */
	pair<int,int> VirtualColumnAlreadyExists(TabID tmp_table, const TabID tab, const AttrID col);

	/*! \brief Checks if exists virtual column defined by expression:
	 * \param tmp_table - id of table for which column is supposed to be created
	 * \param expression
	 * \return number of virtual column if it exists or NULL_VALUE_32 otherwise
	 */
	int VirtualColumnAlreadyExists(TabID tmp_table, MysqlExpression* expression);

	/*! \brief Checks if exists virtual column defined by subquery
	 * \param tmp_table - id of table for which column is supposed to be created
	 * \param subquery - id of subquery
	 * \return number of virtual column if it exists or NULL_VALUE_32 otherwise
	 */
	int VirtualColumnAlreadyExists(TabID tmp_table, TabID subquery);

	int Item2CQTerm(Item *an_arg, CQTerm & term, TabID tmp_table, FilterType filter_type, bool negative = false, Item *left_expr_for_subselect = NULL, Operator *oper_for_subselect = NULL);

	int FilterNotSubselect(Item *conds, TabID tmp_table, FilterType filter_type, FilterID *and_me_filter = 0);

	/*! \brief Create filter from field or function that has no condition attached.
	 * \param conds - condition (a field or function).
	 * \param tmp_table - required for Item2CQTerm.
	 * \param filter_type - type of filter context (WHERE, HAVING).
	 * \param and_me_filter - ?
	 * \return filter number (non-negative) or error indication (negative)
	 */
	int FilterNumberFromNaked(COND* conds, TabID tmp_table, FilterType filter_type, FilterID* and_me_filter);
	int FilterNumberFromMultipleEquality(Item_equal *conds, const TabID & tmp_table, FilterType filter_type, FilterID *and_me_filter = 0);
	int FilterNumberFromComparison(Item *conds, const TabID & tmp_table, FilterType filter_type, FilterID *and_me_filter = 0, bool negative = false);

	/*! \brief Checks type of operator involved in condition
	 * \param conds - conditions
	 * \param op - output param.: type of operator
	 * \param negative - input information about negation of condition
	 * \return void
	 */
	void ExtractOperatorType(Item*& conds, Operator& op, bool& negative);

	/*! \brief Adds ANY modifier to an operator
	 * \param op - operator
	 * \return void
	 */
	static void MarkWithAny(Operator& op);

	/*! \brief Adds ALL modifier to an operator
	* \param op - operator
	* \return void
	*/
	static void MarkWithAll(Operator& op);

	int FilterNumber(COND *conds, TabID tmp_table, FilterType filter_type, FilterID *and_me_filter = 0);
	int BuildFilterIfPossible(COND *conds, const TabID & tmp_table, JoinType join_type);

public:
	/*! \brief Removes ALL/ANY modifier from an operator
	* \param op - operator
	* \return void
	*/
	static void UnmarkAllAny(Operator& op);

	/*! \brief Checks if item is a aggregation with distinct modifier
	 * \param item
	 * \return true if yes
	 */
	static bool IsAggregationDistinct(Item* item);

	/*! \brief Checks if item has a aggregation as one of its arguments.
	 * \param item
	 * \return true if yes
	 */
	static bool HasAggregation(Item* item);

	/*! \brief Checks if an aggregation represents COUNT(*)
	 * \param item_sum - aggregation item
	 * \return true if yes
	 */
	static bool IsCountStar(Item* item_sum);

	/*! \brief Checks if an item is of type FieldItem
	 * \param item
	 * \return true if yes
	 */
	static bool IsFieldItem(Item* item);

	/*! \brief Checks if an item is of type Item_sum_func
	 * \param item
	 * \return true if yes
	 */
	static bool IsAggregationItem(Item * item){ return item->type() == Item::SUM_FUNC_ITEM; }

	/*! \brief Checks if an item represents a constant
	 * \param item
	 * \return true if yes
	 */
	static bool IsConstItem(Item* item);

	/*! \brief Returns true if an item is aggregation over field item
	 * \param item
	 * \return true if yes
	 */
	static bool IsAggregationOverFieldItem(Item * item);

	/*! \brief Checks if given join contains left or right outer component.
	* \param join - join to be checked
	* \return bool - true if contains left or right outer join
	*/
	static bool IsLOJ(List<TABLE_LIST> *join);

	static const string GetItemName(Item* item);
private:

	/*! \brief Creates AddColumn step in compilation by creating, if does not exist, Virtual Column based on PhysicalColumn
	 * \param item - what Item
	 * \param tmp_table - for which TempTable
	 * \param oper - defines type of added column
	 * \param distinct - flag for AddColumn operation
	 * \param group_by - indicates if it is column for group by query
	 * \return column number
	 */
	int AddColumnForPhysColumn(Item* item, const TabID tmp_table, const ColOperation oper, const bool distinct, bool group_by);

	/*! \brief Creates AddColumn step in compilation by creating, if does not exist, Virtual Column based on expression
	 * \param mysql_expression - pointer to expression
	 * \param tmp_table - for which TempTable
	 * \param alias - defines alias of added column
	 * \param oper - defines type of added column
	 * \param distinct - flag for AddColumn operation
	 * \param group_by - indicates if it is column for group by query
	 * \return column number
	 */
	int AddColumnForMysqlExpression(MysqlExpression* mysql_expression, const TabID tmp_table, const char* alias, const ColOperation oper, const bool distinct, bool group_by = false);

	/*! \brief Computes identifier of a column created by AddColumn operation
	 * \param vc - for which virtual column
	 * \param tmp_table -  for which TempTable
	 * \param oper - type of column
	 * \param distinct - modifier of AddColumn
	 * \return column number if it exists or NULL_VALUE_32 otherwise
	 */
	int GetAddColumnId(AttrID vc, const TabID tmp_table, const ColOperation oper, const bool distinct);

	/*! \brief Changes type of AddColumn step in compilation from LISTING to GROUP_BY
	 * \param tmp_table - for which TempTable
	 * \param attr - for which column
	 */
	void CQChangeAddColumnLIST2GROUP_BY(const TabID tmp_table, int attr);

	/*! \brief Creates MysqlExpression object that wraps expression tree of MySQL not containing aggregations.
	 * All Item_field items are substituted with Item_bhfield items
	 * \param item - root of MySQL expression tree to be wrapped
	 * \param tmp_table - alias of TempTable which is checked if it contains aggregation fields (should not)
	 * \param expr - expression to store the result tree
	 * \param in_aggregation - the expression is used in WHERE (true) or elsewhere (having, select list = false)
	 * \return WrapStatus::SUCCESS on success, WrapStatus::OPTIMIZE if an aggregation was encountered, WrapStatus::FAILURE if
	 * there was any problem with wrapping, e.g., not acceptable type of expression
	 */
	WrapStatus::wrap_status_t WrapMysqlExpression(Item *item, TabID tmp_table, MysqlExpression*& expr, bool in_where, bool aggr_used);
//
//	/*! \brief Creates MysqlExpression object that wraps full expression tree of MySQL. All Item_field items are
//	 * substituted with Item_bhfield items. In case of aggregation it is substituted with its whole subtree by a single
//	 * Item_bhfield, however, there is an AddColumn step added to compilation with this aggregation and Item_bhfield refers to the
//	 * column created by this step.
//	 * \param item - root of MySQL expression tree to be wrapped
//	 * \param tmp_table - alias of TempTable to which AddColumn is added in case of aggregation
//	 * \param expr - expression to store the result tree
//	 * \param is_const_or_aggr - optional pointer to bool variable that is set to true if expression has no variables or
//	 * it contains at least one aggregation (false otherwise)
//	 * \return WrapStatus::SUCCESS on success, WrapStatus::FAILURE if there was any problem with wrapping,
//	 *  e.g., not acceptable type of expression
//	 */
//	WrapStatus::wrap_status_t WrapMysqlExpressionWithAggregations(Item *item, TabID tmp_table, MysqlExpression*& expr, bool* is_const_or_aggr = NULL);

	/*! \brief Generates AddColumn compilation steps for every field on SELECT list
	 * \param fields - list of fields
	 * \param tmp_table - alias of TempTable for which AddColumn steps are added
	 * \param group_by_clause - indicates whether it is group by query
	 * \param ignore_minmax - indicates if field of typy Min/Max should be transformed to LISTING
	 * \return returns RETURN_QUERY_TO_MYSQL_ROUTE in case of any problem and RCBASE_QUERY_ROUTE otherwise
	 */
	int AddFields(List<Item> & fields, const TabID & tmp_table, const bool group_by_clause, int& num_of_added_fields, bool ignore_minmax, bool& aggr_used);

	/*! \brief Generates AddColumn compilation steps for every field on GROUP BY list
	 * \param fields - pointer to GROUP BY fields
	 * \param tmp_table - alias of TempTable for which AddColumn steps are added
	 * \return returns RETURN_QUERY_TO_MYSQL_ROUTE in case of any problem and RCBASE_QUERY_ROUTE otherwise
	 */
	int AddGroupByFields(ORDER *group_by, const TabID & tmp_table);

	//! is this item representing a column local to the temp table (not a parameter)
	bool IsLocalColumn(Item* item, TabID tmp_table) ;
	int AddOrderByFields(ORDER *order_by, TabID const&tmp_table, int const group_by_clause);
	int AddGlobalOrderByFields(SQL_LIST *global_order, std::map<std::string,int> & field_alias2column_number, const TabID & tmp_table, int max_col);
	int AddJoins(List<TABLE_LIST> join, TabID & tmp_table, bool first_table = true, bool is_left = false);
	int DoWhereClause(CQTerm & term, Item *item);
	static bool ClearSubselectTransformation(Operator & oper_for_subselect, Item *& field_for_subselect, COND *& conds, Item *& having, Item *& cond_removed, List<Item> *& list_to_reinsert, Item *left_expr_for_subselect);
	void UpdateJoinCondition(MultiIndex & mind, RoughMultiIndex & rmind, TempTable *table, std::vector<Descriptor> & desc, bool count_only_now,_int64 limit_now);
	void SyntacticalDescriptorListPreprocessing(std::vector<Descriptor> & desc, MultiIndex & mind, TempTable* table);
	void DescriptorListOrdering(std::vector<Descriptor> & desc);
	void DescriptorJoinOrdering(std::vector<Descriptor> & desc, MultiIndex & mind);
	void ApplyDescriptor(MultiIndex & mind, RoughMultiIndex & rmind, TempTable *table, Descriptor & desc, int desc_number);
	bool RoughUpdateMultiIndex(RoughMultiIndex & rmind, MultiIndex & mind, std::vector<Descriptor> & desc);
	void RoughMakeProjections(RoughMultiIndex &rmind, MultiIndex &mind, vector<Descriptor> &desc);
	void RoughMakeProjections(RoughMultiIndex &rmind, MultiIndex &mind, vector<Descriptor> &desc, int dim, bool update_reduced = true);
	double EvaluateConditionNonJoinWeight(Descriptor & d);
	double EvaluateConditionJoinWeight(Descriptor &d, MultiIndex &mind);

	/*!
	 * \brief Are the variables constant (i.e. they are parameters) in the context of the query represented by TempTable t
	 */
	bool IsConstExpr(MysqlExpression::SetOfVars& vars, TabID t);

	ConnectionInfo* m_conn;

	//void EvalCQTerm(bool cplx, CQTerm& term, MysqlExpression::BufOfVars bufs, CQTerm& t2, DataType eval_type, std::map<int,ValueOrNull> & columns);
	//void PrepareCQTerm(CQTerm& term, bool& cplx_val2, MysqlExpression::SetOfVars& val2_vars, std::vector<int>& aliases, std::map<int,ValueOrNull>& columns, DataType& attr_eval_type);

	static JoinType GetJoinTypeAndCheckExpr(uint outer_join, Item *on_expr);

	void DisplayJoinResults(MultiIndex& mind, DimensionVector &all_involved_dims,
						JoinAlgType cur_join_type, bool is_outer, int conditions_used);

	/*! \brief Checks if condition represented by \e it is a negation of other condition
	 * \param it - item
	 * \param is_there_not - output param.: true if negation was found
	 * \return Item* - returns pointer to argument of negation or original item if there was no negation
	 */
	Item* FindOutAboutNot(Item* it, bool& is_there_not);

	VirtualColumn*	CreateColumnFromExpression(Query::expressions_t const& exprs,
			TempTable* temp_table, int temp_table_alias, MultiIndex* mind);

	/*! \brief Delays WHERE conditions that are dependent on NULLs for LOJ queries. Potential candidates are marked
	 * by compilation with outer_dim = -2. The function changes it to regular value -1.
	 * \param desc - vector of descriptors to be checked.
	 * \return - void
	 */
	void DelayWhereConditions(vector<Descriptor> &desc);

	bool IsParameterFromWhere(TabID params_table);

	/*! \brief true if the desc vector contains at least 2 1-dimensional descriptors defined for different dimensions
	 *  e.g. true if contains T.a=1, U.b=7, false if T.a=1, T.b=7
	 *
	 */
	bool DimsWith1dimFilters(vector<Descriptor> &desc, MultiIndex& mind);
};

#endif // not CORE_QUERY_H_INCLUDED

