/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */
#include "shared/core/RCEngine.h"
#include "Query.h"
#include "MysqlExpression.h"
#include "compilation_tools.h"
#include "common/mysql_gate.h"

using namespace std;

int Query::FieldUnmysterify(Item * item, const char *& database_name, const char *& table_name,
		const char *& table_alias, const char*& table_path, const char *& field_name, const char *& field_alias)
{
	table_alias = EMPTY_TABLE_CONST_INDICATOR;
	database_name = NULL;
	table_name = NULL;
	table_path = NULL;
	field_name = NULL;
	field_alias = NULL;

	item = UnRef(item);

	Item_field* ifield;
	switch(static_cast<int>(item->type())) {
		case Item_bhfield::BHFIELD_ITEM:
			ifield = dynamic_cast<Item_bhfield *>(item)->OriginalItem();
			if(IsAggregationItem(ifield)) {
				Item_sum* is = (Item_sum*)ifield;
				if (is->arg_count > 1)
					return RETURN_QUERY_TO_MYSQL_ROUTE;
				Item* tmp_item = UnRef(is->args[0]);
				if(tmp_item->type() == Item::FIELD_ITEM)
					ifield = (Item_field*)tmp_item;
				else if ( static_cast<int>( tmp_item->type() ) == static_cast<int>( Item_bhfield::BHFIELD_ITEM ) )
					ifield = dynamic_cast<Item_bhfield *>(tmp_item)->OriginalItem();
				else {
					return RETURN_QUERY_TO_MYSQL_ROUTE;
				}
			}
			break;
		case Item::FIELD_ITEM:  // regular select
			ifield = (Item_field*)item;
			break;

		case Item::SUM_FUNC_ITEM: // min(k), max(k), count(), avg(k), sum
		{
			Item_sum* is = (Item_sum*)item;
			//if ((is->sum_func()==Item_sum::COUNT_FUNC || is->sum_func()==Item_sum::COUNT_DISTINCT_FUNC)
			//	&& is->args[0]->type()==Item::INT_ITEM)
			//{
			//	field_alias = EMPTY_ALIAS;
			//	return RCBASE_QUERY_ROUTE;
			//}
			if (is->arg_count > 1) return RETURN_QUERY_TO_MYSQL_ROUTE;
			Item* tmp_item = UnRef(is->args[0]);
			if(tmp_item->type() == Item::FIELD_ITEM)
				ifield = (Item_field*)tmp_item;
			else if ( static_cast<int>( tmp_item->type() ) == static_cast<int>( Item_bhfield::BHFIELD_ITEM ) ) /* *CAUTION* comparision of enumerators from different enums */
				ifield = dynamic_cast<Item_bhfield *>(tmp_item)->OriginalItem();
			else {
				// it is a constant term, not a column
//				field_alias = item->name;
//				table_alias = EMPTY_TABLE_CONST_INDICATOR;
//
//				if(WrapMysqlExpression(tmp_item, &not_a_table_column) == WrapStatus::SUCCESS)
//					return RCBASE_QUERY_ROUTE;
//				return RETURN_QUERY_TO_MYSQL_ROUTE;
////				if (Item2ConstTerm(tmp_item, &not_a_table_column))
////					return RCBASE_QUERY_ROUTE;

				return RETURN_QUERY_TO_MYSQL_ROUTE;
			}
			break;
		}
		case Item::FUNC_ITEM: // complex expressions
			//if(WrapMysqlExpression(item, &not_a_table_column) == WrapStatus::SUCCESS)
				return RCBASE_QUERY_ROUTE;
			return RETURN_QUERY_TO_MYSQL_ROUTE;
		default:
			//if(WrapMysqlExpression(item, &not_a_table_column) == WrapStatus::SUCCESS)
				return RCBASE_QUERY_ROUTE;
			return RETURN_QUERY_TO_MYSQL_ROUTE;

//			if (Item2ConstTerm(item, &not_a_table_column)) {
//				field_alias = item->name;
//				table_alias = EMPTY_TABLE_CONST_INDICATOR;
//
//				return RCBASE_QUERY_ROUTE;
//			}
//
//			return RETURN_QUERY_TO_MYSQL_ROUTE;
	};

	/*
	 * By MW. Related to bug 1073.
	 *
	 * For views, 'table_alias' must be created as a concatenation
	 * of original view(s) name(s) and the table name
	 * - currently it's just the table name, which leads to ambiguity and errors
	 * when the same table is used in another place in the query.
	 * Note that there can be several nested views.
	 *
	 * To retrieve name(s) of the view(s) from which the current 'ifield' comes
	 * you may try the following expression:
	 *
	 *     ifield->cached_table->table_list->belong_to_view->alias  or  ...->table_name
	 * or  ifield->cached_table->table_list->referencing_view->alias  or  ...->table_name
	 *
	 * Here, 'belong_to_view' and 'referencing_view' are different
	 * if there are nested views.
	 *
	 * Probably ifield->cached_table could be also used to find
	 * 'database_name', 'table_name' and 'table_path'
	 * in a simpler way than currently.
	 */

	Field* f = ifield->result_field;

	ASSERT_MYSQL_STRING(f->table->s->db);
	ASSERT_MYSQL_STRING(f->table->s->table_name);
	ASSERT_MYSQL_STRING(f->table->s->path);

	table_alias = ifield->table_name;
	database_name = f->table->s->db.str;
	table_name = f->table->s->table_name.str;
	table_path = f->table->s->path.str;;

	field_name = f->field_name;
//	field_alias = item->name;   //this should be better but for some reason doesn't work in all cases!!!!!
	field_alias = ifield->name;

	return RCBASE_QUERY_ROUTE;
}

int Query::FieldUnmysterify(Item * item, TabID &tab, AttrID &col, const char *& field_alias, const char *& table_alias)
{
	const char * table_name;
	const char * table_path;
	const char * field_name;
	const char * database_name;

	if (!FieldUnmysterify(item, database_name, table_name, table_alias, table_path, field_name, field_alias))
		return RETURN_QUERY_TO_MYSQL_ROUTE;

	//if(strcmp(field_alias, "d") == 0) table_alias = "t1v";

	if (table_alias == EMPTY_TABLE_CONST_INDICATOR)
		return RCBASE_QUERY_ROUTE; //SZN:indication that
							//we found a constant or aggregation over a constant (or a count over *)
							//in this case we simply set a 'not_a_table_column' field to this constant

	map<string, TABLE*>::iterator it = table_alias2ptr->find(table_alias);
	if (it == table_alias2ptr->end())
		return RETURN_QUERY_TO_MYSQL_ROUTE;

	TABLE* mysql_table = (*table_alias2ptr)[table_alias];
	int field_num = 0;//rctable->AttrNum(field_name);
	for (; mysql_table->field[field_num]; field_num++)
		if (mysql_table->field[field_num]->field_name==field_name)
			break;
	if (!mysql_table->field[field_num])
		return RETURN_QUERY_TO_MYSQL_ROUTE;

	if (mysql_table->derived_select_number) col = AttrID(-field_num-1);
	else col = AttrID(field_num);
	tab = TabID((*table_alias2index)[table_alias]);

	return RCBASE_QUERY_ROUTE;
}

bool Query::FieldUnmysterify(Item * item, TabID &tab, AttrID &col)
{
	Item_field* ifield;
	if(item->type() == Item_bhfield::BHFIELD_ITEM) {
		ifield = dynamic_cast<Item_bhfield *>(item)->OriginalItem();
		if(IsAggregationItem(ifield)) {
			Item_sum* is = (Item_sum*)ifield;
			if(is->arg_count > 1)
				return RETURN_QUERY_TO_MYSQL_ROUTE;
			Item* tmp_item = UnRef(is->args[0]);
			if(tmp_item->type() == Item::FIELD_ITEM)
				ifield = (Item_field*)tmp_item;
			else if(tmp_item->type() == Item_bhfield::BHFIELD_ITEM)
				ifield = dynamic_cast<Item_bhfield *>(tmp_item)->OriginalItem();
			else if(tmp_item->type() == Item::FUNC_ITEM) {
				Item_bhfield * bhi = dynamic_cast<Item_bhfield *>(item);
				tab.n = bhi->varID.tab;
				col.n = bhi->varID.tab;
				return true;
			} else
				return false;
		}
	} else if(item->type() == Item::SUM_FUNC_ITEM) { // min(k), max(k), count(), avg(k), sum
		Item_sum* is = (Item_sum*)item;
		if(is->arg_count > 1)
			return false;
		Item* tmp_item = UnRef(is->args[0]);
		if(tmp_item->type() == Item::FIELD_ITEM)
			ifield = (Item_field*)tmp_item;
		else if(tmp_item->type() == Item_bhfield::BHFIELD_ITEM)
			ifield = dynamic_cast<Item_bhfield *>(tmp_item)->OriginalItem();
		else
			return false;
	} else if(item->type() == Item::FIELD_ITEM)
		ifield = (Item_field*)item;
	else
		return false;

	if(!ifield->table_name) {
		// union results have no name, but refer to the temp table
		if(field_alias2num.find(TabIDColAlias((*table_alias2index)[ifield->table_name], ifield->name)) == field_alias2num.end())
			return false;
		col.n = field_alias2num[TabIDColAlias((*table_alias2index)[ifield->table_name], ifield->name)];
		tab.n = NULL_VALUE_32;
		return true;
	}
	map<string, TABLE*>::iterator it = table_alias2ptr->find(ifield->table_name);
	if(it == table_alias2ptr->end())
		return false;
	TABLE* mysql_table = (*table_alias2ptr)[ifield->table_name];

	int field_num; //rctable->AttrNum(field_name);

	tab = TabID((*table_alias2index)[ifield->table_name]);
	if(mysql_table->derived_select_number == 0) {
		// Physical table in FROM - RCTable
		for(field_num = 0; mysql_table->field[field_num]; field_num++)
			if(mysql_table->field[field_num]->field_name == ifield->result_field->field_name)
				break;
		if(!mysql_table->field[field_num])
			return false;
		col = AttrID(field_num);
	} else {
		// subselect in FROM - TempTable
		if(field_alias2num.find(TabIDColAlias(tab.n, ifield->result_field->field_name)) == field_alias2num.end())
			return false;
		col.n = field_alias2num[TabIDColAlias(tab.n, ifield->result_field->field_name)];
	}

	return true;
}

int Query::SomeSpecialColumnsRemoval(const char * table_alias, ColOperation oper, CQTerm & const_term, AttrID &col, bool group_by_clause)
{
	/*it removes
	1. Changes Count over constants to count over empty terms.
	2. avg and sum on non-numeric constant terms or columns types (ERROR)
	3. COMMENTED code: Additional group_by columns (not specified in GROUP_BY clause
	*/
/*
	if (oper == COUNT && table_alias == EMPTY_TABLE_CONST_INDICATOR)
		const_term = CQTerm();    //SZN: in case of a count over a constant
	//it doesn't really matter over what value the count is done
	//all are equivalent to count(*) or count(0) if you prefer
	// TODO: (JW): The following 2 lines were commented to enable constants in GROUP BY.
	//       Are there any problems around it?
//	else if (table_alias == EMPTY_TABLE_CONST_INDICATOR && group_by_clause) //oper != COUNT
//			return RETURN_QUERY_TO_MYSQL_ROUTE;  //as things stand now, the constants other than COUNT(*) or other COUNTs
												 //cannot be present if group by is present


	if (table_alias != EMPTY_TABLE_CONST_INDICATOR) {
		Field* f = (*table_alias2ptr)[table_alias]->field[col.n];
		if (f && col.n >= 0 && ((oper == SUM || oper == AVG) && !ATI::IsNumericType(rceng->GetCorrespondingType(*f)))) {
			my_message(ER_SYNTAX_ERROR, "Brighthouse specific error: Aggregation (SUM and AVG) can be executed only on numerical columns", MYF(0));
			throw ReturnMeToMySQLWithError();
		};
		////this time, since we are after GROUP_BY clause analysis, we cannot accept
		////any new columns with the group_by status
		//if (oper == GROUP_BY) {
		//	my_message(ER_SYNTAX_ERROR, "Brighthouse specific error: Non-aggregate columns after select must be present after GROUP_BY clause", MYF(0));
		//	return RETURN_QUERY_TO_MYSQL_ROUTE;
		//};
	}
	else {
		if ((oper == SUM || oper == AVG) && const_term.val_n == NULL_VALUE_64) {
			//const_term.val_t != NULL itself means both: it is a constant AND not a numerical AND not null,
			//so checking for val_n in case we KNOW it is a constant is better
			my_message(ER_SYNTAX_ERROR, "Brighthouse specific error: Aggregation (SUM and AVG) can be executed only on numerical constants", MYF(0));
			throw ReturnMeToMySQLWithError();
		}

	};
	*/
	return RCBASE_QUERY_ROUTE;
}

int Query::GetTableIDFromFieldItem(COND * conds)
{
	conds = UnRef(conds);
	if(!conds)
		return RCBASE_QUERY_ROUTE;
	Item::Type cond_type = conds->type();
	//BHASSERT_WITH_NO_PERFORMANCE_IMPACT(cond_type == Item::FIELD_ITEM);

	const char *field_alias= NULL;
	const char *table_alias= NULL;
	const char *database_name= NULL;
	const char *table_name= NULL;
	const char *table_path= NULL;
	const char *field_name= NULL;

	if(!FieldUnmysterify(conds, database_name, table_name, table_alias, table_path, field_name, field_alias))
		return 0;

	if(table_alias == EMPTY_TABLE_CONST_INDICATOR) //if it is a constant - no reason to worry
		return 0;

	map<string, int>::const_iterator it_tab;
	it_tab = table_alias2index->find(table_alias);
	if(it_tab == table_alias2index->end())
		return 0;
	return it_tab->second;
}

int Query::SimpleCaseOfLeftJoin(COND * conds, int to_be_joined_table, int &already_joined_table)
{
	conds = UnRef(conds);
	if(!conds)
		return RCBASE_QUERY_ROUTE;
	Item::Type cond_type = conds->type();
	FilterID filter;
	switch(cond_type) {
		case Item::COND_ITEM: {
			Item_cond *cond_cond= (Item_cond*) conds;
			Item_func::Functype func_type = cond_cond->functype();
			switch(func_type) {
				case Item_func::COND_AND_FUNC:
				case Item_func::COND_OR_FUNC: {
					List_iterator_fast<Item> li(*(cond_cond->argument_list()));
					Item * item;
					while((item = li++)) {
						int ret = SimpleCaseOfLeftJoin(item, to_be_joined_table, already_joined_table);
						if(!ret)
							return RETURN_QUERY_TO_MYSQL_ROUTE;
						if(ret == TABLE_YET_UNSEEN_INVOLVED)
							return TABLE_YET_UNSEEN_INVOLVED;
					};
					break;
				}
					;
				case Item_func::COND_XOR_FUNC: //we don't handle xor as yet
				default:
					return -1; //unknown function type

			}
			break;
		}
		case Item::FUNC_ITEM: {
			Item_func *cond_func= (Item_func*) conds;
			Item_func::Functype func_type = cond_func->functype();

			switch(func_type) {
				case Item_func::BETWEEN:
				case Item_func::LIKE_FUNC:
				case Item_func::IN_FUNC:
				case Item_func::ISNULL_FUNC:
				case Item_func::ISNOTNULL_FUNC:
					return RETURN_QUERY_TO_MYSQL_ROUTE;
				case Item_func::EQ_FUNC: // =
				case Item_func::NE_FUNC: // <>
				case Item_func::LE_FUNC: // <=
				case Item_func::GE_FUNC: // >=
				case Item_func::GT_FUNC: // >
				case Item_func::LT_FUNC: { // <
					int first, second;
					Item ** args = cond_func->arguments();
					// there should be 2 arguments
					Item * an_arg = UnRef(args[0]);
					first = GetTableIDFromFieldItem(an_arg);
					an_arg = UnRef(args[1]);
					second = GetTableIDFromFieldItem(an_arg);
					if(!first || !second || first == second)
						return RETURN_QUERY_TO_MYSQL_ROUTE;
					if(first == to_be_joined_table) {
						if(!already_joined_table)
							already_joined_table = second;
						else if(already_joined_table != second)
							return RETURN_QUERY_TO_MYSQL_ROUTE;
					} else if(second == to_be_joined_table) {
						if(!already_joined_table)
							already_joined_table = first;
						else if(already_joined_table != first)
							return RETURN_QUERY_TO_MYSQL_ROUTE;
					} else
						return RETURN_QUERY_TO_MYSQL_ROUTE;
					break;
				}
				case Item_func::MULT_EQUAL_FUNC: {

					//return RETURN_QUERY_TO_MYSQL_ROUTE;
					Item_equal_iterator li(*(Item_equal *) conds);

					int first = 0, second = 0;

					Item_field * ifield;
					while(ifield = li++) {
						int ret = GetTableIDFromFieldItem(ifield);
						if(!ret)
							return RETURN_QUERY_TO_MYSQL_ROUTE;
						if(!first)
							first = ret;
						else if(!second && ret != first)
							second = ret;
						else if(ret != first && ret != second) // some third table involved
							return RETURN_QUERY_TO_MYSQL_ROUTE;
					}
					// first and second are set
					if(!first || !second || first == second)
						return RETURN_QUERY_TO_MYSQL_ROUTE;
					if(first == to_be_joined_table) {
						if(!already_joined_table)
							already_joined_table = second;
						else if(already_joined_table != second)
							return RETURN_QUERY_TO_MYSQL_ROUTE;
					} else if(second == to_be_joined_table) {
						if(!already_joined_table)
							already_joined_table = first;
						else if(already_joined_table != first)
							return RETURN_QUERY_TO_MYSQL_ROUTE;
					} else
						return RETURN_QUERY_TO_MYSQL_ROUTE;
					break;
				}
				default:
					return RETURN_QUERY_TO_MYSQL_ROUTE; //unknown function type
			}
			break;
		}
		default:
			return RETURN_QUERY_TO_MYSQL_ROUTE;
	}
	return RCBASE_QUERY_ROUTE;
}

int Query::AddJoins(List<TABLE_LIST> join, TabID& tmp_table, bool first_table, bool is_left)
{
	// on first call first_table = true. It indicates if it is the first table to be added
	// is_left is true iff it is nested left join which needs to be flatten (all tables regardless of
	//                                                        their join type need to be left-joined)
	TABLE_LIST* join_ptr;
	List_iterator<TABLE_LIST> li(join);
	vector<TABLE_LIST*> reversed;

	if(!join.elements)
		return RETURN_QUERY_TO_MYSQL_ROUTE; // no tables in table list in this select
	//if the table list was empty altogether, we wouldn't even enter Compilation(...)
	//it must be sth. like `select 1 from t1 union select 2` and we are in the second
	//select in the union

	while(join_ptr = li++)
		reversed.push_back(join_ptr);
	size_t size = reversed.size();
	for(unsigned int i = 0; i < size; i++) {
		join_ptr = reversed[size - i - 1];
		const char* database_name = 0;
		const char* table_name = 0;
		const char* table_alias = 0;
		const char* table_path = 0;
		if(join_ptr->nested_join) {
			// for now there is no support for hard case of joins
			return RETURN_QUERY_TO_MYSQL_ROUTE;

			//			TabID tab;
			//			if (!AddJoins(join_ptr->nested_join->join_list, cq, table_alias2index, table_alias2ptr, field_aliases2index, tab))
			//				return RETURN_QUERY_TO_MYSQL_ROUTE;
			//			if (first_table) {
			//				tmp_table = tab;
			//				first_table = false;
			//			}
			//			else {
			//				cq->Join(tmp_table, tab);
			//
			//				JoinType join_type = GetJoinTypeAndCheckExpr(join_ptr->outer_join, join_ptr->on_expr);
			//
			//				if(!BuildFilterIfPossible						return RETURN_QUERY_TO_MYSQL_ROUTE;
			//			(join_ptr->on_expr, cq, table_alias2index, table_alias2ptr,
			//						field_aliases2index, tmp_table, join_type))
			//					return RETURN_QUERY_TO_MYSQL_ROUTE;
			//			}
		} else {
			/*
			* By MW. Related to bug 1073.
			* Below, another condition is necessary for tables that
			* come from a view that was merged into the query (as a join).
			* Such a table can be recognized by inspecting:
			*     join_ptr->belong_to_view
			* or  join_ptr->referencing_view
			*
			* For such a table, 'table_alias' must be created in a more
			* sophisticated way: by concatenating view(s)'s name(s)
			* and the table name.
			*/
			assert(join_ptr->table && "We require that the table is defined if it is not a nested join");
			TabID tab;
			if(join_ptr->derived) {
				map<string, int> table_alias2index;
				map<string, TABLE *> table_alias2ptr;
				if(!Compile(cq, join_ptr->derived->first_select(), join_ptr->derived->union_distinct, &table_alias2index, &table_alias2ptr, &tab))
					return RETURN_QUERY_TO_MYSQL_ROUTE;
				table_alias = join_ptr->alias;
			} else {
				if(!TableUnmysterify(join_ptr, database_name, table_name, table_alias, table_path))
					return RETURN_QUERY_TO_MYSQL_ROUTE;
				int tab_num = path2num[table_path]; //number of a table on a list in `this` QUERY object
				cq->TableAlias(tab, TabID(tab_num), table_name);
			}
			(*table_alias2ptr)[table_alias] = join_ptr->table;
			(*table_alias2index)[table_alias] = tab.n;
			if(first_table) {
				assert(((!join_ptr->on_expr) && (!is_left))
						&& "It is not possible to join the first table with the LEFT direction");
				cq->TmpTable(tmp_table, tab);
				first_table = false;
			} else {
				cq->Join(tmp_table, tab);
				JoinType join_type = GetJoinTypeAndCheckExpr(join_ptr->outer_join, join_ptr->on_expr);
				if(join_type == JO_LEFT && join_ptr->on_expr && dynamic_cast<Item_cond_or*>(join_ptr->on_expr))
					return RETURN_QUERY_TO_MYSQL_ROUTE;
				if(!BuildFilterIfPossible(join_ptr->on_expr, tmp_table, join_type))
					return RETURN_QUERY_TO_MYSQL_ROUTE;
				if(join_ptr->on_expr)
					cq->SetLOJOuterDim(tmp_table, tab, i);
			}
		}
	}
	return RCBASE_QUERY_ROUTE;
}

int Query::AddFields(List<Item> &fields, TabID const&tmp_table, bool const group_by_clause, int& num_of_added_fields, bool ignore_minmax, bool& aggregation_used)
{
	List_iterator_fast<Item> li(fields);
	Item * item;
	int added = 0;
	item = li++;
	while(item) {
		WrapStatus::wrap_status_t ws;
		ColOperation oper;
		bool distinct;
		if(!OperationUnmysterify(item, oper, distinct, group_by_clause))
			return RETURN_QUERY_TO_MYSQL_ROUTE;

		if(IsAggregationItem(item))
			aggregation_used = true;

		// in case of transformed subquery sometimes we need to revert back transformation to MIN/MAX
		if(ignore_minmax && (oper == MIN || oper == MAX))
			oper = LISTING;

		// select PHYSICAL COLUMN or AGGREGATION over PHYSICAL COLUMN
		if((IsFieldItem(item) || IsAggregationOverFieldItem(item)) && IsLocalColumn(item, tmp_table))
			AddColumnForPhysColumn(item, tmp_table, oper, distinct, false);
		//REF to FIELD_ITEM
		else if (item->type() == Item::REF_ITEM ) {
			item = UnRef(item);
			continue;
		}
		//			if ((UnRef(item)->type() == Item_bhfield::BHFIELD_ITEM || UnRef(item)->type() == Item_bhfield::FIELD_ITEM ) && IsLocalColumn(UnRef(item), tmp_table) )
		//				AddColumnForPhysColumn(UnRef(item), tmp_table, oper, distinct, false, false);
		//			else {
		//				//
		//			}
		else if (IsAggregationItem(item) && (((Item_sum*)item)->args[0])->type() == Item::REF_ITEM
			&& (UnRef(((Item_sum*)item)->args[0])->type() == Item_bhfield::BHFIELD_ITEM || (UnRef(((Item_sum*)item)->args[0])->type() == Item_bhfield::FIELD_ITEM))
			&& IsLocalColumn(UnRef(((Item_sum*)item)->args[0]), tmp_table))
			//AGGR on REF to FIELD_ITEM
			AddColumnForPhysColumn(UnRef(((Item_sum*)item)->args[0]), tmp_table, oper, distinct, false);
		else if(IsAggregationItem(item)) {
			// select AGGREGATION over EXPRESSION
			Item_sum* item_sum = (Item_sum*)item;
			if(item_sum->arg_count > 1  || HasAggregation(item_sum->args[0]))
				return RETURN_QUERY_TO_MYSQL_ROUTE;
			if(IsCountStar(item_sum)) { // count(*) doesn't need any virtual column
				AttrID at;
				cq->AddColumn(at, tmp_table, CQTerm(), oper, item_sum->name, false);
				field_alias2num[TabIDColAlias(tmp_table.n, item_sum->name)] = at.n;
			} else {
				MysqlExpression* expr;
				ws = WrapMysqlExpression(item_sum->args[0], tmp_table, expr, false, false);
				if(ws == WrapStatus::FAILURE)
					return RETURN_QUERY_TO_MYSQL_ROUTE;
				AddColumnForMysqlExpression(expr, tmp_table, ignore_minmax ? item_sum->args[0]->name : item_sum->name, oper, distinct);
			}
		} else if(item->type() == Item::SUBSELECT_ITEM) {
			CQTerm term;
			AttrID at;
			if (Item2CQTerm(item, term, tmp_table, group_by_clause ? HAVING_FILTER : WHERE_FILTER) == RETURN_QUERY_TO_MYSQL_ROUTE)
				return RETURN_QUERY_TO_MYSQL_ROUTE;
			cq->AddColumn(at, tmp_table, term, LISTING, item->name, distinct);
			field_alias2num[TabIDColAlias(tmp_table.n, item->name)] = at.n;
			coltab_to_vc[std::pair<int,int>(at.n, tmp_table.n)] = term.vc_id;
		} else {
			// select EXPRESSION
			if(HasAggregation(item))
				oper = DELAYED;
			MysqlExpression* expr(NULL);
			ws = WrapMysqlExpression(item, tmp_table, expr, false, oper == DELAYED);
			if(ws == WrapStatus::FAILURE)
				return RETURN_QUERY_TO_MYSQL_ROUTE;
			AddColumnForMysqlExpression(expr, tmp_table, item->name, oper, distinct);
		}
		added++;
		item = li++;
	}
	num_of_added_fields = added;
	return RCBASE_QUERY_ROUTE;
}

int Query::AddOrderByFields(ORDER *order_by, TabID const&tmp_table, int const group_by_clause)
{
	for(; order_by; order_by = order_by->next) {
		std::pair<int,int> vc;
		Item* item = *(order_by->item);
		CQTerm my_term;
		int result;
		// at first we need to check if we don't have non-deterministic expression (e.g., rand())
		// in such case we should order by output column in TempTable
		if(!IsFieldItem(item) && !IsAggregationItem(item) && item->type() != Item::SUBSELECT_ITEM) {
			MysqlExpression* expr = NULL;
			WrapStatus::wrap_status_t ws = WrapMysqlExpression(item, tmp_table, expr, false, false);
			if(ws == WrapStatus::FAILURE)
				return RETURN_QUERY_TO_MYSQL_ROUTE;
			if(!expr->IsDeterministic()) {
				int col_num = AddColumnForMysqlExpression(expr, tmp_table, NULL, LISTING, false, true);
				vc = VirtualColumnAlreadyExists(tmp_table, tmp_table, AttrID(-col_num - 1));
				if(vc.first == NULL_VALUE_32) {
					vc.first = tmp_table.n;
					cq->CreateVirtualColumn(vc.second, tmp_table, tmp_table, AttrID(col_num));
					phys2virt.insert(make_pair(std::make_pair(tmp_table.n, -col_num - 1), vc));
				}
				cq->Add_Order(tmp_table, AttrID(vc.second), !(order_by->asc));
				continue;
			}
		}
		if(group_by_clause)
			// we can reuse transformation done in case of HAVING
			result = Item2CQTerm(item, my_term, tmp_table, HAVING_FILTER);
		else
			result = Item2CQTerm(item, my_term, tmp_table, WHERE_FILTER);
		if(result != RCBASE_QUERY_ROUTE)
			return RETURN_QUERY_TO_MYSQL_ROUTE;
		vc.second = my_term.vc_id;
//			if(IsFieldItem(item) && IsLocalColumn(item, tmp_table)) {
//				AttrID col;
//				TabID tab;
//				if(!FieldUnmysterify(item, tab, col))
//					return RETURN_QUERY_TO_MYSQL_ROUTE;
//				vc.n = VirtualColumnAlreadyExists(tab, col);
//				if(vc.n == NULL_VALUE_32) {
//					cq->CreateVirtualColumn(vc, tmp_table, tab, col);
//					phys2virt[std::pair<int,int>(tab.n, col.n)] = vc.n;
//				}
//			} else if(IsAggregationItem(item)) {
//				continue; // no group by + aggregation/const = nothing to sort
//			} else if(item->type() == Item::SUBSELECT_ITEM) {
//				CQTerm term;
//				if (!Item2CQTerm(item, term, tmp_table, WHERE_FILTER))
//					return RETURN_QUERY_TO_MYSQL_ROUTE;
//				vc.n = term.vc_id;
//			} else { // expression
//				MysqlExpression* expr = NULL;
//				WrapStatus::wrap_status_t ws = WrapMysqlExpression(item, tmp_table, expr, false);
//				if(ws == WrapStatus::OPTIMIZE)
//					continue; // no group by + aggregation/const = nothing to sort
//				if(ws == WrapStatus::FAILURE)
//					return RETURN_QUERY_TO_MYSQL_ROUTE;
//				vc.n = VirtualColumnAlreadyExists(tmp_table, expr);
//				if(vc.n == NULL_VALUE_32) {
//					cq->CreateVirtualColumn(vc, tmp_table, expr, TabID(NULL_VALUE_32));
//					tab_id2expression.insert( make_pair(tmp_table, std::make_pair(vc.n, expr)) );
//				}
//			}
//		}
//
		cq->Add_Order(tmp_table, AttrID(vc.second), !(order_by->asc));
	}
	return RCBASE_QUERY_ROUTE;
}

int Query::AddGlobalOrderByFields(SQL_LIST *global_order, map<string, int> &field_alias2column_number, const TabID& tmp_table, int max_col)

{
	if(!global_order)
		return RCBASE_QUERY_ROUTE;

	ORDER * order_by;
	for(unsigned int i=0; i<global_order->elements; i++) {
		order_by = (i==0 ? (ORDER*)(global_order->first) : order_by->next);
		// the way to traverse 'global_order' list maybe is not very orthodox, but it works

		if ((*(order_by->item))->type() == Item::INT_ITEM) {
//			case Item::INT_ITEM: {
				int col_num = int((*(order_by->item))->val_int());
				if(col_num<=0 || col_num > max_col) { //not found
					my_message(ER_SYNTAX_ERROR,
							"Brighthouse specific error: Unknown column number in ORDER BY clause in UNION syntax", MYF(0));
					throw ReturnMeToMySQLWithError();
				};
				std::pair<int,int> vc;
				vc = VirtualColumnAlreadyExists(tmp_table, tmp_table, AttrID(col_num -1));
				if(vc.first == NULL_VALUE_32) {
					vc.first = tmp_table.n;
					cq->CreateVirtualColumn(vc.second, tmp_table, tmp_table, AttrID(col_num -1));
					phys2virt.insert(make_pair(std::make_pair(tmp_table.n, col_num -1), vc));
				}
				cq->Add_Order(tmp_table, AttrID(vc.second), !order_by->asc);
//				break;
		} else {
			std::pair<int,int> vc;
			AttrID col;
			TabID tab;
			CQTerm my_term;
			Item* item = *(order_by->item);
			if(!item->name)
				return false;

			if(field_alias2num.find(TabIDColAlias(tmp_table.n, item->name)) == field_alias2num.end())
				return false;
			col.n = field_alias2num[TabIDColAlias(tmp_table.n, item->name)];
			tab = tmp_table;
			vc = VirtualColumnAlreadyExists(tmp_table, tmp_table, col);
			if(vc.first == NULL_VALUE_32) {
				vc.first = tmp_table.n;
				cq->CreateVirtualColumn(vc.second, tmp_table, tmp_table, col);
				phys2virt.insert(make_pair(std::make_pair(tmp_table.n, col.n), vc));
			}
			cq->Add_Order(tmp_table, AttrID(vc.second), !order_by->asc);
		};
	}
	return RCBASE_QUERY_ROUTE;

}

//Query::WrapStatus::wrap_status_t Query::WrapMysqlExpressionWithAggregations(
//		Item *item, TabID tmp_table, MysqlExpression*& expr, bool* is_const_or_aggr /*= NULL*/)
//{
//	set<Item*> items;
//	WrapStatus::wrap_status_t ws;
//	bool has_aggr, is_const = true;
//	if(!MysqlExpression::SanityAggregationCheck(item, items, true, &has_aggr))
//		return WrapStatus::FAILURE;
//
//	AttrID at, vc;
//	MysqlExpression::Item2VarID item2varid;
//	for(set<Item*>::iterator it = items.begin(); it != items.end(); it++) {
//		if(IsAggregationItem(*it)) {
//			is_const = false;
//			Item_sum* aggregation = (Item_sum*)*it;
//			if(aggregation->arg_count > 1)
//				return WrapStatus::FAILURE;
//
//			if(IsCountStar(aggregation)) { // count(*) doesn't need any virtual column
//				at.n = GetAddColumnId(AttrID(NULL_VALUE_32), tmp_table, COUNT, false);
//				if(at.n == NULL_VALUE_32) // doesn't exist yet
//					cq->AddColumn(at, tmp_table, CQTerm(), COUNT, NULL, false);
//			} else {
//				AttrID col;
//				TabID tab;
//				if(!FieldUnmysterify(aggregation, tab, col))
//					return WrapStatus::FAILURE;
//				if(!cq->ExistsInTempTable(tab, tmp_table)) {
////					bool is_group_by;
////					tmp_table = cq->FindSourceOfParameter(tab, tmp_table, is_group_by);
//					ws = WrapMysqlExpression(aggregation->args[0], tmp_table, expr);
//					ColOperation oper;
//					bool distinct;
//					OperationUnmysterify(aggregation, oper, distinct, true);
//					at.n = AddColumnForMysqlExpression(expr, tmp_table, aggregation->name, oper, distinct, true);
//				} else {
//					ColOperation oper;
//					bool distinct;
//					OperationUnmysterify(aggregation, oper, distinct, true);
//					if(IsFieldItem(aggregation->args[0])) {
//						// PHYSICAL COLUMN
//						at.n = AddColumnForPhysColumn(aggregation, tmp_table, oper, distinct, true);
//					} else {
//						// EXPRESSION
//						ws = WrapMysqlExpression(aggregation->args[0], tmp_table, expr);
//						at.n = AddColumnForMysqlExpression(expr, tmp_table, aggregation->name, oper, distinct, true);
//					}
//				}
//			}
//			item2varid[*it] = VarID(tmp_table.n, at.n);
//		} else if(IsFieldItem(*it)) {
//			is_const = false;
//			AttrID col;
//			TabID tab;
//			if(!FieldUnmysterify(*it, tab, col))
//				return WrapStatus::FAILURE;
//			if(!cq->ExistsInTempTable(tab, tmp_table)) {
//				ColOperation oper;
//				bool distinct;
//				OperationUnmysterify(*it, oper, distinct, true);
//				ws = WrapMysqlExpression(*it, tmp_table, expr);
//				at.n = AddColumnForMysqlExpression(expr, tmp_table, (*it)->name, oper, distinct, true);
//				item2varid[*it] = VarID(tmp_table.n, at.n);
////				bool is_group_by;
////				tmp_table = cq->FindSourceOfParameter(tab, tmp_table, is_group_by);
////				ColOperation oper;
////				bool distinct;
////				OperationUnmysterify(*it, oper, distinct, true);
////				at.n = AddColumnForPhysColumn(*it, tmp_table, oper, distinct, true);
////				if(is_group_by)
////					item2varid[*it] = VarID(tmp_table.n, at.n);
////				else
////					item2varid[*it] = VarID(tab.n, col.n);
//			} else if(has_aggr) {
//				ColOperation oper;
//				bool distinct;
//				OperationUnmysterify(*it, oper, distinct, true);
//				at.n = AddColumnForPhysColumn(*it, tmp_table, oper, distinct, true);
//				item2varid[*it] = VarID(tmp_table.n, at.n);
//			} else {
//				item2varid[*it] = VarID(tab.n, col.n);
//			}
//		} else
//			assert(0); // unknown item type?
//	}
//	if(is_const_or_aggr) {
//		if(is_const || has_aggr)
//			*is_const_or_aggr = true;
//		else
//			*is_const_or_aggr = false;
//	}
//	gc_expressions.push_back(expr = new MysqlExpression(item, item2varid));
//	return WrapStatus::SUCCESS;
//}

Query::WrapStatus::wrap_status_t Query::WrapMysqlExpression(Item *item, TabID tmp_table, MysqlExpression*& expr, bool in_where, bool aggr_used)
{
	// Check if the expression doesn't contain any strange items that we don't want to see.
	// By the way, collect references to all Item_field objects.
	set<Item*> ifields;
	MysqlExpression::Item2VarID item2varid;
	if(!MysqlExpression::SanityAggregationCheck(item, ifields))
		return WrapStatus::FAILURE;

	//this large "if" can be removed to use common code, but many small "ifs" must be created then
	if(in_where) {
		// create a map:  [Item_field pointer] -> VarID
		for(set<Item*>::iterator it = ifields.begin(); it != ifields.end(); it++) {
			if(IsAggregationItem(*it)) {
				// a few checkings for aggregations
				Item_sum* aggregation = (Item_sum*)*it;
				if(aggregation->arg_count > 1)
					return WrapStatus::FAILURE;
				if(IsCountStar(aggregation)) { // count(*) doesn't need any virtual column
					assert(!"Count(*) cannot be here");
					return WrapStatus::FAILURE;
				} 
			}
			AttrID col, at;
			TabID tab;
			// find [tab] and [col] which identify column in BH
			if(!FieldUnmysterify(*it, tab, col))
				return WrapStatus::FAILURE;
			if(!cq->ExistsInTempTable(tab, tmp_table)) {
				bool is_group_by;
				TabID params_table = cq->FindSourceOfParameter(tab, tmp_table, is_group_by);
				ColOperation oper;
				bool distinct;
				OperationUnmysterify(*it, oper, distinct, true);
				if(is_group_by && !IsParameterFromWhere(params_table)) {
					col.n = AddColumnForPhysColumn(*it, params_table, oper, distinct, true);
					item2varid[*it] = VarID(params_table.n, col.n);
				} else
					item2varid[*it] = VarID(tab.n, col.n);
			} else {
				// aggregation in WHERE not possible unless it is a parameter
				BHASSERT_WITH_NO_PERFORMANCE_IMPACT(!IsAggregationItem(*it));
				item2varid[*it] = VarID(tab.n, col.n);
			}
		}
	} else { // !in_where
		bool is_const = true;
		WrapStatus::wrap_status_t ws;
		AttrID at, vc;
		for(set<Item*>::iterator it = ifields.begin(); it != ifields.end(); it++) {
			if(IsAggregationItem(*it)) {
				is_const = false;
				Item_sum* aggregation = (Item_sum*)*it;
				if(aggregation->arg_count > 1)
					return WrapStatus::FAILURE;

				if(IsCountStar(aggregation)) { // count(*) doesn't need any virtual column
					at.n = GetAddColumnId(AttrID(NULL_VALUE_32), tmp_table, COUNT, false);
					if(at.n == NULL_VALUE_32) // doesn't exist yet
						cq->AddColumn(at, tmp_table, CQTerm(), COUNT, NULL, false);
				} else {
					ColOperation oper;
					bool distinct;
					OperationUnmysterify(aggregation, oper, distinct, true);
					AttrID col;
					TabID tab;
					if(IsFieldItem(aggregation->args[0]) && FieldUnmysterify(aggregation, tab, col) && cq->ExistsInTempTable(tab, tmp_table)) {
						// PHYSICAL COLUMN
						at.n = AddColumnForPhysColumn(aggregation->args[0], tmp_table, oper, distinct, true);
					} else {
						// EXPRESSION
						ws = WrapMysqlExpression(aggregation->args[0], tmp_table, expr, in_where, false);
						if(ws == WrapStatus::FAILURE)
							return ws;
						at.n = AddColumnForMysqlExpression(expr, tmp_table, aggregation->name, oper, distinct, true);
					}
				}
				item2varid[*it] = VarID(tmp_table.n, at.n);
			} else if(IsFieldItem(*it)) {
				is_const = false;
				AttrID col;
				TabID tab;
				if(!FieldUnmysterify(*it, tab, col))
					return WrapStatus::FAILURE;
				if(!cq->ExistsInTempTable(tab, tmp_table)) {
					bool is_group_by;
					TabID params_table = cq->FindSourceOfParameter(tab, tmp_table, is_group_by);
					ColOperation oper;
					bool distinct;
					OperationUnmysterify(*it, oper, distinct, true);
					if(is_group_by && !IsParameterFromWhere(params_table)) {
						col.n = AddColumnForPhysColumn(*it, params_table, oper, distinct, true);
						item2varid[*it] = VarID(params_table.n, col.n);
					} else
						item2varid[*it] = VarID(tab.n, col.n);
				} else if(aggr_used) {
					ColOperation oper;
					bool distinct;
					OperationUnmysterify(*it, oper, distinct, true);
					at.n = AddColumnForPhysColumn(*it, tmp_table, oper, distinct, true);
					item2varid[*it] = VarID(tmp_table.n, at.n);
				} else {
					item2varid[*it] = VarID(tab.n, col.n);
				}
			} else
				assert(0); // unknown item type?
		}
	}
	gc_expressions.push_back(expr = new MysqlExpression(item, item2varid));
	return WrapStatus::SUCCESS;
}

int Query::AddColumnForPhysColumn(Item* item, const TabID tmp_table, const ColOperation oper, const bool distinct, bool group_by)
{
	std::pair<int,int> vc;
	AttrID col, at;
	TabID tab;
	if(!FieldUnmysterify(item, tab, col))
		return NULL_VALUE_32;
	if(tab.n == NULL_VALUE_32)
		tab = tmp_table; 	//table name not contained in item - must be the result temp_table

	assert(cq->ExistsInTempTable(tab, tmp_table));
//		// treat it as a constant with parameter
//		MysqlExpression* expr;
//		WrapStatus::wrap_status_t ws = WrapMysqlExpression(item, tmp_table, expr, in_where);
//		assert(ws == WrapStatus::SUCCESS);
//		return AddColumnForMysqlExpression(expr, tmp_table, item->name, oper, distinct, group_by );
//	} else {
	if(item->type() == Item_bhfield::BHFIELD_ITEM && IsAggregationItem(dynamic_cast<Item_bhfield *>(item)->OriginalItem()))
		return ((Item_bhfield*)item)->varID.col;

	vc = VirtualColumnAlreadyExists(tmp_table, tab, col);
	if(vc.first == NULL_VALUE_32) {
		vc.first = tmp_table.n;
		cq->CreateVirtualColumn(vc.second, tmp_table, tab, col);
		phys2virt.insert(make_pair(std::make_pair(tab.n, col.n), vc));
	} else {
		int attr = GetAddColumnId(AttrID(vc.second), tmp_table, oper, distinct);
		if(attr != NULL_VALUE_32) {
			if(group_by) // do not add column - not needed duplicate
				return attr;
			//vc.n = col_to_vc[attr];
		} else if(group_by && oper == GROUP_BY && (attr = GetAddColumnId(AttrID(vc.second), tmp_table, LISTING, distinct)) != NULL_VALUE_32) {
			// modify existing column
			CQChangeAddColumnLIST2GROUP_BY(tmp_table, attr);
			return attr;
		} else if(group_by && oper == LISTING && (attr = GetAddColumnId(AttrID(vc.second), tmp_table, GROUP_BY, distinct)) != NULL_VALUE_32) {
			// don;t add unnecessary column to select list
			return attr;
		}
	}
	if(!item->name && item->type() == Item::SUM_FUNC_ITEM)
		cq->AddColumn(at, tmp_table, CQTerm(vc.second), oper, group_by ? NULL : ((Item_field*)(((Item_sum*)item)->args[0]))->name, distinct);
	else
		cq->AddColumn(at, tmp_table, CQTerm(vc.second), oper, group_by ? NULL : item->name, distinct);
	if(!group_by && item->name)
		field_alias2num[TabIDColAlias(tmp_table.n, item->name)] = at.n;
	coltab_to_vc[std::pair<int,int>(at.n, tmp_table.n)] = vc.second;
	return at.n;
}

int Query::AddColumnForMysqlExpression(MysqlExpression* mysql_expression,
		const TabID tmp_table, const char* alias, const ColOperation oper, const bool distinct, bool group_by /*= false*/)
{
	AttrID at, vc;
	vc.n = VirtualColumnAlreadyExists(tmp_table, mysql_expression);
	if(vc.n == NULL_VALUE_32) {
		cq->CreateVirtualColumn(vc, tmp_table, mysql_expression, (oper == DELAYED ? tmp_table : TabID(NULL_VALUE_32)));
		tab_id2expression.insert(make_pair(tmp_table, std::make_pair(vc.n, mysql_expression)));
	} else {
		int attr = GetAddColumnId(vc, tmp_table, oper, distinct);
		if(attr != NULL_VALUE_32) {
			if(group_by) // do not add column - not needed duplicate
				return attr;
			//vc.n = col_to_vc[attr];
		} else if(group_by && oper == GROUP_BY && (attr = GetAddColumnId(vc, tmp_table, LISTING, distinct)) != NULL_VALUE_32) {
			// modify existing column
			CQChangeAddColumnLIST2GROUP_BY(tmp_table, attr);
			return attr;
		} else if(group_by && oper == LISTING && (attr = GetAddColumnId(vc, tmp_table, GROUP_BY, distinct)) != NULL_VALUE_32) {
			// don;t add unnecessary column to select list
			return attr;
		}
	}

//	bool parametrized = false;
//	if(group_by) {
//		MysqlExpression::SetOfVars& vars = mysql_expression->GetVars();
//		for(MysqlExpression::SetOfVars::iterator iter = vars.begin(); iter != vars.end(); iter++) {
//			if(! cq->ExistsInTempTable(TabID(iter->tab), tmp_table)){
//				parametrized = true;
//				break;
//			}
//		}
//	}
	//if (parametrized)
	//	cq->AddColumn(at, tmp_table, CQTerm(vc.n), DELAYED, group_by ? NULL : alias, distinct);
	//else
		cq->AddColumn(at, tmp_table, CQTerm(vc.n), oper, group_by ? NULL : alias, distinct);
	if(!group_by && alias)
		field_alias2num[TabIDColAlias(tmp_table.n, alias)] = at.n;
	coltab_to_vc[std::pair<int,int>(at.n, tmp_table.n)] = vc.n;
	return at.n;
}

int Query::AddGroupByFields(ORDER * group_by, const TabID& tmp_table)
{
	for(; group_by; group_by = group_by->next) {
		if(!group_by->asc) {
			my_message(ER_SYNTAX_ERROR,
				"Brighthouse specific error: Using DESC after GROUP BY clause not allowed. Use ORDER BY to order the result",
				MYF(0));
			throw ReturnMeToMySQLWithError();
		}

		Item* item = *(group_by->item);
		// group by PHYSICAL COLUMN
		if((IsFieldItem(item) || (IsAggregationItem(item) && IsFieldItem(((Item_sum*)item)->args[0]))) && IsLocalColumn(item, tmp_table))
			AddColumnForPhysColumn(item, tmp_table, GROUP_BY, false, true);
		else if (item->type() == Item::SUBSELECT_ITEM) {
			CQTerm term;
			AttrID at;
			if (Item2CQTerm(item, term, tmp_table, WHERE_FILTER) == RETURN_QUERY_TO_MYSQL_ROUTE)
				return RETURN_QUERY_TO_MYSQL_ROUTE;
			cq->AddColumn(at, tmp_table, term, GROUP_BY, 0);
//			field_alias2num[TabIDColAlias(tmp_table.n, item->name)] = at.n;
			coltab_to_vc[std::pair<int,int>(at.n, tmp_table.n)] = term.vc_id;
		} else { // group by COMPLEX EXPRESSION
			MysqlExpression* expr = 0;
			if(WrapStatus::FAILURE == WrapMysqlExpression(item, tmp_table, expr, true, true))
				return RETURN_QUERY_TO_MYSQL_ROUTE;
			AddColumnForMysqlExpression(expr, tmp_table, item->name, GROUP_BY, false, true);
		}
	}
	return RCBASE_QUERY_ROUTE;
}

bool Query::IsLocalColumn(Item* item, TabID tmp_table)
{
	assert(IsFieldItem(item) || IsAggregationItem(item));
	AttrID col;
	TabID tab;
	if(!FieldUnmysterify(item, tab, col))
		return false;
	return cq->ExistsInTempTable(tab, tmp_table);
}

int JudgeErrors(SELECT_LEX * sl)
{
	if(!sl->join) {
		return RETURN_QUERY_TO_MYSQL_ROUTE;
	}

	if(sl->join && sl->join->procedure) {
		my_message(ER_SYNTAX_ERROR, "Brighthouse specific error: No PROCEDURE syntax supported", MYF(0));
		throw ReturnMeToMySQLWithError();
	}

	if(sl->offset_limit)
		if(sl->offset_limit->type() != Item::INT_ITEM /*|| sl->offset_limit->val_int()*/) {
			//my_message(ER_SYNTAX_ERROR, "Brighthouse specific error: Only OFFSET=0 supported in LIMIT clause", MYF(0));
			my_message(ER_SYNTAX_ERROR, "Brighthouse specific error: Only numerical OFFSET supported", MYF(0));
			throw ReturnMeToMySQLWithError();
		}

	if(sl->select_limit)
		if(sl->select_limit->type() != Item::INT_ITEM) {
			my_message(ER_SYNTAX_ERROR, "Brighthouse specific error: Only numerical LIMIT supported", MYF(0));
			throw ReturnMeToMySQLWithError();
		}
	if(sl->olap == ROLLUP_TYPE) {
		/*my_message(ER_SYNTAX_ERROR, "Brighthouse specific error: WITH ROLLUP not supported", MYF(0));
		 throw ReturnMeToMySQLWithError();*/
		return RETURN_QUERY_TO_MYSQL_ROUTE;
	}

	return RCBASE_QUERY_ROUTE;
}

void SetLimit(SELECT_LEX * sl, SELECT_LEX * gsl, _int64 &offset_value, _int64 &limit_value)
{

	if(sl->select_limit && (!gsl || sl->select_limit != gsl->select_limit)) {
		limit_value = sl->select_limit->val_int();
		if(limit_value == UINT_MAX) {/* this value seems to be sometimes automatically set by MYSQL to UINT_MAX*/
			limit_value = -1; //no limit set
			offset_value = -1;
		};
	};

	if(limit_value)
		offset_value = 0;

	if(sl->offset_limit && (!gsl || sl->offset_limit != gsl->offset_limit))
		offset_value = sl->offset_limit->val_int();

}

bool Query::IsLOJ(List<TABLE_LIST> *join)
{
	TABLE_LIST* join_ptr;
	List_iterator<TABLE_LIST> li(*join);
	while(join_ptr = li++) {
		JoinType join_type = GetJoinTypeAndCheckExpr(join_ptr->outer_join, join_ptr->on_expr);
		if(join_ptr->on_expr && (join_type == JO_LEFT || join_type == JO_RIGHT))
			return true;
	}
	return false;
}

// Used in Query::Compile() to break compilation in the middle and make cleanup before returning
class CompilationError {};

int Query::Compile(CompiledQuery* compiled_query, SELECT_LEX* selects_list, SELECT_LEX* last_distinct,
		map<string, int>* outer_table_alias2index, map<string, TABLE*>* outer_table_alias2ptr, TabID* res_tab, bool ignore_limit,
		Item* left_expr_for_subselect, Operator* oper_for_subselect, bool ignore_minmax)
{
	// at this point all tables are in RCBase engine, so we can proceed with the query

	///////////(remove it) my experiments: push_warning(current_thd, MYSQL_ERROR::WARN_LEVEL_ERROR, 7, "we start compilation");

	/*Item_func
	 |
	 --Item_int_func  <-  arguments are kept in an array accessible through arguments()
	 |
	 --Item_bool_func
	 |   |
	 |   ---Item_cond  <- arguments are kept in a list accessible through argument_list()
	 |   |     |
	 |   |     ---Item_cond_and   <- when negated OR of negated items is created
	 |   |     |
	 |   |     ---Item_cond_or    <- when negated AND of negated items is created
	 |   |     |
	 |   |     ---Item_cond_xor
	 |   |
	 |   ---Item_equal  <- arguments are kept in a list accessible through argument_list()
	 |   |                 + const_item (accesible through get_const() )
	 |   |           (multiple equality)
	 |   |
	 |   ---Item_func_not
	 |   |            (???)
	 |   |
	 |   ---Item func_isnull     <- when negated IS NOT NULL is created
	 |
	 --Item_func_opt_neg  <-  arguments are kept in an array accessible through arguments(), if negated
	 |   |                     this information is kept additionally (in a field named 'negated')
	 |   |
	 |   |
	 |   ---Item_func_in
	 |   |
	 |   |
	 |   ---Item_func_between
	 |
	 |
	 --Item_bool_func2
	 |
	 |
	 ---Item_bool_rowready_func2  <-arguments are kept in an array accessible through arguments(), if negated
	 |                          an object of a corresponding class is created
	 |                           (e.q. ~Item_func_lt => Item_func_ge)
	 |
	 ----Item_func_eq
	 |
	 |
	 ----Item_func_ne
	 |
	 |
	 ----Item_func_ge
	 |
	 |
	 ----Item_func_le
	 |
	 |
	 ----Item_func_gt
	 |
	 |
	 ----Item_func_lt
	 |
	 |
	 ----Item_func_equal   <- This is mystery so far

	 There are 3 equality functions:
	 Item_equal -> multiple equality (many fields and optional additional constant value)
	 Item_func_equal -> ???
	 Item_func_eq -> pairwise equality


	 */

	//	if (tables ->next_leaf)    //make sure there are no JOINs
	//		return RETURN_QUERY_TO_MYSQL_ROUTE;

	bool union_all = (last_distinct==NULL);
	TabID prev_result;

	SQL_LIST *global_order = 0;

	//  the following 2 variables are important in UNIONs and are used by ORDER BY for the UNION syntax
	//they contain the names of columns & the column-count in the first SELECT in a UNION (respectively)
	map<string, int> field_alias2column_number;
	int col_count = 0;
	_int64 global_limit_value = -1;
	_int64 global_offset_value = -1;

	// local copy of current this->cq, to be restored on exit
	CompiledQuery* _cq = this->cq;
	this->cq = compiled_query;

	for(SELECT_LEX *sl= selects_list; sl; sl= sl->next_select()) {
		if(sl!=selects_list) { //only in case of unions this is set
			SetLimit(sl->join->unit->global_parameters, 0, global_offset_value, (_int64&)global_limit_value);
			global_order = &(sl->join->unit->global_parameters->order_list);
		};

		_int64 limit_value = -1;
		_int64 offset_value = -1;

		if(!JudgeErrors(sl))
			return RETURN_QUERY_TO_MYSQL_ROUTE;
		SetLimit(sl, sl==selects_list ? 0 : sl->join->unit->global_parameters, offset_value, limit_value);

		List<Item>* fields = &sl->join->fields_list;
		COND *conds = sl->join->conds;
		ORDER *order = sl->join->order;

		//if (order) global_order = 0;   //we want to zero global order (which seems to be always present)
		//if we find a local order by clause
		//  The above is not necessary since global_order is set only in case of real UNIONs

		ORDER *group = sl->join->group_list;
		COND *having = sl->join->having;
		List<TABLE_LIST> *join_list = sl->join->join_list;
		bool zero_result = sl->join->zero_result_cause != NULL;

		Item* field_for_subselect;
		Item* cond_to_reinsert= NULL;
		List<Item>* list_to_reinsert= NULL;

		// keep local copies of class fields to be changed
		map<string,int>* _table_alias2index = this->table_alias2index;
		map<string,TABLE *>* _table_alias2ptr = this->table_alias2ptr;

		// structures that will be used throughout compilation
		map<string, int> local_table_alias2index = *outer_table_alias2index;
		map<string, TABLE *> local_table_alias2ptr = *outer_table_alias2ptr;

		// assign pointers to above structures onto class fields
		this->table_alias2index = &local_table_alias2index;
		this->table_alias2ptr = &local_table_alias2ptr;

		TabID tmp_table;

		try {
			// partial optimization of LOJ conditions, JOIN::optimize(part=3)
			// necessary due to already done basic transformation of conditions
			// see comments in sql_select.cc:JOIN::optimize()
			if(IsLOJ(join_list))
				sl->join->optimize(3);

			if(left_expr_for_subselect)
				if(!ClearSubselectTransformation(*oper_for_subselect, field_for_subselect, conds, having,
						cond_to_reinsert, list_to_reinsert, left_expr_for_subselect))
					throw CompilationError();

			if(having && !group) //we cannot handle the case of a having without a group by
				throw CompilationError();

			//TABLE_LIST *tables = sl->leaf_tables;
			TABLE_LIST *tables = sl->leaf_tables ? sl->leaf_tables : (TABLE_LIST*)sl->table_list.first;
			for(TABLE_LIST* table_ptr = tables; table_ptr; table_ptr = table_ptr->next_leaf) {
				if(!table_ptr->derived) {
					if(!RCEngine::IsBHTable(table_ptr->table))
						throw CompilationError();
					string path = TablePath(table_ptr);
					if(path2num.find(path)==path2num.end()) {
						path2num[path] = no_tab;
						Add(rceng->GetTable(path, table_ptr->table->s));
					}
				}
			}
			if(!AddJoins(*join_list, tmp_table))
				throw CompilationError();

//			if(!BuildFilter(conds, cq, table_alias2index, table_alias2ptr, tmp_table, WHERE_FILTER, zero_result))
//				throw CompilationError();

			List<Item> field_list_for_subselect;
			if(left_expr_for_subselect && field_for_subselect) {
				field_list_for_subselect.push_back(field_for_subselect);
				fields = &field_list_for_subselect;
			}
			bool aggr_used = false;
			if(!AddFields(*fields, tmp_table, group != NULL, col_count, ignore_minmax, aggr_used ))
				throw CompilationError();

			if(!AddGroupByFields(group, tmp_table))
				throw CompilationError();

			if(!AddOrderByFields(order, tmp_table, group != NULL || sl->join->select_distinct || aggr_used))
				throw CompilationError();

			if(!BuildFilter(conds, cq, table_alias2index, table_alias2ptr, tmp_table, WHERE_FILTER, zero_result))
				throw CompilationError();

			if(!BuildFilter(having, cq, table_alias2index, table_alias2ptr, tmp_table, HAVING_FILTER))
				throw CompilationError();
		}
		catch(...) {
			// restore original values of class fields (necessary if this method is called recursively)
			this->cq = _cq;
			this->table_alias2index = _table_alias2index;
			this->table_alias2ptr = _table_alias2ptr;
			if(cond_to_reinsert && list_to_reinsert)
				list_to_reinsert->push_back(cond_to_reinsert);
			return RETURN_QUERY_TO_MYSQL_ROUTE;
		}

		// restore original values of class fields (necessary if this method is called recursively)
		this->table_alias2index = _table_alias2index;
		this->table_alias2ptr = _table_alias2ptr;

		if(sl->join->select_distinct)
			cq->Mode(tmp_table, TM_DISTINCT);
		if(!ignore_limit && limit_value >= 0)
			cq->Mode(tmp_table, TM_TOP, offset_value, limit_value);

		if(sl==selects_list)
			prev_result = tmp_table;
		else
			cq->Union(prev_result, prev_result, tmp_table, union_all);
		if(sl==last_distinct)
			union_all = true;
		if(cond_to_reinsert && list_to_reinsert)
			list_to_reinsert->push_back(cond_to_reinsert);
	}

	if(!AddGlobalOrderByFields(global_order, field_alias2column_number, prev_result, col_count))
		return RETURN_QUERY_TO_MYSQL_ROUTE;

	if(!ignore_limit && global_limit_value>=0)
		cq->Mode(prev_result, TM_TOP, global_offset_value, global_limit_value);

	if(res_tab!=NULL)
		*res_tab = prev_result;
	else
		cq->Result(prev_result);

	this->cq = _cq;

	return RCBASE_QUERY_ROUTE;

}

JoinType Query::GetJoinTypeAndCheckExpr(uint outer_join, Item *on_expr)
{
	if(outer_join)
		BHASSERT(on_expr != 0, "on_expr shouldn't be null when outer_join != 0");

	JoinType join_type;

	if((outer_join & JOIN_TYPE_LEFT) && (outer_join & JOIN_TYPE_RIGHT))
		join_type = JO_FULL;
	else if(outer_join & JOIN_TYPE_LEFT)
		join_type = JO_LEFT;
	else if(outer_join & JOIN_TYPE_RIGHT)
		join_type = JO_RIGHT;
	else
		join_type = JO_INNER;

	return join_type;
}

