/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include <boost/bind.hpp>

#include "common/CommonDefinitions.h"
#include "RCAttr.h"
#include "RCAttrPack.h"
#include "RCAttrTypeInfo.h"
#include "tools.h"
#include "system/ConnectionInfo.h"
#include "m_string.h"
#include "types/ValueParserForText.h"
#include "DataPackImpl.h"
#include "DPN.h"
#include "shared/common/DataFormat.h"
#include "DataCache.h"
#include "Transaction.h"

#ifdef USE_NEW_ENGINE
#include "types/AValue.h"
#include "types/VString.h"
#endif

#include "util/BHString.h"
#include "system/IBFile.h"

using namespace bh;
using namespace std;
using namespace boost;

enum { VAL_BUF_MIN_COUNT = 24 };

void RCAttr::CalcPackFileSizeLimit(int a_field_size)
{
	// Determine pack file size limit
	// Smaller file => more system overload and size limitations (file naming!).
	// Larger size  => harder to backup; cannot exceed 2.1 GB!
	if (a_field_size > 16000)
		file_size_limit = 1000000000;
	else
		// Default size: the safe distance from 2 GB
		file_size_limit = 2000000000 - 2*65536*size_t(a_field_size);

	try
	{
		size_t new_file_size_limit = 1000000 *
			size_t(Configuration::GetProperty(Configuration::ClusterSize));
		if(new_file_size_limit < file_size_limit)
			file_size_limit = new_file_size_limit;
	}
	catch(...)
	{
		// Just use the default values
	}
	if (file_size_limit < 10000000)
		file_size_limit = 10000000;
	return;
}

RCAttr::RCAttr(char *a_name, int a_num, int t_num,
	AttributeType a_type, int a_field_size,
	int a_dec_places, char *a_path,
	uint param, uint s_id, DTCollation collation)
	: packs_omitted(0)
{
	fill(&fname_buf[0], &fname_buf[FNAME_BUFFER_SIZE], 0);
	fill(&path[0], &path[MAX_PATH], 0);

	SetFileFormat(10);
	CalcPackFileSizeLimit(a_field_size);

	SetNoPack(0);
	SetNoPackPtr(0);
	pack_mode = NULL;
	SetTotalPackFile(1);

	pack_file = (int*)alloc(sizeof(int), BLOCK_TEMPORARY);

        // Number of file where i-th pack is located. Here: "no objects".
	pack_file[0] = PF_NO_OBJ;
        SetSaveFileLoc(0, 0);
	SetSaveFileLoc(1, 1);
	SetSavePosLoc(0, 0);
	SetSavePosLoc(1, 0);
	SetCurReadLocation(0);
	SetCurSaveLocation(0);

	SetSessionId(0);
	SetSaveSessionId(s_id);
        // Address (byte) of the beginning of i-th pack in the file
        pack_addrs = NULL;
        rsi_hist_update = NULL;
	rsi_cmap_update = NULL;

	packs = 0;
	SetLastPackIndex(0);
	SetNoObj(0);

	NullMode nulls;
	bool is_lookup;

	is_lookup = (param & 0x00000002) != 0;
    // automatic LOOKUP for small columns - DISABLED AS TOO RISKY
//    if((a_type == RC_STRING || a_type == RC_VARCHAR) && a_field_size <= 2)
//		is_lookup = true;
	if(a_type != RC_STRING && a_type != RC_VARCHAR)
		is_lookup = false;				// note: is_lookup  =>  RC_STRING || RC_VARCHAR
    if(param & 0x00000001)
    	nulls = NO_NULLS;
	else
    	nulls = AS_MISSED;

    if(param & 0x00000004)
    	inserting_mode = 1;
    else
    	inserting_mode = 0;

	if (a_field_size <= 0)
	{
		unsigned short int fsize;
                // maximum size of double for gcvt(...,15,...)
        if(a_type==RC_REAL) fsize = 23;
        else if(a_type==RC_FLOAT) fsize = 15;
		else if(a_type==RC_BYTEINT) fsize = 4;
		else if(a_type==RC_SMALLINT) fsize = 6;
		else if(a_type==RC_MEDIUMINT) fsize = 8;
		else if(a_type==RC_INT) fsize = 11;
		else if(a_type==RC_BIGINT) fsize = 20;
		else if(a_type==RC_NUM) fsize = 18;
		else if(a_type==RC_DATE) fsize = 10;
		else if(a_type==RC_TIME) fsize = 10;
		else if(a_type==RC_YEAR) fsize = 4;
		else if(a_type == RC_DATETIME || a_type == RC_TIMESTAMP) fsize = 19;
		a_field_size  = fsize;
	}

	ct.Initialize(a_type, nulls, is_lookup, a_field_size, a_dec_places, collation);

	if(Type().IsLookup() || ATI::IsNumericType(a_type) || ATI::IsDateTimeType(a_type))
		SetPackType(PackN);
	else
		SetPackType(PackS);

	int val_buf_size = Type().GetDisplaySize() + 1;
	if (val_buf_size < VAL_BUF_MIN_COUNT)
		val_buf_size=VAL_BUF_MIN_COUNT;		// minimal RC_NUM buffer size

	SetMinInt64(0);
	SetMaxInt64(0);
	current_state = 1;				// read/write session

	SetNoNulls(0);
	SetNaturalSizeSaved(0);
	SetCompressedSize(0);
	is_stored = 0;
	local_min = NULL;
	local_max = NULL;				// Statistics (in 1-level codes) for data packs, or some indexing values for TEXT, RC_BIN
									// level_1_value = local_min + level_2_value
	local_sum = NULL;				// sum in 1-level _int64 encoding
	local_size = NULL;				// maximum byte size of string in pack
	loc_no_obj = NULL;				// number of objects
	loc_no_nulls = NULL;			// number of nulls
	rough_selectivity = -1;
	SetUnique(true);				// all non-null values are different (because there are no values at all)
	SetUniqueUpdated(true);		// true if the information in "is_unique" is up to date

	if(Type().IsLookup())
		dic.Init(Type().GetPrecision());

	attr_number = a_num;	// column number (for file naming purposes etc.)
	table_number = t_num;

	SetName(a_name);
	SetDescription(NULL);

	strcpypath(path, a_path);

	SetDictOffset(0);
	SetPackOffset(0);
	SetPackInfoCollaped((GetDictOffset() || NoPack()));
	file = FILE_READ_SESSION;

	//pthread_mutexattr_t mattr;
	//pthread_mutexattr_init(&mattr);
	//pthread_mutexattr_settype(&mattr, PTHREAD_MUTEX_RECURSIVE);
	//pthread_mutex_init(&cs, &mattr);

	ComputeNaturalSize();
}

bool RCAttr::HasUnCommitedSession()
{
	char* attrf_name = AttrFileName(fname_buf, 1024, FILE_SAVE_SESSION);
	return (DoesFileExist(attrf_name));
}

unsigned int RCAttr::ReadUnCommitedSessionID()
{
	IBFile fattr_save;
	char* attrf_name;
	int fid_attrsave = 0;
	unsigned int session_id = INVALID_TRANSACTION_ID;
	unsigned char a_file_format;
	char* tmp_buf = new char [10000];

	attrf_name = AttrFileName(fname_buf, 1024, FILE_SAVE_SESSION);
	fid_attrsave = fattr_save.OpenReadOnly(attrf_name);
	fattr_save.Read(tmp_buf, CLMD_HEADER_LEN);
	a_file_format = FILEFORMAT(tmp_buf);
	if (a_file_format == CLMD_FORMAT_INVALID_ID)
	{
		fattr_save.Close();
		delete [] tmp_buf;
		rclog << lock << "Unsupported version (not an attribute), or file does not exist: " << attrf_name << unlock;
		string mess = (string)"Unsupported version (not an attribute), or file does not exist: " + attrf_name;
		throw DatabaseRCException(mess.c_str());
	}

	uint sp_offset = *((uint*)(tmp_buf + 42));
	if (sp_offset > 0)										// Note that it is assumed that the first special block contain session info
	{
		fattr_save.Seek(sp_offset + 22, SEEK_SET);
		fattr_save.Read(tmp_buf, 46);
		session_id = *((uint*)(tmp_buf + 1));
	}
	fattr_save.Close();

	delete [] tmp_buf;
	return session_id;
}

RCAttr::RCAttr(int a_num, int t_num, char *a_path, int conn_mode, uint s_id, DTCollation collation) throw(DatabaseRCException)
	:	packs_omitted(0)
{
	fill(&fname_buf[0], &fname_buf[FNAME_BUFFER_SIZE], 0);
	fill(&path[0], &path[MAX_PATH], 0);
	ct.Initialize(RC_INT, AS_MISSED, false, 0, 0, collation);
	rsi_hist_update = NULL;
	rsi_cmap_update = NULL;
	attr_number = a_num;				// index of the column in table (for file naming purposes etc.)
	table_number = t_num;

	strcpypath(path, a_path);
	current_state = conn_mode;			// conn_mode is used to determine initial state; current_state may change

	// Determine currently opened save session_id and save_loc (if any)
	SetSessionId(0);
	SetSaveSessionId(INVALID_TRANSACTION_ID);		// this value indicates that there is no session opened
	SetCurSaveLocation(0);

	int new_save_session = 0;			// this flag will indicate that this is a new session and we should switch save locations

	//IBFile fattr_read;
	char* attrf_name;

	attrf_name = AttrFileName(fname_buf, FNAME_BUFFER_SIZE, FILE_SAVE_SESSION);

	// Normal read session
	int open_fname = FILE_READ_SESSION;

	if(current_state == SESSION_WRITE) {

		if (HasUnCommitedSession()) {
			// There is a load applied but not yet committed
			SetSaveSessionId(ReadUnCommitedSessionID());

			if (GetSaveSessionId() == s_id) {
				// Continue previous save session since this has same session id as current
				open_fname = FILE_SAVE_SESSION;
			}
			else {
				rclog << lock << "Error - previous session (id=" << GetSaveSessionId() << ") was not properly committed!" << unlock;
				new_save_session = 1;
			}
		}
		else {
			new_save_session = 1;
		}

		if (new_save_session) {
			SetSaveSessionId(s_id);
			// open read file to copy (by saving attr.) to the second file type
			open_fname = FILE_READ_SESSION;
		}
	}

	// packs
	pack_mode = NULL;
	pack_file = NULL;
	pack_addrs = NULL;
	packs = 0;
	is_stored = 0;
	local_min = NULL;
	local_max = NULL;
	local_sum = NULL;
	local_size = NULL;
	loc_no_obj = NULL;
	loc_no_nulls = NULL;
	rough_selectivity = -1;

	file = open_fname;

	// Load file with attribute description
	attrf_name = AttrFileName(fname_buf, 1024, open_fname);
	Load(attrf_name, current_state);

	///// Welcome to the Brighthouse Museum of Curiosities /////
	// Sit back, relax, and look deeper into the code below
	////////////////////////////////////////////////////////////

	SetCurReadLocation(GetCurSaveLocation());

	if (new_save_session == 1)
		SetCurSaveLocation(1 - GetCurSaveLocation());
	BHASSERT(GetCurSaveLocation() == 0 || GetCurSaveLocation() == 1, "Invalid current save location!");

	//pthread_mutexattr_t mattr;
	//pthread_mutexattr_init(&mattr);
	//pthread_mutexattr_settype(&mattr, PTHREAD_MUTEX_RECURSIVE);
	//pthread_mutex_init(&cs, &mattr);

	CalcPackFileSizeLimit(Type().GetPrecision());		// Determine pack file size limit
	ComputeNaturalSize();

	// need to add it manually since Load above can reset collation
	ct.SetCollation(collation);
}

RCAttr::~RCAttr()
{
   MEMORY_LOCK_GUARD

	// must be saved and released manually!
	BHASSERT(rsi_hist_update == NULL && rsi_cmap_update == NULL,
					"Invalid Histogram or CMAP object. Either one was not released/freed properly!");
	if(NoPack() > 0 && packs) {
//		for(int i = 0, end = NoPack() - packs_omitted; i < end; i++)
//			delete packs[i];

		free(is_stored);
		free(packs);
		free(pack_mode);

		dealloc(pack_addrs);
		dealloc(local_min);
		dealloc(local_max);
		dealloc(local_sum);
		dealloc(local_size);
		dealloc(loc_no_obj);
		dealloc(loc_no_nulls);
	}

	dealloc(pack_file);
	//pthread_mutex_destroy(&cs);
}

//int RCAttr::ColumnSize() const
//{
//	return ATI::TextSize(TypeName(), Type().GetPrecision(), Precision());
//}

void RCAttr::ComputeNaturalSize()
{
	AttributeType a_type = TypeName();
	_uint64 na_size;

	na_size = (Type().GetNullsMode()  != NO_NULLS ? 1 : 0) * NoObj() / 8;
	if(a_type == RC_STRING || a_type == RC_BYTE || a_type == RC_DATE)
		na_size += Type().GetPrecision() * NoObj();
	else if(a_type == RC_TIME || a_type == RC_YEAR || a_type == RC_DATETIME || a_type == RC_TIMESTAMP)
		na_size += Type().GetDisplaySize() * NoObj();
	else if(a_type == RC_NUM)
		na_size += (Type().GetPrecision() + (Type().GetScale() ? 1 : 0)) * NoObj();
	else if(ATI::IsRealType(a_type))
		na_size += 8 * NoObj();
	else if(a_type == RC_FLOAT)
		na_size += 4 * NoObj();
	else if(a_type == RC_INT)
		na_size += 4 * NoObj();
	else if(a_type == RC_BIGINT)
		na_size += 8 * NoObj();
	else if(a_type == RC_MEDIUMINT)
		na_size += 3 * NoObj();
	else if(a_type == RC_SMALLINT)
		na_size += 2 * NoObj();
	else if(a_type == RC_BYTEINT)
		na_size += 1 * NoObj();
	else if(a_type == RC_VARCHAR)
		na_size += (_int64)GetNaturalSizeSaved();
	else if(a_type == RC_VARBYTE || a_type == RC_BIN)
		na_size += (_int64)GetNaturalSizeSaved();
	SetNaturalSize(na_size);
}

void RCAttr::SaveHeader()
{
	_int64 css = ComputeCompressedSize();

	if (current_state == SESSION_READ)
		file = FILE_READ_SESSION;
	if(current_state == SESSION_WRITE)
		file = FILE_SAVE_SESSION;
	BHASSERT(FileFormat() == CLMD_FORMAT_RSC10_ID, "Invalid Attribute data file format!");
	char* fn = AttrFileName(fname_buf, FNAME_BUFFER_SIZE, file);
	Save(fn, &dic, css);
}

//GenerateColumnMetaDataFileName()
char* RCAttr::AttrFileName(char* filename_buf, unsigned buf_size, int ftype, bool oppositeName/*=false*/) const
{
	char fnm[] = {"TA00000.ctb"};

	if(ftype==1) {
		fnm[1]='S';			// save file:   TS000...
    }
    else {
        ABSwitcher absw;
        ABSwitch cur_ab_switch_state = absw.GetState(path);

        if (oppositeName) {
            if (cur_ab_switch_state==ABS_A_STATE) fnm[1] = 'B';
        }
        else {
            if (cur_ab_switch_state==ABS_B_STATE) fnm[1] = 'B';
        }
    }

	BHASSERT(ftype!=2, "Trying to open file using invalid session mode!");
	fnm[6]=(char)('0'+attr_number%10);
	fnm[5]=(char)('0'+(attr_number/10)%10);
	fnm[4]=(char)('0'+(attr_number/100)%10);
	fnm[3]=(char)('0'+(attr_number/1000)%10);
	fnm[2]=(char)('0'+(attr_number/10000)%10);
	strcpy(filename_buf, path);
	strcat(filename_buf, fnm);
	return filename_buf;
}

_int64 RCAttr::ComputeCompressedSize()
{
	_int64 tsize = 0;
	_int64 size = 0;

	if (GetFileSize(DPNFileName(fname_buf, FNAME_BUFFER_SIZE), size))
		tsize = size;
	// for all pack file
	for(uint p_f = 0; p_f < GetTotalPackFile(); p_f++)
	{
		if (GetFileSize(AttrPackFileNameDirect(fname_buf, 1024, p_f), size))
			tsize += size;
	}
	return tsize;
}

void RCAttr::UpgradeFormat()
{
	if(file!=0)
	{
		rclog << lock << "Error: cannot upgrade. RCAttr constructor connects to working file." << unlock;
		throw;
	}
	LoadPackInfo();
	if (PackType()==PackS)
		for (int i = 0; i < NoPack(); i++)
		{
			local_min[i] = 0;
			local_max[i] = -1;
		}
	SetFileFormat(10);
	CreateDPNFile();
	char* fn = AttrFileName(fname_buf, FNAME_BUFFER_SIZE, file);
	char backup_name[FNAME_BUFFER_SIZE];
	strcpy(backup_name, fn);
	strcpy(backup_name+strlen(backup_name)-3, "bck");
    try {
	    RenameFile(fn,backup_name);
	} catch(...) {
		rclog << lock << "Internal upgrade error: unable to rename " << fn << " to " << backup_name << "." << unlock;
		throw;
	}
	SaveHeader();
	try {
        RemoveFile(backup_name);
	} catch(...) {
		rclog << lock << "Internal upgrade error: unable to remove " << fn << " backup file (change *.bck to *.ctb before next startup)." << unlock;
		throw;
	}
}

void RCAttr::CreateDPNFile()
{
	if (NoPack() == 0)
		return;
	const int buf_size = 10000;
	char buf[buf_size];
	char* fn = DPNFileName(fname_buf, FNAME_BUFFER_SIZE);
	IBFile fdpn;

	fdpn.OpenCreateEmpty(fn);

	ushort buffer_pos = 0;
	for(uint slot = 0; slot < NoPack() + 1; slot++)
	{
		if (buffer_pos + 37 > buf_size)
		{
			fdpn.Write(buf, buffer_pos);
			buffer_pos = 0;
		}
		if (slot < 2) StoreDPN(NoPack()-1, buf+buffer_pos);		// double slot for last pack info for rollback handling
		else StoreDPN(slot-2, buf+buffer_pos);
		buffer_pos += 37;
	}
	fdpn.Write(buf, buffer_pos);
	fdpn.Close();
}

void RCAttr::StoreDPN(uint pack_no, char* buf)
{
	if(loc_no_nulls[pack_no] == (uint)(loc_no_obj[pack_no]) + 1)
	{
		*((int*)(buf))			= PF_NULLS_ONLY;
		*((ushort*)(buf+34))	= 0;		// no_nulls (0 is special value here)
	}
	else
	{
		*((int*)(buf))			= pack_file[pack_no];
		*((ushort*)(buf+34))	= (ushort)loc_no_nulls[pack_no];
	}
	*((uint*)(buf+4))			= pack_addrs[pack_no];

	*((_uint64*)(buf+8))		= local_min[pack_no];
	*((_uint64*)(buf+16))		= local_max[pack_no];
	if(PackType() == PackN)
		*((_uint64*)(buf+24)) 	= local_sum[pack_no];
	else
		*((_uint64*)(buf+24)) 	= local_size[pack_no];

	*((ushort*)(buf+32))		= loc_no_obj[pack_no];
	*((uchar*)(buf+36))			= 0;
}

char *RCAttr::AttrPackFileName(char* filename_buf, unsigned buf_size, int n_pack)
{
	LoadPackInfo();
	int n_file=pack_file[n_pack];
	AttrPackFileNameDirect(filename_buf, buf_size, n_file);
	return filename_buf;
}

char *RCAttr::AttrPackFileNameDirect(char* filename_buf, unsigned buf_size, int n_file)
{
	BHASSERT(n_file>=0, "Invalid input pack number!");
	char fnm[] = {"TA00000000000000.ctb"};
	fnm[15]=(char)('0'+n_file%10);
	fnm[14]=(char)('0'+(n_file/10)%10);
	fnm[13]=(char)('0'+(n_file/100)%10);
	fnm[12]=(char)('0'+(n_file/1000)%10);
	fnm[11]=(char)('0'+(n_file/10000)%10);
	fnm[10]=(char)('0'+(n_file/100000)%10);
	fnm[9]=(char)('0'+(n_file/1000000)%10);
	fnm[8]=(char)('0'+(n_file/10000000)%10);
	fnm[7]=(char)('0'+(n_file/100000000)%10);
	fnm[6]=(char)('0'+attr_number%10);
	fnm[5]=(char)('0'+(attr_number/10)%10);
	fnm[4]=(char)('0'+(attr_number/100)%10);
	fnm[3]=(char)('0'+(attr_number/1000)%10);
	fnm[2]=(char)('0'+(attr_number/10000)%10);
	strcpy(filename_buf, path);
	strcat(filename_buf, fnm);
	return filename_buf;
}

char* RCAttr::DPNFileName(char* filename_buf, unsigned buf_size) const
{
	char fnm[] = {"TA00000DPN.ctb"};
	fnm[6]=(char)('0'+attr_number%10);
	fnm[5]=(char)('0'+(attr_number/10)%10);
	fnm[4]=(char)('0'+(attr_number/100)%10);
	fnm[3]=(char)('0'+(attr_number/1000)%10);
	fnm[2]=(char)('0'+(attr_number/10000)%10);
	fill(filename_buf, filename_buf + buf_size, 0);
	strcpy(filename_buf, path);
	strcat(filename_buf, fnm);
	return filename_buf;
}

time_t RCAttr::UpdateTime() const
{
	char fname[1000];
	AttrFileName(fname, 1000, current_state);
    int result = GetFileTime(fname);
	if (result < 0) {
		rclog << "Error: " << fname << " status can not be obtained" << endl;
		return 0;
	}
	return (time_t) result;
}

void RCAttr::CommitSaveSession(bool same_file)
{
  //WaitForSaveThreads();
	SetSaveSessionId(INVALID_TRANSACTION_ID);
	char fname0[1000];
	char fname1[1000];
	AttrFileName(fname0, 1000, 0, !same_file); // AB Switch: generate alternative name
	AttrFileName(fname1, 1000, 1);
	// if save file is not accessible for writing (or not exists), skip the attribute
	// WARNING: dangerous code  --> if(access(fname1,W_OK|F_OK)!=0) return;
	//pthread_mutex_lock(&cs);
    int res=0;
    try {
        IBGuard guard(attr_mutex);
	    RemoveFile(fname0);
	    RenameFile(fname1, fname0);
    } catch(...) {
		rclog << lock << Name() << ": CommitWork failed! Database may be corrupted!" << unlock;
		throw;
	}
	//pthread_mutex_unlock(&cs);
	file = 0;
}

void RCAttr::Rollback(uint s_id)
{
	DeleteRSIs();
	if(GetSaveSessionId() != INVALID_TRANSACTION_ID)
	{
		if(s_id == 0 || s_id == GetSaveSessionId())
		{
			char fname1[1000];
            AttrFileName(fname1, sizeof(fname1), 1);
			RemoveFile(fname1);
			SetSaveSessionId(INVALID_TRANSACTION_ID);
		}
		else
			rclog << lock << "Rollback error - invalid save session_id = " << GetSaveSessionId() << unlock;
	}
}

void RCAttr::DeleteRSIs()
{
	if(rsi_manager && rsi_hist_update) {
		rsi_manager->UpdateIndex(rsi_hist_update, 0, true);
		rsi_hist_update = 0;
	}
	if(rsi_manager && rsi_cmap_update) {
		rsi_manager->UpdateIndex(rsi_cmap_update, 0, true);
		rsi_cmap_update = 0;
	}
}


void RCAttr::LoadPackInfo_physical()
{
#ifdef FUNCTIONS_EXECUTION_TIMES
	FETOperator fet("RCAttr::LoadPackInfo()");
#endif
    IBGuard guard( attr_mutex);
	//pthread_mutex_lock(&cs);

	if(!GetPackInfoCollaped())
	{
		//pthread_mutex_unlock(&cs);
		return;
	}
	if(NoPack() > GetNoPackPtr())
		SetNoPackPtr(NoPack());
	if(NoPack()>0)
	{
		pack_mode = (char*)malloc(sizeof(char) * GetNoPackPtr());
		pack_file = (int*)alloc(sizeof(int)*GetNoPackPtr(), BLOCK_TEMPORARY);
		pack_addrs = (uint*)alloc(sizeof(uint)*GetNoPackPtr(), BLOCK_TEMPORARY);
		packs = (AttrPack**)malloc(sizeof(AttrPack*) * GetNoPackPtr());
		is_stored = (bool*)malloc(sizeof(bool) * GetNoPackPtr());
		local_min = (_int64*)alloc(sizeof(_int64)*GetNoPackPtr(), BLOCK_TEMPORARY);
		local_max = (_int64*)alloc(sizeof(_int64)*GetNoPackPtr(), BLOCK_TEMPORARY);
		if(PackType() == PackN)
			local_sum = (_int64*)alloc(sizeof(_int64)*GetNoPackPtr(), BLOCK_TEMPORARY);
		else
			local_size = (unsigned short*)alloc(sizeof(unsigned short)*GetNoPackPtr(), BLOCK_TEMPORARY);
		loc_no_obj = (ushort*)alloc(sizeof(ushort)*GetNoPackPtr(), BLOCK_TEMPORARY);
		loc_no_nulls = (uint*)alloc(sizeof(uint)*GetNoPackPtr(), BLOCK_TEMPORARY);

		if(!local_min || !local_max || (!local_sum && !local_size) || !loc_no_obj || !loc_no_nulls || !pack_mode || !pack_file || !pack_addrs)
			rclog << lock << "Error: out of memory (" << NoPack() << " bytes failed). (22)" << unlock;
	}
	// load packs descriptions (file of a pack, localization in file, statistics, ...)
	if (FileFormat() != 9) LoadAllDPN();
	if (FileFormat() == 9 || GetDictOffset() != 0)
	{
		IBFile fattr;
		char* buf_ptr = NULL;
		try
		{
			fattr.OpenReadOnly(AttrFileName(fname_buf, FNAME_BUFFER_SIZE, file));

			int fin_size = fattr.Seek(0, SEEK_END);
			buf_ptr = new char [(fin_size - GetPackOffset()) + 1];
			fattr.Seek(GetPackOffset(), SEEK_SET);
			fattr.Read(buf_ptr, fin_size - GetPackOffset());
			if (FileFormat() == 9)
			{
				char* buf = buf_ptr;
				for(int i=0;i<NoPack();i++)
				{
					RestoreDPN(buf, i);
					buf=buf+36;
				}
			}
			if (GetDictOffset() != 0 && (fin_size - GetDictOffset()))
				LoadDictionaries(buf_ptr);
			delete [] buf_ptr;
			fattr.Close();
		}
		catch (DatabaseRCException& e)
		{
			if (buf_ptr)
				delete [] buf_ptr;
			SetPackInfoCollaped(false);
			//pthread_mutex_unlock(&cs);
			rclog << lock << "Internal error: unable to open/read " << AttrFileName(fname_buf, FNAME_BUFFER_SIZE) << " file." << unlock;
			throw e;	//newly added
		}
	}
	SetPackInfoCollaped(false);
	for(int i = 0; i< GetNoPackPtr(); i++)
		is_stored[i] = ShouldExist(i);
	//pthread_mutex_unlock(&cs);
}

void RCAttr::LoadDictionaries(const char* buf_ptr)
{
	AttributeType a_type = TypeName();
	// load dictionary for the whole attribute
	const char* buf = buf_ptr + (GetDictOffset() - GetPackOffset());
	// dictionary for symbolic (text) values - min, max, FTree object
	if(Type().IsLookup())
	{
		SetMinInt64(*((_uint64*)(buf)));
		SetMaxInt64(*((_uint64*)(buf+8)));
		buf+=16;
		dic.Init((uchar*&)buf);		// note that dic.Init will shift buf to the end of minicompressor stream
	}
	// dictionary for numerical values - min, max
	else if(
		ATI::IsNumericType(a_type) ||
		ATI::IsDateTimeType(a_type) || ATI::IsDateTimeNType(a_type))
	{
		SetMinInt64(*((_uint64*)(buf)));
		SetMaxInt64(*((_uint64*)(buf+8)));
		buf+=16;
	}
	else
	{
		// ... other types of dictionary ...
	}
}

void RCAttr::LoadAllDPN()
{
	if (NoPack() == 0)
		return;

	char* dpnf_name;
	char* buf_ptr = NULL;
	dpnf_name = DPNFileName(fname_buf, FNAME_BUFFER_SIZE);
	IBFile fdpn;

	try
	{
		fdpn.OpenReadOnly(dpnf_name);

		buf_ptr = new char [(NoPack()+1)*37 + 1];
		fdpn.Read(buf_ptr, (NoPack()+1)*37 + 1);

		char* buf = buf_ptr;
		for(int slot = 0; slot<NoPack(); slot++)
		{
			if (slot == 0)
			{
				if (GetCurReadLocation() == 1)
					buf += 37;
				RestoreDPN(buf, NoPack()-1);
			}
			else
				RestoreDPN(buf, slot-1);
			if (slot == 0 && GetCurReadLocation() == 0)
				buf += 37;
			buf += 37;
		}

		fdpn.Close();
		delete [] buf_ptr;
		buf_ptr = NULL;
	}
	catch (DatabaseRCException& e)
	{
		if (buf_ptr)
			delete [] buf_ptr;
		rclog << lock << "Internal error: unable to open " << DPNFileName(fname_buf, FNAME_BUFFER_SIZE) << " file." << unlock;
		throw e;		//newly added
	}
	return;
}

void RCAttr::RestoreDPN(char* buf, uint pack_no)
{
	packs[pack_no] = 0;
	pack_file[pack_no] = *((int*)buf);
	pack_addrs[pack_no] = *((uint*)(buf+4));
	local_min[pack_no] = *((_uint64*)(buf+8));
	local_max[pack_no] = *((_uint64*)(buf+16));
	if(PackType() == PackN)
		local_sum[pack_no] = *((_uint64*)(buf+24));
	else
		local_size[pack_no] = ushort(*((_uint64*)(buf+24)));
	loc_no_obj[pack_no] = *((ushort*)(buf+32));
	loc_no_nulls[pack_no] = *((ushort*)(buf+34));
	if(pack_file[pack_no] == PF_NULLS_ONLY)
		loc_no_nulls[pack_no] = loc_no_obj[pack_no] + 1;
	if(pack_file[pack_no] == PF_NULLS_ONLY ||
		(PackType() == PackN && local_min[pack_no] == local_max[pack_no] && loc_no_nulls[pack_no]==0)) {
		pack_mode[pack_no] = 0;					// trivial data (only nulls or all values are the same), no physical pack
		is_stored[pack_no] = false;
	}
	else if(pack_file[pack_no] == PF_NO_OBJ) {
		pack_mode[pack_no] = 3;					// empty pack, no objects
		is_stored[pack_no] = false;
	}
	else {
		pack_mode[pack_no] = 1;					// data on disk
		is_stored[pack_no] = true;
	}
	if(FileFormat() > 9)
	{
		// restore 37-th byte
	}
}

void RCAttr::WriteUnique(RSValue v)
{
	IBFile fattr;

	if(v!=RS_UNKNOWN)
	{
		SetUniqueUpdated(true);
		if(v==RS_ALL) SetUnique(true);
		if(v==RS_NONE) SetUnique(false);

		//char flags = char(nulls_mode)+(declared_unique?4:0)+(is_primary?8:0)+(is_unique?16:0)+(is_unique_updated?32:0);
		char flags = char(Type().GetNullsMode()) +(IsUnique()?16:0)+(IsUniqueUpdated()?32:0);

        AttrFileName(fname_buf, sizeof(fname_buf), 0);
		fattr.OpenReadWriteWithThreadAffinity(fname_buf);
		fattr.Seek(25,SEEK_SET);
		fattr.Write(&flags,1);
		fattr.CloseWithThreadAffinity(fname_buf);
	}
}

PackOntologicalStatus RCAttr::GetPackOntologicalStatus(int pack_no)
{
	LoadPackInfo();
	if(pack_no < 0 || pack_file[pack_no] == PF_NULLS_ONLY || loc_no_nulls[pack_no]-1 == loc_no_obj[pack_no])
		return NULLS_ONLY;
	if(PackType() == PackN)
	{
		if(local_min[pack_no] == local_max[pack_no])
		{
			if(loc_no_nulls[pack_no] == 0)
				return UNIFORM;
			return UNIFORM_AND_NULLS;
		}
	}
	return NORMAL;
}

bool RCAttr::ShouldExist(int pack_no)
{
	LoadPackInfo();
	if(pack_file[pack_no] == PF_NULLS_ONLY || pack_file[pack_no] == PF_NO_OBJ || loc_no_nulls[pack_no]-1 == loc_no_obj[pack_no])
		return false;
	if(PackType() == PackN)
	{
		if(loc_no_nulls[pack_no] == 0 && local_min[pack_no] == local_max[pack_no])
			return false;
	}
	else
		return pack_mode[pack_no] != 0;
	return true;
}

void RCAttr::LoadPack(int n)
{
#ifdef FUNCTIONS_EXECUTION_TIMES
	std::stringstream s1;
	s1 << "a[" << attr_number << "].LoadPack(...)";
	FETOperator fet1(s1.str());
	FETOperator fet2("RCAttr::LoadPack(...)");
#endif

//	LoadPackInfo();
	BHASSERT(n < NoPack(), "Invalid input pack number. It is greater than the total number counter of packs!");
	if(pack_mode[n]!=1 || !is_stored[n])
		return;
	assert(packs[n]->IsLocked());
	LoadPackInherited(n);
////	in PhysicalColumn connectioInfo is not passed, pack loading statistics should be counted in LockPack
//	if(conn)
//	{
//		conn->NotifyPackLoad();
//		if(conn->killed()) throw KilledRCException();
//	}
//	else
//		BHASSERT(0, "Invalid ConnectionInfo object!");
}

void RCAttr::LoadPackInherited(int n)
{	
	if (pack_file[n] < 0)
		rclog << lock << "INTERNAL ERROR: attempting to open wrong file (LoadPack), pack_file[" << n << "]=" << pack_file[n] << unlock;
    {
        IBGuard guard(attr_mutex);

	    if(pack_mode[n] != 1)
	    	return;

	    BHASSERT(pack_mode[n] == 1, "Invalid pack_mode!");
	    int er = 0;
		IBFile fattr;

	    try {
    		fattr.OpenReadOnly(AttrPackFileName(fname_buf, FNAME_BUFFER_SIZE, n));
    		fattr.Seek(pack_addrs[n],SEEK_SET);
    		packs[n]->LoadData(&fattr);
			fattr.Close();
    	} catch (DatabaseRCException& e) {
    		rclog << lock << "Error: corrupted " << AttrPackFileName(fname_buf, FNAME_BUFFER_SIZE, n) << unlock;
    		throw e;
    	}
    	packs[n]->Uncompress();
    	pack_mode[n]=2;
    	is_stored[n] = true;
    }
}

//old name GetTable_S
RCBString RCAttr::GetValueString(const _uint64 obj, bool bin_as_hex)
{
	if(obj == NULL_VALUE_64)
		return RCBString();
	LoadPackInfo();
	int pack = (int) (obj >> 16);

	if(PackType() == PackS) {
		if(pack_mode[pack] == 0 || pack_file[pack] == PF_NULLS_ONLY)
			return RCBString();
		assert(packs[pack]->IsLocked());
//		if(pack_mode[pack] != 2)
//			LoadPack(pack);
		if(packs[pack]->IsNull((int) (obj & 65535)))
			return RCBString();

		int len = ((AttrPackS*) packs[pack])->GetSize((int) (obj & 65535));
		if(ATI::IsBinType(TypeName()) && bin_as_hex) {
			len *= 2;
			RCBString s(NULL, len, true);
			if(len)
				((AttrPackS*) packs[pack])->GetVal_T((int) (obj & 65535), s.val);
			return s;
		} else if(len)
			return RCBString(((AttrPackS*) packs[pack])->GetVal((int) (obj & 65535)), len);
		else
			return ZERO_LENGTH_STRING;
}
	_int64 v = GetValueInt64(obj);
	return DecodeValue_S(v);
}

//int RCAttr::GetValueText(_uint64 obj, char *val_buf)		// original 0-level value (text, string, date, time etc.), without '\0'
//{
//	AttributeType a_type = TypeName();
//	BHASSERT_WITH_NO_PERFORMANCE_IMPACT( NoObj()>=(_int64)obj );
//	LoadPackInfo();
//	int pack=(int)(obj >> 16);									// simplified version: all packs are full
//
//	if(pack_file[pack]==PF_NULLS_ONLY)
//	{
//		return 0;
//	}
//	if(PackType() == PackS || ATI::IsBinType(a_type))
//	{
//		BHASSERT_WITH_NO_PERFORMANCE_IMPACT(packs);
//		if(pack_mode[pack]==0) return 0;
//		assert(packs[pack]->IsLocked());
////		if(pack_mode[pack]!=2) LoadPack(pack);
//		if(ATI::IsBinType(a_type))
//			return ((AttrPackS*)packs[pack])->GetVal_T((int)(obj&65535),val_buf);
//		else
//		{
//			int len=((AttrPackS*)packs[pack])->GetSize((int)(obj&65535));
//			memcpy(val_buf,((AttrPackS*)packs[pack])->GetVal((int)(obj&65535)),len);
//			return len;
//		}
//	}
//	else if(ATI::IsInteger32Type(a_type) || a_type==RC_STRING || a_type==RC_VARCHAR)
//	{
//		BHASSERT_WITH_NO_PERFORMANCE_IMPACT( PackType() == PackN );
//		int v=GetValueInt32(obj);
//		if(v==NULL_VALUE)
//		{
//			return 0;
//		}
//		return DecodeValue_T(v,val_buf);
//	}
//	else if(a_type==RC_NUM || a_type == RC_BIGINT || ATI::IsDateTimeType(a_type))
//	{
//		_int64 v = GetValueInt64(obj);
//		if(v == NULL_VALUE_64) return 0;
//		return DecodeValue_T(v,val_buf);
//	}
//	else if(ATI::IsRealType(a_type))
//	{
//		_int64 v=GetValueInt64(obj);
//		if(v==NULL_VALUE_D) return 0;
//		//_gcvt_s(val_buf, ColumnSize(), *(double*)(&v), type == RC_REAL ? 15 : 6);
//		gcvt(*(double*)(&v), a_type == RC_REAL ? 15 : 6, val_buf);
//		size_t s = strlen(val_buf);
//		if(s && val_buf[s-1] == '.')
//			val_buf[s-1] = 0;
//		return (int)strlen(val_buf);
//	}
//	return 0;
//}

void RCAttr::GetValueBin(_uint64 obj, int& size, char* val_buf)		// original 0-level value (text, string, date, time etc.)
{
	if(obj == NULL_VALUE_64)
		return;
	AttributeType a_type = TypeName();
	size = 0;
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(NoObj()>=(_int64)obj);
	LoadPackInfo();
	int pack = (int)(obj >> 16);	// simplified version: all packs are full
	if(pack_file[pack] == PF_NULLS_ONLY)
		return;
	if(ATI::IsStringType(a_type)) {
		if(PackType() == PackN) {
			_int64 res = GetValueInt64(obj);
			if(res == NULL_VALUE_64)
				return;
			size = dic.ValueSize((int)res);
			memcpy(val_buf, dic.GetBuffer((int)res), size);
			return;
		} else {			// no dictionary
			if(pack_mode[pack] == 0)
				return;
			assert(packs[pack]->IsLocked());
//			if(pack_mode[pack] != 2) LoadPack(pack);
			size = ((AttrPackS*)packs[pack])->GetSize((int)(obj & 65535));
			memcpy(val_buf, ((AttrPackS*)packs[pack])->GetVal((int)(obj & 65535)), size);
			return;
		}
	} else if(ATI::IsInteger32Type(a_type)) {
		size = 4;
		int v = GetValueInt32(obj);
		if(v == NULL_VALUE_32)
			return;
		*(int*)val_buf = v;
		val_buf[4] = 0;
		return;
	} else if(a_type == RC_NUM || a_type == RC_BIGINT || ATI::IsRealType(a_type) || ATI::IsDateTimeType(a_type)) {
		size = 8;
		_int64 v = GetValueInt64(obj);
		if(v == NULL_VALUE_64)
			return;
		*(_int64*)(val_buf) = v;
		val_buf[8] = 0;
		return;
	}
	return;
}

//unused
//double RCAttr::GetTable_N(_uint64 obj,ConnectionInfo *conn)		// numerical value of object in 0-level encoding (RC_NUM and converted RC_INT/UINT/...)
//{
//	_int64 v=GetValueInt64(obj);
//	if(v==NULL_VALUE_64) return NULL_VALUE_D;
//	if(TypeName()==RC_NUM)	return DecToDouble(v, Type().GetScale());
//	if(ATI::IsRealType(TypeName()))	return *((double*)v);
//	return (double)v;
//}


RCValueObject RCAttr::GetValue(_uint64 obj, bool lookup_to_num)
{
	if(obj == NULL_VALUE_64)
		return RCValueObject();
	AttributeType a_type = TypeName();
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(NoObj()>=(_int64)obj);
	RCValueObject ret;
	if(!IsNull(obj))
	{
		if(ATI::IsTxtType(a_type) && !lookup_to_num)
			ret = GetValueString(obj,false);
		else if(ATI::IsBinType(a_type))
		{
			int tmp_size = GetLength(obj);
			RCBString rcbs(NULL, tmp_size, true);
			GetValueBin(obj, tmp_size, rcbs.val);
			rcbs.null = false;
			ret = rcbs;
		}
		else if(ATI::IsIntegerType(a_type))
			ret = RCNum(GetValueInt64(obj), -1, false, a_type);
		else if(ATI::IsDateTimeType(a_type))
		{
			ret = RCDateTime(this->GetValueInt64(obj), a_type);
		}
		else if(ATI::IsRealType(a_type))
			ret = RCNum(this->GetValueInt64(obj), 0, true, a_type);
		else if(lookup_to_num || a_type == RC_NUM)
			ret = RCNum((_int64)GetValueInt64(obj), Type().GetScale());
	}
	return ret;
}

RCDataType& RCAttr::GetValue(_uint64 obj, RCDataType& value, bool lookup_to_num)
{
	if(obj == NULL_VALUE_64 || IsNull(obj))
		value = ValuePrototype(lookup_to_num);
	else {
		AttributeType a_type = TypeName();
		BHASSERT_WITH_NO_PERFORMANCE_IMPACT(NoObj()>=(_int64)obj);
		if(ATI::IsTxtType(a_type) && !lookup_to_num)
			value = GetValueString(obj, false);
		else if(ATI::IsBinType(a_type)) {
			int tmp_size = GetLength(obj);
			value = RCBString(NULL, tmp_size, true);
			GetValueBin(obj, tmp_size, ((RCBString&)value).val);
			value.null = false;
		} else if(ATI::IsIntegerType(a_type))
			value = RCNum(GetValueInt64(obj), -1, false, a_type);
		else if(ATI::IsDateTimeType(a_type)) {
			value = RCDateTime(this->GetValueInt64(obj), a_type);
		} else if(ATI::IsRealType(a_type))
			value = RCNum(this->GetValueInt64(obj), 0, true, a_type);
		else
			value = RCNum((_int64)GetValueInt64(obj), Type().GetScale());
	}
	return value;
}

#ifdef USE_NEW_ENGINE

void RCAttr::GetAValue(uint64 obj, AValue& value, ConnectionInfo *conn)
{
	AttributeType a_type = TypeName();
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(NoObj()>=(_int64)obj);
	if(!IsNull(obj, conn))
	{
		if(ATI::IsTxtType(a_type))
		{
			BHASSERT_WITH_NO_PERFORMANCE_IMPACT(&dynamic_cast<const VString&>(value));
			RCBString rcs = GetTable_S(obj, false, conn);
			((VString&)value).Set(rcs.val, rcs.size());
		}
		else if(ATI::IsIntegerType(a_type))
		{
			BHASSERT_WITH_NO_PERFORMANCE_IMPACT(&dynamic_cast<const VInteger&>(value));
			((VInteger&)value).Set(GetTable64(obj, conn));
		}
	}
	else
		value.SetNull();
}

void RCAttr::GetMinMaxBounds(VInteger& min, VInteger& max, ConnectionInfo *conn)
{
	AttributeType a_type = TypeName();
	BHASSERT(ATI::IsNumericType(a_type), "The data type must be numeric!");
	LoadPackInfo();
	if(ATI::IsIntegerType(a_type))
	{
		if(this->NoNulls() != NoObj() && NoObj() != 0)
		{
			min = VInteger(GetIMin());
			max = VInteger(GetIMax());
		}
		else
		{
			switch(a_type)
			{
			case RC_BYTEINT :
				min = VInteger(BH_TINYINT_MIN);
				max = VInteger(BH_TINYINT_MAX);
				break;
			case RC_SMALLINT :
				min = VInteger(BH_SMALLINT_MIN);
				max = VInteger(BH_SMALLINT_MAX);
				break;
			case RC_MEDIUMINT :
				min = VInteger(BH_MEDIUMINT_MIN);
				max = VInteger(BH_MEDIUMINT_MAX);
				break;
			case RC_INT :
				min = VInteger(BH_INT_MIN);
				max = VInteger(BH_INT_MAX);
				break;
			}
		}
	}
	else
	{
		BHASSERT(0, "Unsupported data type!");
	}
}

#endif

_int64 RCAttr::GetNoNulls(int pack)
{
	LoadPackInfo();
	if(pack == -1)
		return NoNulls();
	return loc_no_nulls[pack];
}

unsigned int RCAttr::GetNoValues(int pack)
{
	LoadPackInfo();
//	if(pack_mode[pack] == 3)
//		return 0;
	return loc_no_obj[pack] + 1;
}

ushort RCAttr::GetActualSize(int pack)
{
	if(GetPackOntologicalStatus(pack) == NULLS_ONLY)
		return 0;
	if(Type().IsLookup() || PackType() != PackS || local_size[pack] == 0)
		return Type().GetPrecision();
	return local_size[pack];
}

_int64 RCAttr::GetSum(int pack)
{
	LoadPackInfo();
	if(GetPackOntologicalStatus(pack) == NULLS_ONLY || local_sum == NULL)
		return NULL_VALUE_64;
	if(local_min[pack] < (MINUS_INF_64 / 65536) || local_max[pack] > (PLUS_INF_64 / 65536))
		return NULL_VALUE_64;								// conservative overflow test
	return local_sum[pack];
}

double RCAttr::GetSumD(int pack)
{
	LoadPackInfo();
	if(GetPackOntologicalStatus(pack) == NULLS_ONLY)
		return NULL_VALUE_64;
	return *(double*)(&local_sum[pack]);
}

_int64 RCAttr::GetMinInt64(int pack)
{
	LoadPackInfo();
	if(GetPackOntologicalStatus(pack) == NULLS_ONLY)
		return NULL_VALUE_64;
	return local_min[pack];
}

_int64 RCAttr::GetMaxInt64(int pack)
{
	LoadPackInfo();
	if(GetPackOntologicalStatus(pack) == NULLS_ONLY)
		return NULL_VALUE_64;
	return local_max[pack];
}

/*
RCBString RCAttr::GetMaxString(int pack)
{
	assert(0);
	return RCBString();
}

RCBString RCAttr::GetMinString(int pack)
{
	assert(0);
	return RCBString();
}

RCValueObject RCAttr::GetMinValue()
{
	AttributeType a_type = TypeName();
	RCValueObject ret;
	if(this->NoNulls() != NoObj() && NoObj() != 0)
	{
		LoadPackInfo();
		if(ATI::IsIntegerType(a_type))
			ret = RCNum(GetMinInt64());
		else if(ATI::IsDateTimeType(a_type))
			ret = RCDateTime(GetMinInt64(), a_type);
		else if(ATI::IsRealType(a_type))
			ret = RCNum(GetMinInt64(), 0, true);
		else if(a_type == RC_NUM)
			ret = RCNum(GetMinInt64(), Type().GetScale());
		else {
			Filter f(NoObj());
			f.Set();
			ret = MinS(&f);
		}
	}
	return ret;
}

RCValueObject RCAttr::GetMaxValue()
{
	AttributeType a_type = TypeName();
	RCValueObject ret;
	if(this->NoNulls() != NoObj() && NoObj() != 0) {
		LoadPackInfo();
		if(ATI::IsIntegerType(a_type))
			ret = RCNum(GetMaxInt64());
		else if(ATI::IsDateTimeType(a_type))
			ret = RCDateTime(GetMaxInt64(), a_type);
		else if(ATI::IsRealType(a_type))
			ret = RCNum(GetMaxInt64(), 0, true);
		else if(a_type == RC_NUM)
			ret = RCNum(GetMaxInt64(), Type().GetScale());
		else {
			Filter f(NoObj());
			f.Set();

			ret = MaxS(&f);
		}
	}
	return ret;
}

RCValueObject RCAttr::GetMinValue(int pack)
{
	AttributeType a_type = TypeName();
	RCValueObject ret;
	if(this->NoNulls() != NoObj() && NoObj() != 0) {
		LoadPackInfo();
		if(ATI::IsIntegerType(a_type))
			ret = RCNum(local_min[pack]);
		else if(ATI::IsDateTimeType(a_type))
			ret = RCDateTime(local_min[pack], a_type);
		else if(ATI::IsRealType(a_type))
			ret = RCNum(local_min[pack], 0, true);
		else if(a_type == RC_NUM)
			ret = RCNum(local_min[pack], Type().GetScale());
		else {
			Filter f(NoObj());
			f.SetBlock(pack);
			ret = MinS(&f);
		}
	}
	return ret;
}

RCValueObject RCAttr::GetMaxValue(int pack)
{
	AttributeType a_type = TypeName();
	RCValueObject ret;
	if(this->NoNulls() != NoObj() && NoObj() != 0)
	{
		LoadPackInfo();
		if(ATI::IsIntegerType(a_type))
			ret = RCNum(local_max[pack]);
		else if(ATI::IsDateTimeType(a_type))
			ret = RCDateTime(local_max[pack], a_type);
		else if(ATI::IsRealType(a_type))
			ret = RCNum(local_max[pack], 0, true);
		else if(a_type == RC_NUM)
			ret = RCNum(local_max[pack], Type().GetScale());
		else
		{
			Filter f(NoObj());
			f.SetBlock(pack);
			ret = MaxS(&f);
		}
	}
	return ret;
}
*/

std::auto_ptr<DPN> RCAttr::GetDPN(const DataPackId& dpid)
{
	LoadPackInfo();
	int pack_no = (int)dpid - packs_omitted;
	return std::auto_ptr<DPN>(new DPN(dpid, loc_no_obj[pack_no], loc_no_nulls[pack_no],
											pack_file[pack_no], pack_addrs[pack_no],
											local_min[pack_no], local_max[pack_no],
											(local_sum ? local_sum[pack_no] : 0),
											(local_size ? local_size[pack_no] : 0) ));
}

std::auto_ptr<DataPack> RCAttr::GetDataPack(const DataPackId& dpid, ConnectionInfo& conn)
{
	LoadPackInfo();
	std::auto_ptr<DataPack> dp;
	if(ATI::IsStringType(TypeName()))
		dp = auto_ptr<DataPack>(new DataPackImpl<RCBString>(*GetDPN(dpid)));
	else if(ATI::IsDateTimeType(TypeName()))
		dp = auto_ptr<DataPack>(new DataPackImpl<RCDateTime>(*GetDPN(dpid)));
	else
		dp = auto_ptr<DataPack>(new DataPackImpl<RCNum>(*GetDPN(dpid)));

	DataPackLock lock(*this, dpid);

	_uint64 i = (((_uint64)dpid - packs_omitted) << 16);
	_uint64 last = i + GetNoValues(((_uint64)dpid - packs_omitted));
	ushort id = 0;
	for(; i < last; i++)
		GetValue(i, (*dp)[id++]);

	return dp;
}

int	RCAttr::GetLength(_uint64 obj)		// size of original 0-level value (text/binary, not null-terminated)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(NoObj()>=(_int64)obj);
	LoadPackInfo();
	int pack = (int)(obj >> 16);									// simplified version: all packs are full
	if(pack_file[pack]==PF_NULLS_ONLY)	return 0;
	if(PackType() != PackS) return Type().GetDisplaySize();
//	if(pack_mode[pack]!=2) LoadPack(pack);
	return ((AttrPackS*)packs[pack])->GetSize((int)(obj&65535));
}

int RCAttr::DecodeValue_T(_int64 code, char *val_buf)			// original 0-level value for a given 1-level code
{
	AttributeType a_type = TypeName();
	if(code==NULL_VALUE_64)
	{
		return 0;
	}
	int v=(int)code;
	if(a_type==RC_STRING || a_type==RC_VARCHAR)
	{
		BHASSERT_WITH_NO_PERFORMANCE_IMPACT(PackType() == PackN);
		memcpy(val_buf,dic.GetBuffer(v),dic.ValueSize(v));
		return dic.ValueSize(v);
	}
	else if(ATI::IsIntegerType(a_type))
	{
		RCNum rcn(code, -1, false, a_type);
		RCBString local_rcb(val_buf, TEMP_VALUE_BUF_SIZE);
		return (int)strlen(rcn.ToRCString(local_rcb));
	}
	else if(ATI::IsRealType(a_type))
	{
		//_gcvt_s(val_buf, TEMP_VALUE_BUF_SIZE, *((double*)(&code)), type == RC_REAL ? 15 : 6);
		gcvt(*((double*)(&code)), a_type == RC_REAL ? 15 : 6, val_buf);
		size_t s = strlen(val_buf);
		if(s && val_buf[s-1] == '.')
			val_buf[s-1] = 0;
		return (int)strlen(val_buf);
	}
	else if(a_type==RC_NUM)
	{
		Text(code,val_buf,Type().GetScale());
		return (int)strlen(val_buf);
	}
	else if(ATI::IsDateTimeType(a_type))
	{
		RCDateTime rcdt(code, a_type);
		RCBString local_rcb(val_buf, TEMP_VALUE_BUF_SIZE);
		return (int)strlen(rcdt.ToRCString(local_rcb));
	}
	return 0;
}

RCBString RCAttr::DecodeValue_S(_int64 code)			// original 0-level value for a given 1-level code
{
	if(code == NULL_VALUE_64)
	{
		return RCBString();
	}
	int v = (int)code;
	if(TypeName()==RC_STRING || TypeName()==RC_VARCHAR)
	{
		BHASSERT_WITH_NO_PERFORMANCE_IMPACT(PackType() == PackN);
		return dic.GetRealValue(v);
	}
	char buf[TEMP_VALUE_BUF_SIZE];					// the longest possible number or date
	int len = DecodeValue_T(code,buf);
	return RCBString(buf,len,true);
}

int RCAttr::EncodeValue_T(const RCBString& rcbs, int new_val, BHReturnCode* bhrc)	// 1-level code value for a given 0-level (text) value
													// if new_val=1, then add to dictionary if not present
{
	if(bhrc)
		*bhrc = BHRC_SUCCESS;
	if(rcbs.IsNull()) return NULL_VALUE_32;
	if(ATI::IsStringType(TypeName()))
	{
		BHASSERT_WITH_NO_PERFORMANCE_IMPACT(PackType() == PackN);
		LoadPackInfo();
		int vs;
		if(new_val==0)	vs=dic.GetEncodedValue(rcbs);
		else			vs=dic.Add(rcbs);
		if(vs<0) return NULL_VALUE_32;
		return vs;
	}
	char const* val = rcbs.val;
	if(val == 0)
		val = ZERO_LENGTH_STRING;
	if(ATI::IsDateTimeType(TypeName()) || TypeName() == RC_BIGINT)
	{
		BHASSERT(0, "Wrong data type!");
	}
	else
	{
		RCNum rcn;
		BHReturnCode tmp_bhrc = RCNum::Parse(rcbs, rcn, TypeName());
		if(bhrc)
			*bhrc = tmp_bhrc;
		return (int)(_int64)rcn;
	}
	return NULL_VALUE_32;
}

_int64 RCAttr::EncodeValue64(char const*val, int new_val, BHReturnCode* bhrc)	// 1-level code value for a given 0-level (text) value
														// if new_val=1, then add to dictionary if not present
{
	if(bhrc)
		*bhrc = BHRC_SUCCESS;
	_int64 res = NULL_VALUE_64;
	if(val == 0)
		return res;
	BHReturnCode tmp_bhrc = BHRC_SUCCESS;

	if(ATI::IsRealType(TypeName()))
	{
		RCNum rcn;
		tmp_bhrc = RCNum::Parse(val, rcn, TypeName());
		if(bhrc)
			*bhrc = tmp_bhrc;
		return rcn.Value();
	}
	if(ATI::IsDateTimeType(TypeName()))
	{
		res = 0;
		RCDateTime rcdt;
		if(BHReturn::IsError(tmp_bhrc = RCDateTime::Parse(val, rcdt, TypeName())))
		{
			if(bhrc)
				*bhrc = tmp_bhrc;
			throw DataTypeConversionRCException(BHERROR_DATACONVERSION);
		}
		if(bhrc)
			*bhrc = tmp_bhrc;
		return rcdt.GetInt64();
	}
	else if(TypeName() == RC_NUM)
	{
		RCNum rcn;
		tmp_bhrc = RCNum::ParseNum(val, rcn, Type().GetScale());
		res = rcn.Value();
		if(res > 0 && res >= (_int64)Uint64PowOfTen(Type().GetPrecision()))
		{
			res = Uint64PowOfTen(Type().GetPrecision()) - 1;
			if(bhrc)
				*bhrc = tmp_bhrc;
		}
		else if(res < 0 && res <= (_int64)Uint64PowOfTen(Type().GetPrecision()) * -1)
		{
			res = Uint64PowOfTen(Type().GetPrecision()) * -1 + 1;
			if(bhrc)
			*bhrc = tmp_bhrc;
		}
		return res;
	}
	else if(TypeName() == RC_BIGINT) {
		RCNum rcn;
		BHReturnCode tmp_bhrc = ValueParserForText::ParseBigInt(val, rcn);
		if(bhrc)
			*bhrc = tmp_bhrc;
		return rcn.Value();
	}

	res = EncodeValue_T(val, new_val, &tmp_bhrc);
	if(bhrc)
		*bhrc = tmp_bhrc;
	if((int)res == NULL_VALUE_32) return NULL_VALUE_64;
	return res;

}

_int64 RCAttr::EncodeValue64(const RCBString& val, int new_val, BHReturnCode* bhrc)	// 1-level code value for a given 0-level (text) value
// if new_val=1, then add to dictionary if not present
{
	if(bhrc)
		*bhrc = BHRC_SUCCESS;
	_int64 res = NULL_VALUE_64;
	if(val.IsNull())
		return res;
	BHReturnCode tmp_bhrc = BHRC_SUCCESS;

	if(ATI::IsRealType(TypeName()))
	{
		RCNum rcn;
		tmp_bhrc = RCNum::Parse(val, rcn, TypeName());
		if(bhrc)
			*bhrc = tmp_bhrc;
		return rcn.Value();
	}
	if(ATI::IsDateTimeType(TypeName()))
	{
		res = 0;
		RCDateTime rcdt;
		if(BHReturn::IsError(tmp_bhrc = RCDateTime::Parse(val, rcdt, TypeName())))
		{
			if(bhrc)
				*bhrc = tmp_bhrc;
			throw DataTypeConversionRCException(BHERROR_DATACONVERSION);
		}
		if(bhrc)
			*bhrc = tmp_bhrc;
		return rcdt.GetInt64();
	}
	else if(TypeName() == RC_NUM)
	{
		RCNum rcn;
		tmp_bhrc = RCNum::ParseNum(val, rcn, Type().GetScale());
		res = rcn.Value();
		if(res > 0 && res >= (_int64)Uint64PowOfTen(Type().GetPrecision()))
		{
			res = Uint64PowOfTen(Type().GetPrecision()) - 1;
			if(bhrc)
				*bhrc = tmp_bhrc;
		}
		else if(res < 0 && res <= (_int64)Uint64PowOfTen(Type().GetPrecision()) * -1)
		{
			res = Uint64PowOfTen(Type().GetPrecision()) * -1 + 1;
			if(bhrc)
				*bhrc = tmp_bhrc;
		}
		return res;
	} else if(TypeName() == RC_BIGINT) {
		RCNum rcn;
		BHReturnCode tmp_bhrc = ValueParserForText::ParseBigInt(val, rcn);
		if(bhrc)
			*bhrc = tmp_bhrc;
		res = rcn.Value();
	}

	res = EncodeValue_T(val, new_val, &tmp_bhrc);
	if(bhrc)
		*bhrc = tmp_bhrc;
	if((int)res == NULL_VALUE_32) return NULL_VALUE_64;
	return res;
}

// transform a value (CQTerm) into 1-level code, take into account the precision etc.
// no changes for REAL; rounded=true iff v has greater precision than the column and the returned result is rounded down
_int64 RCAttr::EncodeValue64(CQTerm v, bool &rounded, BHReturnCode* bhrc)
{
	// TODO: implement for VCs!
	return NULL_VALUE_64;
/*	rounded = false;

	if(bhrc)
		*bhrc = BHRC_SUCCESS;

	if(v.IsNull())
		return NULL_VALUE_64;

	if(v.IsEncoded())
		return v.val_n;


	if(v.val_t != NULL)
	{
		if(Type().IsLookup())
		{
			size_t s = strlen(v.val_t);
			if(s)
				return EncodeValue_T(v.val_t, 0, bhrc);
			else
				return EncodeValue_T("", 0, bhrc);
		}
		else if(ATI::IsDateTimeType(TypeName()) || ATI::IsNumericType(TypeName()) || ATI::IsDateTimeNType(TypeName()))
		{
			size_t s = strlen(v.val_t);
			if(s)
				return EncodeValue64(v.val_t, 0, bhrc);
			else
				return EncodeValue64("", 0, bhrc);

		}
		throw NotImplementedRCException("Input parameter cannot be interpreted.");
	}
	if(v.val_n == NULL_VALUE_64)
		return NULL_VALUE_64;

	if(ATI::IsRealType(TypeName()))
	{
		if(v.scale==-1) return v.val_n;					// already stored as double
		double res=double(v.val_n);
		//for(int i=0;i<v.precision;i++) res/=10;
		res = (double)RCNum(v.val_n, v.scale);
		return *(_int64*)(&res);								// encode
	}
	else if(ATI::IsDateTimeType(TypeName()))
	{
		if( v.val_n == 0 )			// special value: null in not-null column
			return 0;
		RCDateTime rcdt;
		BHReturnCode tmp_bhrc = BHRC_SUCCESS;
		if(BHReturn::IsError(tmp_bhrc = RCDateTime::Parse(v.val_n, rcdt, TypeName())))
		{
			if(bhrc)
				*bhrc = tmp_bhrc;
			throw DataTypeConversionRCException();
		}
		else
		{
			if(bhrc)
				*bhrc = tmp_bhrc;
			return rcdt.GetInt64();
		}
	}
	BHASSERT(PackType() == PackN, "Pack type must be numeric!");
	if(v.scale==-1)				// v is double
	{
		double vd=*(double*)(&v.val_n);
		unsigned char dplaces = Type().GetScale();
		for(int i=0;i<dplaces;i++)
			vd*=10;				// translate into _int64 of proper precision
		if(vd>PLUS_INF_64) return PLUS_INF_64;
		if(vd<MINUS_INF_64) return MINUS_INF_64;
		_int64 res=_int64(vd);
		if(fabs(vd-double(res))>0.01)	rounded=true;		// ignore errors which are 2 digits less than declared precision
		return res;
	}
	int vprec=v.scale;
	_int64 res=v.val_n;
	unsigned char dplaces = Type().GetScale();
	while(vprec<dplaces)
	{
		if(res<MINUS_INF_64/10) return MINUS_INF_64;
		if(res>PLUS_INF_64/10) return PLUS_INF_64;
		res*=10;
		vprec++;
	}
	while(vprec>dplaces)
	{
		if(res%10!=0) rounded=true;
		res/=10;
		vprec--;
	}
	return res;
	*/
}

// transform a RCNum value into 1-level code, take into account the precision etc.
// no changes for REAL; rounded=true iff v has greater precision than the column and the returned result is rounded down
_int64 RCAttr::EncodeValue64(RCDataType *v, bool &rounded, BHReturnCode* bhrc)
{
	rounded = false;
	if(bhrc)
		*bhrc = BHRC_SUCCESS;
	if(!v || v->IsNull())
		return NULL_VALUE_64;

	if((Type().IsLookup() && v->Type() != RC_NUM)) {
		return EncodeValue_T(v->ToRCString(), 0, bhrc);
	} else if(ATI::IsDateTimeType(TypeName()) || ATI::IsDateTimeNType(TypeName())) {
		return ((RCDateTime*) v)->GetInt64();
	}
	BHASSERT(PackType() == PackN, "Pack type must be numeric!");

	_int64 vv = ((RCNum*) v)->Value();
	int vp = ((RCNum*) v)->Scale();
	if(ATI::IsRealType(TypeName())) {
		if(((RCNum*)v)->IsReal())
			return vv; // already stored as double
		double res = double(vv);
		res /= Uint64PowOfTen(vp);
		//for(int i=0;i<vp;i++) res*=10;
		return *(_int64*) (&res); // encode
	}
	if(((RCNum*)v)->IsReal()) // v is double
	{
		double vd = *(double*)(&vv);
		vd *= Uint64PowOfTen(Type().GetScale());				// translate into _int64 of proper precision
		if(vd > PLUS_INF_64)
			return PLUS_INF_64;
		if(vd < MINUS_INF_64)
			return MINUS_INF_64;
		_int64 res = _int64(vd);
		if(fabs(vd - double(res)) > 0.01)
			rounded = true; // ignore errors which are 2 digits less than declared precision
		return res;
	}
	unsigned char dplaces = Type().GetScale();
	while(vp < dplaces) {
		if(vv < MINUS_INF_64 / 10)
			return MINUS_INF_64;
		if(vv > PLUS_INF_64 / 10)
			return PLUS_INF_64;
		vv *= 10;
		vp++;
	}
	while(vp > dplaces) {
		if(vv % 10 != 0)
			rounded = true;
		vv /= 10;
		vp--;
	}
	return vv;
}

_int64 RCAttr::EncodeValue64(const RCValueObject& v, bool& rounded, BHReturnCode* bhrc)
{
	return EncodeValue64(v.Get(), rounded, bhrc);
}

RCBString RCAttr::GetPrefix(int pack)
{
	LoadPackInfo();

	if(GetPackOntologicalStatus(pack) == NULLS_ONLY)
		return 0;

	char* min_s = (char*)&local_min[pack];
	char* max_s = (char*)&local_max[pack];


	int dif_pos = 0;
	for(; dif_pos < sizeof(_uint64) && min_s[dif_pos] == max_s[dif_pos]; dif_pos++);

	if(dif_pos == 0)
		return ZERO_LENGTH_STRING;

	RCBString s(min_s, dif_pos, true);
	return s;
}

uint AttributeTypeInfo::ExternalSize(RCAttr* attr, EDF edf)
{
	AttributeType attrt = attr->TypeName();
	return AttributeTypeInfo::ExternalSize(attrt, attr->Type().GetPrecision(), attr->Type().GetScale(), edf);
}

uint AttributeTypeInfo::ExtrnalSize(EDF edf)
{
	return AttributeTypeInfo::ExternalSize(attrt, precision, scale, edf);
}

uint AttributeTypeInfo::ExternalSize(AttributeType attrt, int precision, int scale, EDF edf)
{
	return DataFormat::GetDataFormat(edf)->ExtrnalSize(attrt, precision, scale);
}

int AttributeTypeInfo::TextSize()
{
	return DataFormat::GetDataFormat(TXT_VARIABLE)->ExtrnalSize(attrt, precision, scale, &collation);
}

bool AttributeTypeInfo::IsVariableSize(AttributeType attr_type)
{
	if(attr_type == RC_VARCHAR || attr_type == RC_VARBYTE || attr_type == RC_BIN)
		return true;
	return false;
}

bool AttributeTypeInfo::IsFixedSize(AttributeType attr_type, EDF edf)
{
	return !DataFormat::GetDataFormat(edf)->IsVariableSize(attr_type);
}

void RCAttr::LockPackForUse(unsigned pack_no, ConnectionInfo& conn)
{
	MEASURE_FET("RCAttr::LockPackForUse(...)");
	assert((int)pack_no<NoPack());

	if(conn.killed())
		throw KilledRCException();

	LoadPackInfo();
	MEMORY_LOCK_GUARD;
	packs[pack_no] = conn.GetTransaction()->GetAttrPackRO(table_number, attr_number, pack_no + packs_omitted);
	if(packs[pack_no]) {
		if(pack_mode[pack_no] == 1)
			pack_mode[pack_no] = 2;
		packs[pack_no]->Lock();
	} else if(is_stored[pack_no]) {
		if(PackType() == PackS)
			packs[pack_no] = new AttrPackS(*this, pack_no);
		else
			packs[pack_no] = new AttrPackN(*this, pack_no);

		conn.NotifyPackLoad();
		packs[pack_no]->Lock();
	}
	LoadPack(pack_no);
}

void RCAttr::LockPackForUse(unsigned pack_no)
{
	LockPackForUse(pack_no, ConnectionInfoOnTLS.Get());
}

void RCAttr::UnlockPackFromUse(unsigned pack_no)
{
	MEASURE_FET("RCAttr::UnlockPackFromUse(...)");
	assert((int)pack_no<NoPack());
	LoadPackInfo();
	MEMORY_LOCK_GUARD;
	if(packs[pack_no]) {
		packs[pack_no]->Unlock();
		if(!packs[pack_no]->IsLocked() ) {
			if(packs[pack_no]->IsEmpty() || CachingLevel == 0)  
				SynchronizedDataCache::GlobalDataCache().ReleasePack(table_number, attr_number, pack_no + packs_omitted);
			packs[pack_no] = 0;
			if(pack_mode[pack_no] == 2)
				pack_mode[pack_no] = 1;
		}

	}
}

void RCAttr::ForgetPack(int pack_no)
{
//	pthread_mutex_lock(&cs);
	pack_mode[pack_no] = 1;
	packs[pack_no] = NULL;
//	pthread_mutex_unlock(&cs);
}

bool RCAttr::OldFormat()
{
	return (FileFormat() < 10);
}

/******************************************
 * ABSwitcher implementation
 ******************************************/

ABSwitch ABSwitcher::GetState(const char* path)
{
    const size_t MAX_PATH_LENGTH = 1024;
    char name [ MAX_PATH_LENGTH ];
    GenerateName(name, sizeof(name), path);
    return (DoesFileExist(name)) ? ABS_B_STATE : ABS_A_STATE;
}

int ABSwitcher::FlipState(const char* path)
{
    const size_t MAX_PATH_LENGTH = 1024;
    char name [ MAX_PATH_LENGTH ];
    bool is_file_exist;

    int res = GenerateName(name, sizeof(name), path);
    if (res) return -1;

    is_file_exist = DoesFileExist(name);
    if (is_file_exist) {
        // File exists. Delete it now.
        try {
            RemoveFile(name);
        } catch(...) {
            rclog << lock << "ERROR: Failed to delete AB switch file " << name << unlock;
            return -1;
        }
    }
    else {
        // File does not exist. Create it now.
        try {
            IBFile ibfl;
            ibfl.OpenCreate(name);
        } catch(...) {
            rclog << lock << "ERROR: Failed to create AB switch file " << name << unlock;
            return -1;
        }
    }

    try {
        FlushDirectoryChanges(path);
    } catch(DatabaseRCException& e) {
        rclog << lock << e.what() << " DatabaseRCException " << __FILE__ << __LINE__ << unlock;
        return -1;
    }

    return 0;
}

int ABSwitcher::GenerateName(char* name, size_t sz, const char* path)
{
    size_t len;

    strncpy (name, path, sz);
    strcat(name,"/");
    strncat (name, AB_SWITCH_FILE_NAME, sz);
    len = strlen (name);
    if (len >= sz-2) {
        rclog << lock << "ERROR: AB switch file path is too long" << unlock;
        return -1;
    }

    return 0;
}

/*static*/
void ABSwitcher::UnitTest()
{
    int res;
    const char *test_path = "/tmp/";
    const char *brocken_path = "/1234/";
    ABSwitcher absw;

    rclog << lock << "TEST: ABSwitcher::unitTest()" << unlock;

    ABSwitch stateOne = absw.GetState(test_path);
    rclog << lock << "INFO: get AB switch value " << SwitchName (stateOne) << unlock;

    res =  absw.FlipState(test_path);
    if (res) {
		rclog << lock << "ERROR: Failed to flip AB switch on path " << test_path << unlock;
        return;
    }

    ABSwitch stateTwo = absw.GetState(test_path);
    if (stateOne == stateTwo) {
        rclog << lock << "ERROR: State has not changed after switch was flipped" << unlock;
        return;
    }

    res = absw.FlipState(brocken_path);
    if (res == 0) {
        rclog << lock << "ERROR: Flipped AB switch  in the bogus directory" << unlock;
        return;
    }
}

/*static*/
const char* ABSwitcher::SwitchName( ABSwitch value )
{
    if (value == ABS_A_STATE) return "ABS_A_STATE";
    return "ABS_B_STATE";
}

/////// RCAttr metadata methods


bool RCAttr::Load(const char* cmd_file, int open_session_type)
{
	char* tmp_buf = NULL;
	char* buf;
	uint bufsize;
	IBFile fattr_read;
	uint bytes_processed;

	tmp_buf = new char [10000];
	if (!tmp_buf) {
		rclog << lock << "Error: Memroy allocation failed!" << unlock;
		return false;
	}

	try {
		fattr_read.OpenReadOnly(cmd_file);
		if (fattr_read.Read(tmp_buf, CLMD_HEADER_LEN) != CLMD_HEADER_LEN)
			return BH_FILE_ERROR;

		bytes_processed = LoadHeader(tmp_buf);

		bufsize = fattr_read.Read(tmp_buf, 10000);

		bytes_processed = LoadColumnInfo(tmp_buf);
		LoadSessionInfo(tmp_buf + bytes_processed, bufsize, &fattr_read, open_session_type);

		fattr_read.Close();
	}
	catch(DatabaseRCException& e)
	{
		delete [] tmp_buf;
		rclog << lock << "Error while opening file " << cmd_file<< " : " << e.what() << unlock;
		throw e;
	}
	delete [] tmp_buf;
    return true;
}

uint RCAttr::LoadHeader(const char* a_buf)
{
	natural_size_saved = 0;
	compressed_size = 0;

	bool is_lookup;
	NullMode nulls_mode;
	int scale = 0;
	int precision = 0;
	AttributeType	type;

	file_format = FILEFORMAT(a_buf);
	if (file_format == CLMD_FORMAT_INVALID_ID)
		return RCATTR_BADFORMAT;

	if ((uchar)a_buf[8] < 128) {
		type = (AttributeType)a_buf[8];
		is_lookup = true;
	} else {
		type = (AttributeType)((uchar)a_buf[8]-128);
		is_lookup = false;
	}

	if (((type == RC_VARCHAR ||
	  type == RC_STRING) && is_lookup) ||
	  ATI::IsNumericType(type) ||
	  ATI::IsDateTimeType(type))
		pack_type = PackN;
	else
		pack_type = PackS;

	no_obj = *((_int64*)(a_buf + 9));
	no_nulls = *((_int64*)(a_buf + 17));

	nulls_mode = (NullMode)(a_buf[25] % 4);

	SetUnique(((a_buf[25] & 0x10) ? true : false));
	SetUniqueUpdated(((a_buf[25] & 0x20) ? true : false));

	if(a_buf[25] & 0x40)
		inserting_mode = 1;
	else
		inserting_mode = 0;

	precision = *((ushort*)(a_buf + 26));
	scale = a_buf[28];

	ct.Initialize(type, nulls_mode, is_lookup, precision, scale);

	no_pack = *((int*)(a_buf + 30));

	packs_offset = *((uint*)(a_buf + 34));
	dict_offset = *((uint*)(a_buf + 38));
	special_offset = *((uint*)(a_buf + 42));

	i_min = 0;
	i_max = 0;

	last_pack_index = no_pack != 0 ? no_pack - 1 : 0;
	no_pack_ptr = 0;

    return 46;  // 42 + 4, where 42 is an offset for "special_offset" and 4 is a size of of "special_offset"
}

uint RCAttr::LoadColumnInfo(const char* a_buf)
{
	const char* buf = a_buf;
//	size_t const name_ct(strlen(buf) + 1);

//	char name_tmp [name_ct];

	SetName(buf);
//	strcpy(name, buf);
	buf += strlen(buf) + 1;

	SetDescription(buf);
//	size_t const desc_ct(strlen(buf)+1);
//	desc = new char [desc_ct];
//	strcpy(desc, buf);
	buf += strlen(buf) + 1;

	total_p_f =* ((int*)buf);
	buf += 4;

	// the remainings after table of files' locations (if not default)
	int p_f = *((int*)buf);
	if (p_f != PF_END)
	{
		rclog << lock << "Error in attribute file: bad format" << unlock;
		assert(0);
	}
	buf += 4;
	return (a_buf - buf);
}

uint RCAttr::LoadSessionInfo(const char* cur_buf, uint actual_size, IBFile* fcmd_file, int open_session_type)
{
	char* tmp_buf = new char [10000];

	// internal representation (which PackX class to use): numerical, text or binary?
	pack_info_collapsed = (dict_offset || no_pack);

	// load special blocks
	if(actual_size < special_offset + 27 + 2 * sizeof(_uint64)) {
		fcmd_file->Seek(special_offset, SEEK_SET);
		fcmd_file->Read(tmp_buf, 27 + 2 * sizeof(_uint64));
		cur_buf = tmp_buf;
	}
	else
		cur_buf = cur_buf + special_offset;

	if (special_offset>0 && *(uchar*)(cur_buf+5))
		fcmd_file->Seek(special_offset, SEEK_SET);

	while (special_offset > 0) {
		int b_len = *(int*)cur_buf;
		uchar b_type = *(uchar*)(cur_buf + 4);
		uchar b_next = *(uchar*)(cur_buf + 5);

		if(b_type == 0)	{ 								// type 0: the first special block, rollback and session info
			savefile_loc[0] = *((int*)(cur_buf + 6));		// the first special block: rollback and session info
			savefile_loc[1] = *((int*)(cur_buf + 10));
			savepos_loc[0] = *((uint*)(cur_buf + 14));
			savepos_loc[1] = *((uint*)(cur_buf + 18));
			current_save_loc = cur_buf[22];
			if (open_session_type == 0 || open_session_type == 3)	// only in case of read (normal/backup) sessions, otherwise read_session_id remains 0
				last_save_session_id = *((uint*)(cur_buf + 23));
			natural_size_saved = *((_uint64*)(cur_buf + 23 + sizeof(uint)));
			compressed_size = *((_uint64*)(cur_buf + 23 + sizeof(uint) + sizeof(_uint64)));
		}
		else
			rclog << lock << "Warning: special block (unknown type " << int(b_type) << ") ignored in attribute " << Name() << unlock;

		if (b_next == 0)
			special_offset = 0;
		else {
			fcmd_file->Seek(b_len, SEEK_CUR);
			fcmd_file->Read(&b_len, 4);
			fcmd_file->Seek(-4, SEEK_CUR);
			fcmd_file->Read(tmp_buf, b_len);
			cur_buf = tmp_buf;
		}
	}
	current_read_loc = current_save_loc;
	delete [] tmp_buf;
	return 1;
}

bool RCAttr::Save(const char* cmd_file, FTree* dic, _int64 comprs_size)
{
	IBFile fattr;
	fattr.OpenCreateEmpty(cmd_file);

	const int buf_size = 10000;
	char buf[buf_size];

	strcpy(buf, CLMD_FORMAT_RSC10);

	// Reserved for security encryption info
	*((uint*)(buf + 4)) = 0;

	// The first bit (0x80) indicates string interpretation: 0 for dictionary, 1 for long text (>=RSc4)
	buf[8] = (char)((char)TypeName() + (Type().IsLookup() ? 0 : 0x80));
	*((_int64*)(buf + 9)) = no_obj;
	*((_int64*)(buf + 17)) = no_nulls;

	// buf[25] stores attribute's flags: null_mode + 4*declared_unique + 8*is_primary  + 16*is_unique + 32*is_unique_updated
	// Now declared_unique, is_primary are obsolete. Can be removed in future file format
	buf[25] = char(Type().GetNullsMode()) + (IsUnique() ?16:0) + (IsUniqueUpdated() ?32:0) + (inserting_mode == 1? 0x40:0);
	*((ushort*)(buf + 26)) = Type().GetPrecision();
	buf[28] = Type().GetScale();

	// Now it is obsolete field. Can be removed in future file format
	buf[29] = (uchar)0;
	*((int*)(buf + 30)) = no_pack;

	// Actual values will be written at the end
	packs_offset = 0;
	dict_offset = 0;
	special_offset = 0;

	// Actual values will be written at the end
	*((uint*)(buf+34)) = packs_offset;
	*((uint*)(buf+38)) = dict_offset;
	*((uint*)(buf+42)) = special_offset;

	// Temporary, because there is no offset values yet
	fattr.Write(buf, 46);

	fattr.Write(Name(), (int)strlen(Name()) + 1);
	if (Description())
		fattr.Write(Description(),(int)strlen(Description()) + 1);
	else
		fattr.Write("\0", 1);

	// List of all nontrivial path names
	fattr.Write((char*)(&total_p_f),4);

	int p_f=PF_END;
	fattr.Write((char*)(&p_f),4);

	packs_offset = fattr.Tell();
	dict_offset = fattr.Tell();

	AttributeType a_type = TypeName();
	if(Type().IsLookup()){
		*((_uint64*)(buf)) = i_min;
		*((_uint64*)(buf + 8)) = i_max;
		fattr.Write(buf, 16);
		uchar *dic_buf = new uchar [dic->ByteSize()];
		if (!dic_buf)
			rclog << lock << "Error: out of memory (" << dic->ByteSize() << " bytes failed). (23)" << unlock;
		uchar *p_dic_buf = dic_buf;
		dic->SaveData(p_dic_buf);
		fattr.Write((char*)dic_buf, dic->ByteSize());
		delete [] dic_buf;
	}
	else if(ATI::IsNumericType(a_type)  ||
		ATI::IsDateTimeType(a_type) || ATI::IsDateTimeNType(a_type)){
		*((_uint64*)(buf)) = i_min;
		*((_uint64*)(buf+8)) = i_max;
		fattr.Write(buf,16);
	}
	// ... other types of dictionary ...
	else {
		// e.g. string_no_dict==1
		dict_offset = 0;
	}

	special_offset = fattr.Tell();

	// saving session and rollback info
	uint bs = 27 + 2*sizeof(_uint64);
	*((uint*)(buf)) = bs;
	buf[4] = 0;	// = block type
	buf[5] = 0;	// = last block

	*((int*)(buf+6)) = savefile_loc[0];
	*((int*)(buf+10)) = savefile_loc[1];

	*((uint*)(buf+14)) = savepos_loc[0];
	*((uint*)(buf+18)) = savepos_loc[1];

	buf[22] = current_save_loc;
	*((uint*)(buf+23)) = save_session_id;

	*((_uint64*)(buf + 23 + sizeof(uint))) = natural_size_saved;
	*((_uint64*)(buf + 23 + sizeof(uint) + sizeof(_uint64))) = compressed_size;
	fattr.Write(buf, bs);

	// Update compressed data size
	compressed_size = comprs_size + fattr.Tell();
	fattr.Seek(-(long)sizeof(_uint64), SEEK_CUR);
	fattr.Write(&compressed_size, sizeof(_uint64));

	// Update addresses of sections
	*((uint*)(buf)) = packs_offset;
	*((uint*)(buf+4)) = dict_offset;
	*((uint*)(buf+8)) = special_offset;

	//seekp(34,ios::beg);
	fattr.Seek(34, SEEK_SET);
	fattr.Write(buf, 12);
	fattr.Close();
	return true;
}
