/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef RCATTRTYPEINFO_H_INCLUDED
#define RCATTRTYPEINFO_H_INCLUDED

#include <vector>

#include "common/CommonDefinitions.h"

class RCAttr;

typedef std::vector<enum_field_types> fields_t;

class AttributeTypeInfo
{
public:
	AttributeTypeInfo(AttributeType attrt, bool notnull, ushort precision = 0, ushort scale = 0, bool lookup = false, bool for_insert_ = false, DTCollation collation = DTCollation())
		: attrt(attrt), precision(precision), scale(scale), notnull(notnull), lookup(lookup), for_insert(for_insert_), collation(collation)
	{
	}

	uint	ExtrnalSize(EDF edf);
	int		TextSize();

	static bool IsVariableSize(AttributeType attr_type);
	static bool IsFixedSize(AttributeType attr_type, EDF edf);
	static uint ExternalSize(RCAttr* attr, EDF edf);
	static uint ExternalSize(AttributeType attrt, int precision, int scale, EDF edf);
	static ushort ObjPrefixSizeByteSize(AttributeType attr_type, EDF edf);
	static int	TextSize(AttributeType attrt, ushort precision, ushort scale, DTCollation col = DTCollation());

	static bool IsIntervalType(AttributeType attr_type);
	static bool IsInteger32Type(AttributeType attr_type);
	static bool IsIntegerType(AttributeType attr_type);
	static bool IsNumericType(AttributeType attr_type);
	static bool IsFixedNumericType(AttributeType attr_type);
	static bool IsRealType(AttributeType attr_type);
	static bool IsBinType(AttributeType attr_type);
	static bool IsTxtType(AttributeType attr_type);
	static bool IsCharType(AttributeType attr_type);
	static bool IsStringType(AttributeType attr_type);
	static bool IsDateTimeType(AttributeType attr_type);
	static bool IsDateTimeNType(AttributeType attr_type);

	static int NULL32(AttributeType attrt);
	static long long NULL64(AttributeType attrt);

	/*static _int64 MaxValue(AttributeType attrt);
	static _int64 MinValue(AttributeType attrt);
	static _int64 ErrorValue(AttributeType attrt);*/

	AttributeType Type() const	{ return attrt; }
	ushort Precision() const 	{ return precision; }
	ushort Scale() const		{ return scale; }
	ushort CharLen() const		{ return precision / collation.collation->mbmaxlen; }
	bool NotNull() const		{ return notnull; }
	bool Lookup() const			{ return lookup; }
	uint Params() const
	{
		uint parameter = 0;
		if(lookup)
			parameter |= 0x00000002;
		if(notnull)
			parameter |= 0x00000001;
		if(for_insert)
			parameter |= 0x00000004;
		return parameter;
	}

	operator AttributeType()	{ return Type(); }
	DTCollation GetCollation() { return collation; }
private:
	AttributeType attrt;
	int precision;
	int scale;
	bool notnull;
	bool lookup;
	bool for_insert;
	uint parameter;
	DTCollation collation;
};
typedef AttributeTypeInfo ATI;

#endif /* not RCATTRTYPEINFO_H_INCLUDED */

