/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "common/CommonDefinitions.h"
#include "system/Channel.h"
#include "system/RCException.h"
#ifndef __BH_COMMUNITY__
#	include "enterprise/local.h"
#else /* not __BH_COMMUNITY__ */
#	include "community/local.h"
#endif /* __BH_COMMUNITY__ */
#include "PackGuardian.h"
#include "RCAttr.h"
#include "RSI_PackPack.h"
#include "vc/VirtualColumn.h"

using namespace std;

_uint64 RCTableImpl::NoObj()
{
	for(int i=0;i<no_attr;i++) {
		LoadAttribute(i);
		if(a[i]) return (_uint64)a[i]->NoObj();
	}
	return 0;
}

_uint64 RCTableImpl::NoNulls(int attr)
{
	assert(attr <= no_attr);
	if(!a[attr])
		LoadAttribute(attr);
	return a[attr]->NoNulls();
}

RCBString RCTableImpl::GetTable_S(_uint64 obj, int attr, bool bin_as_hex)
{
	assert(attr <= no_attr);
	if(!a[attr])
		LoadAttribute(attr);
	assert((_int64)obj <= a[attr]->NoObj());
	return  a[attr]->GetValueString(obj, bin_as_hex);
}

//int RCTableImpl::GetTable_T(_uint64 obj,int attr,char *val_buf)
//{
//	assert(attr <= no_attr);
//	if(!a[attr]) LoadAttribute(attr);
//	assert((_int64)obj <= a[attr]->NoObj());
//	return a[attr]->GetValueText(obj,val_buf);
//}

void RCTableImpl::GetTable_B(_uint64 obj, int attr, int &size, char *val_buf)
{
	assert(attr <= no_attr);
	if(!a[attr]) LoadAttribute(attr);
	assert((_int64)obj <= a[attr]->NoObj());
	a[attr]->GetValueBin(obj, size, val_buf);
}

int RCTableImpl::GetTable32(_uint64 obj,int attr)
{
	assert(attr <= no_attr && (_int64)obj <= a[attr]->NoObj());
	if(!a[attr]) LoadAttribute(attr);
	return a[attr]->GetValueInt32(obj);
}

_int64 RCTableImpl::GetTable64(_uint64 obj,int attr)
{
	assert(attr <= no_attr);
	if(!a[attr]) LoadAttribute(attr);
	assert((_int64)obj <= a[attr]->NoObj());
	return a[attr]->GetValueInt64(obj);
}

//inline double RCTableImpl::GetTable_N(_uint64 obj,int attr,ConnectionInfo *conn)
//{
//	assert(attr <= no_attr);
//	if(!a[attr]) LoadAttribute(attr);
//	assert((_int64)obj <= a[attr]->NoObj());
//	return a[attr]->GetTable_N(obj,conn);
//}

bool RCTableImpl::IsNull(_uint64 obj,int attr)
{
	assert(attr <= no_attr);
	if(!a[attr]) LoadAttribute(attr);
	assert((_int64)obj <= a[attr]->NoObj());
	return ( a[attr]->IsNull(obj) ? true : false );
}

RCValueObject RCTableImpl::GetValue(_uint64 obj,int attr, ConnectionInfo *conn)
{
	assert(attr <= no_attr);
	if(!a[attr]) LoadAttribute(attr);
	assert((_int64)obj <= a[attr]->NoObj());
	return a[attr]->GetValue(obj, false);
}

//int	RCTableImpl::GetTableOrderedLookup(_uint64 obj, int attr, ConnectionInfo *conn)
//{
//	assert(attr <= no_attr);
//	if(!a[attr])
//		LoadAttribute(attr);
//	assert( a[attr]->Type().IsLookup() );
//	assert((_int64)obj <= a[attr]->NoObj());
//	_int64 res = a[attr]->GetValueInt64(obj);
//	if(res == NULL_VALUE_64)
//		return NULL_VALUE_32;
//	return a[attr]->dic.GetSortPosition(res);
//}

int RCTableImpl::GetLength(_uint64 obj, int attr, ConnectionInfo *conn)
{
	assert(attr <= no_attr);
	if(!a[attr]) LoadAttribute(attr);
	assert((_int64)obj <= a[attr]->NoObj());
	return a[attr]->GetLength(obj);
}

RCBString RCTableImpl::DecodeValue_S(_int64 code,int attr)
{
	assert(attr <= no_attr);
	if(!a[attr])
		LoadAttribute(attr);
	return a[attr]->DecodeValue_S(code);
}

//_int64 RCTableImpl::DecodeOrderedLookup(_int64 code, int attr)
//{
//	assert(attr <= no_attr);
//	if(!a[attr])
//		LoadAttribute(attr);
//	if(code == NULL_VALUE_64)
//		return NULL_VALUE_64;
//	return a[attr]->dic.GetSortValue(code);
//}

_int64 RCTableImpl::EncodeValue_T(char *val,int attr)
{
	assert(attr >= 0 && attr <= no_attr);
	if(!a[attr])
		LoadAttribute(attr);
	return a[attr]->EncodeValue64(val);
}

_int64 RCTableImpl::EncodeValue_T(RCBString& val,int attr,bool sort_order)
{
	assert(attr >= 0 && attr <= no_attr);
	if(!a[attr])
		LoadAttribute(attr);
	int res = a[attr]->EncodeValue_T(val,0);
	if(res==NULL_VALUE_32) return NULL_VALUE_64;
	if(sort_order) return a[attr]->dic.GetSortPosition(res);
	return res;
}

RSValue	RCTableImpl::AttrIsDistinct(int n_a)
{
	assert(n_a >= 0 && n_a <= no_attr);
	LoadAttribute(n_a);
	return a[n_a]->PhysicalColumn::IsDistinct();
}

void RCTableImpl::SetDistinct(int n_a, RSValue v)
{
	assert(n_a >= 0 && n_a <= no_attr);
	LoadAttribute(n_a);
	a[n_a]->WriteUnique(v);
}

_uint64 RCTableImpl::ApproxAnswerSize(int attr,Descriptor &d)	// provide the most probable approximation of number of objects matching the condition
{
	assert(attr <= no_attr);
	if(!a[attr]) LoadAttribute(attr);
	if(a[attr]==NULL) return 0;
	return a[attr]->ApproxAnswerSize(d);
}

ushort RCTableImpl::MaxStringSize(int n_a, Filter *f)
{
	assert(n_a >= 0 && n_a <= no_attr);
	_int64 res = 0;
	if(NoObj() == 0)
		return 1;
	LoadAttribute(n_a);
	if(a[n_a])
		return a[n_a]->MaxStringSize(f);
	return 1;
}

RSIndex_PP* RCTableImpl::LoadRSI_PP(RCTableImpl *t2, int a1, int a2, bool &flipped)		// remember to use ReleaseRSI() after use, before someone will delete sec
{
	if(rsi_manager==NULL) return NULL;
	RSIndex_PP *res;
	if(tab_num>t2->tab_num || (tab_num==t2->tab_num && a1>a2))	// switch order?
	{
		flipped = true;
		res = (RSIndex_PP *)(rsi_manager->GetIndex(RSIndexID(RSI_JPP, t2->tab_num, a2, tab_num, a1),0));
		if(res)
			res->Verify(t2->NoObj(),NoObj());
		return res;
	}
	flipped = false;
	res = (RSIndex_PP *)(rsi_manager->GetIndex(RSIndexID(RSI_JPP, tab_num, a1, t2->tab_num, a2),0));
	if(res)
		res->Verify(NoObj(),t2->NoObj());
	return res;
}

void RCTableImpl::ReleaseIndex(RSIndex *rsi)
{
	if(rsi_manager && rsi)	rsi_manager->ReleaseIndex(rsi);
}
