/* Copyright (C)  2005-2008 Infobright Inc.

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License version 2.0 as
 published by the Free  Software Foundation.

 This program is distributed in the hope that  it will be useful, but
 WITHOUT ANY WARRANTY; without even  the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License version 2.0 for more details.

 You should have received a  copy of the GNU General Public License
 version 2.0  along with this  program; if not, write to the Free
 Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
 02111-1307 USA  */

#ifndef __BH_COMMUNITY__
#	include "enterprise/local.h"
#else /* not __BH_COMMUNITY__ */
#	include "community/local.h"
#endif /* __BH_COMMUNITY__ */
#include "ParametrizedFilter.h"
#include "TempTable.h"
#include "RSI_CMap.h"
#include "MysqlExpression.h"
#include "tools.h"
#include "ValueSet.h"
#include "vc/VirtualColumn.h"
#include "vc/SingleColumn.h"
#include "vc/ExpressionColumn.h"
#include "vc/ConstColumn.h"
#include "Query.h"
#include "common/stdafx.h"
#include "GroupDistinctTable.h"
#include "ColumnBinEncoder.h"

using namespace std;
using namespace boost;

template<class T>
TempTable::AttrBuffer<T>::AttrBuffer(uint page_size, uint elem_size, ConnectionInfo *conn) :
	CachedBuffer<T> (page_size, elem_size, conn)
{
	constant_value = false;
}

template<class T>
TempTable::AttrBuffer<T>::AttrBuffer(_uint64 size, T value, uint page_size, uint elem_size, bool constant,
		ConnectionInfo *conn) :
	CachedBuffer<T> (constant ? 1 : size, value, page_size, elem_size, conn), constant_value(constant)
{
	constant_value = constant;
}

template<class T>
TempTable::AttrBuffer<T>::~AttrBuffer()
{
}

template<class T>
void TempTable::AttrBuffer<T>::Set(_uint64 idx, T value)
{
	if(constant_value && value != CachedBuffer<T>::Get(0)) {
		// it is not the same value any longer, needs to be propagated
		constant_value = false;
		T val = CachedBuffer<T>::Get(0);
		for(_uint64 i = 1; i < idx; i++)
			CachedBuffer<T>::Set(i, val);
	}
	if(constant_value)
		CachedBuffer<T>::Set(0, value);
	else
		CachedBuffer<T>::Set(idx, value);
}

template class TempTable::AttrBuffer<char>;
template class TempTable::AttrBuffer<short>;
template class TempTable::AttrBuffer<int>;
template class TempTable::AttrBuffer<_int64>;
template class TempTable::AttrBuffer<double>;
template class TempTable::AttrBuffer<RCBString>;

TempTable::Attr::Attr() :
	alias(NULL), mode(LISTING), term()
{
	assert(false);
}

TempTable::Attr::Attr(const Attr &a) :
	PhysicalColumn(a)
{
	mode = a.mode;
	distinct = a.distinct;
	term = a.term;
	dim = a.dim;
	assert(a.buffer == NULL); // otherwise we cannot copy Attr !
	buffer = a.buffer;
	no_obj = a.no_obj;
	if(a.alias) {
		alias = new char[strlen(a.alias) + 1];
		strcpy(alias, a.alias);
	} else
		alias = NULL;

	page_size = a.page_size;
}

TempTable::Attr & TempTable::Attr::operator=(const Attr &a)
{
	assert(false);

	//Attr tmp_attr(a);
	//std::swap(tmp_attr, *this);
	return *this;
}

TempTable::Attr::Attr(CQTerm t, ColOperation m, bool dis, char* a, int dim, AttributeType type, uint scale, uint no_digits, NullMode n, DTCollation collation)
	: mode(m), distinct(dis), term(t), dim(dim)
{
	ct.Initialize(type, n, false, no_digits, scale, collation);
	buffer = NULL;
	no_obj = 0;
	if(a) {
		alias = new char[strlen(a) + 1];
		strcpy(alias, a);
	} else
		alias = NULL;
	page_size = 1;
}

void TempTable::Attr::DeleteBuffer()
{
	switch(TypeName()) {
		case RC_INT:
		case RC_MEDIUMINT:
			delete (AttrBuffer<int>*) (buffer);
			break;
		case RC_BYTEINT:
			delete (AttrBuffer<char>*) (buffer);
			break;
		case RC_SMALLINT:
			delete (AttrBuffer<short>*) (buffer);
			break;
		case RC_STRING:
		case RC_VARCHAR:
		case RC_BIN:
		case RC_BYTE:
		case RC_VARBYTE:
			delete (AttrBuffer<RCBString>*) (buffer);
			break;
		case RC_BIGINT:
		case RC_NUM:
		case RC_YEAR:
		case RC_TIME:
		case RC_DATE:
		case RC_DATETIME:
		case RC_TIMESTAMP:
			delete (AttrBuffer<_int64>*) (buffer);
			break;
		case RC_REAL:
		case RC_FLOAT:
			delete (AttrBuffer<double>*) (buffer);
			break;
	}
	buffer = NULL;
	no_obj = 0;
}

TempTable::Attr::~Attr()
{
	DeleteBuffer();
	delete[] alias;
}

int TempTable::Attr::operator==(const TempTable::Attr &sec)
{
	return mode == sec.mode && distinct == sec.distinct && term == sec.term && Type() == sec.Type() && dim == sec.dim;
}

void TempTable::Attr::CreateBuffer(_uint64 size, ConnectionInfo *conn)
{
	// do not create larger buffer than size
	if(size < page_size)
		page_size = (uint) size;
	no_obj = size;
	assert(buffer == NULL);		// otherwise memory leaks may occur
	switch(TypeName()) {
		case RC_INT:
		case RC_MEDIUMINT:
			buffer = new AttrBuffer<int> (size, NULL_VALUE_32, page_size, 0, false, conn);
			break;
		case RC_BYTEINT:
			buffer = new AttrBuffer<char> (size, (char) NULL_VALUE_C, page_size, 0, false, conn);
			break;
		case RC_SMALLINT:
			buffer = new AttrBuffer<short> (size, (short) NULL_VALUE_SH, page_size, 0, false, conn);
			break;
		case RC_STRING:
		case RC_VARCHAR:
		case RC_BIN:
		case RC_BYTE:
		case RC_VARBYTE:
			buffer = new AttrBuffer<RCBString> (size, RCBString(), page_size, this->Type().GetInternalSize() + 4,
					false, conn);
			break;
		case RC_BIGINT:
		case RC_NUM:
		case RC_YEAR:
		case RC_TIME:
		case RC_DATE:
		case RC_DATETIME:
		case RC_TIMESTAMP:
			buffer = new AttrBuffer<_int64> (size, NULL_VALUE_64, page_size, 0, false, conn);
			break;
		case RC_REAL:
		case RC_FLOAT:
			buffer = new AttrBuffer<double> (size, NULL_VALUE_D, page_size, 0, false, conn);
			break;
	}
}

void TempTable::Attr::SetValueInt64(_uint64 obj, _int64 val)
{
	no_obj = obj >= no_obj ? obj + 1 : no_obj;
	switch(TypeName()) {
		case RC_INT:
		case RC_MEDIUMINT:
			if(val == NULL_VALUE_64)
				((AttrBuffer<int> *) buffer)->Set(obj, NULL_VALUE_32);
			else
				((AttrBuffer<int> *) buffer)->Set(obj, (int) val);
			break;
		case RC_BYTEINT:
			if(val == NULL_VALUE_64)
				((AttrBuffer<char> *) buffer)->Set(obj, NULL_VALUE_C);
			else
				((AttrBuffer<char> *) buffer)->Set(obj, (char) val);
			break;
		case RC_SMALLINT:
			if(val == NULL_VALUE_64)
				((AttrBuffer<short> *) buffer)->Set(obj, NULL_VALUE_SH);
			else
				((AttrBuffer<short> *) buffer)->Set(obj, (short) val);
			break;
		case RC_BIGINT:
		case RC_NUM:
		case RC_YEAR:
		case RC_TIME:
		case RC_DATE:
		case RC_DATETIME:
		case RC_TIMESTAMP:
			((AttrBuffer<_int64> *) buffer)->Set(obj, val);
			break;
		case RC_REAL:
		case RC_FLOAT:
			if(val == NULL_VALUE_64)
				((AttrBuffer<double> *) buffer)->Set(obj, NULL_VALUE_D);
			else
				((AttrBuffer<double> *) buffer)->Set(obj, *(double *) &val);
			break;
		default:
			assert(0);
			break;
	}
}

void TempTable::Attr::SetValueString(_uint64 obj, RCBString & val)
{
	no_obj = obj >= no_obj ? obj + 1 : no_obj;
	_int64 val64 = 0;
	double valD = 0.0;
	switch(TypeName()) {
		case RC_INT:
		case RC_MEDIUMINT:
			((AttrBuffer<int> *) buffer)->Set(obj, atoi(val.Value()));
			break;
		case RC_BYTEINT:
			((AttrBuffer<char> *) buffer)->Set(obj, val.Value()[0]);
			break;
		case RC_SMALLINT:
			((AttrBuffer<short> *) buffer)->Set(obj, (short) atoi(val.Value()));
			break;
		case RC_BIGINT:
			((AttrBuffer<_int64> *) buffer)->Set(obj, strtoll(val.Value(),NULL, 10));
			break;
		case RC_BIN:
		case RC_BYTE:
		case RC_VARBYTE:
		case RC_STRING:
		case RC_VARCHAR:
			((AttrBuffer<RCBString> *) buffer)->Set(obj, val);
			break;
		case RC_NUM:
		case RC_YEAR:
		case RC_TIME:
		case RC_DATE:
		case RC_DATETIME:
		case RC_TIMESTAMP:
			val64 = *(_int64 *) (val.Value());
			((AttrBuffer<_int64> *) buffer)->Set(obj, val64);
			break;
		case RC_REAL:
		case RC_FLOAT:
			if(!val.IsNullOrEmpty())
				valD = *(double *) (val.Value());
			else
				valD = NULL_VALUE_D;
			((AttrBuffer<double> *) buffer)->Set(obj, valD);
			break;
	}
}

RCValueObject TempTable::Attr::GetValue(_uint64 obj, bool lookup_to_num )
{
	if(obj == NULL_VALUE_64)
		return RCValueObject();
	RCValueObject ret;
	if(ATI::IsStringType(TypeName()))
		ret = GetValueString(obj);
	else if(ATI::IsIntegerType(TypeName()))
		ret = RCNum(GetValueInt64(obj));
	else if(ATI::IsDateTimeType(TypeName()))
		ret = RCDateTime(this->GetValueInt64(obj), TypeName()/*, precision*/);
	else if(ATI::IsRealType(TypeName()))
		ret = RCNum(this->GetValueInt64(obj), 0, true);
	else if(TypeName() == RC_NUM)
		ret = RCNum((_int64) GetValueInt64(obj), Type().GetScale());
	return ret;
}

RCBString& TempTable::Attr::GetValueString(_uint64 obj)
{
	if(obj == NULL_VALUE_64) {
		bufS = RCBString();
		return bufS;
	}
	_int64 *val64_p = NULL;
	double *d_p = NULL;
	switch(TypeName()) {
		case RC_BIN:
		case RC_BYTE:
		case RC_VARBYTE:
		case RC_STRING:
		case RC_VARCHAR:
			return (*(AttrBuffer<RCBString> *) buffer)[obj];
			break;
		case RC_BYTEINT: {
			RCNum rcn((_int64) (*(AttrBuffer<char> *) buffer)[obj], -1, false, TypeName());
			bufS = rcn.ToRCString();
			break;
		}
		case RC_SMALLINT: {
			RCNum rcn((_int64) (*(AttrBuffer<short>*) buffer)[obj], -1, false, TypeName());
			bufS = rcn.ToRCString();
			break;
		}
		case RC_INT:
		case RC_MEDIUMINT: {
			RCNum rcn((int) (*(AttrBuffer<int>*) buffer)[obj], -1, false, TypeName());
			bufS = rcn.ToRCString();
			break;
		}
		case RC_BIGINT: {
			RCNum rcn((_int64) (*(AttrBuffer<_int64>*) buffer)[obj], -1, false, TypeName());
			bufS = rcn.ToRCString();
			break;
		}
		case RC_NUM: {
			RCNum rcn((*(AttrBuffer<_int64>*) buffer)[obj], Type().GetScale());
			bufS = rcn.ToRCString();
			break;
		}
		case RC_YEAR:
		case RC_TIME:
		case RC_DATE:
		case RC_DATETIME:
		case RC_TIMESTAMP: {
			RCDateTime rcdt((*(AttrBuffer<_int64> *) buffer)[obj], TypeName());
			bufS = rcdt.ToRCString();
			break;
		}
		case RC_REAL:
		case RC_FLOAT: {
			d_p = &(*(AttrBuffer<double> *) buffer)[obj];
			RCNum rcn(*(_int64*) d_p, 0, true, TypeName());
			bufS = rcn.ToRCString();
			break;
		}
	}
	return bufS;
}

_int64 TempTable::Attr::GetSum(int pack)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(ATI::IsNumericType(ct.GetTypeName()));
	_uint64 start = pack * 65536;
	_uint64 stop = (pack + 1) * 65536;
	stop = stop > no_obj ? no_obj : stop;
	_int64 sum = 0, val;
	double sum_d = 0.0;
	bool not_null_exists = false;
	for(_uint64 i = start; i < stop; i++) {
		if(!IsNull(i)) {
			not_null_exists = true;
			if(ATI::IsRealType(ct.GetTypeName())) {
				val = GetValueInt64(i);
				sum_d += *(double *)&val;
			} else
				sum += GetValueInt64(i);
		}
	}
	if(ATI::IsRealType(ct.GetTypeName()))
		sum = *(_int64*)&sum_d;
	return not_null_exists ? sum : NULL_VALUE_64;
}

_int64 TempTable::Attr::GetNoNulls(int pack)
{
	_uint64 start;
	_uint64 stop;
	if(pack == -1) {
		start = 0;
		stop = no_obj;
	} else {
		start = pack * 65536;
		stop = (pack + 1) * 65536;
		stop = stop > no_obj ? no_obj : stop;
	}
	_int64 no_nulls = 0;
	for(_uint64 i = start; i < stop; i++) {
		if(IsNull(i))
			no_nulls++;
	}
	return no_nulls;
}

RCBString TempTable::Attr::GetMinString(int pack)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(ATI::IsStringType(ct.GetTypeName()));
	_uint64 start = pack * 65536;
	_uint64 stop = (pack + 1) * 65536;
	stop = stop > no_obj ? no_obj : stop;
	RCBString* val = NULL;
	RCBString min;
	bool first = true;
	for(_uint64 i = start; i < stop; i++) {
		if(!IsNull(i)) {
			if(first) {
				min = *val;
				first = false;
				continue;
			}
			val = &GetValueString(i);
			if(*val < min)
				min = *val;
		}
	}
	return min;
}

RCBString TempTable::Attr::GetMaxString(int pack)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(ATI::IsStringType(ct.GetTypeName()));
	_uint64 start = pack * 65536;
	_uint64 stop = (pack + 1) * 65536;
	stop = stop > no_obj ? no_obj : stop;
	RCBString* val = NULL;
	RCBString max;
	bool first = true;
	for(_uint64 i = start; i < stop; i++) {
		if(!IsNull(i)) {
			if(first) {
				max = *val;
				first = false;
				continue;
			}
			val = &GetValueString(i);
			if(*val > max)
				max = *val;
		}
	}
	return max;
}

_int64 TempTable::Attr::GetMaxInt64(int pack)
{
	_uint64 start = pack * 65536;
	_uint64 stop = (pack + 1) * 65536;
	stop = stop > no_obj ? no_obj : stop;
	_int64 val, max = BH_BIGINT_MIN;
	for(_uint64 i = start; i < stop; i++) {
		if(!IsNull(i)) {
			val = GetValueInt64(i);
			if( (ATI::IsRealType(ct.GetTypeName()) && *(double *)&val > *(double *)&max) ||
				(!ATI::IsRealType(ct.GetTypeName()) && val > max) )
				max = val;
		}
	}
	return (max == BH_BIGINT_MIN ? NULL_VALUE_64 : max);
}

_int64 TempTable::Attr::GetMinInt64(int pack)
{
	_uint64 start = pack * 65536;
	_uint64 stop = (pack + 1) * 65536;
	stop = stop > no_obj ? no_obj : stop;
	_int64 val, min = BH_BIGINT_MAX;
	for(_uint64 i = start; i < stop; i++) {
		if(!IsNull(i)) {
			val = GetValueInt64(i);
			if( (ATI::IsRealType(ct.GetTypeName()) && *(double *)&val < *(double *)&min) ||
				(!ATI::IsRealType(ct.GetTypeName()) && val < min) )
				min = val;
		}
	}
	return (min == BH_BIGINT_MAX ? NULL_VALUE_64 : min);
}

void TempTable::Attr::GetValueBin(_uint64 obj, int& size, char* buf)
{
	size = 0;
	if(IsNull(obj))
		return;
	switch(TypeName()) {
		case RC_INT:
		case RC_MEDIUMINT:
			size = sizeof(int);
			memcpy(buf, &(*(AttrBuffer<int> *) buffer)[obj], size);
			break;
		case RC_BYTEINT:
			size = sizeof(char);
			memcpy(buf, &(*(AttrBuffer<char> *) buffer)[obj], size);
			break;
		case RC_SMALLINT:
			size = sizeof(short);
			memcpy(buf, &(*(AttrBuffer<short> *) buffer)[obj], size);
			break;
		case RC_STRING:
		case RC_VARCHAR:
		case RC_BIN:
		case RC_BYTE:
		case RC_VARBYTE: {
			RCBString* val_s = &GetValueString(obj);
			size = val_s->size();
			memcpy(buf, val_s->Value(), size);
			break;
		}
		case RC_BIGINT:
		case RC_NUM:
		case RC_YEAR:
		case RC_TIME:
		case RC_DATE:
		case RC_DATETIME:
		case RC_TIMESTAMP:
		case RC_REAL:
		case RC_FLOAT:
			size = sizeof(_int64);
			memcpy(buf, &(*(AttrBuffer<_int64> *) buffer)[obj], size);
			break;
		default:
			assert(0);
	}
}

int TempTable::Attr::GetValueInt32(_uint64 row) const
{
	return (int)GetValueInt64(row);
}

_int64 TempTable::Attr::GetValueInt64(_uint64 obj) const
{
	_int64 res = NULL_VALUE_64;
	if(obj == NULL_VALUE_64)
		return res;
	switch(TypeName()) {
		case RC_INT:
		case RC_MEDIUMINT:
			if(!IsNull(obj))
				res = (*(AttrBuffer<int> *) buffer)[obj];
			break;
		case RC_BYTEINT:
			if(!IsNull(obj))
				res = (*(AttrBuffer<char> *) buffer)[obj];
			break;
		case RC_SMALLINT:
			if(!IsNull(obj))
				res = (*(AttrBuffer<short> *) buffer)[obj];
			break;
		case RC_STRING:
		case RC_VARCHAR:
			assert(0);
			break;
		case RC_BIGINT:
		case RC_NUM:
		case RC_YEAR:
		case RC_TIME:
		case RC_DATE:
		case RC_DATETIME:
		case RC_TIMESTAMP:
			if(!IsNull(obj))
				res = (*(AttrBuffer<_int64> *) buffer)[obj];
			break;
		case RC_REAL:
		case RC_FLOAT:
			if(!IsNull(obj))
				res = *(_int64 *) &(*(AttrBuffer<double> *) buffer)[obj];
			else
				res = *(_int64 *) &NULL_VALUE_D;
			break;
		case RC_BIN:
		case RC_BYTE:
		case RC_VARBYTE:
			assert(0);
			break;
	}
	return res;
}

bool TempTable::Attr::IsNull(const _uint64 obj) const
{
	if(obj == NULL_VALUE_64)
		return true;
	bool res = false;
	switch(TypeName()) {
		case RC_INT:
		case RC_MEDIUMINT:
			res = (*(AttrBuffer<int> *) buffer)[obj] == NULL_VALUE_32;
			break;
		case RC_BYTEINT:
			res = (*(AttrBuffer<char> *) buffer)[obj] == NULL_VALUE_C;
			break;
		case RC_SMALLINT:
			res = (*(AttrBuffer<short> *) buffer)[obj] == NULL_VALUE_SH;
			break;
		case RC_BIN:
		case RC_BYTE:
		case RC_VARBYTE:
		case RC_STRING:
		case RC_VARCHAR:
			res = (*(AttrBuffer<RCBString> *) buffer)[obj].IsNull();
			break;
		case RC_BIGINT:
		case RC_NUM:
		case RC_YEAR:
		case RC_TIME:
		case RC_DATE:
		case RC_DATETIME:
		case RC_TIMESTAMP:
			res = (*(AttrBuffer<_int64> *) buffer)[obj] == NULL_VALUE_64;
			break;
		case RC_REAL:
		case RC_FLOAT:
			res = (*(AttrBuffer<double> *) buffer)[obj] == NULL_VALUE_D;
			break;
	}
	return res;
}

void TempTable::Attr::ApplyFilter(MultiIndex & mind, _uint64 offset, _uint64 no_obj)
{
	assert(mind.NoDimensions() == 1);

	if(mind.NoDimensions() != 1)
		throw NotImplementedRCException("MultiIndex has too many dimensions.");
	if(mind.ZeroTuples() || no_obj == 0) {
		DeleteBuffer();
		return;
	}

	if(no_obj > mind.NoTuples())
		no_obj = mind.NoTuples();

	void *old_buffer = buffer;
	buffer = NULL;
	CreateBuffer(no_obj, mind.m_conn);

	MIIterator mit(&mind);
	for(_uint64 i = 0; i < offset; i++)
		++mit;
	_uint64 idx = 0;
	for(_uint64 i = 0; i < no_obj; i++, ++mit) {
		idx = mit[0];
		assert(idx != NULL_VALUE_64); // null object should never appear in a materialized temp. table
		switch(TypeName()) {
			case RC_INT:
			case RC_MEDIUMINT:
				((AttrBuffer<int> *) buffer)->Set(i, (*(AttrBuffer<int> *) old_buffer)[idx]);
				break;
			case RC_BYTEINT:
				((AttrBuffer<char> *) buffer)->Set(i, (*(AttrBuffer<char> *) old_buffer)[idx]);
				break;
			case RC_SMALLINT:
				((AttrBuffer<short> *) buffer)->Set(i, (*(AttrBuffer<short> *) old_buffer)[idx]);
				break;
			case RC_STRING:
			case RC_VARCHAR:
			case RC_BIN:
			case RC_BYTE:
			case RC_VARBYTE:
				((AttrBuffer<RCBString> *) buffer)->Set(i, (*(AttrBuffer<RCBString> *) old_buffer)[idx]);
				break;
			case RC_BIGINT:
			case RC_NUM:
			case RC_YEAR:
			case RC_TIME:
			case RC_DATE:
			case RC_DATETIME:
			case RC_TIMESTAMP:
				((AttrBuffer<_int64> *) buffer)->Set(i, (*(AttrBuffer<_int64> *) old_buffer)[idx]);
				break;
			case RC_REAL:
			case RC_FLOAT:
				((AttrBuffer<double> *) buffer)->Set(i, (*(AttrBuffer<double> *) old_buffer)[idx]);
				break;
			default:
				assert(0);
		}
	}

	switch(TypeName()) {
		case RC_INT:
		case RC_MEDIUMINT:
			delete (AttrBuffer<int> *) old_buffer;
			break;
		case RC_BYTEINT:
			delete (AttrBuffer<char> *) old_buffer;
			break;
		case RC_SMALLINT:
			delete (AttrBuffer<short> *) old_buffer;
			break;
		case RC_STRING:
		case RC_VARCHAR:
		case RC_BIN:
		case RC_BYTE:
		case RC_VARBYTE:
			delete (AttrBuffer<RCBString> *) old_buffer;
			break;
		case RC_BIGINT:
		case RC_NUM:
		case RC_YEAR:
		case RC_TIME:
		case RC_DATE:
		case RC_DATETIME:
		case RC_TIMESTAMP:
			delete (AttrBuffer<_int64> *) old_buffer;
			break;
		case RC_REAL:
		case RC_FLOAT:
			delete (AttrBuffer<double> *) old_buffer;
			break;
		default:
			assert(0);
	}
}

//void TempTable::Attr::InferComplexType(MysqlExpression::TypOfVars* vartypes)
//{
//	try {
//		ct = term.c_term->EvalType(vartypes);
//	} catch (UnsupportedDataTypeRCException &x) {
//		throw NotImplementedRCException("Couldn't evaluate the type of the result column in complex expression");
//	}
//}

ValueOrNull TempTable::Attr::GetComplexValue(const _uint64 obj)
{
	if(IsNull(obj))
		return ValueOrNull();

	if(ct.IsFixed() || ct.IsFloat() || ct.IsDateTime())
		return ValueOrNull(GetValueInt64(obj));
	if(ct.IsString()) {
		ValueOrNull val;
		RCBString s(GetValueString(obj));
		val.SetString(s);
		return val;
	}
	BHERROR("Unrecognized type in TempTable::Attr::GetComplexValue()");
	return ValueOrNull();
}
void TempTable::Attr::SetComplexValue(_uint64 obj, ValueOrNull val)
{
	if(ct.IsFixed() || ct.IsFloat() || ct.IsDateTime())
		if(val.IsNull())
			SetValueInt64(obj, NULL_VALUE_64 );
		else {
			_int64 v = val.Get64();
			assert(v != NULL_VALUE_64);
			SetValueInt64(obj, v);
		}

	else if(ct.IsString()) {
		RCBString s;
		if(!val.IsNull())
			val.GetString(s);
		SetValueString(obj, s);
	} else
		BHERROR("Unrecognized type in TempTable::Attr::SetComplexValue()");
}

TempTable::TempTable(const TempTable& t, bool subq)
	: filter(t.filter), output_mind(t.output_mind), m_conn(t.m_conn)
{
	no_obj = t.no_obj;
	materialized = t.materialized;
	aliases = t.aliases;
	group_by = t.group_by;
	no_cols = t.no_cols;
	mode = t.mode;
	if(t.tree)
		tree = new DescTree(*t.tree);
	else
		tree = NULL;
	tables = t.tables;
	join_types = t.join_types;
	buf_index = t.buf_index;
	order_by = t.order_by;
	manage_used_tables = t.manage_used_tables;
	has_temp_table = t.has_temp_table;
	virt_cols = t.virt_cols;
	virt_cols_for_having = t.virt_cols_for_having;
	query = t.query;
	for_subq = subq;
	no_global_virt_cols = t.virt_cols.size();
	if(for_subq) {
		filter.table = this;
		for(uint i = 0; i < filter.descriptors.size(); i++)
			filter.descriptors[i].table = this;
		for(uint i = 0; i < filter.parametrized_desc.size(); i++)
			filter.parametrized_desc[i].table = this;
	}
	for(uint i = 0; i< t.attrs.size(); i++) {
		if(subq) {
			assert(t.attrs[i]->buffer == NULL);
			attrs.push_back(t.attrs[i]);
		} else
			attrs.push_back(new Attr(*t.attrs[i]));
	}
}

TempTable::~TempTable()
{
	// remove all temporary tables used by *this
	// but only if *this should manage those tables
	if(manage_used_tables && has_temp_table) {
		for(uint i = 0; i < tables.size(); i++) {
			if(tables[i] && tables[i]->TableType() == TEMP_TABLE)
				delete tables[i];
		}
	}
	for(uint i = 0; i < attrs.size(); i++) {
		if(!for_subq)
			delete attrs[i];
		else
			attrs[i]->DeleteBuffer();
	}
	delete tree;
	if(!for_subq) {
		for(uint i = 0; i < virt_cols.size(); i++)
			delete virt_cols[i];
	} else {
		for(uint i = no_global_virt_cols; i < virt_cols.size(); i++)
			delete virt_cols[i];
	}
}

TempTablePtr TempTable::CreateMaterializedCopy(bool translate_order)			// move all buffers to a newly created TempTable, make this VC-pointing to it; the new table must be deleted by DeleteMaterializedCopy()
{
	// Copy buffers in a safe place for a moment
	vector<void*> copy_buf;
	map<PhysicalColumn *, PhysicalColumn *> attr_translation;	// new attr (this), old attr (working_copy)
	for(uint i = 0; i < attrs.size(); i++) {
		copy_buf.push_back(attrs[i]->buffer);
		attrs[i]->buffer = NULL;
		attrs[i]->mode = LISTING;
	}
	TempTablePtr working_copy = Create(*this, false);		// Original VCs of this will be copied to working_copy, and then deleted in its destructor
	working_copy->mode = TableMode();			// reset all flags (distinct, limit) on copy
	for(uint i = 0; i < attrs.size(); i++) {
		working_copy->attrs[i]->buffer = copy_buf[i];		// copy the data buffers
		attrs[i]->no_obj = 0;
		attr_translation[attrs[i]] = working_copy->attrs[i];
	}
	// VirtualColumns are copied, and we should replace them by references to the temporary source
	delete filter.m_index;
	filter.m_index = new MultiIndex();
	filter.m_index->AddDimension_cross(no_obj);
	virt_cols.clear();
	for(uint i = 0; i < attrs.size(); i++) {
		VirtualColumn *new_vc = new SingleColumn(working_copy->attrs[i], filter.m_index, 0, 0, working_copy.get(), 0);
		virt_cols.push_back(new_vc);
		attrs[i]->term.vc = new_vc;
		attrs[i]->dim = 0;
	}
	if(translate_order) {
		order_by.clear();				// translation needed: virt_cols should point to working_copy as a data source
		for(uint i = 0; i < working_copy->virt_cols.size(); i++) {
			// TODO: redesign it for more universal solution
			if(dynamic_cast<SingleColumn*>(working_copy->virt_cols[i]))
				static_cast<SingleColumn*>(working_copy->virt_cols[i])->TranslateSourceColumns(attr_translation);
			else if(working_copy->virt_cols[i])/*if(dynamic_cast<ExpressionColumn*>(working_copy->virt_cols[i]))*/ {
				//ExpressionColumn *evc = static_cast<ExpressionColumn*>(working_copy->virt_cols[i]);
				VirtualColumn::var_maps_t& var_maps = working_copy->virt_cols[i]->GetVarMap();
				for(uint i = 0; i < var_maps.size(); i++) {
					if(var_maps[i].tab.lock().get() == this)
						var_maps[i].tab = working_copy;
				}
			} /*else if(dynamic_cast<ConstColumn*>(working_copy->virt_cols[i])) {
				// do nothing
			} else {
				BHASSERT(false, "ORDER BY too complex");
			}*/
		}
		for(uint i = 0; i < working_copy->order_by.size(); i++) {		
			SortDescriptor sord;
			sord.vc = working_copy->order_by[i].vc;
			sord.dir = working_copy->order_by[i].dir;
			order_by.push_back(sord);
		}
	} else {	// Create() did moved vc's used in order_by into working_copy, so we should bring them back
		for(uint i = 0; i < order_by.size(); i++) {
			// find which working_copy->vc is used in ordering
			uint j = 0;
			for(j = 0; j < working_copy->virt_cols.size(); j++) {
				if(working_copy->order_by[i].vc == working_copy->virt_cols[j])
					break;
			}
			if(j < working_copy->virt_cols.size()) {
				virt_cols.push_back(working_copy->virt_cols[j]);
				order_by[i].vc = working_copy->virt_cols[j];
				working_copy->virt_cols[j] = NULL;				// moving vc - now it is back in this
			}
		}
	}
	return working_copy;	// must be deleted by DeleteMaterializedCopy()
}

void TempTable::DeleteMaterializedCopy(TempTablePtr old_t)		// delete the external table and remove VC pointers, make this fully materialized
{
	for(uint i = 0; i < attrs.size(); i++) {		// Make sure VCs are deleted before the source table is deleted
		delete virt_cols[i];
		virt_cols[i] = NULL;
	}
	old_t.reset();
}

////////////////////////////////////////////////////////////////////////////////////////////////////

TempTable::Attr * TempTable::GetDisplayableAttrP(uint attr)
{
	int idx = -1;
	uint i;
	for(i = 0; i < attrs.size(); i++)
		if(attrs[i]->alias) {
			idx++;
			if(idx == attr)
				break;
		}
	assert(i < attrs.size());
	return attrs[i];
}

uint TempTable::GetDisplayableAttrIndex(uint attr)
{
	int idx = -1;
	uint i;
	for(i = 0; i < attrs.size(); i++)
		if(attrs[i]->alias) {
			idx++;
			if(idx == attr)
				break;
		}
	assert(i < attrs.size());
	return i;
}

void TempTable::Filter(ParameterizedFilter *f)
{
	if(!f->tree) {// on result = false
		if(!f->m_index->NoDimensions())
			return;
		if(f->m_index->NoDimensions() != filter.m_index->NoDimensions())
			throw NotImplementedRCException("TempTable::Filter - not compatible MultiIndex applied.");
		filter.m_index->Canibalize(*f->m_index);
		filter.descriptors = f->descriptors;
		filter.parametrized_desc = f->parametrized_desc;
		filter.parameters = f->parameters;
		filter.aliases = f->aliases;
		filter.table = f->table;

		for(size_t i = 0; i < filter.filters.size(); i++)
			delete filter.filters[i];
		filter.filters.clear();

		for(size_t i = 0; i < f->filters.size(); i++)
			filter.filters.push_back(new ParameterizedFilter(*f->filters[i]));

		filter.log_connectors = f->log_connectors;
	} else { // on result = true
		delete tree;
		tree = new DescTree(*f->tree);
	}
}

void TempTable::SetMode(TMParameter mode, _int64 mode_param1, _int64 mode_param2)
{
	switch(mode) {
		case TM_DISTINCT:
			this->mode.distinct = true;
			break;
		case TM_TOP:
			this->mode.top = true;
			this->mode.param1 = mode_param1; // offset
			this->mode.param2 = mode_param2; // limit
			break;
		case TM_EXISTS:
			this->mode.exists = true;
			break;
		default:
			assert(false);
	}

}
bool TempTable::GetMode(TMParameter mode)
{
	switch(mode) {
		case TM_DISTINCT:
			return this->mode.distinct;
		case TM_TOP:
			return this->mode.top;
		case TM_EXISTS:
			return this->mode.exists;
		default:
			return false;
	}
}

int TempTable::AddColumn(CQTerm e, ColOperation mode, char * alias, bool distinct)
{

	if(alias)
		no_cols++;
	AttributeType type; // type of column
	NullMode is_nullable = NO_NULLS; // true = NULLs are allowed
	int dim = -2; // -2 is non-allowed value of dimension
	uint scale = 0; // number of decimal places
	uint precision = 0; // number of all places

	// new code for VirtualColumn
	if (e.vc_id != NULL_VALUE_32) {
		assert(e.vc);
		// enum ColOperation {LISTING, COUNT, SUM, MIN, MAX, AVG, GROUP_BY};
		if(mode == COUNT || mode == BIT_AND || mode == BIT_OR || mode == BIT_XOR) {
			type = RC_NUM; // 64 bit, Decimal(18,0)
			precision = 18;
			is_nullable = NO_NULLS;
		} else if(mode == AVG || mode == VAR_POP || mode == VAR_SAMP || mode == STD_POP || mode == STD_SAMP) {
			type = RC_REAL; // 64 bit
			precision = 18;
			is_nullable = AS_MISSED;
		} else {
			scale = e.vc->Type().GetScale();
			precision = e.vc->Type().GetPrecision();
			type = e.vc->TypeName();
			is_nullable = e.vc->Type().GetNullsMode();
			if(mode == SUM) {
				switch(e.vc->TypeName()) {
					case RC_INT:
					case RC_NUM:
					case RC_SMALLINT:
					case RC_BYTEINT:
					case RC_MEDIUMINT:
						type = RC_NUM;
						precision = 18;
						is_nullable = AS_MISSED;
						break;
					case RC_BIGINT:
						type = RC_BIGINT;
						precision = 19;
						is_nullable = AS_MISSED;
						break;
					case RC_REAL:
					case RC_FLOAT:
					default:			// all other types summed up as double (parsed from text)
						type = RC_REAL;
						is_nullable = AS_MISSED;
						break;
//					default:
//						throw NotImplementedRCException("SUM performed on non-numerical column.");
				}
			} else if(mode == MIN || mode == MAX) {
				is_nullable = AS_MISSED;
			} else if(mode == LISTING || mode == GROUP_BY) {
				if(mode == GROUP_BY)
					group_by = true;
				is_nullable =  e.vc->Type().GetNullsMode();
			}
		}
	} else if(mode == COUNT) {
		type = RC_NUM;
		precision = 18;
		is_nullable = NO_NULLS;
	} else {
		// illegal execution path: neither VC is set nor mode is COUNT
		BHASSERT_WITH_NO_PERFORMANCE_IMPACT(!"wrong execution path");
		//throw NotImplementedRCException("Invalid column on SELECT list.");
	}
	attrs.push_back(new Attr(e, mode, distinct, alias, -2, type, scale, precision, is_nullable, e.vc ? e.vc->GetCollation() : DTCollation()));
	return -(int) attrs.size(); // attributes are counted from -1, -2, ...
}

void TempTable::AddOrder(VirtualColumn *vc, int direction)
{
	// don't sort by const expressions
	if(vc->IsConst())
		return;
	SortDescriptor d;
	d.vc = vc;
	d.dir = direction;
	bool already_added = false;
	for(uint j = 0; j < order_by.size(); j++) {
		if(order_by[j] == d) {
			already_added = true;
			break;
		}
	}
	if(!already_added)
		order_by.push_back(d);
}

void TempTable::Union(TempTable* t, int all)
{
	assert( NoDisplaybleAttrs() == t->NoDisplaybleAttrs() );
	if(NoDisplaybleAttrs() != t->NoDisplaybleAttrs())
		throw NotImplementedRCException("UNION of tables with different number of columns.");
	if(this->IsParametrized() || t->IsParametrized())
		throw NotImplementedRCException("Materialize: not implemented union of parameterized queries.");
	this->Materialize();
	t->Materialize();
	if((!t->NoObj() && all) || (!this->NoObj() && !t->NoObj())) // no objects = no union
		return;

	::Filter first_f(NoObj()), first_mask(NoObj()); // mask of objects to be added to the final result set
	::Filter sec_f(t->NoObj()), sec_mask(t->NoObj());
	first_mask.Set();
	sec_mask.Set();
	if(!all && NoObj() && t->NoObj()) {
		::Filter first_f(NoObj());
		first_f.Set();
		::Filter sec_f(t->NoObj());
		sec_f.Set();
		GroupDistinctTable dist_table;
		vector<ColumnBinEncoder> encoder;
		typedef shared_ptr<VirtualColumn> vc_ptr_t;
		typedef vector<vc_ptr_t> vcs_t;
		vcs_t first_vcs;
		vcs_t sec_vcs;
		uchar* input_buf = NULL;
		_int64 size = 0;
		for(uint i = 0; i < NoDisplaybleAttrs(); i++) {
			first_vcs.push_back(vc_ptr_t( new SingleColumn(GetDisplayableAttrP(i), &output_mind, 0, -i - 1, this, 0) ));
			sec_vcs.push_back(vc_ptr_t( new SingleColumn(t->GetDisplayableAttrP(i), t->GetOutputMultiIndexP(), 1, -i - 1, t, 0) ));
			encoder.push_back(ColumnBinEncoder());
			if(encoder[i].PrepareEncoder(first_vcs[i].get(), sec_vcs[i].get()) == false) {
				stringstream ss;
				ss << "UNION of non-matching columns (column no " << i<< ") .";
				throw NotImplementedRCException(ss.str());
			}
			encoder[i].SetPrimaryOffset(size);
			size +=  encoder[i].GetPrimarySize();
		}
		input_buf = new uchar[size];
		dist_table.InitializeB(size, NoObj());
		MIIterator first_mit(&output_mind);
		MIIterator sec_mit(t->GetOutputMultiIndexP());
		// check all objects from the first table
		do {
			first_mit.Rewind();
			dist_table.Clear();
			while(first_mit.IsValid()) {
				if(first_f.Get(**first_mit)) {
					for(int i = 0; i < encoder.size(); i++)
						encoder[i].Encode(input_buf, first_mit);
					GDTResult res = dist_table.Add(input_buf);
					if(res == GDT_EXISTS)
						first_mask.Reset(**first_mit);
					if(res != GDT_FULL)
						first_f.Reset(**first_mit);
				}
				++first_mit;
				if(m_conn.killed())	
					throw KilledRCException();
			}
			sec_mit.Rewind();
			while(sec_mit.IsValid()) {
				if(sec_f.Get(**sec_mit)) {
					for(int i = 0; i < encoder.size(); i++)
						encoder[i].Encode(input_buf, sec_mit, sec_vcs[i].get());
					GDTResult res = dist_table.Add(input_buf);	
					if(res == GDT_EXISTS)
						sec_mask.Reset(**sec_mit);
					if(res != GDT_FULL)
						sec_f.Reset(**sec_mit);
				}
				++sec_mit;
				if(m_conn.killed())
					throw KilledRCException();
			}
		} while(!first_f.IsEmpty() && !sec_f.IsEmpty());
		delete [] input_buf;
	}
	_int64 new_no_obj = first_mask.NoOnes() + sec_mask.NoOnes();
	int new_page_size = CalculatePageSize(new_no_obj);
	for(uint i = 0; i < NoDisplaybleAttrs(); i++) {
		//if(!t->NoObj())
		//	continue;
		Attr* f = GetDisplayableAttrP(i);
		Attr* s = t->GetDisplayableAttrP(i);
		ColumnType new_type = GetUnionType(f->Type(), s->Type());
		Attr* new_attr = new Attr(CQTerm(), LISTING, false, f->alias, 0, new_type.GetTypeName(), new_type.GetScale(), new_type.GetPrecision(), new_type.GetNullsMode(), new_type.GetCollation());
		new_attr->page_size = new_page_size;
		new_attr->CreateBuffer(new_no_obj, &m_conn);
		if(f->TypeName() == RC_NUM && s->TypeName() == RC_NUM && f->Type().GetScale() != s->Type().GetScale()) {
			uint max_scale = new_attr->Type().GetScale(); 
			// copy attr from first table to new_attr
			double multiplier = PowOfTen(max_scale - f->Type().GetScale());
			first_mask.Rewind();
			for(_uint64 j = 0; j < first_mask.NoOnes(); j++) {
				_int64 pos = first_mask.GetNext1();
				if(!f->IsNull(pos))
					new_attr->SetValueInt64(j, f->GetValueInt64(pos) * (_int64) multiplier);
				else
					new_attr->SetValueInt64(j, NULL_VALUE_64);
			}
			// copy attr from second table to new_attr 
			multiplier = PowOfTen(max_scale - s->Type().GetScale());
			sec_mask.Rewind();
			for(_uint64 j = 0; j < sec_mask.NoOnes(); j++) {
				_int64 pos = sec_mask.GetNext1();
				if(!s->IsNull(pos))
					new_attr->SetValueInt64(first_mask.NoOnes() + j, s->GetValueInt64(pos) * (_int64) multiplier);
				else
					new_attr->SetValueInt64(first_mask.NoOnes() + j, NULL_VALUE_64 );
			}
		} else if(ATI::IsStringType(new_attr->TypeName())) {
			first_mask.Rewind();
			for(_uint64 j = 0; j < first_mask.NoOnes(); j++)
				new_attr->SetValueString(j, f->GetValueString(first_mask.GetNext1()));
			sec_mask.Rewind();
			for(_uint64 j = 0; j < sec_mask.NoOnes(); j++)
				new_attr->SetValueString(first_mask.NoOnes() + j, s->GetValueString(sec_mask.GetNext1()));
		} else if(new_attr->Type().IsFloat()) {
			first_mask.Rewind();
			for(_uint64 j = 0; j < first_mask.NoOnes(); j++) {
				_int64 pos = first_mask.GetNext1();
				if(f->IsNull(pos))
					new_attr->SetValueInt64(j, NULL_VALUE_64);
				else if(f->Type().IsFloat())
					new_attr->SetValueInt64(j, f->GetValueInt64(pos));
				else {
					double v = (double)f->GetValueInt64(pos) / PowOfTen(f->Type().GetScale());
					new_attr->SetValueInt64(j, *(_int64*)&v);
				}
			}
			sec_mask.Rewind();
			for(_uint64 j = 0; j < sec_mask.NoOnes(); j++) {
				_int64 pos = sec_mask.GetNext1();
				if(s->IsNull(pos))
					new_attr->SetValueInt64(first_mask.NoOnes() + j, NULL_VALUE_64);
				else if(s->Type().IsFloat())
					new_attr->SetValueInt64(first_mask.NoOnes() + j, s->GetValueInt64(pos));
				else {
					double v = (double)s->GetValueInt64(pos) / PowOfTen(s->Type().GetScale());
					new_attr->SetValueInt64(first_mask.NoOnes() + j, *(_int64*)&v);
				}
			}
		} else {
			double multiplier = PowOfTen(new_attr->Type().GetScale() - f->Type().GetScale());
			first_mask.Rewind();
			for(_uint64 j = 0; j < first_mask.NoOnes(); j++) {
				_int64 pos = first_mask.GetNext1();
				if(f->IsNull(pos))
					new_attr->SetValueInt64(j, NULL_VALUE_64);
				else
					new_attr->SetValueInt64(j, f->GetValueInt64(pos) * multiplier);
			}
			multiplier = PowOfTen(new_attr->Type().GetScale() - s->Type().GetScale());
			sec_mask.Rewind();
			for(_uint64 j = 0; j < sec_mask.NoOnes(); j++) {
				_int64 pos = sec_mask.GetNext1();
				if(s->IsNull(pos))
					new_attr->SetValueInt64(first_mask.NoOnes() + j, NULL_VALUE_64);
				else
					new_attr->SetValueInt64(first_mask.NoOnes() + j, s->GetValueInt64(pos) * multiplier);
			}
		}
		attrs[GetDisplayableAttrIndex(i)] = new_attr;
		delete f;
	}
	this->no_obj = new_no_obj;
	//this->Display();
	output_mind = MultiIndex();
	output_mind.AddDimension_cross(no_obj);
}

void TempTable::Display(ostream & out)
{
	out << "No obj.:" << no_obj << ", No attrs.:" << this->NoDisplaybleAttrs() << endl << "-----------" << endl;
	for(_uint64 i = 0; i < no_obj; i++) {
		for(uint j = 0; j < this->NoDisplaybleAttrs(); j++) {
			if(!attrs[j]->alias)
				continue;
			if(!IsNull(i, j))
				out << *GetTable_pS(i, j) << " ";
			else
				out << "NULL" << " ";
		}
		out << endl;
	}
	out << "-----------" << endl;
}

int TempTable::GetDimension(int alias)
{
	for(uint i = 0; i < aliases.size(); i++) {
		if(aliases[i] == alias)
			return i;
	}
	return NULL_VALUE_32;
}

class bufOfVarsWrapper
{
private:
	typedef std::map<int, MysqlExpression::BufOfVars> mapOfBufOfVars;
	std::map<int, bool> isBuffersSet;
	mapOfBufOfVars buforsForComplex;
public:
	void registerColumn(int complexColumn, int dependantColumn)
	{
		buforsForComplex[complexColumn][VarID(0, dependantColumn)] = new ValueOrNull;
		isBuffersSet[complexColumn] = false;
	}
	ValueOrNull *giveBuffer(int complexColumn, int dependantColumn,
			MysqlExpression &complexColumnTerm)
	{
		if(!isBuffersSet[complexColumn]) {
//			complexColumnTerm.SetBufs(&(buforsForComplex[complexColumn]));
			isBuffersSet[complexColumn] = true;
		}
		return buforsForComplex[complexColumn][VarID(0, dependantColumn)];
	}
	~bufOfVarsWrapper()
	{
		for(mapOfBufOfVars::iterator it = buforsForComplex.begin(); it != buforsForComplex.end(); it++)
			for(MysqlExpression::BufOfVars::iterator it2 = it->second.begin(); it2 != it->second.end(); it2++)
				delete it2->second;
	}

};

void TempTable::MaterializeDelayedAttrs()
{
	MultiIndex mind;
	mind.AddDimension_cross(no_obj);
	MIIterator it(&mind);
	RCBString vals;
	for(uint i = 0; i < attrs.size(); i++) {
		if(attrs[i]->mode != DELAYED)
			continue;
		Attr* attr = attrs[i];
		VirtualColumn *vc = attr->term.vc;
		_int64 row = 0;
		it.Rewind();
		while (it.IsValid()) { // go thru all rows
			if(it.PackrowStarted()) {
				vc->LockSourcePacks(it);
			}
			switch(attr->TypeName()) {
				case RC_STRING:
				case RC_VARCHAR:
					vc->GetValueString(vals, it, false);
					attr->SetValueString(row, vals);
					break;
				case RC_BIN:
				case RC_BYTE:
				case RC_VARBYTE:
					if( !vc->IsNull(it) )
						vc->GetValueString(vals, it, false);
					else
						vals = RCBString();
					attr->SetValueString(row, vals);
					break;
				default:
					attr->SetValueInt64(row, vc->GetValueInt64(it));
			}
			++it;
			++row;
		}
	}
}

void TempTable::Materialize(bool in_subq)
{
	MEASURE_FET("TempTable::Materialize()");
	int page_size = CalculatePageSize();
	_int64 offset = 0; // controls the first object to be materialized
	_int64 limit = -1;	// if(limit>=0)  ->  for(row = offset; row < offset + limit; row++) ....
	bool limits_present = mode.top;

	if(limits_present) {
		offset = (mode.param1 >= 0 ? mode.param1 : 0);
		limit = (mode.param2 >= 0 ? mode.param2 : 0);
		// applied => reset to default values
		mode.top = false;
		mode.param2 = -1;
		mode.param1 = -1;
	}
	_int64 local_offset = 0; // controls the first object to be materialized in a given algorithm
	_int64 local_limit = -1;

	if(materialized && (order_by.size() > 0 || limits_present) && no_obj) {
		// if TempTable is already materialized but some additional constraints were specified, e.g., order by or limit
		// this is typical case for union, where constraints are specified for the result of union after materialization
		if(limits_present) {
			local_offset = offset;
			local_limit = min(limit, (_int64)no_obj - offset);
		} else
			local_limit = no_obj;

		if(order_by.size() != 0 && no_obj > 1) {
			TempTablePtr temporary_source_table = CreateMaterializedCopy(true);		// true: translate definition of ordering
			OrderByAndMaterialize(order_by, local_limit, local_offset);
			DeleteMaterializedCopy(temporary_source_table);
		}
		else if(limits_present)
			ApplyOffset(local_limit, local_offset);
		return;
	} 
	if(materialized)
		return;

	bool table_distinct = this->mode.distinct;
	bool distinct_on_materialized = false;
	for(uint i = 0; i < NoAttrs(); i++)
		if(attrs[i]->mode != LISTING)
			group_by = true;
	if(table_distinct && group_by) {
		distinct_on_materialized = true;
		table_distinct = false;
	}

	bool no_rows_too_large = filter.m_index->TooManyTuples();
	no_obj = -1; // no_obj not calculated yet - wait for better moment
	VerifyAttrsSizes(); // resize attr[i] buffers basing on the current multiindex state

	// the case when there is no grouping of attributes, check also DISTINCT modifier of TT
	if(!group_by && !table_distinct) {
		if(limits_present && !distinct_on_materialized) {
			if(no_rows_too_large && order_by.size() == 0)
				no_obj = offset + limit; // offset + limit in the worst case
			else
				no_obj = filter.m_index->NoTuples();
			if(no_obj <= offset) {
				no_obj = 0;
				materialized = true;
				return;
			}
			local_offset = offset;
			local_limit = min(limit, (_int64)no_obj - offset);
		} else {
			no_obj = filter.m_index->NoTuples();
			local_limit = no_obj;
		}
		page_size = CalculatePageSize();		// recalculate, as no_obj might changed
		// fill buffers with values = compute result table (ordered or not)

		// perform order by: in this case it can be done on source tables, not on the result
		bool materialized_by_ordering = this->OrderByAndMaterialize(order_by, local_limit, local_offset);	// false if no sorting used

		if(!materialized_by_ordering) 		// not materialized yet?
			FillMaterializedBuffers(local_limit, local_offset);		// materialize without aggregations/ordering

	} else {
		// GROUP BY or DISTINCT -  compute aggregations
		if(limits_present && tree == NULL && !distinct_on_materialized && order_by.size() == 0) {
			local_offset = offset;
			local_limit = limit;
		}

		no_obj = Aggregate(table_distinct, local_limit, local_offset);
		materialized = true;
		// compute all DELAYED columns, expressions over aggregations
		if(HasDelayedAttrs())
			MaterializeDelayedAttrs();

		// apply conditions defined by HAVING clause (conditions are stored in this->tree)
		if(tree && no_obj && (! IsParametrized(tree->root) || in_subq)) {
			output_mind.Clear();
			output_mind.AddDimension_cross(no_obj);
			if(in_subq)
				tree->root->Simplify();
			ApplyDescTree(output_mind, tree->root);
			for(uint i = 0; i < attrs.size(); i++)			// all attrs, not only visible, because of further operations (e.g. sort)
				attrs[i]->ApplyFilter(output_mind, 0, output_mind.NoTuples());
			no_obj = output_mind.NoTuples();
			if(!distinct_on_materialized && order_by.size() == 0 && limits_present) {
				local_offset = offset;
				local_limit = min(limit, (_int64)no_obj - offset);
				ApplyOffset(local_limit, local_offset);
			}
		}
		if(no_obj == 0)
			return;
	}

	output_mind.Clear();
	output_mind.AddDimension_cross(no_obj);	// an artificial dimension for result
	local_offset = 0;
	local_limit = -1;

	// DISTINCT after grouping
	if(distinct_on_materialized && no_obj > 1) {
		if(limits_present && order_by.size() == 0) {
			local_offset = offset;
			local_limit = min(limit, (_int64)no_obj - offset);
		} else
			local_limit = no_obj;
		TempTablePtr temporary_source_table = CreateMaterializedCopy(false);
		no_obj = Aggregate(true, local_limit, local_offset);	// true => select-level distinct
		DeleteMaterializedCopy(temporary_source_table);
	} // end of distinct part

	// ORDER BY, if not sorted until now
	if(order_by.size() != 0 && no_obj > 1) {
		if(limits_present) {
			local_offset = offset;
			local_limit = min(limit, (_int64)no_obj - offset);
		} else
			local_limit = no_obj;
		TempTablePtr temporary_source_table = CreateMaterializedCopy(true);		// true: translate definition of ordering
		OrderByAndMaterialize(order_by, local_limit, local_offset);
		DeleteMaterializedCopy(temporary_source_table);
	}
	materialized = true;
}

_int64 TempTable::GetTable64(_uint64 obj, int attr)
{
	if(!materialized || no_obj == 0)
		return NULL_VALUE_64;
	assert(obj < no_obj && (uint)attr < attrs.size());
	return attrs[attr]->GetValueInt64(obj);
}

int TempTable::GetTable32(_uint64 obj, int _attr)
{
	uint attr = (uint) _attr;
	if(!materialized)
		return NULL_VALUE_32;
	assert(obj < no_obj && attr < attrs.size());
	_int64 res = GetTable64(obj, attr);
	return (res == NULL_VALUE_64 ? ATI::NULL32(AttrType(attr)) : (int) res);
}

RCBString *TempTable::GetTable_pS(_uint64 obj, uint attr)
{
	assert(materialized && obj < no_obj && attr < attrs.size());
	return &attrs[attr]->GetValueString(obj);
}

RCBString TempTable::GetTable_S(_uint64 obj, int _attr, bool bin_as_hex)
{
	if(!materialized || no_obj == 0)
		return RCBString();
	uint attr = (uint) _attr;
	assert(obj < no_obj && attr < attrs.size());
	return attrs[attr]->GetValueString(obj);
}

void TempTable::GetTable_B(_uint64 obj, int _attr, int & size, char *val_buf)
{
	throw NotImplementedRCException("TempTable::GetTable_B.");
	assert(0);
	// TODO: get table B
	return;
}

bool TempTable::IsNull(_uint64 obj, int attr)
{
	if (no_obj == 0)
		return true;
	assert(obj < no_obj && (uint)attr < attrs.size());
	return attrs[attr]->IsNull(obj);
}

_uint64 TempTable::Attr::ApproxDistinctVals(bool incl_nulls, ::Filter *f) // provide the best upper approximation of number of diff. values (incl. null)
{
	// TODO: can it be done better?
	if(f)
		return f->NoOnes();
	return no_obj;
}

ushort TempTable::Attr::MaxStringSize(::Filter *f)			// maximal byte string length in column
{
	return ct.GetPrecision();
}

_uint64 TempTable::ApproxAnswerSize(int attr, Descriptor &d) // provide the most probable approximation of number of objects matching the condition
{
	// TODO: can it be done better?
	return no_obj / 2;
}

RCValueObject TempTable::GetTable(_uint64 obj, uint attr)
{
	if (no_obj == 0 )
		return RCValueObject();
	assert(obj < no_obj && (uint)attr < attrs.size());
	return attrs[attr]->GetValue(obj);
}

bool TempTable::IsAttrNumeric(uint a)
{
	assert(a < no_cols);
	return attrs[a]->Type().IsNumeric();
}

int TempTable::CalculatePageSize(_int64 _no_obj)
{
	_int64 new_no_obj = _no_obj == -1 ? no_obj : _no_obj;
	int size_of_one_record = 0;
	for(uint i = 0; i < attrs.size(); i++)
		if(attrs[i]->TypeName() == RC_BIN || attrs[i]->TypeName() == RC_BYTE || attrs[i]->TypeName() == RC_VARBYTE
				|| attrs[i]->TypeName() == RC_STRING || attrs[i]->TypeName() == RC_VARCHAR)
			size_of_one_record += attrs[i]->Type().GetInternalSize() + 4; // 4 bytes describing length
		else
			size_of_one_record += attrs[i]->Type().GetInternalSize();
	int raw_size = (int)new_no_obj;
	if(size_of_one_record < 1)
		size_of_one_record = 1;
	if(CACHE_SIZE / size_of_one_record <= new_no_obj) {
		raw_size = CACHE_SIZE / size_of_one_record;
		raw_size = 1 << (CalculateBinSize((uint)raw_size)); // round up to 2^k
	}
	if(raw_size < 1)
		raw_size = 1;
	for(uint i = 0; i < attrs.size(); i++)
		attrs[i]->page_size = raw_size;
	return raw_size;
}

void OrderedList::Insert(_int64 key) // inserts key to a list, if exists increases count
{
	if(size == 0) {
		head = tail = curr = new OrderedListNode(key);
		size = 1;
	} else if(key < head->key) {
		curr = new OrderedListNode(key);
		curr->next = head;
		head->prev = curr;
		head = curr;
		size++;
	} else if(key > tail->key) {
		curr = new OrderedListNode(key);
		curr->prev = tail;
		tail->next = curr;
		tail = curr;
		size++;
	} else {
		for(curr = head; curr; curr = curr->next) {
			if(curr->key == key) {
				curr->count++;
				break;
			} else if(curr->key > key) { //insert before
				OrderedListNode *temp = new OrderedListNode(key);
				temp->next = curr;
				temp->prev = curr->prev;
				curr->prev = temp;
				temp->prev->next = temp;
				size++;
			}
		}
	}
}

OrderedList::~OrderedList()
{
	OrderedListNode *temp;
	if(size)
		for(curr = head; curr;) {
			temp = curr->next;
			delete curr;
			curr = temp;
		}
}

DescTree::DescTree(CQTerm e1, Operator op, CQTerm e2, CQTerm e3, TempTable* t)
{
	curr = root = new DescTreeNode(e1, op, e2, e3, t);
}

DescTree::DescTree(DescTree &t)
{
	curr = root = Copy(t.root);
}

DescTreeNode * DescTree::Copy(DescTreeNode * node)
{
	if(!node)
		return NULL;
	DescTreeNode *res = new DescTreeNode(*node);
	if(node->left) {
		res->left = Copy(node->left);
		res->left->parent = res;
	}
	if(node->right) {
		res->right = Copy(node->right);
		res->right->parent = res;
	}
	return res;
}

void DescTree::Release(DescTreeNode * &node)
{
	if(node && node->left)
		Release(node->left);
	if(node && node->right)
		Release(node->right);
	delete node;
	node = NULL;
}

void DescTree::Release()
{
	Release(root);
	curr = root;
}

// make _lop the root, make current tree the left child, make descriptor the right child
void DescTree::AddDescriptor(LogicalOperator lop, CQTerm e1, Operator op, CQTerm e2, CQTerm e3, TempTable* t)
{
	curr = new DescTreeNode(lop,t);
	curr->left = root;
	curr->left->parent = curr;
	curr->right = new DescTreeNode(e1, op, e2, e3, t);
	curr->right->parent = curr;
	root = curr;
}

// make _lop the root, make current tree the left child, make tree the right child
void DescTree::AddTree(LogicalOperator lop, DescTree *tree)
{
	if (tree->root) {
		curr = new DescTreeNode(lop, tree->root->desc.table);
		curr->left = root;
		root->parent = curr;
		curr->right = tree->root;
		tree->root->parent = curr;
		root = curr;
		tree->root = NULL;
		tree->curr = NULL;
	}
}

void DescTree::Display()
{
	Display(root);
}

void DescTree::Display(DescTreeNode * node)
{
	if(node && node->left)
		Display(node->left);
	if(node && node->right)
		Display(node->right);
	cout << "------------------------" << endl;
	if(node && node->left)
		cout << (node->desc.lop ? "OR" : "AND") << endl;
	else {
		char buf[50];
		cout << node->desc.attr.ToString(buf, 0) << endl;
		cout << "........" << endl;
		if(node->desc.op)
			cout << node->desc.op << endl;
		else
			cout << "=" << endl;
		cout << "........" << endl;
		cout << node->desc.val1.ToString(buf, 0) << endl;
	}
	cout << "------------------------" << endl;
}


void DescTreeNode::Simplify()
{
	if(left)
		left->Simplify();
	if(right)
		right->Simplify();
	desc.Simplify();
}

bool DescTreeNode::CheckCondition(MIIterator& mit)
{
	if(left) { // i.e., not a leaf
		assert(right); // if left is not empty so should be right
		if(desc.lop == O_AND) {
			if(!left->CheckCondition(mit))
				return false;
			if(!right->CheckCondition(mit))
				return false;
			return true;
		} else {
			if(left->CheckCondition(mit))
				return true;
			if(right->CheckCondition(mit))
				return true;
			return false;
		}
	} else { // i.e., a leaf
		desc.LockSourcePacks(mit);
		if(RequiresUTFConversions(desc.GetCollation())) {
			desc.InitUTFBuffers();
			return desc.CheckCondition_UTF(mit);
		} else
			return desc.CheckCondition(mit);
	}
}

/*! Filter out rows from the given multiindex according to the tree of descriptors
 *
 */
void TempTable::ApplyDescTree(MultiIndex& mind, DescTreeNode* tree)
{
	MIUpdatingIterator mit(&mind);
	while(mit.IsValid()){
		if(!tree->CheckCondition(mit))
			mit.ResetCurrent();
		++mit;
	}
	mit.Commit();
	mind.UpdateNoTuples();
}

bool TempTable::uses(TempTable * what)
{
	for(uint i = 0; i < tables.size(); i++)
		if(tables[i] == what)
			return true;
		else if(tables[i]->TableType() == TEMP_TABLE && ((TempTable *) tables[i])->uses(what))
			return true;
	return uses(&filter, what);
}

bool TempTable::uses(ParameterizedFilter *f, TempTable * what)
{
	// TODO: implement for VCs!
	//Descriptor * curr;
	//for(uint i = 0; i < f->parametrized_desc.size(); i++) {
	//	curr = &f->parametrized_desc[i];
	//	if(curr->attr.tab_p == what)
	//		return true;
	//	if(curr->val1.tab_p == what)
	//		return true;
	//	if(curr->val2.tab_p == what)
	//		return true;
	//}
	//for(uint i = 0; i < f->filters.size(); i++)
	//	if(uses(f->filters[i], what))
	//		return true;
	return false;
}

void TempTable::DisplayRSI()
{
	for(uint i = 0; i < tables.size(); i++) {
		if(tables[i]->TableType() == RC_TABLE)
			((RCTable*) tables[i])->DisplayRSI();
	}
}

////////////////////////////////////////////////////
// Refactoring: extracted methods

void TempTable::UnlockLastPacks(unsigned* last_pack)
{
	//for(unsigned j = 0; j < NoAttrs(); j++) {
	//	if(attrs[j]->dim >= 0) {
	//		JustATable* curr_table = tables[attrs[j]->dim];
	//		if(curr_table && curr_table->TableType() == RC_TABLE && last_pack[j] != 0xffffffff)
	//			curr_table->UnlockPackFromUse(attrs[j]->term.attr_id.n, last_pack[j]);
	//	}
	//}
	filter.m_index->UnlockAllFromUse();
}

void TempTable::RemoveFromManagedList(const RCTable* tab)
{
	tables.erase(std::remove(tables.begin(), tables.end(), tab), tables.end());
}

template<>
JustATable::RecordIteratorImplementation<JustATable::TempTableLockPolicy>::RecordIteratorImplementation(
		JustATable::TempTableLockPolicy::initializer_t const& init, _uint64 pos) :
	_data(JustATable::TempTableLockPolicy::data_t(init.table)), _currentRNo(pos), _conn(init.conn)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT( _data.table != 0 );
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT( _currentRNo <= _data.table->NoObj() );
}

template<>
RCValueObject JustATable::RecordImplementation<JustATable::RecordIteratorImplementation<JustATable::TempTableLockPolicy> >::do_get_item(
		uint i)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT( _it._data.table );
	bool buf_used = true;
	AttributeType attrt_tmp = _it._data.table->GetDisplayableAttrP(i)->TypeName();
	if(attrt_tmp == RC_INT || attrt_tmp == RC_MEDIUMINT) {
		int& v = (*(TempTable::AttrBuffer<int>*) _it._data.table->GetDisplayableAttrP(i)->buffer)[_atRow];
		if(v == NULL_VALUE_32)
			return RCValueObject();
		return RCValueObject( RCNum(v, 0, false, attrt_tmp) );
	} else if(attrt_tmp == RC_SMALLINT) {
		short& v = (*(TempTable::AttrBuffer<short>*) _it._data.table->GetDisplayableAttrP(i)->buffer)[_atRow];
		if(v == NULL_VALUE_SH)
			return RCValueObject();
		return RCValueObject( RCNum(v, 0, false, attrt_tmp) );
	} else if(attrt_tmp == RC_BYTEINT) {
		char& v = (*(TempTable::AttrBuffer<char>*) _it._data.table->GetDisplayableAttrP(i)->buffer)[_atRow];
		if(v == NULL_VALUE_C)
			return RCValueObject();
		return RCValueObject( RCNum(v, 0, false, attrt_tmp) );
	} else if(ATI::IsRealType(attrt_tmp)) {
		double& v = (*(TempTable::AttrBuffer<double>*) _it._data.table->GetDisplayableAttrP(i)->buffer)[_atRow];
		if(v == NULL_VALUE_D)
			return RCValueObject();
		return RCValueObject( RCNum(v) );
	} else if(attrt_tmp == RC_NUM || attrt_tmp == RC_BIGINT ){
		_int64& v = (*(TempTable::AttrBuffer<_int64>*) _it._data.table->GetDisplayableAttrP(i)->buffer)[_atRow];
		if(v == NULL_VALUE_64)
			return RCValueObject();
		return RCValueObject( RCNum(v, _it._data.table->GetDisplayableAttrP(i)->Type().GetScale(), false, attrt_tmp) );
	} else if (ATI::IsDateTimeType(attrt_tmp)) {
		_int64& v = (*(TempTable::AttrBuffer<_int64>*) _it._data.table->GetDisplayableAttrP(i)->buffer)[_atRow];
		if(v == NULL_VALUE_64)
			return RCValueObject();
		return RCValueObject( RCDateTime(v, attrt_tmp) );

	}
	BHASSERT( ATI::IsStringType(attrt_tmp), "not all possible attr_types checked" );
	RCBString rcs = (*(TempTable::AttrBuffer<RCBString>*) _it._data.table->GetDisplayableAttrP(i)->buffer)[_atRow];
	return RCValueObject(rcs);
}

JustATable::TempTableLockPolicy::iterator_t& JustATable::TempTableLockPolicy::Assign(
		JustATable::TempTableLockPolicy::iterator_t& to, JustATable::TempTableLockPolicy::iterator_t const& from)
{
	to._data.table = from._data.table;
	to._currentRNo = from._currentRNo;
	to._conn = from._conn;
	return to;
}

void JustATable::TempTableLockPolicy::Next(iterator_t& it)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT( it._currentRNo < it._data.table->NoObj() );
	++it._currentRNo;
}

JustATable::RecordIterator TempTable::do_find(ConnectionInfo *conn, _uint64 pos,
		JustATable::RcTableLockPolicy::attr_uses_t const&)
{
	return (JustATable::RecordIterator(JustATable::RecordIteratorInterface::iterator_ptr_t(
			new JustATable::RecordIteratorImplementation<JustATable::TempTableLockPolicy>(
					JustATable::TempTableLockPolicy::initializer_t(this, conn), pos))));
}

JustATable::RecordIterator TempTable::do_find(ConnectionInfo *conn, _uint64 pos, ::Filter const*,
		JustATable::RcTableLockPolicy::attr_uses_t const&)
{
	return (JustATable::RecordIterator(JustATable::RecordIteratorInterface::iterator_ptr_t(
			new JustATable::RecordIteratorImplementation<JustATable::TempTableLockPolicy>(
					JustATable::TempTableLockPolicy::initializer_t(this, conn), pos))));
}

JustATable::RecordIterator TempTable::do_begin(ConnectionInfo *conn, JustATable::RcTableLockPolicy::attr_uses_t const&)
{
	return (JustATable::RecordIterator(JustATable::RecordIteratorInterface::iterator_ptr_t(
			new JustATable::RecordIteratorImplementation<JustATable::TempTableLockPolicy>(
					JustATable::TempTableLockPolicy::initializer_t(this, conn)))));
}

JustATable::RecordIterator TempTable::do_begin(ConnectionInfo *conn, ::Filter const& filter,
		JustATable::RcTableLockPolicy::attr_uses_t const&)
{
	return (JustATable::RecordIterator(JustATable::RecordIteratorInterface::iterator_ptr_t(
			new JustATable::RecordIteratorImplementation<JustATable::TempTableLockPolicy>(
					JustATable::TempTableLockPolicy::initializer_t(this, conn)))));
}

JustATable::RecordIterator TempTable::do_end(ConnectionInfo *conn)
{
	return (JustATable::RecordIterator(JustATable::RecordIteratorInterface::iterator_ptr_t(
			new JustATable::RecordIteratorImplementation<JustATable::TempTableLockPolicy>(
					JustATable::TempTableLockPolicy::initializer_t(this, conn), NoObj()))));
}

void TempTable::ApplyOffset(_int64 limit, _int64 offset)
{
	// filter out all unwanted values from buffers
	no_obj = limit;
	for(uint i = 0; i < attrs.size(); i++) {
		if(attrs[i]->alias)
			attrs[i]->ApplyFilter(output_mind, offset, offset + limit);
		else
			attrs[i]->DeleteBuffer();
	}
}

uint JustATable::TempTableLockPolicy::DataTag::Size()
{
	return (table->NoDisplaybleAttrs());
}

const PhysicalColumn* TempTable::GetSourceColumn(int alias, int col_no)
{
	uint i;
	for(i=0 ; i< aliases.size(); i++)
		if(aliases[i] == alias)
			return tables[i]-> GetColumn(col_no) ;
	assert(0 && "Source column in TempTable not found");
	return NULL;
}

bool TempTable::HasDelayedAttrs()
{
	for(uint i = 0; i < attrs.size(); i++) {
		if(attrs[i] && attrs[i]->mode == DELAYED)
			return true;
	}
	return false;
}

void TempTable::ProcessParameters()
{
	for(uint i = 0; i< virt_cols.size(); i++)
		virt_cols[i]->RequestEval();
	filter.ProcessParameters();
	query->UpdateParamFilter(&filter, false, -1, false);
	// filter.m_index->UpdateNoTuples();
}

bool TempTable::IsParametrized(DescTreeNode *n)
{
	if(n == NULL)
		return false;
	bool p = false;
	if (n->desc.attr.vc)
		p |= n->desc.attr.vc->IsParameterized();
	if (n->desc.val1.vc)
		p |=n->desc.val1.vc->IsParameterized();
	if (n->desc.val2.vc)
		p |=n->desc.val2.vc->IsParameterized();
	p |= IsParametrized(n->left) || IsParametrized(n->right);
	return p;
}

bool TempTable::IsParametrized()
{
	for(int i = 0; i < virt_cols.size(); i++)
		if(virt_cols[i] && virt_cols[i]->IsParameterized())
			return true;
	return false;
}

int TempTable::AddVirtColumn(VirtualColumn* vc)
{
	 virt_cols.push_back(vc);
	 virt_cols_for_having.push_back(vc->GetMultiIndex() == &output_mind);
	 return virt_cols.size()-1;
}

int	TempTable::AddVirtColumn(VirtualColumn* vc, int no)
{
	assert(no < virt_cols.size());
	virt_cols_for_having[no] = (vc->GetMultiIndex() == &output_mind);
	virt_cols[no] = vc;
	return no;
}

TempTablePtr TempTable::Create(const TempTable& t, bool for_subq)
{
	TempTablePtr tnew = boost::shared_ptr<TempTable>(new TempTable(t, for_subq));
	if(for_subq)
		for(uint i = 0; i< tnew->virt_cols.size(); i++) {
			if(tnew->virt_cols_for_having[i])
				tnew->virt_cols[i]->SetMultiIndex(&tnew->output_mind, tnew);
			else
				tnew->virt_cols[i]->SetMultiIndex(tnew->filter.m_index);
		}
	return tnew;
}

TempTablePtr TempTable::Create(JustATable* const t, int alias, Query* q)
{
	return boost::shared_ptr<TempTable>(new TempTable(t, alias, q));
}

ColumnType TempTable::GetUnionType(ColumnType type1, ColumnType type2)
{
	AttributeType tn1 = type1.GetTypeName();
	AttributeType tn2 = type2.GetTypeName();
	if(type1.IsFloat() || type2.IsFloat())
		return type1.IsFloat() ? type1 : type2;
	if(type1.IsString() && !type2.IsString())
		return type1;
	if(type2.IsString() && !type1.IsString())
		return type2;
	if(type1.GetTypeName() == RC_NUM && type2.GetTypeName() == RC_NUM && type1.GetScale() != type2.GetScale()) {
		uint max_scale = type1.GetScale() > type2.GetScale() ? type1.GetScale() : type2.GetScale();
		uint max_ints = (type1.GetPrecision() - type1.GetScale()) > (type2.GetPrecision()
					- type2.GetScale()) ? (type1.GetPrecision() - type1.GetScale())
					: (type2.GetPrecision() - type2.GetScale());
		if(max_ints + max_scale > 18) {
			stringstream ss;
			ss << "UNION of non-matching columns (column no " << 0 << ") .";
			throw NotImplementedRCException(ss.str());
		}
		return type1.GetScale() > type2.GetScale() ? type1 : type2;
	}
	if(type1.GetInternalSize() < type2.GetInternalSize())
		return type2;
	else if(type1.GetInternalSize() == type2.GetInternalSize())
		return type1.GetPrecision() > type2.GetPrecision() ? type1 : type2;
	else
		return type1;
	//while(1) {
	//	if (f->Type() == s->Type() && !f->Type().IsString()) //== for strings ignores precision
	//		break;
	//	else if(f->Type().IsString() && s->Type().IsString())
	//		if (f->Type().GetPrecision() < s->Type().GetPrecision()) {
	//			stringstream ss;
	//			ss << "Cannot UNION with a column wider then the current practical column width (column " << i << "). Try changing order of union'ed tables)";
	//			throw NotImplementedRCException(ss.str());
	//		} else
	//			break;
	//		if(f->Type().IsFixed() && s->Type().IsFixed()) {
	//			if(f->Type().GetInternalSize() < s->Type().GetInternalSize()) {
	//				break;
	//				stringstream ss;
	//				ss << "Cannot UNION with a wider column (column " << i << "). Try changing order of union'ed tables)";
	//				throw NotImplementedRCException(ss.str());
	//			} else {
	//				if(f->Type().GetScale() == 0 && s->Type().GetScale() == 0)
	//					break;
	//				if(f->TypeName() == RC_NUM && s->TypeName() == RC_NUM)
	//					break; //handled below
	//			}
	//		}
	//		//otherwise
	//		break;
	//		stringstream ss;
	//		ss << "UNION of non-matching columns (column no " << i<< ") .";
	//		throw NotImplementedRCException(ss.str());
	//}
}
