/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef TRANSACTION_H
#define TRANSACTION_H

#include "DataCache.h"
#include "RCAttrPack.h"
#ifdef __GNUC__
	#include <ext/hash_map>
	#ifndef stdext
	#define stdext __gnu_cxx
	#endif
#else
	#include <hash_map>
#endif

// NOTE: These classes are still undergoing significant changes
class RCEngine;
class RCAttr;
class AttrDataCache;

class Transaction
{
	// This is the parent in a nested transaction
	Transaction *m_parent;

	DataCache m_localCache;
	SynchronizedDataCache* m_globalCache;
	RCEngine &m_rcengine;
	stdext::hash_map<std::string,RCTablePtr> m_tables;
	IBMutex m_tablesMutex;
	public:

	Transaction( Transaction *p, RCEngine &r, SynchronizedDataCache *d )
		: m_parent(p), m_globalCache(d), m_rcengine(r) {}
	~Transaction();
	
	Transaction *getParent() { return m_parent; }

    void Commit();
	// void Rollback()
	
	RCTable * GetTable(const std::string & table_path, bool locked = false);
	void DropTable(const std::string &);
	
	//--- for use by RCAttr during initial integration
	// RO = ReadOnly
	// RW = ReadWrite
	AttrPack * GetAttrPackRO(int table, int column, int dp);
	AttrPack * GetAttrPackRW(int table,int column, int dp) ;

	void PutAttrPackRO(int table,int column, int dp, AttrPack *dpp) ;
	void PutAttrPackRW(int table,int column,int dp, AttrPack *dpp);

	void ReleasePack(int table, int column, int dp);
	void ReleasePackRow(int table, int pack_row);
	#if 0
	boost::shared_ptr<AttrDataCache> GetAttrDataCache(RCAttr& attr)
	{
		return DataCache::GlobalDataCache().GetAttrDataCache(attr);
	}
#endif
	// ReleaseAttrPack(RCAttrPack *)
	// PersistAttrPack(RCAttrPack *) -- maybe done automatically by Commit ???
	//--- for use by RCAttr

};

#endif
