/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef _VALUESET_H_
#define _VALUESET_H_

#include "common/CommonDefinitions.h"
#include "types/RCDataTypes.h"

#ifdef __GNUC__
#ifndef stdext
#define stdext __gnu_cxx
#endif
#endif

class RCDataType;
class RCValueObject;
class Filter;


class ValueSet
{
public:

#ifdef __GNUC__
	typedef stdext::hash_set<RCDataType*, rc_hash_compare<RCDataType*>, rc_hash_compare<RCDataType*> > RCDTptrHashSet;
#else
	typedef stdext::hash_set<RCDataType*, rc_hash_compare<RCDataType*> > RCDTptrHashSet;
#endif
	typedef RCDTptrHashSet::const_iterator const_iterator;
	ValueSet();
	ValueSet(const ValueSet& sec);
	~ValueSet();

	void Add(std::auto_ptr<RCDataType> rcdt);
	void Add(const RCValueObject& rcv);
	void AddNull()	{ contains_nulls = true; }
	void Prepare(AttributeType at, int scale, DTCollation);
	bool Contains(const RCDataType& v, DTCollation);
	bool Contains(const RCValueObject& v, DTCollation);

	void Clear( void );
	int 		NoVals() const;
	bool 		IsEmpty() const		{ return (NoVals() == 0 && !contains_nulls); }
	bool 		IsText() const			{ return is_text; }
	bool 		ContainsNulls() const 	{ return contains_nulls; }
	RCDataType* Min() const				{ return min; }
	RCDataType* Max() const				{ return max; }
	std::string ToText();
	const_iterator begin() const;
	const_iterator end() const;
	_int64 AsInt64(const_iterator const&) const;
	/////////////////////////////////////////////////////////////////////////////////////////////////
private:
	bool contains_nulls;
	int no_obj;					// number of values added
	bool is_text;
	bool prepared;
	AttributeType prep_type;	// type to which it is prepared
	int prep_scale;				// scale to which it is prepared
	DTCollation prep_collation;
	bool IsPrepared(AttributeType at, int scale, DTCollation);
	int scope;

	RCDTptrHashSet* values;
	RCDataType* min;
	RCDataType* max;

};



inline bool ValueSet::IsPrepared(AttributeType at, int scale, DTCollation coll)
{
	return (prepared && prep_type == at && prep_scale == scale && 
		(prep_collation.collation == coll.collation || 
		(!RequiresUTFConversions(prep_collation) && !RequiresUTFConversions(coll))
		));
}



#endif

