/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include <boost/bind.hpp>
#include "Query.h"
#include "compilation_tools.h"
#include "CQTerm.h"
#include "ValueSet.h"
#include "CompiledQuery.h"
#include "MysqlExpression.h"

using namespace std;

int Query::Item2CQTerm(Item* an_arg, CQTerm& term, TabID tmp_table, FilterType filter_type, bool negative, Item* left_expr_for_subselect, Operator* oper_for_subselect) {
	an_arg = UnRef(an_arg);
	if(an_arg->type() == Item::SUBSELECT_ITEM) {
		Item_subselect* item_subs = dynamic_cast<Item_subselect*>(an_arg);
		assert(item_subs && "The cast to (Item_subselect*) was unsuccessful");

		bool ignore_limit = false;
		if(dynamic_cast<Item_maxmin_subselect*>(item_subs) != NULL || dynamic_cast<Item_in_subselect*>(item_subs) != NULL)
			ignore_limit = true;
		TabID subselect;
		st_select_lex_unit* select_unit = item_subs->get_unit();

		// needs to check if we can relay on subquery transformation to min/max
		bool ignore_minmax = (dynamic_cast<Item_maxmin_subselect*>(item_subs)==NULL
			&& dynamic_cast<Item_in_subselect*>(item_subs)==NULL
			&& negative
			&& item_subs->substype()==Item_subselect::SINGLEROW_SUBS);
		subqueries_in_where.push_back(make_pair(tmp_table, filter_type == WHERE_FILTER));
		int res = Compile(cq, select_unit->first_select(), select_unit->union_distinct, table_alias2index, table_alias2ptr,
			&subselect, ignore_limit, left_expr_for_subselect, oper_for_subselect, ignore_minmax);
		subqueries_in_where.pop_back();
		if(res == RCBASE_QUERY_ROUTE) {
			AttrID at, vc;
			vc.n = VirtualColumnAlreadyExists(tmp_table, subselect);
			if(vc.n == NULL_VALUE_32) {
				cq->CreateVirtualColumn(vc, tmp_table, subselect, filter_type == HAVING_FILTER ? true : false);
				tab_id2subselect.insert(make_pair(tmp_table, std::make_pair(vc.n, subselect)));
			}
			if(oper_for_subselect) {
				if(dynamic_cast<Item_maxmin_subselect*>(item_subs) != NULL || dynamic_cast<Item_in_subselect*>(item_subs) != NULL) {
					if(negative) {
						MarkWithAll(*oper_for_subselect);
						if(dynamic_cast<Item_in_subselect*>(item_subs) != NULL && *oper_for_subselect == O_IN)
							*oper_for_subselect = O_EQ_ALL;
					} else {
						MarkWithAny(*oper_for_subselect);
						if(dynamic_cast<Item_allany_subselect*>(item_subs) != NULL && dynamic_cast<Item_allany_subselect*>(item_subs)->all == 1)
							*oper_for_subselect = O_EQ_ALL;
					}
				} else {
					if(negative) {
						//if(item_subs->substype() != Item_subselect::SINGLEROW_SUBS)
						//	return RETURN_QUERY_TO_MYSQL_ROUTE;
						MarkWithAll(*oper_for_subselect);
					} else
						UnmarkAllAny(*oper_for_subselect);
				}
			}
			term = CQTerm(vc.n);
//			if(filter_type == HAVING_FILTER) {
//				AttrID at;
//				int attr = GetAddColumnId(vc, tmp_table, LISTING, false);
//				if(attr == NULL_VALUE_32)
//					attr = GetAddColumnId(vc, tmp_table, GROUP_BY, false);
//				if(attr == NULL_VALUE_32) {
//					cq->AddColumn(at, tmp_table, term, DELAYED, NULL, false);
//					coltab_to_vc[std::pair<int,int>(at.n, tmp_table.n)] = term.vc_id;
//				}
//				vc.n = VirtualColumnAlreadyExists(tmp_table, at);
//				if(vc.n == NULL_VALUE_32) {
//					cq->CreateVirtualColumn(vc, tmp_table, tmp_table, at);
//					phys2virt[std::pair<int,int>(tmp_table.n, at.n)] = vc.n;
//				}
//				term = CQTerm(vc.n);
//			}
		}
		return res;
	}

	if(filter_type == HAVING_FILTER) {
		ColOperation oper;
		bool distinct;
		if(!OperationUnmysterify(an_arg, oper, distinct, true)) //is_having_clause may be true only in case group by clause was present
			return RETURN_QUERY_TO_MYSQL_ROUTE;

		std::pair<int,int> phys_vc;
		AttrID col, at, vc;
		TabID tab;
		if((IsFieldItem(an_arg) || IsAggregationOverFieldItem(an_arg))&& !FieldUnmysterify(an_arg, tab, col))
			return RETURN_QUERY_TO_MYSQL_ROUTE;
		if(IsAggregationItem(an_arg) && HasAggregation(((Item_sum*)an_arg)->args[0]))
			return RETURN_QUERY_TO_MYSQL_ROUTE;
		if((IsFieldItem(an_arg) || IsAggregationOverFieldItem(an_arg)) && cq->ExistsInTempTable(tab, tmp_table)) {
			int col_num = AddColumnForPhysColumn(an_arg, tmp_table, oper, distinct, true);
			phys_vc = VirtualColumnAlreadyExists(tmp_table, tmp_table, AttrID(-col_num - 1));
			if(phys_vc.first == NULL_VALUE_32) {
				phys_vc.first = tmp_table.n;
				cq->CreateVirtualColumn(phys_vc.second, tmp_table, tmp_table, AttrID(col_num));
				phys2virt.insert(make_pair(std::pair<int,int>(tmp_table.n, -col_num - 1), phys_vc));
			}
			vc.n = phys_vc.second;
		} else if(IsCountStar(an_arg)) {
			at.n = GetAddColumnId(AttrID(NULL_VALUE_32), tmp_table, COUNT, false);
			if(at.n == NULL_VALUE_32) // doesn't exist yet
				cq->AddColumn(at, tmp_table, CQTerm(), COUNT, NULL, false);
			phys_vc = VirtualColumnAlreadyExists(tmp_table, tmp_table, at);
			if(phys_vc.first == NULL_VALUE_32) {
				phys_vc.first = tmp_table.n;
				cq->CreateVirtualColumn(phys_vc.second, tmp_table, tmp_table, at);
				phys2virt.insert(make_pair(std::pair<int,int>(tmp_table.n, at.n), phys_vc));
			}
			vc.n = phys_vc.second;
		} else if (an_arg->type() == Item::VARBIN_ITEM) {
			String str;
			String* tmp = an_arg->val_str(&str); //sets null_value
			if(!an_arg->null_value) {
				if(an_arg->max_length <= 8) {
					Item* int_item = new Item_int((ulonglong)an_arg->val_int());
					MysqlExpression* mysql_expression = NULL;
					MysqlExpression::Item2VarID item2varid;
					gc_expressions.push_back(mysql_expression = new MysqlExpression(int_item, item2varid));
					vc.n = VirtualColumnAlreadyExists(tmp_table, mysql_expression);
					if(vc.n == NULL_VALUE_32) {
						cq->CreateVirtualColumn(vc, tmp_table, mysql_expression);
						tab_id2expression.insert(make_pair(tmp_table, std::make_pair(vc.n, mysql_expression)));
					}
				} else
					return RETURN_QUERY_TO_MYSQL_ROUTE; //too large binary to be treated as BIGINT
			} else {
				return RETURN_QUERY_TO_MYSQL_ROUTE;
			}
		} else {
			bool is_const_or_aggr;
			MysqlExpression* expr;
			MysqlExpression::SetOfVars vars;
			if(WrapMysqlExpression(an_arg, tmp_table, expr, false, true) == WrapStatus::FAILURE)
				return RETURN_QUERY_TO_MYSQL_ROUTE;
			if(IsConstExpr(expr->GetVars(), tmp_table)) {
				vc.n = VirtualColumnAlreadyExists(tmp_table, expr);
				if(vc.n == NULL_VALUE_32) {
					cq->CreateVirtualColumn(vc, tmp_table, expr, tmp_table);
					tab_id2expression.insert( make_pair(tmp_table, std::make_pair(vc.n, expr)) );
				}
			} else if(IsAggregationItem(an_arg)) {
				assert(expr->GetItem()->type() == Item_bhfield::BHFIELD_ITEM);
				int col_num = ((Item_bhfield*)expr->GetItem())->varID.col;
				phys_vc = VirtualColumnAlreadyExists(tmp_table, tmp_table, AttrID(-col_num - 1));
				if(phys_vc.first == NULL_VALUE_32) {
					phys_vc.first = tmp_table.n;
					cq->CreateVirtualColumn(phys_vc.second, tmp_table, tmp_table, AttrID(col_num));
					phys2virt.insert(make_pair(std::pair<int,int>(tmp_table.n, -col_num - 1), phys_vc));
				}
				vc.n = phys_vc.second;
			} else {
//				int col_num = AddColumnForMysqlExpression(expr, tmp_table, NULL, DELAYED, distinct, true);
				vc.n = VirtualColumnAlreadyExists(tmp_table, expr);
				if(vc.n == NULL_VALUE_32) {
					cq->CreateVirtualColumn(vc, tmp_table, expr, tmp_table);
					tab_id2expression.insert( make_pair(tmp_table, std::make_pair(vc.n, expr)) );
				}
			}
		}
		term = CQTerm(vc.n);
		return RCBASE_QUERY_ROUTE;
	} else {
		//WHERE FILTER

		std::pair<int,int> phys_vc;
		AttrID vc;
		AttrID col;
		TabID tab;
		if(IsFieldItem(an_arg) && !FieldUnmysterify(an_arg, tab, col))
			return RETURN_QUERY_TO_MYSQL_ROUTE;
		if(IsFieldItem(an_arg) && cq->ExistsInTempTable(tab, tmp_table)) {
			phys_vc = VirtualColumnAlreadyExists(tmp_table, tab, col);
			if(phys_vc.first == NULL_VALUE_32) {
				phys_vc.first = tmp_table.n;
				cq->CreateVirtualColumn(phys_vc.second, tmp_table, tab, col);
				phys2virt.insert(make_pair(std::pair<int,int>(tab.n, col.n), phys_vc));
			}
			vc.n = phys_vc.second;
		} else if(an_arg->type() == Item::VARBIN_ITEM) {
			String str;
			String* tmp = an_arg->val_str(&str); //sets null_value
			if(!an_arg->null_value) {
				if(an_arg->max_length <= 8) {
					Item* int_item= new Item_int((ulonglong)an_arg->val_int());
					MysqlExpression* mysql_expression = NULL;
					MysqlExpression::Item2VarID item2varid;
					gc_expressions.push_back( mysql_expression = new MysqlExpression(int_item, item2varid) );
					vc.n = VirtualColumnAlreadyExists(tmp_table, mysql_expression);
					if(vc.n == NULL_VALUE_32) {
						cq->CreateVirtualColumn(vc, tmp_table, mysql_expression);
						tab_id2expression.insert(make_pair(tmp_table, std::make_pair(vc.n, mysql_expression)));
					}
				} else
					return RETURN_QUERY_TO_MYSQL_ROUTE; //too large binary to be treated as BIGINT
			} else {
				return RETURN_QUERY_TO_MYSQL_ROUTE;
			}
		} else {
			MysqlExpression* expr;
			WrapStatus::wrap_status_t ws = WrapMysqlExpression(an_arg, tmp_table, expr, true, false);
			if(ws != WrapStatus::SUCCESS)
				return RETURN_QUERY_TO_MYSQL_ROUTE;
			vc.n = VirtualColumnAlreadyExists(tmp_table, expr);
			if(vc.n == NULL_VALUE_32) {
				cq->CreateVirtualColumn(vc, tmp_table, expr);
				tab_id2expression.insert( make_pair(tmp_table, std::make_pair(vc.n, expr)) );
			}
		}
		term = CQTerm(vc.n);
		return RCBASE_QUERY_ROUTE;
	}
	return RETURN_QUERY_TO_MYSQL_ROUTE;
}

int Query::FilterNumberFromMultipleEquality(Item_equal * conds, const TabID &tmp_table, FilterType filter_type, FilterID * and_me_filter)
{
	Item_equal_iterator li(*conds);

	CQTerm zero_term, first_term, next_term;
	Item_field* ifield;
	Item* const_item = conds->get_const();
	if(const_item) {
		if(!Item2CQTerm(const_item, zero_term, tmp_table, filter_type))
			return -1;
	} else {
		 ifield = li++;
		 if(!Item2CQTerm(ifield, zero_term, tmp_table, filter_type))
			 return -1;
	}
	ifield = li++;
	if(!Item2CQTerm(ifield, first_term, tmp_table, filter_type))
		return -1;
	FilterID filter;
	if(!and_me_filter)
		cq->CreateFilter(filter, tmp_table, first_term, O_EQ, zero_term, CQTerm(), filter_type);
	else
		cq->And(*and_me_filter, *and_me_filter, tmp_table, first_term,O_EQ, zero_term);
	while(ifield = li++) {
		if(!Item2CQTerm(ifield, next_term, tmp_table, filter_type))
			 return -1;
		if(!and_me_filter)
			cq->And(filter, filter, tmp_table, next_term,O_EQ, zero_term);
		else
			cq->And(*and_me_filter, *and_me_filter, tmp_table, next_term,O_EQ, zero_term);
	}
	if(and_me_filter)
		filter = *and_me_filter;
	return filter.n;
}

Item* Query::FindOutAboutNot(Item* it, bool& is_there_not)
{
	is_there_not = false;

	/*First we try to take care of IN and BETWEEN*/
	Item_func_opt_neg* it_maybe_neg = dynamic_cast<Item_func_opt_neg *>(it);
	if(it_maybe_neg) {
		is_there_not = it_maybe_neg->negated;
		return it;
	}

	/*Then we remove negation in case of LIKE (or sth. else)*/
	if( ((Item_func*)it)->functype() == Item_func::NOT_FUNC ) {
		Item * arg = UnRef(((Item_func *)it)->arguments()[0]);
		//OK, in case of subselects (and possibly other stuff) NOT is handled by other statements,
		//So we narrow the choice of cases in which we handle NOT down to LIKE
		if(arg->type() == Item::FUNC_ITEM && ((Item_func*)arg)->functype() == Item_func::LIKE_FUNC) {
			is_there_not = true;
			return arg;
		}
	}
	return it;
}

int Query::FilterNumberFromComparison(Item* conds, const TabID& tmp_table, FilterType filter_type, FilterID* and_me_filter, bool negative)
{
	FilterID filter;
	Operator op;  /*{	O_EQ, O_NOT_EQ, O_LESS, O_MORE, O_LESS_EQ, O_MORE_EQ,
			O_IS_NULL, O_NOT_NULL, O_BETWEEN, O_LIKE, O_IN, O_ESCAPE etc...};*/
	Item_in_optimizer* in_opt = NULL;	// set if IN expression with subselect

	ExtractOperatorType(conds, op, negative);
	Item_func* cond_func = (Item_func*)conds;
	if(op == O_MULT_EQUAL_FUNC)
		return FilterNumberFromMultipleEquality((Item_equal *) conds, tmp_table, filter_type, and_me_filter);
	else if(op == O_NOT_FUNC) {
		if(cond_func->arg_count != 1 || dynamic_cast<Item_in_optimizer*>(cond_func->arguments()[0]) == NULL)
			return -2;
		return FilterNumberFromComparison(cond_func->arguments()[0], tmp_table, filter_type, and_me_filter, true);
	} else if(op == O_NOT_ALL_FUNC) {
		if(cond_func->arg_count != 1)
			return -1;
		return FilterNumberFromComparison(cond_func->arguments()[0], tmp_table, filter_type, and_me_filter, dynamic_cast<Item_func_nop_all*>(cond_func) == NULL);
	} else if(op == O_UNKNOWN_FUNC) {
		in_opt = dynamic_cast<Item_in_optimizer*>(cond_func);
		if(in_opt == NULL || cond_func->arg_count != 2 || in_opt->arguments()[0]->cols() != 1)
			return -2;
		op = O_IN;
	} else if(op == O_ERROR)
		return -2;   //unknown function type

	if((cond_func->arg_count < 0) || (cond_func->arg_count > 3 && op != O_IN && op!= O_NOT_IN))
		return -1; // argument count error

	CQTerm terms[3];
	vector<MysqlExpression*> exprs;
	vector<int> vcs;

	Item** args = cond_func->arguments();
	for(uint i = 0; i < cond_func->arg_count; i++) {
		Item* an_arg = UnRef(args[i]);
		if((op == O_IN || op == O_NOT_IN) && i > 0) {
			if(i==1 && in_opt) {
				if(!Item2CQTerm(an_arg, terms[i], tmp_table, filter_type, negative, *in_opt->get_cache(), &op))
					return -1;
				if(negative)
					switch(op) {
						case O_EQ:		op = O_NOT_EQ; break;
						case O_EQ_ALL:	op = O_NOT_IN; break;
						case O_EQ_ANY:	op = O_NOT_EQ_ANY; break;
						case O_NOT_EQ:	op = O_EQ; break;
						case O_NOT_EQ_ALL:	op = O_EQ_ALL; break;
						case O_NOT_EQ_ANY:	op = O_EQ_ANY; break;
						case O_LESS_EQ:	op = O_MORE; break;
						case O_LESS_EQ_ALL:	op = O_MORE_ALL; break;
						case O_LESS_EQ_ANY:	op = O_MORE_ANY; break;
						case O_MORE_EQ:	op = O_LESS; break;
						case O_MORE_EQ_ALL:	op = O_LESS_ALL; break;
						case O_MORE_EQ_ANY:	op = O_LESS_ANY; break;
						case O_MORE:	op = O_LESS_EQ; break;
						case O_MORE_ALL:	op = O_LESS_EQ_ALL; break;
						case O_MORE_ANY:	op = O_LESS_EQ_ANY; break;
						case O_LESS:	op = O_MORE_EQ; break;
						case O_LESS_ALL:	op = O_MORE_EQ_ALL; break;
						case O_LESS_ANY:	op = O_MORE_EQ_ANY; break;
						case O_LIKE:	op = O_NOT_LIKE; break;
						case O_IN:		op = O_NOT_IN; break;
						case O_NOT_LIKE:op = O_LIKE; break;
						case O_NOT_IN:	op = O_IN; break;
						default:
							return -1;
					}
			} else {
				CQTerm t;
				if(!Item2CQTerm(an_arg, t, tmp_table, filter_type, an_arg->type() == Item::SUBSELECT_ITEM ? negative : false, NULL, &op))
					return -1;
				vcs.push_back(t.vc_id);
			}
		} else {
			if(!Item2CQTerm(an_arg, terms[i], tmp_table, filter_type, an_arg->type() == Item::SUBSELECT_ITEM ? negative : false, NULL, &op))
				return -1;
			if((op == O_LIKE || op == O_NOT_LIKE) &&
				!(an_arg->field_type() == MYSQL_TYPE_VARCHAR || an_arg->field_type() == MYSQL_TYPE_STRING || an_arg->field_type() == MYSQL_TYPE_VAR_STRING || an_arg->field_type() == MYSQL_TYPE_BLOB)) {
				my_message(ER_SYNTAX_ERROR, "Brighthouse specific error: Both operator and operand must be string data types for LIKE comparison", MYF(0));
				throw ReturnMeToMySQLWithError();
			}
		}
	}

	if((op == O_IN || op == O_NOT_IN) && !in_opt) {
		AttrID vc;
		cq->CreateVirtualColumn(vc, tmp_table, vcs, AttrID(terms[0].vc_id));
		terms[1] = CQTerm(vc.n);
	}

	if(!and_me_filter)
		cq->CreateFilter(filter, tmp_table, terms[0], op, terms[1], terms[2], filter_type);
	else {
		cq->And(*and_me_filter,*and_me_filter, tmp_table, terms[0], op, terms[1], terms[2]);
		filter = *and_me_filter;
	}
	return filter.n;
}

int Query::FilterNotSubselect(Item* conds, TabID tmp_table, FilterType filter_type, FilterID * and_me_filter)
{
	bool propagate_not = dynamic_cast<Item_in_optimizer*>(conds)!=NULL;
	int filter_num = FilterNumberFromComparison(conds, tmp_table, filter_type, propagate_not?and_me_filter:NULL, propagate_not);
	if (filter_num < 0) return -1;
	FilterID filter( filter_num );
	if (!propagate_not)
	{
		cq->Not(filter, filter);
		if (and_me_filter)
		{
			cq->And(*and_me_filter, *and_me_filter, filter, tmp_table);
			filter = *and_me_filter;
		}
	}
	return filter.n;
}

int Query::FilterNumberFromNaked(COND* conds, TabID tmp_table, FilterType filter_type, FilterID* and_me_filter)
{
	FilterID filter;
	CQTerm naked_col;
#ifdef __WIN__
    // WINPORT - explicit bool expression
	if(!Item2CQTerm(conds, naked_col, tmp_table, filter_type, conds->type() == Item::SUBSELECT_ITEM ? (and_me_filter!=NULL) : false))
		return -1;
#else
	if(!Item2CQTerm(conds, naked_col, tmp_table, filter_type, conds->type() == Item::SUBSELECT_ITEM ? and_me_filter : false))
		return -1;
#endif
	bool is_string = conds->result_type() == STRING_RESULT;
	MysqlExpression::Item2VarID item2varid;
	AttrID vc;
	Item* zero_item;
	if(is_string)
		zero_item = new Item_empty_string("", 0);
	else
#ifdef __WIN__
		// WINPORT - explicit bool expression
		zero_item= new Item_int((longlong)0);
#else
		zero_item= new Item_int(0);
#endif
	MysqlExpression* mysql_expression = NULL;
	gc_expressions.push_back( mysql_expression = new MysqlExpression(zero_item, item2varid) );
	// TODO: where mysql_expression & zero_item is destroyed ???
	vc.n = VirtualColumnAlreadyExists(tmp_table, mysql_expression);
	if(vc.n == NULL_VALUE_32) {
		cq->CreateVirtualColumn(vc, tmp_table, mysql_expression);
		tab_id2expression.insert(make_pair(tmp_table, std::make_pair(vc.n, mysql_expression)));
	}
	cq->CreateFilter(filter, tmp_table, naked_col, O_NOT_EQ, CQTerm(vc.n), CQTerm(), filter_type);
	return filter.n;
}

int Query::FilterNumber(COND* conds, TabID tmp_table, FilterType filter_type, FilterID * and_me_filter)
{
	//we know, that conds != 0
	//returns -1 on error
	//        >=0 is a created filter number
	conds = UnRef(conds);
	Item::Type cond_type = conds->type();
	FilterID filter;
	if(cond_type == Item::COND_ITEM) {
		Item_cond *cond_cond= (Item_cond*) conds;
		Item_func::Functype func_type = cond_cond->functype();
		bool first = true;
		switch(func_type) {
			case Item_func::COND_AND_FUNC: {
	//			List_iterator_fast<Item_equal> list_equal(((Item_cond_and*)cond_cond)->cond_equal.current_level);
				List_iterator_fast<Item> li(*(cond_cond->argument_list()));
				Item* item;
				//in first run we skip all OR conditions, in the second run we do them
				for(int doing_ors = 0; doing_ors <= 1; doing_ors++) {
					while((item = li++) /*|| (item = list_equal++)*/) {
						item = UnRef(item);
						Item::Type item_type = item->type();
						bool is_or = false;
						if(item_type == Item::COND_ITEM && ((Item_cond*)item)->functype() == Item_func::COND_OR_FUNC)
							is_or = true;
						if((doing_ors && !is_or) || (!doing_ors && is_or))
							continue;
						if(first && !and_me_filter) {
							int filter_num = FilterNumber(item, tmp_table, filter_type);
							if (filter_num < 0)
								return -1;
							filter = FilterID(filter_num);
							first = false;
							and_me_filter = &filter;
						} else {   //here and_me_filter is NOT NULL in every case
							int filter_num = FilterNumber(item, tmp_table, filter_type, and_me_filter);
							if(filter_num < 0)
								return -1;
							if(filter_num != and_me_filter->n)
								cq->And(*and_me_filter, *and_me_filter, FilterID( filter_num ), tmp_table);
						}
					}
					li.rewind();
				}  //after this while (and for), if only the argument list was not empty
				//1. and_me_filter is NOT NULL
				//2. and_me_filter is the one current
				if(and_me_filter)  //in case of an empty argument list
					filter = *and_me_filter;
				break;
			}
			case Item_func::COND_OR_FUNC: {
				List_iterator_fast<Item> li(*(cond_cond->argument_list()));
				Item* item;
				while(item = li++) {
					FilterID* new_and_me_filter = 0;
					try {
	//to deal with deleting dynamical variables (new_and_me_filter) properly in case of an exception
						if(and_me_filter) {
							new_and_me_filter = new FilterID;
							cq->Copy(*new_and_me_filter, *and_me_filter);
						}
						int filter_num = FilterNumber(item, tmp_table, filter_type, new_and_me_filter);
						if(filter_num < 0) {//to deal with deleting dynamical variables (new_and_me_filter) properly
											//in case of failure
							if(new_and_me_filter)
								delete new_and_me_filter;
							return -1;
						}
						if(first) {
							if(new_and_me_filter && filter_num != new_and_me_filter->n) {
								cq->And(*new_and_me_filter, *new_and_me_filter, FilterID( filter_num ), tmp_table);
								filter = *new_and_me_filter;
							} else
								filter = FilterID( filter_num );
							first = false;
						} else {
							if(new_and_me_filter && filter_num != new_and_me_filter->n) {
								cq->And(*new_and_me_filter, *new_and_me_filter, FilterID( filter_num ), tmp_table);
								cq->Or(filter, filter, FilterID(*new_and_me_filter));
							} else
								cq->Or(filter, filter, FilterID(filter_num));
						}
						if(new_and_me_filter)
							delete new_and_me_filter;
					} catch(...) {//to deal with deleting dynamical variables (new_and_me_filter) properly in case of an exception
						if(new_and_me_filter)
							delete new_and_me_filter;
						throw;
					}
				}
				if(and_me_filter)
					and_me_filter->n = filter.n;    //this condition was already considered,
				//we have a new and_me_filter, the last one can be discarded
				break;
			}
			case Item_func::COND_XOR_FUNC:   //we don't handle xor as yet
			default:
				return -1;    //unknown function type
		} // end switch()
	} else if(cond_type == Item::FUNC_ITEM) {
		Item_func* cond_func = (Item_func*)conds;
		Item_func::Functype func_type = cond_func->functype();
		Item* arg = NULL;
		if(cond_func->arg_count == 1)
			arg = cond_func->arguments()[0];
		if(func_type == Item_func::NOT_FUNC && arg != NULL
			&& arg->type() == Item::SUBSELECT_ITEM && ((Item_subselect*)arg)->substype() == Item_subselect::EXISTS_SUBS) {

			CQTerm term;
			if(!Item2CQTerm(arg, term, tmp_table, filter_type))
				return -1;
			if(!and_me_filter)
				cq->CreateFilter(filter, tmp_table, term, O_NOT_EXISTS, CQTerm(), CQTerm(), filter_type);
			else {
				cq->And(*and_me_filter,*and_me_filter, tmp_table, term, O_NOT_EXISTS, CQTerm(), CQTerm());
				filter = *and_me_filter;
			}
		} else if(func_type == Item_func::COND_XOR_FUNC) {
			return -1;
		} else {
			int val = FilterNumberFromComparison(cond_func, tmp_table, filter_type, and_me_filter);
			if(val == -2)
				val = FilterNumberFromNaked(conds, tmp_table, filter_type, and_me_filter);
			return val;
		}
	} else if(cond_type == Item::SUBSELECT_ITEM && ((Item_subselect*)conds)->substype() == Item_subselect::EXISTS_SUBS) {
		CQTerm term;
		if(!Item2CQTerm(conds, term, tmp_table, filter_type))
			return -1;
		if(!and_me_filter) {
			cq->CreateFilter(filter, tmp_table, term, O_EXISTS,  CQTerm(), CQTerm(), filter_type);
		} else {
			cq->And(*and_me_filter,*and_me_filter, tmp_table, term, O_EXISTS, CQTerm(), CQTerm());
			filter = *and_me_filter;
		}
	} else if(cond_type == Item::FIELD_ITEM ||
			  cond_type == Item::SUM_FUNC_ITEM ||			  
			  cond_type == Item::SUBSELECT_ITEM ||
			  cond_type == Item::INT_ITEM ||
			  cond_type == Item::STRING_ITEM ||
			  cond_type == Item::NULL_ITEM ||
			  cond_type == Item::REAL_ITEM ||
			  cond_type == Item::DECIMAL_ITEM) {
		return FilterNumberFromNaked( conds, tmp_table, filter_type, and_me_filter);
	}
	return filter.n;
}

int Query::BuildFilter(COND* conds, CompiledQuery* cq, map<string,int>* table_alias2index,
		map<string,TABLE *>* table_alias2ptr, const TabID& tmp_table,
		FilterType filter_type, bool is_zero_result, JoinType join_type)
{
	conds = UnRef(conds);
	printItemTree("BuildFiler(), item tree passed in 'conds':", conds);
	if(is_zero_result) {
		FilterID fi;
		cq->CreateFilter(fi, tmp_table, CQTerm(), O_FALSE, CQTerm(), CQTerm(), filter_type);
		cq->Filter(tmp_table, fi);
		return RCBASE_QUERY_ROUTE;
	}
	if(!conds)
		return RCBASE_QUERY_ROUTE;   //No Conditions - no filters. OK

	// keep local copies of class fields to be changed
	CompiledQuery* _cq = this->cq;
	map<string,int>* _table_alias2index = this->table_alias2index;
	map<string,TABLE *>* _table_alias2ptr = this->table_alias2ptr;

	// copy method arguments to class fields
	this->cq = cq;
	this->table_alias2index = table_alias2index;
	this->table_alias2ptr = table_alias2ptr;

	int f_num = FilterNumber(conds, tmp_table, filter_type);

	// restore original values of class fields (may be necessary if this method is called recursively)
	this->cq = _cq;
	this->table_alias2index = _table_alias2index;
	this->table_alias2ptr = _table_alias2ptr;
	if(f_num < 0)
		return RETURN_QUERY_TO_MYSQL_ROUTE;
	cq->Filter(tmp_table, FilterID(f_num));
	return RCBASE_QUERY_ROUTE;
}

bool Query::ClearSubselectTransformation(Operator& oper_for_subselect, Item*& field_for_subselect, COND*& conds, Item*& having, Item*& cond_to_reinsert, List<Item>*& list_to_reinsert, Item* left_expr_for_subselect)
{
	cond_to_reinsert = NULL;
	list_to_reinsert = NULL;
	Item* cond_removed;
	if (having
		&& (having->type()==Item::COND_ITEM
		|| (having->type()==Item::FUNC_ITEM
			&& ((Item_func*)having)->functype()!=Item_func::ISNOTNULLTEST_FUNC
			&& (((Item_func*)having)->functype()!=Item_func::TRIG_COND_FUNC
				|| ((Item_func*)having)->arguments()[0]->type()!=Item::FUNC_ITEM
				|| ((Item_func*)((Item_func*)having)->arguments()[0])->functype()!=Item_func::ISNOTNULLTEST_FUNC))))
	{
		if (having->type() == Item::COND_ITEM)
		{
			Item_cond *having_cond= (Item_cond*) having;
			// if the condition is a complex formula it must be AND
			if (having_cond->functype()!=Item_func::COND_AND_FUNC) return false;
			//the extra condition is in the last argument
			if (having_cond->argument_list()->elements < 2) return false;
			List_iterator<Item> li(*(having_cond->argument_list()));
			while (li++ != NULL) cond_to_reinsert = *li.ref();
			li.rewind();
			while (*li.ref()!=cond_to_reinsert) li++;
			li.remove();
			list_to_reinsert = having_cond->argument_list();
			cond_removed = cond_to_reinsert;
		}
		else
		{
			// if no complex boolean formula the original condition was empty
			cond_removed = having;
			having = NULL;
		}
		cond_removed = UnRef(cond_removed);
		// check if the extra condition was wrapped into trigger
		if (cond_removed->type()==Item::FUNC_ITEM && ((Item_func*)cond_removed)->functype()==Item_func::TRIG_COND_FUNC)
		{
			cond_removed = ((Item_func*)cond_removed)->arguments()[0];
			cond_removed = UnRef(cond_removed);
		}
		// check if the extra condition is a comparison
		if (cond_removed->type()!=Item::FUNC_ITEM || ((Item_func*)cond_removed)->arg_count!=2)
			return false;
		// the right side of equality is the field of the original subselect
		if (dynamic_cast<Item_ref_null_helper*>(((Item_func*)cond_removed)->arguments()[1])==NULL)
			return false;
		field_for_subselect = NULL;
	}
	else if (!having
			|| (having->type()==Item::FUNC_ITEM
				&& (((Item_func*)having)->functype()==Item_func::ISNOTNULLTEST_FUNC
					|| ((Item_func*)having)->functype()==Item_func::TRIG_COND_FUNC)))
	{
		if (!conds) return false;
		if (conds->type()==Item::COND_ITEM && ((Item_cond*)conds)->functype()==Item_func::COND_AND_FUNC)
		{
			// if the condition is a conjunctive formula
			// the extra condition should be in the last argument
			if (((Item_cond*) conds)->argument_list()->elements < 2) return false;
			List_iterator<Item> li(*(((Item_cond*) conds)->argument_list()));
			while (li++ != NULL) cond_to_reinsert = *li.ref();
			li.rewind();
			while (*li.ref()!=cond_to_reinsert) li++;
			li.remove();
			list_to_reinsert = ((Item_cond*) conds)->argument_list();
			cond_removed = cond_to_reinsert;
		}
		else
		{
			// if no conjunctive formula the original condition was empty
			cond_removed = conds;
			conds = NULL;
		}
		if (cond_removed->type()==Item::FUNC_ITEM && ((Item_func*)cond_removed)->functype()==Item_func::TRIG_COND_FUNC)
		{
			// Condition was wrapped into trigger
			cond_removed = (Item_cond*) ((Item_func*)cond_removed)->arguments()[0];
		}
		if (cond_removed->type()==Item::COND_ITEM && ((Item_func*)cond_removed)->functype()==Item_func::COND_OR_FUNC)
		{
			// if the subselect field could have null values
			// equality condition was OR-ed with IS NULL condition
			Item_cond *cond_cond = (Item_cond*) cond_removed;
			List_iterator_fast<Item> li(*(cond_cond->argument_list()));
			cond_removed = li++;
			if (cond_removed==NULL) return false;
			if (li++ == NULL) return false;
			if (li++ != NULL) return false;
			// the original having was empty
			having = NULL;
		}
		// check if the extra condition is a comparison
		if (cond_removed->type()!=Item::FUNC_ITEM || ((Item_func*)cond_removed)->arg_count!=2)
			return false;
		// the right side of equality is the field of the original subselect
		field_for_subselect = ((Item_func*)cond_removed)->arguments()[1];
	}
	else return false;
	// the left side of equality should be the left side of the original expression with subselect
	Item* left_ref = ((Item_func*)cond_removed)->arguments()[0];
	if (dynamic_cast<Item_int_with_ref*>(left_ref)!=NULL)
		left_ref = ((Item_int_with_ref*)left_ref)->real_item();
	if (left_ref->type()!=Item::REF_ITEM
		|| ((Item_ref*)left_ref)->ref_type()!=Item_ref::DIRECT_REF
		|| ((Item_ref*)left_ref)->real_item()!=left_expr_for_subselect)
		return false;
	// set the operation type
	switch (((Item_func*)cond_removed)->functype())
	{
	case Item_func::EQ_FUNC:
		oper_for_subselect = O_IN; /*Operator::O_IN;*/
		break;
	case Item_func::NE_FUNC:
		oper_for_subselect = O_NOT_EQ;
		break;
	case Item_func::LT_FUNC:
		oper_for_subselect = O_LESS;
		break;
	case Item_func::LE_FUNC:
		oper_for_subselect = O_LESS_EQ;
		break;
	case Item_func::GT_FUNC:
		oper_for_subselect = O_MORE;
		break;
	case Item_func::GE_FUNC:
		oper_for_subselect = O_MORE_EQ;
		break;
	default:
		return false;
	}
	return true;
}

int Query::PrefixCheck(COND * conds)
//used for verification if all tables involved in a condition were already seen
//used only in assert macro
{
	conds = UnRef(conds);
	if(!conds)
		return RCBASE_QUERY_ROUTE;
	Item::Type cond_type = conds->type();
	FilterID filter;
	switch(cond_type) {
		case Item::COND_ITEM: {
			Item_cond* cond_cond = (Item_cond*)conds;
			Item_func::Functype func_type = cond_cond->functype();
			switch(func_type) {
				case Item_func::COND_AND_FUNC:
				case Item_func::COND_OR_FUNC: {
		//			List_iterator_fast<Item_equal> list_equal(((Item_cond_and*)cond_cond)->cond_equal.current_level);
					List_iterator_fast<Item> li(*(cond_cond->argument_list()));
					Item* item;
					while((item = li++) /*|| (item = list_equal++)*/) {
						int ret = PrefixCheck(item);
						if(!ret)
							return RETURN_QUERY_TO_MYSQL_ROUTE;
						if(ret == TABLE_YET_UNSEEN_INVOLVED)
							return TABLE_YET_UNSEEN_INVOLVED;
						//for RETURN_TO_RCBASE_ROUTE the next item is evaluated
					}
					break;
				}
				case Item_func::COND_XOR_FUNC:   //we don't handle xor as yet
				default:
					return -1;    //unknown function type
			}
			break;
		}
		case Item::FUNC_ITEM: {
			Item_func* cond_func = (Item_func*)conds;
			Item_func::Functype func_type = cond_func->functype();
			switch (func_type) {
				case Item_func::BETWEEN:
				case Item_func::LIKE_FUNC:
				case Item_func::ISNULL_FUNC:
				case Item_func::ISNOTNULL_FUNC:
				case Item_func::IN_FUNC:
				case Item_func::EQ_FUNC:            // =
				case Item_func::NE_FUNC:            // <>
				case Item_func::LE_FUNC:            // <=
				case Item_func::GE_FUNC:            // >=
				case Item_func::GT_FUNC:            // >
				case Item_func::LT_FUNC: {          // <
					for(int i = 0; (unsigned)i < cond_func->arg_count; i++) {
						Item** args = cond_func->arguments();
						Item* an_arg = UnRef(args[i]);
						int ret = PrefixCheck(an_arg);
						if(!ret)
							return RETURN_QUERY_TO_MYSQL_ROUTE;
						if(ret == TABLE_YET_UNSEEN_INVOLVED)
							return TABLE_YET_UNSEEN_INVOLVED;
					}
					break;
				}
				case Item_func::MULT_EQUAL_FUNC: {
					Item_equal_iterator li(*(Item_equal *)conds);
					Item_field* ifield;
					while(ifield = li++) {
						int ret = PrefixCheck(ifield);
						if(!ret)
							return RETURN_QUERY_TO_MYSQL_ROUTE;
						if(ret == TABLE_YET_UNSEEN_INVOLVED)
							return TABLE_YET_UNSEEN_INVOLVED;
					}
					break;
				}
				default:
					return RETURN_QUERY_TO_MYSQL_ROUTE;   //unknown function type
			}
			break;
		}
		case Item::FIELD_ITEM:  // regular select
		case Item::SUM_FUNC_ITEM: { // min(k), max(k), count(), avg(k), sum
			const char* field_alias = NULL;
			const char* table_alias = NULL;
			const char* database_name = NULL;
			const char* table_name = NULL;
			const char* table_path = NULL;
			const char* field_name = NULL;
			if(!FieldUnmysterify(conds, database_name, table_name, table_alias, table_path, field_name, field_alias))
				return RETURN_QUERY_TO_MYSQL_ROUTE;
			if(table_alias == EMPTY_TABLE_CONST_INDICATOR)  //if it is a constant - no reason to worry
				return RCBASE_QUERY_ROUTE;
			map<string, TABLE *>::const_iterator it_tab;
			it_tab = table_alias2ptr->find(table_alias);
			if(it_tab == table_alias2ptr->end())
				return TABLE_YET_UNSEEN_INVOLVED;
			else
				return RCBASE_QUERY_ROUTE;
			break;
		}
	}
	return RCBASE_QUERY_ROUTE;
}

//#include "sql_select.h"

int Query::BuildFilterIfPossible(COND *conds, const TabID& tmp_table, JoinType join_type)
{
	conds = UnRef(conds);
	if(conds) {
		FilterType filter_type = (join_type == JO_LEFT ? ON_LEFT_FILTER : (join_type == JO_RIGHT ? ON_RIGHT_FILTER : ON_INNER_FILTER));
		// in case of Right join MySQL changes order of tables. Right must be switched back to left!
		if(filter_type == ON_RIGHT_FILTER)
			filter_type = ON_LEFT_FILTER;
		assert(PrefixCheck(conds) != TABLE_YET_UNSEEN_INVOLVED && "Table not yet seen was involved in this condition");

		bool zero_result = conds->type() == Item::INT_ITEM && !conds->val_bool();
		if(!BuildFilter(conds, cq, table_alias2index, table_alias2ptr, tmp_table, filter_type, zero_result, join_type))
			return RETURN_QUERY_TO_MYSQL_ROUTE;
		conds = 0;
	}
	return RCBASE_QUERY_ROUTE;
}

