/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef _BINTOOLS_H_
#define _BINTOOLS_H_

#include <boost/shared_ptr.hpp>
#include <boost/assign/std/vector.hpp>
#include <vector>

#include "common/CommonDefinitions.h"

/////// Number of "1"-s in 32-bit value //////////////////////////////

char CalculateBinSum(unsigned int n);
int CalculateBinSize(unsigned int n);
int CalculateBinSize(_uint64 n);
int CalculateByteSize(_uint64 n);

/////// Text / _int64 conversions //////////////////////////////

void DateToString(const int date, char *val_buf, uint buf_size);
void TimeToString(const _int64 time, char *val_buf, uint buf_size, int precision);
double DecToDouble(const _int64 value, int scale);
//char* Text(_int64 value, char* buf, int scale);
_int64 AtoDec(char *b,int scale);

/////// Additional tools ////////////////////////////////////////////extern ////
_int64  FloatMinusEpsilon(_int64 v);		// modify the number to the nearest down (precision permitting)
_int64  FloatPlusEpsilon(_int64 v);		// modify the number to the nearest up (precision permitting)
_int64  FloatMinusEpsilon(_int64 v);		// modify the number to the nearest down (precision permitting)
_int64  FloatPlusEpsilon(_int64 v);		// modify the number to the nearest up (precision permitting)
_int64  DoubleMinusEpsilon(_int64 v);	// modify the number to the nearest down (precision permitting)
_int64  DoublePlusEpsilon(_int64 v);		// modify the number to the nearest up (precision permitting)
_int64  MonotonicDouble2Int64(_int64 d);	// encode double value (stored bitwise as _int64) into _int64 to prevent comparison directions
_int64  MonotonicInt642Double(_int64 d);	// reverse encoding done by MonotonicDouble2Int64

_int64 SafeMultiplication(_int64 x, _int64 y);	// return a multiplication of two numbers or NULL_VALUE_64 in case of overflow

uint	CRC32(unsigned char const* buf, int n);	// CRC-32 checksum for a given memory buffer, n - number of bytes
#define HASH_FUNCTION_BYTE_SIZE 16
void HashMD5( unsigned char const* buf, int n, unsigned char* hash );

/*
class Nullable<T>		// the class containing value and an information about nulls (e.g. any 64-bit value, incl. the reserved one)
{
public:
	Nullable<T>()			:n(0),is_null(true) {};
	Nullable<T>(T v)		:n(v),is_null(false) {};

	T		n;
	bool	is_null;
};
*/
/////// Table of integer values of arbitrary bit precision (up to 32) ///

class FreeBinTable		// Assumption: table may contain more than 2^32 bits!
						// Up to 2^32 * 32 bits (4 GB)
{
public:
	FreeBinTable(int _b,_uint64 _n);	// bit rate (up to 64 for _uint64), table size
	FreeBinTable(FreeBinTable &sec);	// copy all
	~FreeBinTable();

	_uint64 N()						// number of values in table
	{ return n; }
	unsigned int B()				// bit rate
	{ return b; }
	unsigned int BinSize()			// table size in bytes
	{ return size*4; }
	char *Table()					// table itself
	{ return (char *)t; }
	void ExpandTo(_uint64 _n);		// enlarge the table to the size _n, copying existing values and set 0 to others

	void Set64(_uint64 k, const _uint64 &v);
	_uint64 Get64(_uint64 k);
////// These two functions should be  (HERE, NOT IN .cpp),
	// because they are 30% faster.
	void Set(_uint64 k,unsigned int v)
	{
		assert(b<=32 && b>0);
		assert(k<n);
		b1=(unsigned int)((k*b)/32);
		s1=(unsigned int)((k*b)%32);
		if(s1+b<=32)		// can we use one int?
		{
			// create mask for deleting old value
			mask=0xFFFFFFFF>>(32-b);	// = 000000000...0011<b>11			>> and << safe as b>0
			mask=mask<<(32-s1-b);	// = 00<s1>0011<b>110...00
			t[b1]=t[b1]&(~mask);	// delete old value
			mask=v<<(32-s1-b);		// = 00<s1>00[..v..]0...00
			t[b1]=t[b1]|mask;		// set new value
		}
		else
		{
			// create mask for deleting old value - part one
			mask=0xFFFFFFFF>>s1;	// = 00<s1>000...0011<.>11				>> safe
			t[b1]=t[b1]&(~mask);	// delete old value
			mask=v>>(b-32+s1);		// = 00<s1>00000000[..v..]				>> safe
			t[b1]=t[b1]|mask;		// set new value - part one
			// create mask for deleting old value - part two
			s1=(s1+b)%32;			// s1>0, so << safe
			mask=0xFFFFFFFF<<(32-s1);
										// = 11<s1>1100000000...00
			t[b1+1]=t[b1+1]&(~mask);	// delete old value
			mask=v<<(32-s1);			// = [..v...]00000000...00
			t[b1+1]=t[b1+1]|mask;		// set new value - part two
		}
	};

	unsigned int Get(_uint64 k)
	{
		assert(b<=32 && b>0);
		assert(k<n);
		vv=0;
		b1=(unsigned int)((k*b)/32);
		s1=(unsigned int)((k*b)%32);
		if(s1+b<=32)		// can we use one int?
		{
			vv=t[b1]>>(32-s1-b);							// >> and << safe
			if(b<32) vv=vv%((unsigned int)(1)<<b);
		}
		else
		{
			vv=t[b1]%((unsigned int)(1)<<(32-s1));			// s1>0, so << and >> safe
			s1=(s1+b)%32;
			vv=(vv<<s1)+(t[b1+1]>>(32-s1));
		}
		return vv;
	};
////////////////////////////////////////////////////////////////////

private:
	int b;				// number of bits per value
	int blow,bhi;		// as above, for 64-bit table, blow=(b or 32), bhi=(0 or b-32)
	_uint64 n;			// number of values in table
	unsigned int size;	// number of ints in table
	unsigned int *t;	// table itself

	// technical:
	unsigned int b1,s1,mask,vv;
	_uint64 v64;
};

//////////////////////////////////////////////////////////////////////
////////////////   Binary matrix - arbitrary  ////////////////////////

class BinMatrix
{
public:
	BinMatrix(int size_x,int size_y, bool v = false);
	~BinMatrix();

	void Resize(int new_size_x, int new_size_y, bool v = false);

	void Set(bool v);									// set all to the same value
	void Set(int x,int y,bool v)			// optimal search order: for(x...) for(y...) Set(x,y)
	{
		if(v)	t[x*y_ints+y/32] |=  (0x00000001u<<(y%32));	// set 1		// << and >> safe (never 32)
		else	t[x*y_ints+y/32] &= ~(0x00000001u<<(y%32));	// set 0
	};

	bool Get(int x,int y) const
	{
		return ( (t[x*y_ints+y/32]>>(y%32)) & 0x00000001 )!=0 ;
	};

	void And(BinMatrix &sec);			// must have identical size!
	double Density();					// number of ones (percentage); 1 - full
	bool SumIsFull(BinMatrix &sec);	// true if there are no common 0; must have identical size!
	bool IsFull();

	/////////////////// Unit testing //////////////////////////////
	void Display();

	/////////////////// For e.g. loading/saving: //////////////////
	void *Buf()			{ return t; }
	int BufSize()		{ return x_rows*y_ints*sizeof(uint); }

	int NoRowsY() const		{ return y_rows; }
	int NoRowsX() const		{ return x_rows; }

private:
	uint *t;					// table itself
	int y_ints,x_rows,y_rows;	// size of table (y in ints, x in rows)
};

class BinDecomposer
{
public:
	class JoiningGroup
	{
		friend class BinDecomposer;
	public:
		JoiningGroup();
		void Rewind(short dim);
		int GetNext(short dim);
		int NoPacks(short dim);
	private:
		void AddPack(short dim, int pack_no);
	private:
		int iters[2];
		std::vector<int> t[2];
	};
	typedef boost::shared_ptr<JoiningGroup> JoinGroupPtr;
public:
	BinDecomposer(const BinMatrix& bm);
	bool NextGroup();
	void RewindPack(short dim);
	int GetNextPack(short dim);
	int NoPacks(short dim);
private:
	std::vector<JoinGroupPtr> joining_groups;
	int joining_groups_iter;
};

#endif

