/*************************************************************************
 */

static FILE* f_dump = 0;
static const char* nameOf[] = {
            "FIELD_ITEM", "FUNC_ITEM", "SUM_FUNC_ITEM", "STRING_ITEM", "INT_ITEM",
            "REAL_ITEM", "NULL_ITEM", "VARBIN_ITEM", "COPY_STR_ITEM", "FIELD_AVG_ITEM",
            "DEFAULT_VALUE_ITEM", "PROC_ITEM", "COND_ITEM", "REF_ITEM", "FIELD_STD_ITEM",
            "FIELD_VARIANCE_ITEM", "INSERT_VALUE_ITEM", "SUBSELECT_ITEM", "ROW_ITEM",
            "CACHE_ITEM", "TYPE_HOLDER", "PARAM_ITEM", "TRIGGER_FIELD_ITEM", "DECIMAL_ITEM",
            "XPATH_NODESET", "XPATH_NODESET_CMP", "VIEW_FIXER_ITEM"};

static const char* sumNameOf[] =  { 
    "COUNT_FUNC", "COUNT_DISTINCT_FUNC", "SUM_FUNC", "SUM_DISTINCT_FUNC", "AVG_FUNC",
    "AVG_DISTINCT_FUNC", "MIN_FUNC", "MAX_FUNC", "STD_FUNC",
    "VARIANCE_FUNC", "SUM_BIT_FUNC", "UDF_SUM_FUNC", "GROUP_CONCAT_FUNC" };


class Item_bhwrapper : public Item_func
{
public:
    Item_field* the_field;
    bool* gpq;
    CHARSET_INFO *cs;

    enum { BHWRAPPER_FUNC = 12345 };
    virtual Functype functype() const   { return (Functype) BHWRAPPER_FUNC; }

    virtual double val_real();
    virtual longlong val_int();
    virtual String* val_str(String* s);
    virtual my_decimal *val_decimal(my_decimal *d);

    virtual void fix_length_and_dec() {}
    virtual const char* func_name() const { return "bhwrapper"; }
};

double Item_bhwrapper::val_real() 
{ 
    return *gpq ? the_field->val_real() : 1.0; 
}

longlong Item_bhwrapper::val_int() 
{ 
    return *gpq ? the_field->val_int() : 1L; 
}

String* Item_bhwrapper::val_str(String* s) 
{ 
    if ( *gpq ) {
        String* result = the_field->val_str (s);
        return result;
    }

    s->set_charset(cs);
    String* val_ptr = & str_value;
    val_ptr->set((const char*) "abcde", 5, cs);
    return val_ptr;
}

my_decimal* Item_bhwrapper::val_decimal(my_decimal *d) 
{
    if ( *gpq ) return the_field->val_decimal (d);
    int2my_decimal( E_DEC_FATAL_ERROR, 1L, 0, d);
    return d;
}

static void printSelectLex (THD* thd, bool* gpq, SELECT_LEX* lex);
static void coverFields (THD* thd, bool* gpq, Item* item);


static const char* getName (uint index)
{
    if (index >= sizeof(nameOf)/sizeof(*nameOf)) return "<UNKNOWN>";
    return nameOf[index];
}

static const char* getSumName (uint index)
{
    if (index >= sizeof(sumNameOf)/sizeof(*sumNameOf)) return "<UNKNOWN>";
    return sumNameOf[index];
}

static void coverFields (THD* thd, bool* gpq, Item** args, uint args_count)
{
    for (uint i=0; i < args_count; i++) {
        if (args[i]->type() == Item::FIELD_ITEM) {
            Item_bhwrapper *bhw = new (thd->mem_root) Item_bhwrapper;
            bhw->the_field = static_cast <Item_field*> (args[i]);
            bhw->gpq = gpq;
            bhw->cs = system_charset_info;
            args[i] = bhw;
        }
        else {
            coverFields (thd, gpq, args[i]);
        }
    }
}

static void coverFields (THD* thd, bool* gpq, Item* item)
{
    if (item == 0) return;

    Item::Type type = item->type();

    if (type == Item::FUNC_ITEM) {
        Item_func* func = static_cast <Item_func*> (item);
        coverFields (thd, gpq, func->arguments(), func->argument_count());
    }
    else if  (type == Item::SUM_FUNC_ITEM) {
        Item_sum* sum_func = static_cast <Item_sum*> (item);
        coverFields (thd, gpq, sum_func->args, sum_func->arg_count);
    }
}


static void printItemTree (THD* thd, bool* gpq, Item* item, int indent)
{
    if (item == 0) return;

    for (int j=0; j < indent; j++) fprintf (f_dump, "*");

    Item::Type type = item->type();

    if (type == Item::FUNC_ITEM) {
        Item_func* func = static_cast <Item_func*> (item);

        if (func->functype() == (Item_func::Functype)Item_bhwrapper::BHWRAPPER_FUNC ) {
            Item_bhwrapper* bhw = static_cast <Item_bhwrapper*> (item);
            fprintf (f_dump, " BHWrapper item \n"); fflush  (f_dump);
            for (int j=0; j < indent+2; j++) fprintf (f_dump, "*");
            fprintf (f_dump, " %s %s \n", getName (bhw->the_field->type()), bhw->the_field->full_name()); fflush  (f_dump);
        }
        else {
            Item** args = func->arguments();
            uint args_count = func->argument_count();

            const char* s = "UNKNOWN_RESULT";
            switch (item->result_type()) {
                case STRING_RESULT: s = "STRING_RESULT"; break;
                case INT_RESULT: s = "INT_RESULT"; break;
                case REAL_RESULT: s = "REAL_RESULT"; break;
                case DECIMAL_RESULT: s = "DECIMAL_RESULT"; break;
            }
            fprintf (f_dump, " Function: %s   type=[%s]\n", func->func_name(), s); fflush  (f_dump);
 
            for (uint i=0; i < args_count; i++) printItemTree (thd, gpq, args[i], indent+2);

            if (indent == 2) {
                bool save_gpq = *gpq;
                *gpq = false;
                fprintf (f_dump, "\n>>>>> Executing function. Result: ");

                String str, *strp;
                double x;
                my_decimal mydec;

                switch (item->result_type()) {
                    case STRING_RESULT: 
                        strp = func->val_str (& str);
                        fprintf (f_dump, "(string) [%s]\n", strp->c_ptr_safe());
                        break;
                    case REAL_RESULT:
                        fprintf (f_dump, "(real) %.2f\n", func->val_real());
                        break;
                    case INT_RESULT:
                        fprintf (f_dump, "(int) %d\n", func->val_int());
                        break;
                    case DECIMAL_RESULT:
                        my_decimal2double ((uint)-1, func->val_decimal( & mydec ), & x);
                        fprintf (f_dump, "(decimal)  %.2f\n", x);
                        break;
                }
                fflush (f_dump);

                *gpq = save_gpq;
            }
        }

        return;
    }

    if  (type == Item::SUM_FUNC_ITEM) {
        Item_sum* sum_func = static_cast <Item_sum*> (item);

        Item** args = sum_func->args;
        uint args_count = sum_func->arg_count;

        fprintf (f_dump, " SumFunction: %s \n", getSumName (sum_func->sum_func())); fflush  (f_dump);

        for (uint i=0; i < args_count; i++) printItemTree (thd, gpq, args[i], indent+2);
        return;
    }

    if (type == Item::REF_ITEM) {
        Item_ref* ref = static_cast<Item_ref*> (item);
        Item* real = ref->real_item();

        fprintf (f_dump, " Reference item (0x%u)\n", item); fflush  (f_dump);

        if (ref != real) printItemTree (thd, gpq, real, indent+2);
        return;
    }

    if (type == Item::SUBSELECT_ITEM) {
        Item_subselect* sub = static_cast <Item_subselect*> (item);
        SELECT_LEX* sl = sub->get_select_lex();

        fprintf (f_dump, " Subselect item (0x%u)\n", item); fflush  (f_dump);
        printSelectLex (thd, gpq, sl);
        return;
    }

    fprintf (f_dump, " %s %s \n", getName (item->type()), item->full_name()); fflush  (f_dump);
}

static void printSelectLex (THD* thd, bool* gpq, SELECT_LEX* lex)
{
    Item* item;
    JOIN* join;
    ORDER* ord;

    for (SELECT_LEX* sl=lex; sl; sl = sl->next_select()) {
        join = sl->join;
        if (join == 0) continue;

        fprintf (f_dump, "\n--------- Select -------------\n"); fflush (f_dump);
        List_iterator<Item> li (join->fields_list);
        while ( (item = li++) ) {
            coverFields (thd, gpq, item);
            printItemTree (thd, gpq, item, 2);
        }

        fprintf (f_dump, "\n--------- Where -------------\n"); fflush (f_dump);
        printItemTree (thd, gpq, join->conds, 2);

        fprintf (f_dump, "\n--------- GroupBy -------------\n"); fflush (f_dump);
        for (ord = join->group_list; ord; ord = ord->next) printItemTree (thd, gpq, ord->item_ptr, 2);

        fprintf (f_dump, "\n--------- Having -------------\n"); fflush (f_dump);
        printItemTree (thd, gpq, sl->having, 2);

        fprintf (f_dump, "\n--------- Order -------------\n"); fflush (f_dump);
        for (ord = join->order; ord; ord = ord->next) printItemTree (thd, gpq, ord->item_ptr, 2);

        fprintf (f_dump, "\n===============================\n"); fflush (f_dump);
    }
}

static void traverseAndPrint (THD* thd, SELECT_LEX* lex)
{
    if (f_dump) { fclose (f_dump); f_dump = 0; }
    f_dump = fopen ("/tmp/f_dump.txt", "w");
    if (f_dump == 0) return;

    bool* gpq = (bool*)  alloc_root(thd->mem_root, sizeof(bool));
    *gpq = true;

    printSelectLex (thd, gpq, lex);
}


