/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef _ENGINE_BHQUERY_H_
#define _ENGINE_BHQUERY_H_

#include "Expression.h"
#include "Condition.h"
#include "DataSource.h"
#include "common/bhassert.h"

namespace na
{

	struct TableDescriptor
	{
		TABLE*			mytable;	// pointer to MySQL object which represents the _occurence_ of a table in a Select (for resolving)
		JustATable*		bhtable;
		DataSource*		dstable;	// DSTable created in TransformToTree()

		TableDescriptor()					{ mytable = NULL; bhtable = NULL; dstable = NULL; }
		DataSource* TransformToTree();
	};

	class BHQuery : public BHItem
	{
	public:
		BHQuery()						{ from = NULL; where = NULL; }
		virtual ~BHQuery()				{ delete where; delete from; for(size_t i=0; i<select.size(); i++) delete select[i]; }
		virtual bool IsBHQuery()		{ return true; }

		void SetFrom(TableDescriptor* tab)
		{
			BHASSERT(!from, "??? 'from' should be null");
			delete from;
			from = tab;
		}
		void AddSelect(Expression* expr)			{ select.push_back(expr); }

		TableDescriptor* FindTable(TABLE* mytab)	{ return (from->mytable==mytab) ? from : NULL; }

		// The tree is allocated inside and must be deallocated OUTside.
		// Some structures (expressions, conditions...) are moved from BHQuery to the tree.
		DSSelect* TransformToTree();

	private:
		std::vector<Expression*>	select;
		TableDescriptor*			from;
		Condition*					where;

		friend class QueryCompiler;
	};


}



#endif
