/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef _ENGINE_EXPRESSION_H_
#define _ENGINE_EXPRESSION_H_

#include "BHItem.h"
#include "../common/CommonDefinitions.h"
#include "../types/AValue.h"
namespace na {

struct TableDescriptor;



// Base class for all expressions other than logical (which are called 'conditions').
class Expression : public BHExprCond
{
public:
	virtual bool IsExpression()		{ return true; }
	
	AValue* GetBuffer()	
	{ 
		BHASSERT(val, "'val' should not be null"); 
		return val; 
	}
	virtual void Evaluate()	
	{ 
		BHERROR("method should not be invoked"); 
	} // computes result and stores in 'val'

protected:
	Expression()		:val(NULL) {}
	AValue* val;
};


// Represents variable (i.e. column)
class ExprVar : public Expression
{
public:
	ExprVar(TableDescriptor* t, int c);
	virtual void Evaluate()					{}			// nothing has to be done - the variable value must already reside in 'val' buffer

protected:
	virtual int Proc_LinkVarsToDS(void*);
	virtual int Proc_GetVariables(void* vars)			{ ((SetOfVars*)vars)->insert(Variable(dstab,bhtab,col));	return 0; }
	virtual int Proc_SetBuffers(void* bufs)				{ val = (*(BufOfVars*)bufs)[Variable(dstab,bhtab,col)];	return 0; }

	// descriptor of a column: link to TableDescriptor object maintained by BHQuery,
	// which is then translated into a pointer to DataSource (usually DSTable, maybe DSSelect representing a subquery)
	TableDescriptor*	tabdesc;
	JustATable*			bhtab;
	DataSource*			dstab;			// set by LinkVarsToDS()
	int					col;			// column no. in JustATable
};


// A constant
class ExprConst : public Expression
{
public:
	ExprConst(AValue* v) 
	{ 
		BHASSERT(v, "'v' should not be null"); 
		val = v; 
	}
	virtual void Evaluate()		{}
};


//class ExprArith : public Expression
//{
//	ArithOperator	oper;
//	Expression*		left;
//	Expression*		right;
//};

// Unary minus (negation) operator
class ExprArithNeg : public Expression
{
public:
	ExprArithNeg(Expression* a)		:arg(a) 
	{ 
		BHASSERT(a, "'a' should not be null");; 
	}
	virtual void Evaluate()			{ VInteger::OperNeg(val,argv); }

protected:
	Expression*		arg;
	AValue*			argv;

	virtual int Walk(ItemProcess proc, void* par, bool sub)		{ if(arg->Walk(proc,par,sub)) return 1;   return (this->*proc)(par); }
	virtual int Proc_SetBuffers(void*)							{ val = new VInteger; argv = arg->GetBuffer();   return 0; }
};

// Base class for 2-argument operators: + - * /
class ExprArith2 : public Expression
{
protected:
	Expression		*arg1, *arg2;
	AValue			*arg1v, *arg2v;
};



}	// namespace na
#endif

