/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "TxtDataFormat.h"
#include "shared/exporter/DataExporterTxt.h"
#include "types/ValueParserForText.h"

using namespace std;

TxtDataFormat::TxtDataFormat()
	:	DataFormat("TXT_VARIABLE", TXT_VARIABLE)
{
}

DataExporterAutoPtr TxtDataFormat::CreateDataExporter(uchar delimiter, uchar string_qualifier, string nulls_str) const
{
	return DataExporterAutoPtr(new RCDEforTxtVariable(delimiter, string_qualifier, nulls_str));
}

ValueParserAutoPtr TxtDataFormat::CreateValueParser() const
{
	return ValueParserAutoPtr(new ValueParserForText());
}

//! Number of bytes taken by a value when written out,
//! or number of chars if collation not binary and RC_STRING or RC_VARCHAR
uint TxtDataFormat::ExtrnalSize(AttributeType attrt, int precision, int scale, DTCollation* col) const
{
	if(attrt == RC_NUM)
		return 1 + precision + (scale ? 1 : 0) + (precision == scale ? 1 : 0);
	// because "-0.1" may be declared as DEC(1,1), so 18 decimal places may take 21 characters
	if(attrt == RC_BYTE || attrt == RC_VARBYTE || attrt == RC_BIN) return precision * 2;
	if(attrt == RC_TIME_N || attrt == RC_DATETIME_N) return precision + scale + 1;
	if(attrt == RC_TIME)
		return 10;
	if(attrt == RC_DATETIME || attrt == RC_TIMESTAMP)
		return 19;
	if(attrt == RC_DATE)
		return 10;
	else if(attrt == RC_INT)
		return 11;
	else if(attrt == RC_BIGINT)
		return 20;
	else if(attrt == RC_BYTEINT || attrt == RC_YEAR)
		return 4;
	else if(attrt == RC_SMALLINT)
		return 6;
	else if(attrt == RC_MEDIUMINT)
		return 8;
	else if(attrt == RC_REAL)
		return 23;
	else if(attrt == RC_FLOAT)
		return 15;
	if (col == NULL)
		return precision;
	else
		return precision / col->collation->mbmaxlen;
}

ushort TxtDataFormat::ObjPrefixSizeByteSize(AttributeType attr_type) const
{
	return 0;
}

bool TxtDataFormat::IsVariableSize(AttributeType attr_type) const
{
	if(attr_type == RC_VARCHAR || attr_type == RC_VARBYTE || attr_type == RC_BIN)
		return true;
	return false;
}

