/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef _DATA_EXPORTERTXT_H_
#define _DATA_EXPORTERTXT_H_

#include "shared/exporter/DataExporter.h"

class DEforTxt : public DataExporter
{
	public:
		virtual void PutNull() {
			WriteNull();
			WriteValueEnd();
		}
		virtual void PutText(const RCBString& str) {
			WriteStringQualifier();
			int char_len = deas[cur_attr].GetCollation().collation->cset->numchars(
					deas[cur_attr].GetCollation().collation, str.val, str.val + str.len); //len in chars
			WriteString(str, str.len); //len in bytes
			if((deas[cur_attr].Type() == RC_STRING) && (char_len < deas[cur_attr].CharLen()))
				// it can be necessary to change the WritePad implementation to something like:
				// collation->cset->fill(cs, copy->to_ptr+copy->from_length, copy->to_length-copy->from_length, ' ');
				// if ' ' (space) can have different codes.
				WritePad(deas[cur_attr].CharLen() - char_len);
			WriteStringQualifier();
			WriteValueEnd();
		}
		virtual void PutBin(const RCBString& str) {
			int len = (int)strlen(str);
		//if((rcdea[cur_attr].attrt == RC_BYTE) && (len < rcdea[cur_attr].size))
		//	len = rcdea[cur_attr].size;
			if(len > 0) {
				char* hex = new char[len*2];
				Convert2Hex((const unsigned char*)str.val, len, hex, len*2, false);
				WriteString(hex, len*2);
				delete[] hex;
			}
			WriteValueEnd();
		}
		virtual void PutNumeric(_int64 num) {
			RCNum rcn(num, deas[cur_attr].Scale(), ATI::IsRealType(deas[cur_attr].Type()), deas[cur_attr].Type());
			RCBString rcs = rcn.ToRCString();
			WriteString(rcs);
			WriteValueEnd();
		}

		virtual void PutRowEnd() {
			WriteString("\r\n", 2);
		}

	protected:
		DEforTxt(uchar delim, uchar str_q, std::string nulls_str)
			:	delim(delim), str_q(str_q), nulls_str(nulls_str) {}

		void WriteStringQualifier()		{ buf->WriteIfNonzero(str_q); }
		void WriteDelimiter()			{ buf->WriteIfNonzero(delim); }
		void WriteNull()				{ WriteString(nulls_str.c_str(), (int)nulls_str.length()); }

		void WriteString(const char* str)			{ WriteString(str, (int)strlen(str)); }
		void WriteString(const char* str, int len)	{ strncpy(buf->BufAppend((uint)len), str, (size_t)len); }
		void WriteString(const RCBString& str)			{ WriteString(str, (int)strlen(str)); }
		void WriteString(const RCBString& str, int len)	{ strncpy(buf->BufAppend((uint)len), str, (size_t)len); }
		void WriteChar(char c, uint repeat = 1)		{ memset(buf->BufAppend(repeat), c, repeat); }
		void WritePad(uint repeat)					{ WriteChar(' ', repeat); }

		void WriteValueEnd() {
			if(cur_attr == no_attrs - 1)
				cur_attr = 0;
			else {
				WriteDelimiter();
				cur_attr++;
			}
		}

		uchar	delim, str_q;
		std::string nulls_str;
};

class RCDEforTxtVariable : public DEforTxt
{
public:
	RCDEforTxtVariable(uchar delim, uchar str_q, std::string nulls_str)
		:	DEforTxt(delim, str_q, nulls_str) {}

	virtual void PutDateTime(_int64 dt) {
		RCDateTime rcdt(dt, deas[cur_attr].Type());
		RCBString rcs = rcdt.ToRCString();
		WriteString(rcs);
		WriteValueEnd();
	}
};

#endif
