/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */
//#define USEVALGRIND

#ifdef __GNUC__
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#endif
#include <cstdlib>
#include <ctime>
#include <string>
#include <fstream>

#include "system/RCSystem.h"
#include "BHLoader.h"
#include "BHLoaderApp.h"
#include "common/bhassert.h"
#include "system/SharedMemOut.h"
#include "system/MemoryManagement/Initializer.h"
#include "shared/loader/RCTable_load.h"
#include "core/tools.h"
#include "system/IBFileSystem.h"
#include "system/ib_system.h"

using namespace std;
using namespace bh;

BHLoader::~BHLoader()
{
}

bool BHLoader::Proceed(pid_t& pi)		// return true on success otherwise return false
{
	BHASSERT(!mode, "'mode' should be false");

	std::string shared_obj = GetSystemTempPath();
	shared_obj += shared_obj_name;
    std::string report_file_name = shared_obj + BHLOADER_REPORT_EXTENSTION;

	//iop->PutParameters(IOParamsBuffer(), shared_mem_size - (int)(IOParamsBuffer() - shared_mem));
	ReturnCode()= BHERROR_UNKNOWN;

	try {
		shared_file.OpenCreate(shared_obj.c_str());
	} catch(DatabaseRCException& e)	{
		string err_msg= "Unable to start Loader. Trying to create a memory mapped file but it says - ";
		err_msg+= e.what();
		throw InternalRCException(err_msg.c_str());
	}

    FlushParameters();  // write loader parameters to file
#ifndef RCLOAD_STANDALONE
	{
		TemporalValueReplacement<ProcessType::enum_t> t(process_type, ProcessType::BHLOADER);
		ReturnCode() = (BHErrorCode)LoadData(shared_obj_name, shared_obj);
		ReLoadParameters();  // read loader results from file
		return (ReturnCode() == BHERROR_SUCCESS);
	}
#endif

    IBProcess ibpr;
    int is_finished, is_terminated;
    int return_code = -128;

    try {
        ibpr.Start(bhload_app.c_str(), shared_obj_name.c_str(), shared_obj.c_str());
        pi = (pid_t) ibpr.GetProcess();
        ibpr.Wait(is_finished, is_terminated, return_code);
        ReLoadParameters();  // read loader results from file
    }
    catch(...) {
        shared_file.Close();
        RemoveFile(report_file_name.c_str(),false);
        RemoveFile(shared_obj.c_str(),false);
        throw;
    }

    std::ifstream is( report_file_name.c_str() );
    if ( is ) {
       std::string line;
       while(std::getline(is, line).good())
          rclog << lock << line << unlock;
       is.close(); 
    }
    RemoveFile(report_file_name.c_str(), false);
    DeInit(true);

	if ((return_code < 0 || return_code > BHERROR_ERROR_LIMIT) && is_terminated) {
		ReturnCode() = BHERROR_UNKNOWN;
		std::string err= "Load process was terminated abnormally. This could happen if someone killed the bhloader process, killed the "\
"running load or due to unexpected internal error inside bhloader process.";
		strcpy(ErrorMessage(), err.c_str());
		return false;
	}

	return (ReturnCode() == BHERROR_SUCCESS);
}

#if 0
This part of code is not compiled
bool BHLoader::Proceed(pid_t& pi)		// return true on success otherwise return false
{
	BHASSERT(!mode, "'mode' should be false");
	std:string err = "";
	char* eventname_for_loader = new char[shared_event_name.length()+2];
	strcpy(eventname_for_loader, shared_event_name.c_str());
	strcat(eventname_for_loader, "a");

	key_t sem_ipc_key;
	union semun semctl_val;

	std::string shared_obj;

	shared_obj = TMP_FOLDER;
	shared_obj += shared_obj_name;

	if ((sem_ipc_key = ftok(shared_obj.c_str(), 1)) == (key_t) -1)
	{
		ReturnCode()= BHERROR_UNKNOWN;
		err= "Unable to start BHLoader. It says - ";
		err+= strerror(errno);
		strcpy(ErrorMessage(), err.c_str());
		delete [] eventname_for_loader;
		return false;
	}

    if ((sem_ipc = semget(sem_ipc_key,2,IPC_CREAT | S_IRUSR |
		S_IWUSR | S_IROTH | S_IWOTH)) < 0)
    {
	  ReturnCode()= BHERROR_UNKNOWN;
	  err= "Unable to start BHLoader. It says - ";
	  err+= strerror(errno);
	  strcpy(ErrorMessage(), err.c_str());
	  delete [] eventname_for_loader;
	  return false;
    }

	semctl_val.val = 0;
    if (semctl(sem_ipc, 0, SETVAL, semctl_val)==-1)
	{
	  ReturnCode()= BHERROR_UNKNOWN;
	  err= "Unable to start BHLoader. It says - ";
	  err+= strerror(errno);
	  strcpy(ErrorMessage(), err.c_str());
	  delete [] eventname_for_loader;
	  return false;
	}

    if (semctl(sem_ipc, 1, SETVAL, semctl_val) == -1)
	{
	  ReturnCode()= BHERROR_UNKNOWN;
	  err= "Unable to start BHLoader. It says - ";
	  err+= strerror(errno);
	  strcpy(ErrorMessage(), err.c_str());
	  delete [] eventname_for_loader;
	  return false;
	}

	iop->PutParameters(IOParamsBuffer(), shared_mem_size - (int)(IOParamsBuffer() - shared_mem));
	ReturnCode()= BHERROR_UNKNOWN;

#ifndef RCLOAD_STANDALONE
	ReturnCode() = (BHErrorCode)LoadData(shared_obj_name, shared_obj);
#else

    std::string bhloader_exec;
    pthread_t thd_waitfor_loader;

  	struct stat buf;
  	if(stat(bhload_app.c_str(), &buf) == -1)
  	{
  		ReturnCode() = BHERROR_CANNOT_OPEN_FILE_OR_PIPE;
	    err= "Unable to start BHLoader. It says - BHLoader does not exist in ";
		err+= + bhload_app.c_str();
		strcpy(ErrorMessage(), err.c_str());
		delete [] eventname_for_loader;
		return false;
  	}

	pid_t pid = 0;
#ifdef JBDEBUG
	char cmdline[500];
	sprintf(cmdline,"ddd /home/janb/workspace/brighthouse/sql/bhloader --args /home/janb/workspace/brighthouse/sql/bhloader %s %s\n",shared_obj_name.c_str(), shared_obj.c_str());
	system(cmdline);
#else
	#ifdef USEVALGRIND
	char cmdline[500];
	sprintf(cmdline,"ddd /home/janb/workspace/brighthouse/sql/bhloader --args /home/janb/workspace/brighthouse/sql/bhloader %s %s\n",shared_obj_name.c_str(), shared_obj.c_str());
	system(cmdline);
	#endif
#endif
	if((pid = fork()) == 0)
	{
		// child process
		// execute bhloader, which replaces the child process with bhloader process;
		// this function returns only in case of error
		execl(bhload_app.c_str(), bhload_app.c_str(), shared_obj_name.c_str(), shared_obj.c_str(), (char*)NULL);
		exit(BHERROR_UNKNOWN);
	}
	else if(pid == -1)
	{
		ReturnCode()= BHERROR_UNKNOWN;
		err= "Unable to start BHLoader. It says - ";
		err+= strerror(errno);
		strcpy(ErrorMessage(), err.c_str());
		delete [] eventname_for_loader;
		return false;
	}

	pi = pid;

	struct sembuf op_wait = {0, -1, SEM_UNDO | IPC_WAIT};
	struct sembuf op_signal = {1, 1, SEM_UNDO | IPC_WAIT};

	std::string shmdata;
	int status;
	timespec ts;
	ts.tv_sec = 0;
	ts.tv_nsec = 50000;
	while (true)
	{
		if(waitpid(pid, &status, WNOHANG) == 0)
		{
			if (semtimedop(sem_ipc, &op_wait, 1, &ts))
			{
				if(errno != EAGAIN && ReturnCode() == BHERROR_UNKNOWN)
				{
					ReturnCode() = BHERROR_UNKNOWN;
					err= "Unable to start BHLoader. It says - ";
					err+= strerror(errno);
					strcpy(ErrorMessage(), err.c_str());
					delete [] eventname_for_loader;
					return false;
				}
			}
			else
			{
				shmdata= RCLogMessage();
				rclog << lock << shmdata.c_str() << unlock;
				if (semop(sem_ipc, &op_signal, 1))
				{
					ReturnCode()= BHERROR_UNKNOWN;
					err= "Unable to start BHLoader. It says - ";
					err+= strerror(errno);
					strcpy(ErrorMessage(), err.c_str());
					delete [] eventname_for_loader;
					return false;
				}
			}
			Sleep(50);
		}
		else
		{
			if (!WIFEXITED(status))
			{
				ReturnCode()= BHERROR_UNKNOWN;
				err= "Load process was terminated.";
				strcpy(ErrorMessage(), err.c_str());
				delete [] eventname_for_loader;
				return false;
			}
			break;
		}
	}
#endif

	delete [] eventname_for_loader;
	if (ReturnCode() != BHERROR_SUCCESS)
		return false;

	return true;
}
#endif /*** ifdef 0 ***/

int LoadData(string shared_obj_name, string shared_obj_path)
{
	BHErrorCode bhl_error_code;
	BHLoader bhl;

	try
	{
		bhl.Init(shared_obj_name, shared_obj_path);
		auto_ptr<RCTableLoad> rct;
		try {
            string table_path = bhl.GetIOParameters().TablePath();
            table_path += ".bht";
			rct = RCTableLoadAutoPtr(new RCTableLoad(table_path.c_str(), 0, 0, true));
		} catch(DatabaseRCException&) {
			bhl.ReturnCode() = BHERROR_DATA_ERROR;
			return BHERROR_DATA_ERROR;
		}

		_int64 prev_no_obj = rct->NoObj();
		unsigned long transaction_id = bhl.TransactionID();
		rcdev.setOff();

		try {
			BufferingLevel = Configuration::GetProperty(Configuration::BufferingLevel);
		} catch (Configuration::Exception& configurationException) {
			strcpy(bhl.ErrorMessage(), configurationException.what());
			return bhl.ReturnCode() = BHERROR_BAD_CONFIGURATION;
		}

		uint sesion = 0xFFFFFFFF;
		if (rct->GetSaveSessionId() != 0xFFFFFFFF && rct->GetSaveSessionId() != transaction_id)	{
			bhl.ReturnCode() = BHERROR_CANNOT_CONNECT_TO_THE_DATABES;
			strcpy(bhl.ErrorMessage(), "Unable to start load data session.");
		} else {
			rct->SetWriteMode(transaction_id);
			try {
				rct->LoadData(bhl.GetIOParameters());
				bhl.ReturnCode() = BHERROR_SUCCESS;
				bhl.NoCopied() = bhl.NoRecords() = rct->NoRecordsLoaded();
				bhl.NoDeleted() = 0;
			} catch(FormatRCException& rce) {
				bhl.ReturnCode() = BHERROR_DATA_ERROR;
				if (rce.m_row_no != -1 && rce.m_field_no != -1) {
					stringstream err_msg;
					err_msg << rce.what() << " Row: " << rce.m_row_no << ", field: " << rce.m_field_no << '.';
					strcpy(bhl.ErrorMessage(), err_msg.str().c_str());
				}
			} catch(FileRCException& rce) {
				strcpy(bhl.ErrorMessage(), rce.what());
				bhl.ReturnCode() = BHERROR_CANNOT_OPEN_FILE_OR_PIPE;
			} catch(RCException& rce) {
				bhl.ReturnCode() = BHERROR_UNKNOWN;
				strcpy(bhl.ErrorMessage(), rce.what());
			} catch(...) {
				bhl.ReturnCode() = BHERROR_UNKNOWN;
			}
		}

		#ifdef FUNCTIONS_EXECUTION_TIMES
		fet->PrintToRcdev();
		#endif
		bhl_error_code = bhl.ReturnCode();
	} catch (...) {
		bhl_error_code = BHERROR_UNKNOWN;
	}

    bhl.FlushParameters();

	return bhl_error_code;
}
