/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifdef _MSC_VER

#include <winsock2.h>
#include <Aclapi.h>
#include "RCSystem.h"
#include "IBPipe.h"


IBPipe::IBPipe()
	:	curpipe(INVALID_HANDLE_VALUE)
{
}

IBPipe::~IBPipe() 
{
	// TODO Auto-generated destructor stub
}

bool createSecurityDescritpor(SECURITY_ATTRIBUTES &sa) 
{
	//returns true on success
	class tmpException {};

	unsigned long dwRes;
	PSID pEveryoneSID = NULL;
	PACL pACL = NULL;
	PSECURITY_DESCRIPTOR pSD = NULL;
	EXPLICIT_ACCESS ea;
	SID_IDENTIFIER_AUTHORITY SIDAuthWorld = SECURITY_WORLD_SID_AUTHORITY;
	HKEY hkSub = NULL;

	try {


		// Create a well-known SID for the Everyone group.
		if(!AllocateAndInitializeSid(&SIDAuthWorld, 1,
			SECURITY_WORLD_RID,
			0, 0, 0, 0, 0, 0, 0,
			&pEveryoneSID)) throw tmpException();

		// Initialize an EXPLICIT_ACCESS structure for an ACE.
		// The ACE will allow Everyone read access to the key.
		ZeroMemory(&ea, sizeof(EXPLICIT_ACCESS));
		ea.grfAccessPermissions = GENERIC_READ | GENERIC_WRITE;
		ea.grfAccessMode = SET_ACCESS;
		ea.grfInheritance= NO_INHERITANCE;
		ea.Trustee.TrusteeForm = TRUSTEE_IS_SID;
		ea.Trustee.TrusteeType = TRUSTEE_IS_COMPUTER;
		ea.Trustee.ptstrName  = (LPTSTR) pEveryoneSID;

		// Create a new ACL that contains the new ACEs.
		dwRes = SetEntriesInAcl(1, &ea, NULL, &pACL);
		if (ERROR_SUCCESS != dwRes) throw tmpException();

		// Initialize a security descriptor.  
		pSD = (PSECURITY_DESCRIPTOR) LocalAlloc(LPTR, 
			SECURITY_DESCRIPTOR_MIN_LENGTH); 
		if (NULL == pSD) throw tmpException();

		if (!InitializeSecurityDescriptor(pSD, SECURITY_DESCRIPTOR_REVISION)) throw tmpException();


		// Add the ACL to the security descriptor. 
		if (!SetSecurityDescriptorDacl(pSD, 
			TRUE,     // bDaclPresent flag   
			pACL, 
			FALSE))   // not a default DACL 
			throw tmpException();

		// Initialize a security attributes structure.
		sa.nLength = sizeof (SECURITY_ATTRIBUTES);
		sa.lpSecurityDescriptor = pSD;
		sa.bInheritHandle = FALSE;

		// Use the security attributes to set the security descriptor 
	}
	catch (tmpException &) {
		if (pEveryoneSID) 
			FreeSid(pEveryoneSID);
		if (pACL) 
			LocalFree(pACL);
		if (pSD) 
			LocalFree(pSD);

		return false;
	}

	return true;
}


int IBPipe::Open(std::string const& filename, int flags, ib_pmode mode)
	throw(DatabaseRCException)
{
		SECURITY_ATTRIBUTES sa;
		if (!createSecurityDescritpor(sa))
		{
			rclog << lock << "Error: can't initiate security attributes" << unlock;
			return -1;
		}
		name = filename;
		curpipe = CreateNamedPipe(filename.c_str(),
				mode,
				PIPE_TYPE_BYTE | PIPE_WAIT, //?
				1,
				PIPE_BUFFER_SIZE,
				0,
				PIPE_TIMEOUT,
				&sa);
		return 1;
}

bool IBPipe::IsOpen() 
{
		return curpipe != INVALID_HANDLE_VALUE;
}

int IBPipe::Close()
{
	if(curpipe != INVALID_HANDLE_VALUE)
	{
		FlushFileBuffers(curpipe);
		if(curpipe != INVALID_HANDLE_VALUE)
			CloseHandle(curpipe);
		curpipe = INVALID_HANDLE_VALUE;
		return 1;
	}
	return 0;
}

void WaitForPipeClient(HANDLE curpipe)
{	
	ConnectNamedPipe(curpipe, 0);
}

void WaitForPipe(LargeBuffer::Params* params)
{	
	while(params->pipe == INVALID_HANDLE_VALUE && !params->finish)
		params->pipe = CreateFileA(params->path, params->dwDesiredAccess, 0, NULL , OPEN_EXISTING, 0, NULL);
}


int IBPipe::OpenReadOnly(std::string const& filename, int create, int timeout)
	throw(DatabaseRCException)
{	
	if(create) {
		SECURITY_ATTRIBUTES sa;
		if (!createSecurityDescritpor(sa))
		{
			rclog << lock << "Error: can't initiate security attributes" << unlock;
			return -1;
		}
		curpipe = CreateNamedPipe(filename.c_str(),
			PIPE_ACCESS_INBOUND,
			PIPE_TYPE_BYTE | PIPE_WAIT,
			1,			
			0,
			PIPE_BUFFER_SIZE,
			timeout,
			&sa);
		if(curpipe == INVALID_HANDLE_VALUE){					
			return 0;			
		} else {
			unsigned long threadId;
			HANDLE thread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)WaitForPipeClient, (LPVOID) curpipe, 0, &threadId);
			if(WaitForSingleObject(thread, timeout*1000) == WAIT_OBJECT_0) { 
				return 1;						
			} else {
				TerminateThread(thread, 0);	// warning C6258: Using TerminateThread does not allow proper thread clean up
				CloseHandle(thread);
				if(curpipe != INVALID_HANDLE_VALUE)
					CloseHandle(curpipe);
				curpipe = INVALID_HANDLE_VALUE;
				return 0;
			}
		}
	} else {
		LargeBuffer::Params p;
		size_t path_ct = strlen(filename.c_str()) + 1; 
		p.path = new char [path_ct];
		p.dwDesiredAccess = GENERIC_READ;
		strcpy_s(p.path, path_ct, filename.c_str());
		p.pipe = curpipe;
		p.finish = false;
		unsigned long threadId;
		HANDLE thread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)WaitForPipe, &p,	0, &threadId);      // returns thread ID 
		if(WaitForSingleObject(thread, timeout*1000) == WAIT_OBJECT_0) {
			curpipe = p.pipe;
			CloseHandle(thread);
			return 2;
		} else {
			p.finish = true;
			WaitForSingleObject(thread, INFINITE);
			CloseHandle(thread);
			if(p.pipe != INVALID_HANDLE_VALUE)
				CloseHandle(p.pipe);
			curpipe = INVALID_HANDLE_VALUE;					
			return 0;
		}		
	}
	return 0;
};

int IBPipe::OpenCreateEmpty(std::string const& filename, int create, int timeout)
	throw(DatabaseRCException)
{
	name = filename;	
	if(create)
	{
		SECURITY_ATTRIBUTES sa;
		if (!createSecurityDescritpor(sa)) {
			rclog << lock << "Error: can't initiate security attributes" << unlock;
			return 0;
		}
		
		curpipe = CreateNamedPipe(filename.c_str(),
			PIPE_ACCESS_OUTBOUND,			
			PIPE_TYPE_BYTE | PIPE_WAIT,					
			1,	
			PIPE_BUFFER_SIZE,					
			0,					
			timeout,	
			&sa);			

		if(curpipe == INVALID_HANDLE_VALUE)	{			
			return 0;
		} else {
			unsigned long threadId;
			HANDLE thread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)WaitForPipeClient,(LPVOID) curpipe, 0, &threadId);
			if(WaitForSingleObject(thread, timeout*1000) == WAIT_OBJECT_0) { 
				return 1;
				CloseHandle(thread);
			} else {
				TerminateThread(thread, 0); // warning C6258: Using TerminateThread does not allow proper thread clean up
				CloseHandle(thread);
				if(curpipe != INVALID_HANDLE_VALUE)
					CloseHandle(curpipe);
				curpipe = INVALID_HANDLE_VALUE;
				return 0;
			}
		}
	} else {
		LargeBuffer::Params p;
		size_t path_ct = strlen(filename.c_str()) + 1; 
		p.path = new char [path_ct];
		strcpy_s(p.path, path_ct, filename.c_str());
		p.pipe = curpipe;
		p.finish = false;
		p.dwDesiredAccess = GENERIC_WRITE;
		unsigned long threadId;
		HANDLE thread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)WaitForPipe, &p,	0, &threadId);      // returns thread ID 
		if(WaitForSingleObject(thread, timeout*1000) == WAIT_OBJECT_0) {
			curpipe = p.pipe;
			CloseHandle(thread);
			return 2;
		} else {
			p.finish = true;
			WaitForSingleObject(thread, INFINITE);
			CloseHandle(thread);
			if(p.pipe != INVALID_HANDLE_VALUE)
				CloseHandle(p.pipe);
			curpipe = INVALID_HANDLE_VALUE;								
			return 0;
		}
	}
	return 0;
}


int IBPipe::OpenWriteAppend(std::string const& filename, int create, int timeout)
	throw(DatabaseRCException)
{	
	return OpenCreateEmpty(filename, create, timeout);
}

uint IBPipe::Read(void* buf, uint count)
	throw(DatabaseRCException)
{
	unsigned long no_read_bytes = 0;	
	int ret = ReadFile(curpipe, buf, count, &no_read_bytes, 0);

	//if ( ret == 0)
	//	ThrowError(errno);
	return no_read_bytes;
}

uint IBPipe::WriteExact(const void* buf, uint count)
	throw(DatabaseRCException)
{ 
	unsigned long written_bytes = 0;
	WriteFile(curpipe, buf, count, &written_bytes, 0);
	if ( written_bytes != count)
		ThrowError("write mismatch: write count and expected count did not match");
	return written_bytes;
}

#endif
