/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef _BASIC_HEAP_POLICY_B88F2F43_7761_4fcf_84B9_FD172616BB25
#define _BASIC_HEAP_POLICY_B88F2F43_7761_4fcf_84B9_FD172616BB25

#include <map>
#include <string>

#include "MemoryBlock.h"
#include "SplayTree.h"
#include "common.h"

#define MEM_HANDLE_BH void*

class BasicHeap
{
public:
	BasicHeap();
	virtual ~BasicHeap();

	MemoryBlock<MEM_HANDLE_BH>* CreateHeap(size_t size, std::string hugedir = "");

	/*
		allocate memory block of size [size] and for data of type [type]
		type != BLOCK_FREE
	*/
	MEM_HANDLE_BH	alloc(	size_t size, BLOCK_TYPE type, TrackableObject* owner);
	void		dealloc(MEM_HANDLE_BH mh);
	MEM_HANDLE_BH	rc_realloc(MEM_HANDLE_BH mh, size_t size);

	HEAP_STATUS getHeapStatus(){return m_hs;}

	size_t		getBlockSize(MEM_HANDLE_BH mh);

	/*
		Dummy function to check MEM_HANDLE_BH class consistency (it will compile only
		if MEM_HANDLE_BH = void**
	*/
	void** check_consistency(void** f){return f;}

	//For testing purposes only!
	size_t getMaximalBlock();

private:
	SplayTree<MEM_HANDLE_BH, MemoryBlock<MEM_HANDLE_BH> * > ADDRESS_TREE; //Tree of free memory blocks sorted by addresses
	SplayTree<size_t, MemoryBlock<MEM_HANDLE_BH> * > LENGTH_TREE; //Tree of free memory blocks sorted by lengths
	char* heap_frame; //Frame where heaps memory was allocated

	std::map<MEM_HANDLE_BH, MemoryBlock<MEM_HANDLE_BH>* > alloc_map;
	bool is_huge;
	size_t size;
	char hugefilename[512];
	HEAP_STATUS m_hs;
};



#endif

