/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef _SPLAY_TREE_EE2425FE_E79D_4786_AF3B_B8FEF3CE5198
#define _SPLAY_TREE_EE2425FE_E79D_4786_AF3B_B8FEF3CE5198

#include <list>

template<typename Key, typename Obj>
class SplayNode
{

public:

	SplayNode	*left, *right, *parent;
	std::list<Obj>	*items;
	Key			key;

	bool operator ==(const SplayNode& node);
	bool operator !=(const SplayNode& node);

	SplayNode(): left(0), right(0), parent(0), key(0)
	{
		items = new std::list<Obj>();
	}

	~SplayNode(void)
	{
		if(items)
		{
			items->clear();
			delete items;
			items = NULL;
		}
		/*
		Possible problem with out of memory on recursion stack!
		*/

		//Erase itself from parent references
		if(parent)
		{
			if(parent->key >= key)
				parent->left = NULL;
			else
				parent->right = NULL;
		}

		if(left)
		{
			delete left;
		}
		if(right)
		{
			delete right;
		}
	}
};


/*
SplayTree

Template class representing binary search tree with
splay - function which (takes searched key value to the root node)
and balances tree after this operation
*/


/*
	Due to characteristics of rotation function (especially zigzig)
	tree must consist of node with UNIQUE key values!! Items of new nodes with
	equal key values will be inserted to the existing node's item list

	Constraints of the solution:
	ADDRESS_TREE nodes must contain only ONE item in the list
	LENGTH_TREE nodes with the same key value are grouped in one node
*/

template <typename Key, typename Obj>
class SplayTree
{
public:

	SplayTree();
	SplayTree(		SplayTree<Key, Obj>& splyTree);

	~SplayTree(void);

	void insert(		Key key, Obj item);
	Obj remove(			Key key);
	void removeEx(		Key key, Obj item); //Remove node with a given key and a given item, no other!
	void splay(			Key key);
	SplayNode<Key, Obj>* strictSplay( Key key); //Splays the node with a given key ONLY! if key is not found NULL is returned
	SplayNode<Key, Obj>* predecessor(SplayNode<Key, Obj>* node);
	Obj removeRoot(){ return deleteNode(root);} //Returns first item from the list, if item is the only one than also removes root
	Obj deleteNode(	SplayNode<Key, Obj>* node);

	bool operator ==(const SplayTree<Key, Obj>& tree);

	Key getMostRightKey();

private:

	bool areTreesEqual(	SplayNode<Key, Obj>* nodeA, SplayNode<Key, Obj>* nodeB);
	void zig(			SplayNode<Key, Obj>* node);
	void zigzig(		SplayNode<Key, Obj>* node);
	void zigzag(		SplayNode<Key, Obj>* node);

	SplayNode<Key, Obj>* root;
};

#endif

