/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "RCSystem.h"
#include "MemoryManager.h"
#include "core/RCAttrPack.h"
#include "util/SystemInformation.h"
#include "core/tools.h"

MemoryManager rcmem;

/////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////  MemoryManager   /////////////////////////////////////////////////////////////////

const unsigned MemoryManager::VIRTUAL_MINIMUM = 0x20000000;		// 512 MB (virtual) must be always avail. (0x7e000000 for tests)
const unsigned MemoryManager::VIRTUAL_LIMIT_FOR_US = 2000000000;// 2 GB (virtual) may be occupied at max.
const unsigned MemoryManager::PHYSICAL_MINIMUM = 100000000;		// 100 MB (physical) must be always avail.


MemoryManager::MemoryManager():m_first(NULL),m_last(NULL),m_first_pack(NULL),m_last_pack(NULL)
{
#ifdef _MSC_VER
	MEMORYSTATUSEX statex;
	statex.dwLength = sizeof (statex);
	GlobalMemoryStatusEx(&statex);
	m_total_virtual = statex.ullTotalVirtual;
	if(statex.ullAvailPhys<100000000) rclog << lock << "Error: it is not a good idea to run me on less than 100 MB free physical memory." << unlock;
	srand(unsigned(time(NULL))+(uint)statex.ullAvailPhys);
#else
	_meminfo statex;
	getmeminfo(statex);
	m_total_virtual = statex.swap_total;
	if(statex.ram_free<100000000) rclog << lock << "Error: it is not a good idea to run me on less than 100 MB free physical memory." << unlock;
	srand(unsigned(time(NULL))+(uint)statex.ram_free);
#endif
	//pthread_mutexattr_t mattr;
	//pthread_mutexattr_init(&mattr);
	//pthread_mutexattr_settype(&mattr, PTHREAD_MUTEX_RECURSIVE);
	//pthread_mutex_init(&cs, &mattr);
}

MemoryManager::~MemoryManager()
{
	//pthread_mutex_destroy(&cs);
}

void MemoryManager::LockForUse(ManageableObject& o)
{
	if (o.IsDataPack())
	{
		rclog << lock << "Error: Use LockPackForUse instead of LockForUse to lock data pack against memory manager" << unlock;
		return;
	}
	//pthread_mutex_lock(&cs);
    mem_mngr_mutex.Lock();
	if (o.m_locked==0) RemoveFromLists(&o);
	o.m_locked++;
    mem_mngr_mutex.Unlock();
	//pthread_mutex_unlock(&cs);
}

void MemoryManager::UnlockFromUse(ManageableObject& o) throw(InternalRCException)
{
	if (o.IsDataPack())
	{
		rclog << lock << "Error: Use UnlockPackFromUse instead of UnlockFromUse to unlock data pack in memory manager"  << unlock;
		return;
	}
	//pthread_mutex_lock(&cs);
    mem_mngr_mutex.Lock();
	if (o.m_locked==0)
	{
        mem_mngr_mutex.Unlock();
		//pthread_mutex_unlock(&cs);
		rclog << lock << "Error: Memory resource unlock not preceded by lock" << unlock;
	}
	o.m_locked--;
	if (o.m_locked>0)
	{
        mem_mngr_mutex.Unlock();
		//pthread_mutex_unlock(&cs);
		return;
	}
	InsertLast(&o);
    mem_mngr_mutex.Unlock();
	//pthread_mutex_unlock(&cs);
	if (Overallocated()) Reserve(0);
}
//
//void MemoryManager::LockPackForUse(AttrPack*& pack, unsigned short& locks_counter, bool uncompress)
//{
//    bool shouldUncompress = false;
//	//pthread_mutex_lock(&cs);
//    mem_mngr_mutex.Lock();
//	if (pack!=NULL && locks_counter==0)
//	{
//		if (pack->IsCompressed() && uncompress)
//			shouldUncompress = true;
//		//RemoveFromLists(pack);
//	}
//	locks_counter++;
//    mem_mngr_mutex.Unlock();
//	//pthread_mutex_unlock(&cs);
//	if (shouldUncompress)
//		pack->Uncompress();
//}
//
//void MemoryManager::UnlockPackFromUse(AttrPack*& pack, unsigned short& locks_counter)
//{
//	//pthread_mutex_lock(&cs);
//    mem_mngr_mutex.Lock();
//	if (locks_counter<=0)
//	{
//        mem_mngr_mutex.Unlock();
//		//pthread_mutex_unlock(&cs);
//		rclog << lock << "Error: Data pack unlock not preceded by lock" << unlock;
//	}
//	locks_counter--;
//	if (pack==NULL || locks_counter>0)
//	{
//        mem_mngr_mutex.Unlock();
//		//pthread_mutex_unlock(&cs);
//		return;
//	}
//	//InsertLast(pack);
//	//InsertLastPack(pack);
//    mem_mngr_mutex.Unlock();
//	//pthread_mutex_unlock(&cs);
//	pack->Destroy();
//	if (Overallocated()) Reserve(0);
//}

void MemoryManager::Remove(ManageableObject& o)
{
	//pthread_mutex_lock(&cs);
    IBGuard guard(mem_mngr_mutex);
	RemoveFromLists(&o);
	//pthread_mutex_unlock(&cs);
}

void MemoryManager::InsertLastPack(AttrPack* pack)
{
	pack->m_prev_pack = m_last_pack;
	if (m_last_pack==NULL) m_first_pack = pack;
	else m_last_pack->m_next_pack = pack;
	m_last_pack = pack;
}

AttrPack* MemoryManager::RemoveFirstPack()
{
	if (m_first_pack==NULL) return NULL;
	AttrPack* pack = m_first_pack;
	m_first_pack = pack->m_next_pack;
	if (m_last_pack==pack) m_last_pack = pack->m_prev_pack;
	else pack->m_next_pack->m_prev_pack = pack->m_prev_pack;
	pack->m_prev_pack = NULL;
	pack->m_next_pack = NULL;
	return pack;
}

void MemoryManager::InsertLast(ManageableObject* o)
{
	o->m_prev = m_last;
	if (m_last==NULL) m_first = o;
	else m_last->m_next = o;
	m_last = o;
}

ManageableObject* MemoryManager::RemoveFirst()
{
	if (m_first==NULL) return NULL;
	ManageableObject* o = m_first;
	m_first = o->m_next;
	if (m_last==o) m_last = o->m_prev;
	else o->m_next->m_prev = o->m_prev;
	o->m_prev = NULL;
	o->m_next = NULL;
	return o;
}

void MemoryManager::RemoveFromLists(ManageableObject* o)
{
	if (o->IsDataPack())
	{
		AttrPack* pack = (AttrPack*)o;
		if (pack->m_prev_pack!=NULL || m_first_pack==pack)
		{
			if (m_first_pack==pack) m_first_pack = pack->m_next_pack;
			else pack->m_prev_pack->m_next_pack = pack->m_next_pack;
			if (m_last_pack==pack) m_last_pack = pack->m_prev_pack;
			else pack->m_next_pack->m_prev_pack = pack->m_prev_pack;
			pack->m_prev_pack = NULL;
			pack->m_next_pack = NULL;
		}
	}
	if (o->m_prev!=NULL || m_first==o)
	{
		if (m_first==o) m_first = o->m_next;
		else o->m_prev->m_next = o->m_next;
		if (m_last==o) m_last = o->m_prev;
		else o->m_next->m_prev = o->m_prev;
		o->m_prev = NULL;
		o->m_next = NULL;
	}
}

bool MemoryManager::Overallocated()
{
	//printf( "\tFree memory: %i M\n", VirtualMemoryAvailable()/1024/1024 );
	unsigned long long pma = PhysicalMemoryAvailable();
	if (pma < PHYSICAL_MINIMUM) return true;
	size_t vmu = VirtualMemoryUsage();
	if (m_total_virtual-vmu < VIRTUAL_MINIMUM) return true;
	if (vmu > VIRTUAL_LIMIT_FOR_US) return true;
	return false;
}

bool MemoryManager::Reserve(unsigned bytes)
{
	bool res = true;
	//pthread_mutex_lock(&cs);
    IBGuard guard(mem_mngr_mutex);
	while (res && PhysicalMemoryAvailable()-bytes < PHYSICAL_MINIMUM)
		if (m_first_pack!=NULL) RemoveFirstPack()->StayCompressed();
		else if (m_first!=NULL)
		{
			ManageableObject* o = RemoveFirst();
			if (o->IsDataPack()) delete o;
			else o->Collapse();
		}
		else res = false;
	size_t vmu = VirtualMemoryUsage();
	while (res && (m_total_virtual-vmu-bytes < VIRTUAL_MINIMUM || vmu+bytes > VIRTUAL_LIMIT_FOR_US))
	{
		if (m_first_pack!=NULL) RemoveFirstPack()->StayCompressed();
		else if (m_first!=NULL)
		{
			ManageableObject* o = RemoveFirst();
			if (o->IsDataPack()) delete o;
			else o->Collapse();
		}
		else res = false;
		vmu = VirtualMemoryUsage();
	}
	//pthread_mutex_unlock(&cs);
	return res;
}

ManageableObject::~ManageableObject()
{
	rcmem.Remove(*this);
}



#ifdef _MSC_VER
void ReportLine(Channel& chan)
{
	MEMORYSTATUSEX statex;
	char timebuf[25];
	statex.dwLength = sizeof (statex);
	GlobalMemoryStatusEx(&statex);
	_strdate_s(timebuf, 10);
	timebuf[8]=' ';
	_strtime_s(timebuf+9,10);
	chan << timebuf << " (mem.load " << statex.dwMemoryLoad
		 << "%, used " << uint(VirtualMemoryUsage()/1024/1024)
		 << " MB, ph.free " <<  statex.ullAvailPhys/1024/1024 << " MB)" << flush;

/*	unsigned int total_h,committed_h;
	MemoryOccupied(total_h,committed_h);
	chan << timebuf << " (mem.load " << (unsigned int)statex.dwMemoryLoad
		 << "%, used " << int(total_h/1024.0/1024.0)
		 << " MB, ph.free " <<  int((uint)statex.ullAvailPhys/1024.0/1024.0) << " MB)" << flush;*/
}

void MemReport(ChannelOut& chout)				// temporary
{
	MEMORYSTATUSEX statex;
	statex.dwLength = sizeof (statex);
	GlobalMemoryStatusEx(&statex);
	chout << "Memory usage: " << statex.dwMemoryLoad << "%." << endl;
	chout << "Physical memory: " << (statex.ullTotalPhys/1024/1024) << " MB total, ";
	chout << (statex.ullAvailPhys/1024/1024) << " MB free." << endl;
	chout << "Memory used by RCBase: " << int(VirtualMemoryUsage()/1024.0/1024.0*10)/10.0 << " MB." << endl;

	//chout << Paging file: "  << int((unsigned int)statex.ullAvailPageFile/1024.0/1024.0) << " MB free." << endl;
	//unsigned int total_h,committed_h;
	//MemoryOccupied(total_h,committed_h);
	//chout << "Memory used by RCBase: " << int(total_h/1024.0/1024.0*10.0)/10.0 << " MB." << endl;
	//chout << committed_h << " bytes committed in heap." << endl;
}
#endif
/////////////////////////////////////////////////////////////////////////////////////////////////////

MassiveDiscOperationObject::MassiveDiscOperationObject()
{
	counter=0;
}

MassiveDiscOperationObject::~MassiveDiscOperationObject()
{
	while(counter>0) DiscOpStop();
}

void MassiveDiscOperationObject::DiscOpStart()
{
	counter++;
}

void MassiveDiscOperationObject::DiscOpStop()
{
	if(counter>0)
	{
		counter--;
	}
}

/*int MemoryManager::MemoryAccepted(unsigned d)
{
	uint cd = cur_declared;
	if(cd>200000000)	// check physical memory only when at least 200 MB is already used (to prevent too much delays)
	{
		if(cur_declared > 900000000)
			return 0;
		if(MemoryAvailable()<100000000)	// less than 100 MB free?
			return 0;
	}
	if(d<0)				// Warning! Potential signed/unsigned problems.
	{
		if(cd<(unsigned int)(-d)) return 1;
		if(cd-(unsigned int)(-d)>=rcsystem_memory_limit) return 0;
	}
	else
	{
		if(cd>rcsystem_memory_limit) return 0;
		if((rcsystem_memory_limit-cd)<(unsigned int)d) return 0;
	}
	return 1;
}*/

/*void MemoryManager::MemoryDelta(int d)
{
	EnterCriticalSection(&cs);
	if(d>=0)
	{
		if(cur_declared>0x80000000)		// Warning: 32-bit range may be exceeded
		{
			cur_declared=cur_declared-0x80000000;
			cur_declared+=(unsigned int)d;
			if(cur_declared>=0x80000000) cur_declared=0xFFFFFFFF;		// end of memory!
			else cur_declared=cur_declared+0x80000000;
		}
		else
			cur_declared+=(unsigned int)d;
	}
	else
	{
		if(cur_declared<(unsigned int)(-d)) cur_declared=0;
		else cur_declared=cur_declared-(unsigned int)(-d);
	}
	LeaveCriticalSection(&cs);
}*/

/*void MemoryOccupied(unsigned int &total_h,unsigned int &committed_h)
{
	total_h=0;
	committed_h=0;
	int result=_HEAPOK;
	_HEAPINFO hinfo;
	hinfo._pentry=NULL;
	while(result==_HEAPOK)
	{
		result=_heapwalk(&hinfo);
		if(result==_HEAPOK)
		{
			total_h+=(uint)hinfo._size;
			if(hinfo._useflag)
				committed_h+=(uint)hinfo._size;
		}
	}
}*/

