/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef _SYSTEM_MEMORYMANAGER_H_
#define _SYSTEM_MEMORYMANAGER_H_

#include "RCException.h"
#include "Channel.h"
#include "system/ib_system.h"

//////////////////////////////////////////////////////////////////////////////////////////
///////////////////  MemoryManager:   /////////////////////////////////////////////////////////
//
// Main task:
//	- managing memory usage for large objects (data blocks etc.), including optimization,
//	  statistics, forcing compression; abstract class ManageableObject covers these tasks
//

class ManageableObject;
class AttrPack;

class MemoryManager
{
public:
	MemoryManager();
	~MemoryManager();

	//////////////////////////////////////////////////////////////////////////////////////
	///////////////////  Memory management ///////////////////////////////////////////////

	void LockForUse(ManageableObject& o);									// locks an object for use
	void UnlockFromUse(ManageableObject& o) throw(InternalRCException);		// unlocks an object from use
	//void LockPackForUse(AttrPack*& pack, unsigned short& locks_counter, bool uncompress = true);	// locks data pack for use
	//void UnlockPackFromUse(AttrPack*& pack, unsigned short& locks_counter);	// unlocks data pack from use
	void Remove(ManageableObject& o);										// removes an object from memory manager

	bool Reserve(unsigned bytes);						// makes the requested space
	bool FreeOverallocated() { return Reserve(0); };	// free the resources to the level of the assumed memory usage
	bool Overallocated();								// check whether memory usage is too high

private:
	static const unsigned VIRTUAL_MINIMUM;
	static const unsigned VIRTUAL_LIMIT_FOR_US;
	static const unsigned PHYSICAL_MINIMUM;

	void InsertLastPack(AttrPack* o);
	AttrPack* RemoveFirstPack();
	void InsertLast(ManageableObject* o);
	ManageableObject* RemoveFirst();
	void RemoveFromLists(ManageableObject* o);


	unsigned long long m_total_virtual;
	ManageableObject* m_first, * m_last;
	AttrPack* m_first_pack, * m_last_pack;
	IBMutex mem_mngr_mutex;
	//pthread_mutex_t cs;

	//unsigned int cur_declared = 0;	// old memory usage counter
    //void MemoryDelta(int d);			// declare memory usage or release (positive/negative number), just for information
};

//////////////////////////////////////////////////////////////////////////////////////

class MassiveDiscOperationObject
{
public:
	MassiveDiscOperationObject();
	virtual ~MassiveDiscOperationObject();

	void DiscOpStart();	// Increase a read_flag in MM. When MM finishes current background operation and this flag is set, it doesn�t start another scheduled background disc operation.
	void DiscOpStop();

private:
	int counter;
};

class ManageableObject
{
public:
	ManageableObject() : m_locked(1), m_prev(NULL), m_next(NULL) {}
	virtual ~ManageableObject();

	//virtual int CurrentSize() = 0;

	virtual bool IsDataPack() const { return false; };
	virtual int Collapse() = 0;		// conserving memory: delete all data, must be reloaded from disk on next use

	unsigned m_locked;
	ManageableObject* m_prev;
	ManageableObject* m_next;
};

void ReportLine(Channel& chan);
void MemReport(ChannelOut& chout);				// temporary

#endif //_SYSTEM_MEMORYMANAGER_H_

