/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include <assert.h>
#include "core/CompiledQuery.h"
#include "ConstExpressionColumn.h"
#include "core/MysqlExpression.h"
#include "core/RCAttr.h"

void ConstExpressionColumn::RequestEval()
{
	first_eval = true;
	//TODO: check if parameters were changed before reeval
	if (expr->GetItem()->type() == Item_bhfield::BHFIELD_ITEM) {
		// a special case when a naked column is a parameter
		last_val = ((Item_bhfield*)(expr->GetItem()))->GetCurrentValue();
		last_val.MakeStringOwner();
	} else
		last_val = expr->Evaluate();
}

BHReturnCode ConstExpressionColumn::DoGetValueDateTime(MIIterator const&, RCDateTime& dt, AttributeType t) {
	BHReturnCode rc = BHRC_SUCCESS;

	assert(ATI::IsDateTimeType(TypeName()));
	if(t == RC_UNKNOWN)
		t= TypeName();
	dt = RCDateTime(last_val.GetDateTime64(), t);
	return rc;
}

BHReturnCode ConstExpressionColumn::DoGetValueRCNum(MIIterator const& mit, RCNum& rcn) {
	BHReturnCode rc = BHRC_SUCCESS;
	assert(ATI::IsNumericType(TypeName()));
	if(!IsNull(mit)) {
		rcn = RCNum(last_val.x, Type().GetScale(), Type().GetScale() == -1, TypeName());
	} else
		rcn = RCNum();
	return rc;
}

double ConstExpressionColumn::DoGetValueDouble(const MIIterator& mit) {
	assert(ATI::IsNumericType(TypeName()));
	double val = 0;
	if (last_val.IsNull())
		val = NULL_VALUE_D;
	if (ATI::IsIntegerType(TypeName()))
		val = (double) last_val.Get64();
	else if(ATI::IsFixedNumericType(TypeName()))
		val = ((double) last_val.Get64()) / PowOfTen(ct.GetScale());
	else if(ATI::IsRealType(TypeName())) {
		union { double d; _int64 i;} u;
		u.i = last_val.Get64();
		val = u.d;
	} else
		assert(0 && "conversion to double not implemented");
	return val;
}

void ConstExpressionColumn::DoGetValueBin(const MIIterator& mit, int& size, char* buf) {
	if(ATI::IsStringType((TypeName()))) {
		RCBString s;
		last_val.GetString(s);
		size = s.size();
		memcpy(buf, s.GetDataBytesPointer(), size);
	} else {
		_int64 val = last_val.Get64();
		size = 8;
		memcpy(buf, &val, size);
	}
}

RCValueObject ConstExpressionColumn::DoGetValue(const MIIterator& mit, bool lookup_to_num) {

	if (last_val.null)
		return RCValueObject();

	if(ATI::IsStringType((TypeName()))) {
		RCBString s;
		last_val.GetString(s);
		return s;
	}
	if(ATI::IsIntegerType(TypeName()))
		return RCNum(last_val.Get64(), -1, false, TypeName());
	if(ATI::IsDateTimeType(TypeName()))
		return RCDateTime(last_val.GetDateTime64() , TypeName());
	if(ATI::IsRealType(TypeName()))
		return RCNum(last_val.Get64(), 0, true, TypeName());
	if(lookup_to_num || TypeName() == RC_NUM)
		return RCNum((_int64)last_val.Get64(), Type().GetScale());
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(!"Illegal execution path");
	return RCValueObject();
}

_int64 ConstExpressionColumn::DoGetSum(const MIIterator &mit)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(!ATI::IsStringType(TypeName()));
	if(last_val.IsNull())
		return NULL_VALUE_64;		// note that this is a bit ambiguous: the same is for sum of nulls and for "not implemented"
	if( ATI::IsRealType(TypeName()) ) {
		double res = last_val.GetDouble() * mit.GetPackSizeLeft();
		return *(_int64 *)&res;
	}
	return (last_val.Get64() * mit.GetPackSizeLeft());
}

RCBString ConstExpressionColumn::DoGetMinString(const MIIterator &mit) {
	RCBString s;
	last_val.GetString(s);
	return s;
}

RCBString ConstExpressionColumn::DoGetMaxString(const MIIterator &mit) {
	RCBString s;
	last_val.GetString(s);
	return s;
}

_int64 ConstExpressionColumn::DoGetApproxDistVals(bool incl_nulls)
{
	return 1;
}

ushort ConstExpressionColumn::DoMaxStringSize()		// maximal byte string length in column
{
	return ct.GetPrecision();
}

PackOntologicalStatus ConstExpressionColumn::DoGetPackOntologicalStatus(const MIIterator &mit)
{
	if (last_val.IsNull())
		return NULLS_ONLY;
	return UNIFORM;
}

void ConstExpressionColumn::DoEvaluatePack(MIUpdatingIterator& mit, Descriptor& desc)
{
	assert(0); 	// comparison of a const with a const should be simplified earlier
}

RSValue ConstExpressionColumn::DoRoughCheck(const MIIterator& mit, Descriptor& d)
{
	BHASSERT(false, "RoughCheck on VirtualColumn representing constant. Descriptor should be evaluated earlier.");
	return RS_SOME; //not implemented
}


