/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include <assert.h>
#include "core/CompiledQuery.h"
#include "SubSelectColumn.h"
#include "core/MysqlExpression.h"
#include "ConstColumn.h"
#include "core/RCAttr.h"
#include "core/ValueSet.h"

using namespace boost;

SubSelectColumn::SubSelectColumn(TempTable* subq, MultiIndex* mind, TempTable* temp_table, int temp_table_alias)
	:	MultiValColumn(subq->GetColumnType(0), mind), subq(static_pointer_cast<TempTable>(subq->shared_from_this())),
		min_max_uptodate(false), no_cached_values(0)
{
	const vector<JustATable*>* tables = &temp_table->GetTables();
	const vector<int>* aliases = &temp_table->GetAliases();

	// set col_idx
	col_idx = 0;
	for(uint i = 0; i < subq->NoAttrs(); i++)
		if(subq->IsDisplayAttr(i)) {
			col_idx = i;
			break;
		}
	ct = subq->GetColumnType(col_idx);
	MysqlExpression::SetOfVars all_params;

	for(uint i = 0; i < subq->NoVirtColumns(); i++) {
		MysqlExpression::SetOfVars params = subq->GetVirtualColumn(i)->GetParams();
		all_params.insert(params.begin(), params.end());
		MysqlExpression::bhfields_cache_t bhfields = subq->GetVirtualColumn(i)->GetBHItems();
		if(bhfields.size()) {
			for(MysqlExpression::bhfields_cache_t::const_iterator bhfield = bhfields.begin(); bhfield != bhfields.end(); bhfield++) {
				MysqlExpression::bhfields_cache_t::iterator bhitem = bhitems.find(bhfield->first) ;
				if(bhitem == bhitems.end())
					bhitems.insert(*bhfield);
				else {
					bhitem->second.insert(bhfield->second.begin(), bhfield->second.end());
				}
			}
		}
	}

	for(MysqlExpression::SetOfVars::iterator iter = all_params.begin(); iter != all_params.end(); iter++) {
		vector<int>::const_iterator ndx_it = find(aliases->begin(), aliases->end(), iter->tab);
		if(ndx_it != aliases->end()) {
			int ndx = distance(aliases->begin(), ndx_it);

			var_map.push_back(VarMap(*iter,(*tables)[ndx], ndx));
			var_types[*iter] = (*tables)[ndx]->GetColumnType(var_map[var_map.size()-1].col_ndx);
			var_buf[*iter] = std::vector<MysqlExpression::value_or_null_info_t>(); //now empty, pointers inserted by SetBufs()
		} else if(iter->tab == temp_table_alias) {
			var_map.push_back(VarMap(*iter, temp_table, 0));
			var_types[*iter] = temp_table->GetColumnType(var_map[var_map.size()-1].col_ndx);
			var_buf[*iter] = std::vector<MysqlExpression::value_or_null_info_t>(); //now empty, pointers inserted by SetBufs()
		} else {
			//parameter
			params.insert(*iter);
//			param_buf[*iter].second = NULL; //assigned by SetBufs()
		}
	}
	SetBufs(&var_buf);
//	SetBufs(&param_buf);

	if(IsCorrelated())
		table.reset();
	else
		table = this->subq;

	if(var_map.size() == 1)
		dim = var_map[0].dim;
	else
		dim = -1;
}

SubSelectColumn::~SubSelectColumn( void )
{
//	if(table != subq)
//		delete table;
}

void SubSelectColumn::SetBufs(MysqlExpression::var_buf_t* bufs)
{
	assert(bufs);
	for( MysqlExpression::bhfields_cache_t::iterator it = bhitems.begin(); it != bhitems.end(); ++ it ) {
		MysqlExpression::var_buf_t::iterator buf_set = bufs->find(it->first);
		if(buf_set != bufs->end()) {
			//for each bhitem* in the set it->second put its buffer to buf_set.second
			for(std::set<Item_bhfield*>::iterator bhfield = it->second.begin(); bhfield != it->second.end(); bhfield++) {
				ValueOrNull* von;
				(*bhfield)->SetBuf(von);
				buf_set->second.push_back(MysqlExpression::value_or_null_info_t(ValueOrNull(), von));
			}
		}
	}
}

RCBString SubSelectColumn::DoGetMinString(const MIIterator &mit)
{
	RCBString s;
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT( !"To be implemented." );
	return s;
}

RCBString SubSelectColumn::DoGetMaxString(const MIIterator &mit)
{
	RCBString s;
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT( !"To be implemented." );
	return s;
}

ushort SubSelectColumn::DoMaxStringSize()		// maximal byte string length in column
{
	return ct.GetPrecision();
}

PackOntologicalStatus SubSelectColumn::DoGetPackOntologicalStatus(const MIIterator &mit)
{
	return NORMAL;
}

void SubSelectColumn::DoEvaluatePack(MIUpdatingIterator& mit, Descriptor& desc)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT( !"To be implemented." );
	assert(0);
}

RSValue SubSelectColumn::DoRoughCheck(const MIIterator& mit, Descriptor& d)
{
	return RS_SOME; //not implemented
}


BHTribool SubSelectColumn::DoContains(MIIterator const& mit, RCDataType const& v)
{
	PrepareSubqCopy(mit);
	BHTribool res = false;
	if(!cache) {
		cache = shared_ptr<ValueSet>(new ValueSet());
		cache->Prepare(Type().GetTypeName(), Type().GetScale(), GetCollation());
	}

	if(cache->Contains(v, GetCollation()))
		return true;
	else if(cache->ContainsNulls())
		res = BHTRIBOOL_UNKNOWN;

	if(RequiresUTFConversions(GetCollation()) && Type().IsString()) {
		for(_uint64 i = no_cached_values; i < table->NoObj(); i++, no_cached_values++) {
			if(table->IsNull(i, col_idx)) {
				cache->AddNull();
				res = BHTRIBOOL_UNKNOWN;
			} else {
				RCBString val = table->GetTable_S(i, col_idx);
				val.MakePersistent();
				cache->Add(val);
				if(CollationStrCmp(GetCollation(), val, v.ToRCString()) == 0)
					return true;
			}
		}
	} else {
		for(_uint64 i = no_cached_values; i < table->NoObj(); i++, no_cached_values++) {
			if(table->IsNull(i, col_idx)) {
				cache->AddNull();
				res = BHTRIBOOL_UNKNOWN;
			} else {
				RCValueObject value = table->GetTable(i, col_idx);
				if(RCBString* s = dynamic_cast<RCBString*>(value.Get()))
					s->MakePersistent();
				cache->Add(value);
				if(value == v)
					return true;
			}
		}
	}
	return res;
}

_int64 SubSelectColumn::DoNoValues(MIIterator const& mit)
{
	PrepareSubqCopy(mit);
	if(!table->IsMaterialized())
		table->Materialize();
	return table->NoObj();
}

_int64 SubSelectColumn::DoAtLeastNoDistinctValues(MIIterator const& mit, _int64 const at_least)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(at_least > 0);
	PrepareSubqCopy(mit);
	ValueSet vals;
	vals.Prepare(expected_type.GetTypeName(), expected_type.GetScale(), expected_type.GetCollation());
	if(RequiresUTFConversions(GetCollation()) && Type().IsString()) {
		RCBString buf(NULL, CollationBufLen(GetCollation(), table->MaxStringSize(col_idx)), true);
		for(_uint64 i = 0; vals.NoVals() < at_least && i < table->NoObj(); i++) {
			if(!table->IsNull(i, col_idx)) {
				ConvertToBinaryForm(table->GetTable_S(i, col_idx), buf, GetCollation());
				vals.Add(buf);
			}
		}
	} else {	
		for(_uint64 i = 0; vals.NoVals() < at_least && i < table->NoObj(); i++) {
			if(!table->IsNull(i, col_idx)) {
				RCValueObject val = table->GetTable(i, col_idx);
				vals.Add(val);
			}
		}
	}
	return vals.NoVals();
}

bool SubSelectColumn::DoContainsNull(const MIIterator& mit)
{
	PrepareSubqCopy(mit);
	for(_uint64 i = 0; i < table->NoObj(); ++i)
		if(table->IsNull(i, col_idx))
			return true;
	return false;
}

MultiValColumn::Iterator::impl_t SubSelectColumn::DoBegin(MIIterator const& mit)
{
	PrepareSubqCopy(mit);
	return MultiValColumn::Iterator::impl_t(new IteratorImpl(table->begin()));
}

MultiValColumn::Iterator::impl_t SubSelectColumn::DoEnd(MIIterator const& mit)
{
	PrepareSubqCopy(mit);
	return MultiValColumn::Iterator::impl_t(new IteratorImpl(table->end()));
}

void SubSelectColumn::PrepareSubqCopy(const MIIterator& mit)
{
	if(IsCorrelated()) {
		if(FeedArguments(mit)) {
			table.reset();
			cache.reset();
			no_cached_values = 0;
			table = TempTable::Create(*subq, true);
			FunctionExecutor fe(bind(&TempTable::SuspendDisplay, table), bind(&TempTable::ResumeDisplay, table));
			table->ProcessParameters();
		}
	}
	if(!table->IsMaterialized()) {
		FunctionExecutor fe(bind(&TempTable::SuspendDisplay, table), bind(&TempTable::ResumeDisplay, table));
		table->Materialize(true);
	}
}

bool SubSelectColumn::IsCorrelated() const
{
	if(var_map.size() || params.size())
		return true;
	return false;
}

bool SubSelectColumn::DoIsNull(const MIIterator& mit)
{
	PrepareSubqCopy(mit);
	return table->IsNull(0, col_idx);
}

RCValueObject SubSelectColumn::DoGetValue(const MIIterator& mit, bool lookup_to_num)
{
	PrepareSubqCopy(mit);
	return table->GetTable(0, col_idx);
}

_int64 SubSelectColumn::DoGetValueInt64(const MIIterator& mit)
{
	PrepareSubqCopy(mit);
	return table->GetTable64(0, col_idx);
}

double SubSelectColumn::DoGetValueDouble(const MIIterator& mit)
{
	PrepareSubqCopy(mit);
	_int64 v = table->GetTable64(0, 0);
	return *((double*)(&v));
}

void SubSelectColumn::DoGetValueString(RCBString& s, const MIIterator& mit, bool bin_as_hex)
{
	PrepareSubqCopy(mit);
	s = table->GetTable_S(0, 0, bin_as_hex);
}

RCValueObject SubSelectColumn::DoGetSetMin(MIIterator const& mit)
{
	// assert: this->params are all set
	PrepareSubqCopy(mit);
	if(!min_max_uptodate)
		CalculateMinMax();
	return min;
}

RCValueObject SubSelectColumn::DoGetSetMax(MIIterator const& mit)
{
	PrepareSubqCopy(mit);
	if(!min_max_uptodate)
		CalculateMinMax();
	return max;
}

bool SubSelectColumn::DoIsEmpty(MIIterator const& mit)
{
	PrepareSubqCopy(mit);
	return table->NoObj() == 0;
}

void SubSelectColumn::CalculateMinMax()
{
	if(!table->IsMaterialized())
		table->Materialize();
	if(table->NoObj() == 0) {
		min = max = RCValueObject();
		min_max_uptodate = true;
		return;
	}

	min = max = RCValueObject();
	RCValueObject val;
	bool found_not_null = false;
	if(RequiresUTFConversions(GetCollation()) && Type().IsString()) {
		RCBString val_s, min_s, max_s;
		for(_uint64 i = 0; i < table->NoObj(); i++) {
			val_s = table->GetTable_S(i, col_idx);
			if(val_s.IsNull())
				continue;
			if(!found_not_null && !val_s.IsNull()) {
				found_not_null = true;
				min_s = max_s = val_s;
				min_s.MakePersistent();
				max_s.MakePersistent();
				continue;
			}
			if(CollationStrCmp(GetCollation(), val_s, max_s) > 0) {
				max_s = val_s;
				max_s.MakePersistent();
			} else if(CollationStrCmp(GetCollation(), val_s, min_s) < 0) {
				min_s = val_s;
				min_s.MakePersistent();
			}
		}
		min = min_s;
		max = max_s;
		min.MakePersistent();
		max.MakePersistent();
	} else {
		for(_uint64 i = 0; i < table->NoObj(); i++) {
			val = table->GetTable(i, col_idx);
			if(!found_not_null && !val.IsNull()) {
				found_not_null = true;
				min = max = val;
				min.MakePersistent();
				max.MakePersistent();
				continue;
			}
			if(val > max) {
				max = val;
				max.MakePersistent();
			} else if(val < min) {
				min = val;
				min.MakePersistent();
			}
		}
	}
	min_max_uptodate = true;
}

bool SubSelectColumn::FeedArguments(const MIIterator& mit)
{
	bool diff = first_eval;
	for(vector<VarMap>::const_iterator iter = var_map.begin(); iter != var_map.end(); iter++) {
		ValueOrNull v = iter->GetTabPtr()->GetComplexValue(mit[iter->dim], iter->col_ndx);
		MysqlExpression::var_buf_t::iterator cache = var_buf.find(iter->var);
		BHASSERT_WITH_NO_PERFORMANCE_IMPACT( cache != var_buf.end() );
		diff = diff || (v != cache->second.begin()->first);
		if(diff)
			for(std::vector<MysqlExpression::value_or_null_info_t>::iterator val_it = cache->second.begin(); val_it  != cache->second.end(); val_it ++)
				*((*val_it ).second) = (*val_it ).first = v;
	}
	first_eval = false;
	min_max_uptodate = false;
	return (diff);
}

void SubSelectColumn::DoSetExpectedType(ColumnType const& ct)
{
	expected_type = ct;
}
