/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include <assert.h>
#include "core/CompiledQuery.h"
#include "VirtualColumn.h"
#include "core/MysqlExpression.h"
#include "InSetColumn.h"
#include "SubSelectColumn.h"
#include "core/RCAttr.h"

using namespace std;

VirtualColumn::VirtualColumn(ColumnType const& ct, MultiIndex* mind)
	: 	Column(ct), mind(mind), conn_info(mind ? mind->ConnInfo() : ConnectionInfoOnTLS.Get()), first_eval(true), pguard(this), dim(-1)
{
}

VirtualColumn::VirtualColumn(VirtualColumn const& vc)
	:	 Column(vc.ct), mind(vc.mind), conn_info(vc.conn_info), first_eval(true), pguard(this), dim(vc.dim)
{
}


VirtualColumn::~VirtualColumn()
{
	pguard.UnlockAll();
//	for(ComplexTerm::SetOfVars::iterator iter = vars.begin(); iter != vars.end(); iter++) {
//		delete var_buf[*iter] ;
//	}
}

//void VirtualColumn::SetParam(VarID var, ValueOrNull val)
//{
//	MysqlExpression::var_buf_t::const_iterator it = param_buf.find(var);
//	if(it == param_buf.end())
//		assert(0 && "setting not existing param");
//	else
//		if (val != it->second.first) {
//			*it->second.second = val;
//			first_eval = true;
//		}
//}


void VirtualColumn::LockSourcePacks(const MIIterator& mit)
{
	pguard.LockPackrow(mit);		// TODO: do something with a return code, when it becomes nontrivial
}

void VirtualColumn::UnlockSourcePacks()
{
	pguard.UnlockAll();
}

void  VirtualColumn::MarkUsedDims(DimensionVector& dims_usage)
{
	for(var_maps_t::const_iterator it = var_map.begin(), end = var_map.end(); it != end; ++it)
		dims_usage[it->dim] = true;
}

VirtualColumn::dimensions_t VirtualColumn::GetDimensions()
{
	std::set<int> d;
	for(var_maps_t::const_iterator it = var_map.begin(), end = var_map.end(); it != end;++ it)
		d.insert(it->dim);
	return d;
}

_int64 VirtualColumn::NoTuples(void)
{
	if(mind == NULL)							// constant
		return 1;
	DimensionVector dims(mind);
	MarkUsedDims(dims);
	return MIIterator(mind, dims).NoTuples();
}

bool VirtualColumn::IsConstExpression(MysqlExpression* expr, int temp_table_alias, const vector<int>* aliases)
{
	MysqlExpression::SetOfVars& vars = expr->GetVars(); 		// get all variables from complex term

	for(MysqlExpression::SetOfVars::iterator iter = vars.begin(); iter != vars.end(); iter++) {
		vector<int>::const_iterator ndx_it = find(aliases->begin(), aliases->end(), iter->tab);
		if(ndx_it != aliases->end())
			return false;
		else
			if(iter->tab == temp_table_alias)
				return false;
	}
	return true;
}

void VirtualColumn::SetMultiIndex(MultiIndex* m, JustATablePtr t)
{
	mind = m;
	if (t)
		for(vector<VarMap>::iterator iter = var_map.begin(); iter != var_map.end(); iter++) {
			(*iter).tab = t;
	}
}

RSValue VirtualColumn::RoughCheckBetween(const MIIterator& it, _int64 min, _int64 max) 	// default: use RoughMin / RoughMax
{
	_int64 pack_min = GetMinInt64(it);
	_int64 pack_max = GetMaxInt64(it);
	if(pack_min != NULL_VALUE_64 && pack_max != NULL_VALUE_64 &&
			(pack_min > max || pack_max < min))
		return RS_NONE;
	return RS_SOME;
}
